/*
 *	plugin_aiff.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spInputPluginP.h>

#define SP_NUM_AIFF_TYPE 4
#define SP_AIFF_TYPE_NONE 0
#define SP_AIFF_TYPE_PCM 1
#define SP_AIFF_TYPE_ULAW 2
#define SP_AIFF_TYPE_ALAW 3

#define SP_AIFF_TYPE_NONE_STRING "AIFF"
#define SP_AIFF_TYPE_PCM_STRING "AIFC not compressed"
#define SP_AIFF_TYPE_ULAW_STRING "AIFC ITU-T G.711 mu-law"
#define SP_AIFF_TYPE_ALAW_STRING "AIFC ITU-T G.711 A-law"

typedef struct _spAiffPluginInstance
{
    int type_index;
    spWaveInfo info;
    spSongInfo song_info;
    
    char filename[SP_MAX_PATHNAME];
    FILE *fp;
    long current_pos;
    long total_length;
} *spAiffPluginInstance;

static spBool spInitPluginAiff(char *lang)
{
    return SP_TRUE;
}

static spBool spFreePluginAiff(void)
{
    return SP_TRUE;
}

static void *spInitPluginInstanceAiff(char *lang)
{
    spAiffPluginInstance instance;
    
    instance = xalloc(1, struct _spAiffPluginInstance);
    instance->type_index = SP_AIFF_TYPE_NONE;
    spInitWaveInfo(&instance->info);
    spInitSongInfo(&instance->song_info);
    
    strcpy(instance->filename, "");
    instance->fp = NULL;
    instance->current_pos = 0;
    instance->total_length = 0;

    return (void *)instance;
}

static spBool spFreePluginInstanceAiff(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypeAiff(void *instance, int index)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    switch (index) {
      case SP_AIFF_TYPE_NONE:
      case SP_AIFF_TYPE_PCM:
      case SP_AIFF_TYPE_ULAW:
      case SP_AIFF_TYPE_ALAW:
	pinstance->type_index = index;
	break;
      default:
	return SP_FALSE;
    }
    
    return SP_TRUE;
}

static int spGetPluginFileTypeAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    return pinstance->type_index;
}

static spBool spSetPluginSongInfoAiff(void *instance, spSongInfo *song_info)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    spCopySongInfo(&pinstance->song_info, song_info);

    return SP_TRUE;
}

static spBool spGetPluginSongInfoAiff(void *instance, spSongInfo *song_info)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    spCopySongInfo(song_info, &pinstance->song_info);

    return SP_TRUE;
}

static char *spGetPluginBestSuffixAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (pinstance->type_index == SP_AIFF_TYPE_NONE) {
	return ".aif";
    }

    return ".afc";
}

static spBool spSetPluginSampleBitAiff(void *instance, int samp_bit)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (samp_bit == 8) {
	pinstance->info.samp_bit = 8;
    } else if (samp_bit == 16) {
	pinstance->info.samp_bit = 16;
    } else if (samp_bit == 24) {
	pinstance->info.samp_bit = 24;
    } else if (samp_bit == 32) {
	pinstance->info.samp_bit = 32;
    } else {
	return SP_FALSE;
    }

    return SP_TRUE;
}

static spBool spGetPluginSampleBitAiff(void *instance, int *samp_bit)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    *samp_bit = pinstance->info.samp_bit;

    return SP_TRUE;
}

static spBool spSetPluginChannelAiff(void *instance, int num_channel)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    pinstance->info.num_channel = MAX(num_channel, 1);
    
    return SP_TRUE;
}

static spBool spGetPluginChannelAiff(void *instance, int *num_channel)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    *num_channel = pinstance->info.num_channel;

    return SP_TRUE;
}

static spBool spSetPluginSampleRateAiff(void *instance, double samp_rate)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    pinstance->info.samp_rate = samp_rate;

    return SP_TRUE;
}

static spBool spGetPluginSampleRateAiff(void *instance, double *samp_rate)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    *samp_rate = pinstance->info.samp_rate;

    return SP_TRUE;
}

static spBool spSetPluginOtherInfoAiff(void *instance, char *id, char *data)
{
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoAiff(void *instance, char *id, char *data)
{

    return SP_FALSE;
}
