/*
 *	input_audio.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spAudio.h>

#include <sp/spPluginMain.h>
#include <sp/spOutputPluginP.h>
#include <sp/spInputPluginP.h>

static spBool spInitPluginAudio(char *lang);
static spBool spFreePluginAudio(void);

static void *spInitPluginInstanceAudio(char *lang);
static spBool spFreePluginInstanceAudio(void *instance);

static spBool spGetPluginNumDeviceAudio(void *instance, int *num_device);
static char *spGetPluginDeviceNameAudio(void *instance, int index);
static spBool spSelectPluginDeviceAudio(void *instance, int index);
static spBool spSetPluginBufferSizeAudio(void *instance, int buffer_size);
static spBool spGetPluginBufferSizeAudio(void *instance, int *buffer_size);
static spBool spSetPluginVolumeAudio(void *instance, int channel, int volume);
static spBool spGetPluginVolumeAudio(void *instance, int channel, int *volume);

static spBool spSetPluginSampleBitAudio(void *instance, int samp_bit);
static spBool spGetPluginSampleBitAudio(void *instance, int *samp_bit);
static spBool spSetPluginChannelAudio(void *instance, int num_channel);
static spBool spGetPluginChannelAudio(void *instance, int *num_channel);
static spBool spSetPluginSampleRateAudio(void *instance, double samp_rate);
static spBool spGetPluginSampleRateAudio(void *instance, double *samp_rate);
static spBool spSetPluginOtherInfoAudio(void *instance, char *id, char *data);
static spBool spGetPluginOtherInfoAudio(void *instance, char *id, char *data);

static spPluginError spOpenPluginAudio(void *instance, char *filename, char *mode);
static spBool spClosePluginAudio(void *instance);
static spPluginState spGetPluginStateAudio(void *instance);
static long spGetPluginCurrentPositionAudio(void *instance);
static spBool spStopPluginAudio(void *instance);
static spBool spPausePluginAudio(void *instance);
static spBool spRestartPluginAudio(void *instance);

static long spReadPluginAudio(void *instance, char *data, long length);
static spBool spSeekPluginAudio(void *instance, long pos);
static long spGetPluginTotalLengthAudio(void *instance);


static spInputPluginRec sp_input_plugin_audio = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "Audio",
    2,
    SP_PLUGIN_PRIORITY_NULL,
    0,
    spInitPluginAudio,
    spFreePluginAudio,
    "Audio Input Plugin",
    "Audio Input Plugin Version 0.1\nCopyright(C) Hideki Banno\nE-mail: banno@itakura.nuee.nagoya-u.ac.jp",

    spInitPluginInstanceAudio,
    spFreePluginInstanceAudio,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_AUDIO,
    NULL,
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    0,
    NULL,
    NULL,
    NULL,

    spGetPluginNumDeviceAudio,
    spGetPluginDeviceNameAudio,
    spSelectPluginDeviceAudio,
    spSetPluginBufferSizeAudio,  
    spGetPluginBufferSizeAudio,
    spSetPluginVolumeAudio,      
    spGetPluginVolumeAudio,      
    
    spSetPluginSampleBitAudio,   
    spGetPluginSampleBitAudio,   
    spSetPluginChannelAudio,     
    spGetPluginChannelAudio,     
    spSetPluginSampleRateAudio,  
    spGetPluginSampleRateAudio,  
    spSetPluginOtherInfoAudio,   
    spGetPluginOtherInfoAudio,
    
    spOpenPluginAudio,              
    spClosePluginAudio,             
    spGetPluginStateAudio,          
    spGetPluginCurrentPositionAudio,
    spStopPluginAudio,              
    spPausePluginAudio,             
    spRestartPluginAudio,

    spReadPluginAudio,             
    spSeekPluginAudio,
    spGetPluginTotalLengthAudio,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_audio;
}

typedef struct _spAudioPluginInstance
{
    spAudio input_audio;
} *spAudioPluginInstance;

static spBool spInitPluginAudio(char *lang)
{
    return SP_TRUE;
}

static spBool spFreePluginAudio(void)
{
    return SP_TRUE;
}

static void *spInitPluginInstanceAudio(char *lang)
{
    spAudio audio;
    spAudioPluginInstance instance;

    if ((audio = spInitAudio()) == NULL) {
	return NULL;
    }
    
    instance = xalloc(1, struct _spAudioPluginInstance);
    instance->input_audio = audio;
    
    return (void *)instance;
}

static spBool spFreePluginInstanceAudio(void *instance)
{
    spFreeAudio(((spAudioPluginInstance)instance)->input_audio);
    xfree(instance);
    
    return SP_TRUE;
}

static spBool spGetPluginNumDeviceAudio(void *instance, int *num_device)
{
    return 1;
}

static char *spGetPluginDeviceNameAudio(void *instance, int index)
{
    if (index == 0) {
	return "default";
    }

    return NULL;
}

static spBool spSelectPluginDeviceAudio(void *instance, int index)
{
    if (index == 0) {
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spSetPluginBufferSizeAudio(void *instance, int buffer_size)
{
    return spSetAudioBufferSize(((spAudioPluginInstance)instance)->input_audio, buffer_size);
}

static spBool spGetPluginBufferSizeAudio(void *instance, int *buffer_size)
{
    return spGetAudioBufferSize(((spAudioPluginInstance)instance)->input_audio, buffer_size);
}

static spBool spSetPluginVolumeAudio(void *instance, int channel, int volume)
{
    return SP_FALSE;
}

static spBool spGetPluginVolumeAudio(void *instance, int channel, int *volume)
{

    return SP_FALSE;
}

static spBool spSetPluginSampleBitAudio(void *instance, int samp_bit)
{
    return spSetAudioSampleBit(((spAudioPluginInstance)instance)->input_audio, MAX(samp_bit, 16));
}

static spBool spGetPluginSampleBitAudio(void *instance, int *samp_bit)
{
    return spGetAudioSampleBit(((spAudioPluginInstance)instance)->input_audio, samp_bit);
}

static spBool spSetPluginChannelAudio(void *instance, int num_channel)
{
    return spSetAudioChannel(((spAudioPluginInstance)instance)->input_audio, num_channel);
}

static spBool spGetPluginChannelAudio(void *instance, int *num_channel)
{
    return spGetAudioChannel(((spAudioPluginInstance)instance)->input_audio, num_channel);
}

static spBool spSetPluginSampleRateAudio(void *instance, double samp_rate)
{
    return spSetAudioSampleRate(((spAudioPluginInstance)instance)->input_audio, samp_rate);
}

static spBool spGetPluginSampleRateAudio(void *instance, double *samp_rate)
{
    return spGetAudioSampleRate(((spAudioPluginInstance)instance)->input_audio, samp_rate);
}

static spBool spSetPluginOtherInfoAudio(void *instance, char *id, char *data)
{

    return SP_FALSE;
}

static spBool spGetPluginOtherInfoAudio(void *instance, char *id, char *data)
{

    return SP_FALSE;
}

static long sp_current_pos = 0;

static spPluginError spOpenPluginAudio(void *instance, char *filename, char *mode)
{
    spDebug(10, "spOpenPluginAudio", "in\n");

    if (spOpenAudioDevice(((spAudioPluginInstance)instance)->input_audio, mode) == SP_TRUE) {
	return SP_PLUGIN_ERROR_SUCCESS;
    } else {
	return SP_PLUGIN_ERROR_OPEN;
    }
}

static spBool spClosePluginAudio(void *instance)
{
    return spCloseAudioDevice(((spAudioPluginInstance)instance)->input_audio);
}

static spPluginState spGetPluginStateAudio(void *instance)
{
    if (sp_current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static long spGetPluginCurrentPositionAudio(void *instance)
{

    return sp_current_pos;
}

static spBool spStopPluginAudio(void *instance)
{
    sp_current_pos = 0;

    return spStopAudio(((spAudioPluginInstance)instance)->input_audio);
}

static spBool spPausePluginAudio(void *instance)
{

    return SP_FALSE;
}

static spBool spRestartPluginAudio(void *instance)
{

    return SP_FALSE;
}

static long spReadPluginAudio(void *instance, char *data, long length)
{
    int num_channel = 1;
    long nread;
    
    nread = spReadAudio(((spAudioPluginInstance)instance)->input_audio, data, length);
    spGetAudioChannel(((spAudioPluginInstance)instance)->input_audio, &num_channel);
    
    sp_current_pos += (nread / num_channel);

    return nread;
}

static spBool spSeekPluginAudio(void *instance, long pos)
{

    return SP_TRUE;
}

static long spGetPluginTotalLengthAudio(void *instance)
{

    return 0;
}
