/*
 *	plugin_nist.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spOutputPlugin.h>

#define SP_NUM_NIST_TYPE 4
#define SP_NIST_TYPE_LITTLE 0
#define SP_NIST_TYPE_BIG 1
#define SP_NIST_TYPE_ULAW 2
#define SP_NIST_TYPE_ALAW 3

typedef struct _spNistPluginInstance
{
    char type[SP_MAX_LINE];
    int type_index;
    int samp_bit;
    int num_channel;
    double samp_rate;
    long head_len;
    
    FILE *fp;
    long current_pos;
    long file_length;
} *spNistPluginInstance;

static spBool spInitPluginNist(char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginNist(void)
{

    return SP_TRUE;
}

static void *spInitPluginInstanceNist(char *lang)
{
    spNistPluginInstance instance;
    
    instance = xalloc(1, struct _spNistPluginInstance);
    strcpy(instance->type, "");
    instance->type_index = SP_NIST_TYPE_LITTLE;
    instance->samp_bit = 16;
    instance->num_channel = 1;
    instance->samp_rate = 8000.0;
    instance->head_len = 0;
    
    instance->fp = NULL;
    instance->current_pos = 0;
    instance->file_length = 0;

    return (void *)instance;
}

static spBool spFreePluginInstanceNist(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypeNist(void *instance, int index)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    pinstance->type_index = index;
    
    return SP_TRUE;
}

static int spGetPluginFileTypeNist(void *instance)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    return pinstance->type_index;
}

static spBool spSetPluginSongInfoNist(void *instance, spSongInfo *song_info)
{

    return SP_FALSE;
}

static spBool spGetPluginSongInfoNist(void *instance, spSongInfo *song_info)
{

    return SP_FALSE;
}

static char *spGetPluginBestSuffixNist(void *instance)
{
    return ".wav";
}

static spBool spSetPluginSampleBitNist(void *instance, int samp_bit)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (samp_bit == 8 || samp_bit == 16) {
	pinstance->samp_bit = samp_bit;
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spGetPluginSampleBitNist(void *instance, int *samp_bit)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    *samp_bit = pinstance->samp_bit;
    
    return SP_TRUE;
}

static spBool spSetPluginChannelNist(void *instance, int num_channel)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    pinstance->num_channel = num_channel;
    
    return SP_TRUE;
}

static spBool spGetPluginChannelNist(void *instance, int *num_channel)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    *num_channel = pinstance->num_channel;
    
    return SP_TRUE;
}

static spBool spSetPluginSampleRateNist(void *instance, double samp_rate)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    pinstance->samp_rate = samp_rate;
    
    return SP_TRUE;
}

static spBool spGetPluginSampleRateNist(void *instance, double *samp_rate)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    *samp_rate = pinstance->samp_rate;
    
    return SP_TRUE;
}

static spBool spSetPluginOtherInfoNist(void *instance, char *id, char *data)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (strcaseeq(id, SP_HEADER_LENGTH_ID)) {
	long *len = (long *)data;
	pinstance->head_len = *len;
	return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool spGetPluginOtherInfoNist(void *instance, char *id, char *data)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (strcaseeq(id, SP_HEADER_LENGTH_ID)) {
	long *len = (long *)data;
	*len = pinstance->head_len;
	return SP_TRUE;
    }

    return SP_FALSE;
}
