/*
 *	plugin_raw.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spOutputPluginP.h>

#define SP_NUM_RAW_TYPE 6
#define SP_RAW_TYPE_RAW 0
#define SP_RAW_TYPE_SWAP 1
#define SP_RAW_TYPE_LITTLE 2
#define SP_RAW_TYPE_BIG 3
#define SP_RAW_TYPE_ULAW 4
#define SP_RAW_TYPE_ALAW 5

typedef struct _spRawPluginInstance
{
    char type[SP_MAX_LINE];
    int type_index;
    int samp_bit;
    int num_channel;
    double samp_rate;
    long head_len;

    char filename[SP_MAX_PATHNAME];
    FILE *fp;
    long current_pos;
    long file_length;
} *spRawPluginInstance;

static spBool spInitPluginRaw(char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginRaw(void)
{

    return SP_TRUE;
}

static void *spInitPluginInstanceRaw(char *lang)
{
    spRawPluginInstance instance;
    
    instance = xalloc(1, struct _spRawPluginInstance);
    strcpy(instance->type, "");
    instance->type_index = SP_RAW_TYPE_RAW;
    instance->samp_bit = 16;
    instance->num_channel = 1;
    instance->samp_rate = 8000.0;
    instance->head_len = 0;
    
    strcpy(instance->filename, "");
    instance->fp = NULL;
    instance->current_pos = 0;
    instance->file_length = 0;

    return (void *)instance;
}

static spBool spFreePluginInstanceRaw(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypeRaw(void *instance, int index)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    rinstance->type_index = index;
    
    return SP_TRUE;
}

static int spGetPluginFileTypeRaw(void *instance)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    return rinstance->type_index;
}

static spBool spSetPluginSongInfoRaw(void *instance, spSongInfo *song_info)
{

    return SP_FALSE;
}

static spBool spGetPluginSongInfoRaw(void *instance, spSongInfo *song_info)
{

    return SP_FALSE;
}

static char *spGetPluginBestSuffixRaw(void *instance)
{

    return NULL;
}

static spBool spSetPluginSampleBitRaw(void *instance, int samp_bit)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (samp_bit == 8 || samp_bit == 16 || samp_bit == 64
	|| samp_bit == 24 || samp_bit == 32 || samp_bit == 33) {
	rinstance->samp_bit = samp_bit;
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spGetPluginSampleBitRaw(void *instance, int *samp_bit)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    *samp_bit = rinstance->samp_bit;
    
    return SP_TRUE;
}

static spBool spSetPluginChannelRaw(void *instance, int num_channel)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    rinstance->num_channel = num_channel;
    
    return SP_TRUE;
}

static spBool spGetPluginChannelRaw(void *instance, int *num_channel)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    *num_channel = rinstance->num_channel;
    
    return SP_TRUE;
}

static spBool spSetPluginSampleRateRaw(void *instance, double samp_rate)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    rinstance->samp_rate = samp_rate;
    
    return SP_TRUE;
}

static spBool spGetPluginSampleRateRaw(void *instance, double *samp_rate)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    *samp_rate = rinstance->samp_rate;
    
    return SP_TRUE;
}

static spBool spSetPluginOtherInfoRaw(void *instance, char *id, char *data)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (strcaseeq(id, SP_HEADER_LENGTH_ID)) {
	long *len = (long *)data;
	rinstance->head_len = *len;
	return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool spGetPluginOtherInfoRaw(void *instance, char *id, char *data)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (strcaseeq(id, SP_HEADER_LENGTH_ID)) {
	long *len = (long *)data;
	*len = rinstance->head_len;
	return SP_TRUE;
    }

    return SP_FALSE;
}
