/*
 *	main.c
 *
 *	Last modified: <2002-04-29 12:48:56 hideki>
 */

#define HIGH_LEVEL_EVENT

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spAudioLib.h>
#include <sp/spComponentLib.h>
#include <sp/spComponentMain.h>

#include "swWindow.h"
#include "swDialog.h"

#include "swStringDefs.h"
#include "swStringDefs_C.h"
/*
#if (defined(_WIN32) && !defined(__CYGWIN32__)) || defined(MACOS9)
#include "swStringDefs_SJIS.h"
#else
#include "swStringDefs_EUC.h"
#endif
*/

static spBool sw_is_region_label = SP_FALSE;
static char sw_label_file[SP_MAX_PATHNAME] = "";

static swWaveConfigRec sw_wave_config =
{
    16384,			/* max_read_length */
    262144,			/* read_callback_length */
    8192,			/* normal_read_length */
    4096,			/* play_read_length */
    8192,			/* buffer_length */
    32768,			/* max_num_undo */

    SP_FALSE,			/* thread_safe */
#ifdef SP_SUPPORT_AUDIO
    SP_TRUE,			/* play_use_audio */
#else
    SP_FALSE,			/* play_use_audio */
#endif
#ifdef _WIN32
    SP_TRUE,			/* play_use_wav */
#else
    SP_FALSE,			/* play_use_wav */
#endif
    SP_TRUE,			/* process_use_thread */
    SP_FALSE,			/* linear_spectrum */
    SP_FALSE,			/* normalize_spectrum */
    SP_FALSE,			/* float_normalized */
    
    SW_AUDIO_BUFFER_SIZE,	/* audio_buffer_size */
    "",				/* temp_dir */

    SW_TIME_FORMAT_SEC,		/* label_format */
    SW_TIME_FORMAT_SEC,		/* region_label_format */
    
    32768,			/* sfc_buffer_length */
    0.95,			/* sfc_cutoff */
    50.0,			/* sfc_sidelobe [dB] */
    0.05,			/* sfc_transition */
    2.0,			/* sfc_tolerance [%] */
    1.0,			/* sfc_gain */
    
    NULL,			/* error_func */
    NULL, 			/* option_func */
    NULL, 			/* play_func */
    NULL, 			/* edit_func */
    NULL, 			/* edit_finish_func */
    NULL,			/* read_func */
};

static int debug_level = -1;
static spBool help_flag;
static swConfigRec sw_config =
{
    SP_FALSE,			/* help_flag */
    SP_FALSE,			/* prev_flag */
    SP_FALSE,			/* use_def_format */
    SP_FALSE,			/* use_lowlevel_thread */
    SP_FALSE,			/* use_play_command */
    SP_FALSE,			/* use_loop_play */
    SP_FALSE,			/* use_sync_play */
    SP_TRUE,			/* play_draw */
    SP_FALSE,			/* alt_ctrl_swap */
    SP_FALSE,			/* no_tool_bar */
    SP_TRUE,			/* display_overview */
    SP_TRUE,			/* display_meter */
    SP_TRUE,			/* draw_selection_times */
    SP_TRUE,			/* draw_selection_length */
    SP_FALSE,			/* link_region_label */
    SP_TRUE,			/* erase_label_prompt */
    SP_FALSE,			/* display_info_area */
    SP_FALSE,			/* display_freq_info_area */
    SP_FALSE,			/* analysis_new_window */

    SP_TRUE,			/* scale_flag */
    SP_TRUE,			/* grid_flag */
    SP_TRUE,			/* zero_flag */
    
    SW_TIME_DATA,		/* data_type */
    SW_ANALYSIS_SPECTRUM,	/* analysis_type */
    SW_WINDOW_HAMMING,		/* window_type */
    0,				/* num_channel */
    1,				/* def_num_channel */
    -1,				/* head_size */
    0,				/* def_head_size */
    0,				/* samp_bit */
    16,				/* def_samp_bit */
    SW_AUDIO_BUFFER_SIZE,	/* audio_buffer_size */
    
    800,			/* width */
    200,			/* height */
    400,			/* freq_width */
    300,			/* freq_height */
    50,				/* overview_height */
    60,				/* meter_range */
    256,			/* fftl */
    65536,			/* max_fftl */
    4.0,			/* lifterm */
    8.0,			/* shiftm */
    32.0,			/* framem */
    8.0,			/* wide_framem */
    51.1,			/* narrow_framem */
    0.0,			/* samp_rate */
    8000.0,			/* def_samp_rate */
    
    "",				/* data_type_string */
    "",				/* analysis_type_string */
    "",				/* window_type_string */
    "",				/* format_string */
    "",				/* def_format_string */
    "",				/* play_command */
    "",				/* browser_command */
    "",				/* doc_dir */
    "",				/* setup_file */
    "",				/* temp_dir */

    SP_TRUE,			/* autosave_use_region */
    SP_FALSE,			/* autosave_label_fullpath */
    SP_FALSE,			/* autosave_by_drop */
    0,				/* autosave_id_type */
    "",				/* autosave_dir */
    "",				/* autosave_id_prefix */
    "",				/* autosave_label_suffix */

    "",				/* wave_fg */
    "",				/* wave_fg */
    "",				/* pointer_color */
    "",				/* string_color */
    "",				/* label_color */
    "",				/* scale_color */

    SW_TIME_FORMAT_SEC,		/* time_format */
    SW_FREQ_FORMAT_HZ,		/* freq_format */

    SP_FALSE,			/* format_specified */
    
    &sw_wave_config,		/* wave_config */
    NULL,			/* toplevel */
};

static spOption sw_option[] = {
    {"-f", "-freq", "sampling frequency for raw data [Hz]", NULL,
	 SP_TYPE_DOUBLE, &sw_config.samp_rate, NULL},
    {NULL, NULL, "default sampling frequency [Hz]", "samp_freq",
	 SP_TYPE_DOUBLE, &sw_config.def_samp_rate, "8000.0"},
    {"-c", "-channel", "number of channel for raw data", NULL,
	 SP_TYPE_INT, &sw_config.num_channel, NULL},
    {NULL, NULL, "default number of channel", "num_channel",
	 SP_TYPE_INT, &sw_config.def_num_channel, "1"},
    {"-b", "-bit", "bits per sample for raw data", NULL,
	 SP_TYPE_INT, &sw_config.samp_bit, NULL},
    {NULL, NULL, "default bits per sample", "bits_per_sample",
	 SP_TYPE_INT, &sw_config.def_samp_bit, "16"},
    {"-F", "-format", "file format for raw data (raw|swap|little|big|text)", NULL,
	 SP_TYPE_STRING_A, sw_config.format_string, NULL},
    {NULL, NULL, "default file format (raw|swap|little|big|text)", "file_format",
	 SP_TYPE_STRING_A, sw_config.def_format_string, "raw"},
#if 0
    {"-head", "-header", "header length for raw data", NULL,
	 SP_TYPE_INT, &sw_config.head_size, NULL},
    {NULL, NULL, "default header length", "header_length",
	 SP_TYPE_INT, &sw_config.def_head_size, "0"},
#endif
    {"-play", NULL, "sound play command", "play_command",
	 SP_TYPE_STRING_A, sw_config.play_command, ""},
#ifdef SP_SUPPORT_AUDIO
    {"-useplay", NULL, "use play command forcedly", "use_play",
	 SP_TYPE_BOOLEAN, &sw_config.use_play_command, SP_FALSE_STRING},
    {NULL, NULL, "use loop play initially", "use_loop_play",
	 SP_TYPE_BOOLEAN, &sw_config.use_loop_play, SP_FALSE_STRING},
    {NULL, NULL, "use synchronous play initially", "use_sync_play",
	 SP_TYPE_BOOLEAN, &sw_config.use_sync_play, SP_FALSE_STRING},
    {NULL, NULL, "draw cursor on playing", "play_draw",
	 SP_TYPE_BOOLEAN, &sw_config.play_draw, SP_TRUE_STRING},
    {"-buf", NULL, "audio buffer size", "audio_buffer_size",
	 SP_TYPE_INT, &sw_config.audio_buffer_size, NULL},
#endif
    {NULL, NULL, "use low level thread", "use_lowlevel_thread",
	 SP_TYPE_BOOLEAN, &sw_config.use_lowlevel_thread, SP_FALSE_STRING},
    {NULL, NULL, "use thread for processing", "use_thread",
	 SP_TYPE_BOOLEAN, &sw_wave_config.process_use_thread, SP_TRUE_STRING},
    {"-tf", "-tformat", "time format index (0-3)", "time_format",
	 SP_TYPE_ENUM, &sw_config.time_format, "0"},
    {"-label", NULL, "label file for first input file", NULL,
         SP_TYPE_STRING_A, sw_label_file, ""},
    {"-reg", NULL, "label file specified by \"-label\" is region label", NULL,
	 SP_TYPE_BOOLEAN, &sw_is_region_label, SP_FALSE_STRING},
    {"-lf", "-lformat", "label format index (0-12)", "label_format",
	 SP_TYPE_ENUM, &sw_wave_config.label_format, "0"},
    {"-rf", "-rformat", "region label format index (0-11)", "region_label_format",
	 SP_TYPE_ENUM, &sw_wave_config.region_label_format, "0"},
#ifdef SW_SUPPORT_AUTOSAVE
    {NULL, NULL, "write fullpaths into autosave label files", "autosave_label_fullpath",
	 SP_TYPE_BOOLEAN, &sw_config.autosave_label_fullpath, SP_FALSE_STRING},
    {NULL, NULL, "autosave by drag & drop", "autosave_by_drop",
	 SP_TYPE_BOOLEAN, &sw_config.autosave_by_drop, SP_FALSE_STRING},
    {NULL, NULL, "autosave ID type", "autosave_id_type",
	 SP_TYPE_INT, &sw_config.autosave_id_type, "0"},
    {NULL, NULL, "autosave directory", "autosave_dir",
	 SP_TYPE_STRING_A, sw_config.autosave_dir, ""},
    {NULL, NULL, "autosave ID prefix", "autosave_id_prefix",
	 SP_TYPE_STRING_A, sw_config.autosave_id_prefix, ""},
    {NULL, NULL, "autosave label suffix", "autosave_label_suffix",
	 SP_TYPE_STRING_A, sw_config.autosave_label_suffix, "_cut.txt"},
#endif
    {"-temp", NULL, "temporary directory", "temp_dir",
	 SP_TYPE_STRING_A, sw_config.temp_dir, ""},
    {"-wavefg", NULL, "wave foreground color", "wave_fg",
	 SP_TYPE_STRING_A, sw_config.wave_fg, "black"},
    {"-wavebg", NULL, "wave background color", "wave_bg",
	 SP_TYPE_STRING_A, sw_config.wave_bg, "white"},
    {"-pointfg", NULL, "pointer foreground color", "point_fg",
	 SP_TYPE_STRING_A, sw_config.pointer_color, "red"},
    {"-stringfg", NULL, "string foreground color", "string_fg",
	 SP_TYPE_STRING_A, sw_config.string_color, "red"},
    {"-labelfg", NULL, "label foreground color", "label_fg",
	 SP_TYPE_STRING_A, sw_config.label_color, "blue"},
    {"-scalefg", NULL, "scale foreground color", "scale_fg",
	 SP_TYPE_STRING_A, sw_config.scale_color, "grey50"},
    {"-width", NULL, "window width", "window_width",
	 SP_TYPE_INT, &sw_config.width, "800"},
    {"-height", NULL, "window height", "window_height",
	 SP_TYPE_INT, &sw_config.height, "200"},
#ifdef SW_USE_ANALYSIS
    {"-fwidth", NULL, "window width for frequency data", "freq_width",
	 SP_TYPE_INT, &sw_config.freq_width, "400"},
    {"-fheight", NULL, "window height for frequency data", "freq_height",
	 SP_TYPE_INT, &sw_config.freq_height, "300"},
    {NULL, NULL, "default analysis type", "analysis_type",
	 SP_TYPE_STRING_A, sw_config.analysis_type_string, "spectrum"},
    {NULL, NULL, "default window type", "window_type",
	 SP_TYPE_STRING_A, sw_config.window_type_string, "hamming"},
    {NULL, NULL, "minimum FFT length", "fft_length",
	 SP_TYPE_LONG, &sw_config.fftl, "256"},
    {NULL, NULL, "minimum FFT length", "max_fft_length",
	 SP_TYPE_LONG, &sw_config.max_fftl, "65536"},
    {NULL, NULL, "lifter length for cepstrum analysis [ms]", "lifter",
	 SP_TYPE_DOUBLE, &sw_config.lifterm, "4.0"},
    {NULL, NULL, "frame length for wide band analysis [ms]", "wide_frame",
	 SP_TYPE_DOUBLE, &sw_config.wide_framem, "8.0"},
    {NULL, NULL, "frame length for narrow band analysis [ms]", "narrow_frame",
	 SP_TYPE_DOUBLE, &sw_config.narrow_framem, "51.1"},
#if 0
    {NULL, NULL, "default frame shift [ms]", "shift",
	 SP_TYPE_DOUBLE, &sw_config.shiftm, "8.0"},
    {NULL, NULL, "default frame length [ms]", "frame",
	 SP_TYPE_DOUBLE, &sw_config.framem, "32.0"},
#endif
    {NULL, NULL, "create always new window on analysis", "analysis_new_window",
	 SP_TYPE_BOOLEAN, &sw_config.analysis_new_window, SP_FALSE_STRING},
    {NULL, NULL, "use linear spectrum (not use decibel)", "linear_spectrum",
	 SP_TYPE_BOOLEAN, &sw_wave_config.linear_spectrum, SP_FALSE_STRING},
    {NULL, NULL, "normalize spectrum with maximum amplitude of waveform", "normalize_spectrum",
	 SP_TYPE_BOOLEAN, &sw_wave_config.normalize_spectrum, SP_FALSE_STRING},
#endif
    {NULL, NULL, "floating value is normalized from -1.0 to 1.0", "float_normalized",
	 SP_TYPE_BOOLEAN, &sw_wave_config.float_normalized, SP_FALSE_STRING},
    {NULL, NULL, "buffer length for sampling frequency conversion", "sfc_buffer_length", 
	 SP_TYPE_LONG, &sw_wave_config.sfc_buffer_length, "32768"},
    {NULL, NULL, "normalized cutoff frequency (nyquist = 1)", "cutoff", 
	 SP_TYPE_DOUBLE, &sw_wave_config.sfc_cutoff, "0.95"},
    {NULL, NULL, "height of sidelobe [dB]", "sidelobe", 
	 SP_TYPE_DOUBLE, &sw_wave_config.sfc_sidelobe, "50.0"},
    {NULL, NULL, "normalized transition width", "transition", 
	 SP_TYPE_DOUBLE, &sw_wave_config.sfc_transition, "0.05"},
    {NULL, NULL, "tolerance [%]", "tolerance", 
	 SP_TYPE_DOUBLE, &sw_wave_config.sfc_tolerance, "2.0"},
    {"-usedef", NULL, "use default format when opening file", "use_default",
	 SP_TYPE_BOOLEAN, &sw_config.use_def_format, SP_FALSE_STRING},
    {"-acswap", NULL, "alt-ctrl swap for shortcut keys", "alt_ctrl_swap",
	 SP_TYPE_BOOLEAN, &sw_config.alt_ctrl_swap, SP_FALSE_STRING},
#ifdef SW_USE_TOOL_BAR
    {NULL, NULL, "not use tool bar", "no_tool_bar",
	 SP_TYPE_BOOLEAN, &sw_config.no_tool_bar, SP_FALSE_STRING},
#endif
    {NULL, NULL, "display overview of waveform", "display_overview",
         SP_TYPE_BOOLEAN, &sw_config.display_overview, SP_TRUE_STRING},
    {NULL, NULL, "display level meter", "display_meter",
         SP_TYPE_BOOLEAN, &sw_config.display_meter, SP_TRUE_STRING},
    {NULL, NULL, "draw selection times", "draw_selection_times",
         SP_TYPE_BOOLEAN, &sw_config.draw_selection_times, SP_TRUE_STRING},
    {NULL, NULL, "draw selection length", "draw_selection_length",
         SP_TYPE_BOOLEAN, &sw_config.draw_selection_length, SP_TRUE_STRING},
    {"-lreg", NULL, "link label to selection", "link_region_label",
	 SP_TYPE_BOOLEAN, &sw_config.link_region_label, SP_FALSE_STRING},
    {NULL, NULL, "prompt before erasing labels", "erase_label_prompt",
	 SP_TYPE_BOOLEAN, &sw_config.erase_label_prompt, SP_TRUE_STRING},
    {"-info", NULL, "display information area", "display_info_area",
	 SP_TYPE_BOOLEAN, &sw_config.display_info_area, SP_FALSE_STRING},
    {"-finfo", NULL, "display frequency information area", "display_freq_info_area",
	 SP_TYPE_BOOLEAN, &sw_config.display_freq_info_area, SP_FALSE_STRING},
    {NULL, NULL, "draw scale", "draw_scale",
	 SP_TYPE_BOOLEAN, &sw_config.scale_flag, SP_TRUE_STRING},
    {NULL, NULL, "draw grid", "draw_grid",
	 SP_TYPE_BOOLEAN, &sw_config.grid_flag, SP_TRUE_STRING},
    {"-debug", NULL, "debug level", NULL,
	 SP_TYPE_INT, &debug_level, NULL},
    {"-h", "-help", "display this message", NULL,
	 SP_TYPE_BOOLEAN, &help_flag, SP_FALSE_STRING},
};

static char *sw_filelabel[] = {
    "[filename...]",
};

int spMain(int argc, char *argv[])
{
    char *filename;
    spOptions options;
    spTopLevel toplevel;
    char buf[SP_MAX_MESSAGE];

    spSetApplicationId("spwave");

    //spSetStringTable(SW_JP_LANG, sw_string_table_jp);
    spSetStringTable("C", sw_string_table);

    /* initialize toolkit */
    toplevel = spInitialize(&argc, &argv,
			    SppIconName, "spwave_icon",
			    SppThreadSafe, SP_TRUE,
			    SppUseWindowMenu, SP_TRUE,
			    NULL);
    
    spSetOptionDescOffset(48);
    spSetSetup(SW_SETUP_FILE);
    spSetHelpMessage(&help_flag, SW_HELP_MESSAGE, SW_VERSION_STRING);
    
    /* get option value */
    options = spGetOptions(argc, argv, sw_option, sw_filelabel);
    spGetOptionsValue(argc, argv, options);
#if 0
    spSetDebugLevel(100);
#else
    spSetDebugLevel(debug_level);
#endif
    
    if (spIsThreadLevelVariable() == SP_TRUE) {
	if (sw_config.use_lowlevel_thread == SP_TRUE) {
	    spSetThreadLevel(SP_THREAD_LEVEL_LOW);
	} else {
	    spSetThreadLevel(SP_THREAD_LEVEL_HIGH);
	}
    }
    
    /* set program information (useful on Mac) */
    sprintf(buf, SW_INFO_MESSAGE, SW_VERSION_STRING);
    spSetTopLevelParams(toplevel,
			SppInformation, buf,
			SppAltCtrlSwap, sw_config.alt_ctrl_swap,
			NULL);
    
    swInitialize(toplevel, &sw_config);

    spDebug(10, NULL, "label_file = %s, label_format = %d, region_label_format = %d\n",
	    sw_label_file, sw_wave_config.label_format, sw_wave_config.region_label_format);
    
    while ((filename = spGetFile(options)) != NULL) {
	spDebug(10, NULL, "file = %s\n", filename);
	swCreateWindow(filename, sw_label_file, sw_is_region_label, &sw_config);
    }
    swCreateWindow(NULL, NULL, SP_FALSE, &sw_config);
    
    spDebug(10, NULL, "loop of creating window done\n");

    /* main loop */
    return spMainLoop(toplevel);
}
