/*
 *	swAnalysis.c
 *
 *	Last modified: <2002-02-17 01:40:21 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>

#include "swWindow.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swEdit.h"

#ifdef SW_USE_ANALYSIS
#include <sp/spLib.h>

swAnalysisType swGetAnalysisTypeFromLabel(char *label)
{
    if (streq(label, SW_ANALYSIS_SPECTRUM_LABEL)
	|| streq(label, SW_ANALYSIS_SPECTRUM_PARAM_LABEL)) {
	return SW_ANALYSIS_SPECTRUM;
    } else if (streq(label, SW_ANALYSIS_SMOOTHED_SPECTRUM_LABEL)
	       || streq(label, SW_ANALYSIS_SMOOTHED_SPECTRUM_PARAM_LABEL)) {
	return SW_ANALYSIS_SMOOTHED_SPECTRUM;
    } else if (streq(label, SW_ANALYSIS_PHASE_LABEL)
	       || streq(label, SW_ANALYSIS_PHASE_PARAM_LABEL)) {
	return SW_ANALYSIS_PHASE;
    } else if (streq(label, SW_ANALYSIS_UNWRAPPED_PHASE_LABEL)
	       || streq(label, SW_ANALYSIS_UNWRAPPED_PHASE_PARAM_LABEL)) {
	return SW_ANALYSIS_UNWRAPPED_PHASE;
    } else if (streq(label, SW_ANALYSIS_GROUP_DELAY_LABEL)
	       || streq(label, SW_ANALYSIS_GROUP_DELAY_PARAM_LABEL)) {
	return SW_ANALYSIS_GROUP_DELAY;
    } else if (streq(label, SW_ANALYSIS_CEPSTRUM_LABEL)
	       || streq(label, SW_ANALYSIS_CEPSTRUM_PARAM_LABEL)) {
	return SW_ANALYSIS_CEPSTRUM;
    } else {
	return SW_ANALYSIS_NONE;
    }
}

char *swGetAnalysisTypeParamLabel(swAnalysisType type)
{
    switch (type) {
      case SW_ANALYSIS_SPECTRUM:
	return SW_ANALYSIS_SPECTRUM_PARAM_LABEL;
      case SW_ANALYSIS_SMOOTHED_SPECTRUM:
	return SW_ANALYSIS_SMOOTHED_SPECTRUM_PARAM_LABEL;
      case SW_ANALYSIS_PHASE:
	return SW_ANALYSIS_PHASE_PARAM_LABEL;
      case SW_ANALYSIS_UNWRAPPED_PHASE:
	return SW_ANALYSIS_UNWRAPPED_PHASE_PARAM_LABEL;
      case SW_ANALYSIS_GROUP_DELAY:
	return SW_ANALYSIS_GROUP_DELAY_PARAM_LABEL;
      case SW_ANALYSIS_CEPSTRUM:
	return SW_ANALYSIS_CEPSTRUM_PARAM_LABEL;
      default:
	return NULL;
    }
}

swWindowType swGetWindowTypeFromLabel(char *label)
{
    if (streq(label, SW_WINDOW_RECTANGLE_LABEL)
	|| streq(label, SW_WINDOW_RECTANGLE_PARAM_LABEL)) {
	return SW_WINDOW_RECTANGLE;
    } else if (streq(label, SW_WINDOW_HAMMING_LABEL)
	       || streq(label, SW_WINDOW_HAMMING_PARAM_LABEL)) {
	return SW_WINDOW_HAMMING;
    } else if (streq(label, SW_WINDOW_HANNING_LABEL)
	       || streq(label, SW_WINDOW_HANNING_PARAM_LABEL)) {
	return SW_WINDOW_HANNING;
    } else if (streq(label, SW_WINDOW_BLACKMAN_LABEL)
	       || streq(label, SW_WINDOW_BLACKMAN_PARAM_LABEL)) {
	return SW_WINDOW_BLACKMAN;
    } else if (streq(label, SW_WINDOW_GAUSS_LABEL)
	       || streq(label, SW_WINDOW_GAUSS_PARAM_LABEL)) {
	return SW_WINDOW_GAUSS;
    } else {
	return SW_WINDOW_NONE;
    }
}

char *swGetWindowTypeParamLabel(swWindowType type)
{
    switch (type) {
      case SW_WINDOW_RECTANGLE:
	return SW_WINDOW_RECTANGLE_PARAM_LABEL;
      case SW_WINDOW_HAMMING:
	return SW_WINDOW_HAMMING_PARAM_LABEL;
      case SW_WINDOW_HANNING:
	return SW_WINDOW_HANNING_PARAM_LABEL;
      case SW_WINDOW_BLACKMAN:
	return SW_WINDOW_BLACKMAN_PARAM_LABEL;
      case SW_WINDOW_GAUSS:
	return SW_WINDOW_GAUSS_PARAM_LABEL;
      default:
	return NULL;
    }
}

void swSetDataLabel(swWindow window, char *label)
{
    if (window != NULL) {
	if (window->data_label != NULL) {
	    xfree(window->data_label);
	    window->data_label = NULL;
	}
	if (!strnone(label)) {
	    window->data_label = strclone(label);
	}
    }
    
    return;
}

void swSetDataDimension(swWindow window, char *dim)
{
    if (window != NULL) {
	if (window->data_dim != NULL) {
	    xfree(window->data_dim);
	    window->data_dim = NULL;
	}
	if (!strnone(dim)) {
	    window->data_dim = strclone(dim);
	}
    }
    
    return;
}

swWindow swAnalysisWindow(swWindow window, long offset, long length, spBool new_flag)
{
    long fftl;
    swWave wave;
    swWindow new_window = NULL;
    swDataType data_type = SW_FREQ_DATA;

    if (swIsNoWave(window) == SP_TRUE) return NULL;

    spDebug(10, "swAnalysisWindow", "analysis_type = %d, window_type = %d\n",
	    window->config->analysis_type, window->config->window_type);

    fftl = MAX(length, window->config->fftl);
    fftl = POW2(spNextPow2(fftl));
    if (fftl > window->config->max_fftl) {
	spDisplayError(window->window, NULL, SW_ANALYSIS_LONG_ERROR_MESSAGE);
	return NULL;
    }
    
    swSetMouseCursor(window, SP_CURSOR_WAIT);
    
    wave = swAnalysisWave(window->wave, offset, length,
			  window->config->analysis_type, window->config->window_type,
			  fftl, window->config->lifterm);

    if (wave == NULL) {
	swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
	spDisplayError(window->window, NULL, SW_ANALYSIS_ERROR_MESSAGE);
	return NULL;
    }

    if (window->config->analysis_type == SW_ANALYSIS_CEPSTRUM) {
	data_type = SW_TIME_DATA;
    }
    
    if (window->related_window != NULL
	&& window->related_window->analysis_flag == SP_TRUE
	&& (new_flag == SP_FALSE || window->config->analysis_new_window == SP_FALSE)) {
	new_window = window->related_window;
	
	if (new_window->wave != NULL) {
	    swDestroyWave(new_window->wave);
	}
	new_window->offset = 0;
	new_window->length = 0;
	new_window->analysis_flag = SP_TRUE;
	
	new_window->data_type = data_type;
	
	swSetWave(new_window, wave);
	swSetWindowTitle(new_window);
	swSetSenseLevel(new_window);
	swReloadWave(new_window, SP_FALSE);
    } else {
	/* create new window */
	new_window = swCreateWaveWindow(wave, window->config, data_type);
	new_window->analysis_flag = SP_TRUE;
    }
    window->related_window = new_window;
    new_window->related_window = window;

    spDebug(10, "swAnalysisWindow", "find window done\n");
    
    /* set data label and dimension */
    switch (window->config->analysis_type) {
      case SW_ANALYSIS_SPECTRUM:
      case SW_ANALYSIS_SMOOTHED_SPECTRUM:
	if (window->config->analysis_type == SW_ANALYSIS_SMOOTHED_SPECTRUM) {
	    swSetDataLabel(new_window, "Smoothed Amplitude");
	} else {
	    swSetDataLabel(new_window, "Amplitude");
	}
	if (window->config->wave_config->linear_spectrum == SP_FALSE) {
	    swSetDataDimension(new_window, "[dB]");
	}
	break;
      case SW_ANALYSIS_PHASE:
      case SW_ANALYSIS_UNWRAPPED_PHASE:
	if (window->config->analysis_type == SW_ANALYSIS_UNWRAPPED_PHASE) {
	    swSetDataLabel(new_window, "Unwrapped Phase");
	} else {
	    swSetDataLabel(new_window, "Phase");
	}
	swSetDataDimension(new_window, "[rad]");
	break;
      case SW_ANALYSIS_GROUP_DELAY:
	swSetDataLabel(new_window, "Group Delay");
	swSetDataDimension(new_window, "[ms]");
	break;
      default:
	break;
    }
    
    swSetMouseCursor(window, SP_CURSOR_UNKNOWN);

    spDebug(10, "swAnalysisWindow", "done\n");
    
    return new_window;
}

void swAnalysisRegionCB(spComponent component, swWindow window)
{
    long offset, length;
    swWindow new_window;

    if (swIsNoWave(window) == SP_FALSE) {
	/* get edge */
	if (swGetEdge(window, window->sel_st, window->sel_ed, &offset, &length) != SP_TRUE) {
	    spDisplayError(window->window, NULL, SW_REGION_ERROR_MESSAGE);
	    return;
	}
	
	if ((new_window = swAnalysisWindow(window, offset, length, SP_TRUE)) == NULL) {
	    return;
	} else {
	    spMapWindow(new_window->window);
	}
    }
    return;
}

void swAnalysisFrame(spComponent component, swWindow window, spBool wide_flag)
{
    long framel;
    long ana_offset;
    long offset, length;
    swWindow new_window;

    if (swIsNoWave(window) == SP_FALSE) {
	if (wide_flag == SP_TRUE) {
	    framel = swDimToSamp(window, window->config->wide_framem / 1000.0);
	} else {
	    framel = swDimToSamp(window, window->config->narrow_framem / 1000.0);
	}
	ana_offset = window->point - framel / 2;
	
	/* get edge */
	if (swGetEdge(window, ana_offset, ana_offset + framel - 1, &offset, &length) != SP_TRUE) {
	    spDisplayError(window->window, NULL, SW_REGION_ERROR_MESSAGE);
	    return;
	}

	/* select region */
	swSelectRegion(window, window->wave->selected_channel, offset, offset + length - 1);
	
	/* analysis */
	if ((new_window = swAnalysisWindow(window, offset, length, SP_TRUE)) == NULL) {
	    return;
	} else {
	    spMapWindow(new_window->window);
	}
    }
    return;
}

void swAnalysisWideCB(spComponent component, swWindow window)
{
    swAnalysisFrame(component, window, SP_TRUE);
    return;
}

void swAnalysisNarrowCB(spComponent component, swWindow window)
{
    swAnalysisFrame(component, window, SP_FALSE);
    return;
}

void swAnalysisAll(spComponent component)
{
    long offset, length;
    spComponent next = NULL;
    swWindow window = NULL;
    swWindow next_window = NULL;

    if (spIsCreated(component) == SP_FALSE) return;

    next = component;
    while (1) {
	if (next == NULL) {
	    break;
	}
	
	if ((window = (swWindow)spGetUserData(next)) != NULL) {
	    break;
	}

	next = spGetNextWindow(next, SP_TRUE);
	
	if (next == component) {
	    break;
	}
    }

    if (window == NULL) return;

    next = window->window;
    while (1) {
	if (next == NULL) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && next_window->wave != NULL) {
	    if (next_window->data_type == SW_TIME_DATA
		&& next_window->related_window != NULL) {
		/* get edge */
		if (swGetEdge(next_window, next_window->sel_st, next_window->sel_ed, &offset, &length) == SP_TRUE) {
		    swAnalysisWindow(next_window, offset, length, SP_FALSE);
		}
	    }
	}

	next = spGetNextWindow(next, SP_FALSE);
	
	if (next == window->window) {
	    break;
	}
    }
    
    return;
}

#endif

