/*
 *	swAnalysisDialog.c
 *
 *	Last modified: <2001-03-01 00:30:43 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>

#include "swWindow.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swEdit.h"

#ifdef SW_USE_ANALYSIS
#include "swAnalysis.h"
#include "swAnalysisDialog.h"

static char *analysis_type_list[] =
{
    SW_ANALYSIS_SPECTRUM_LABEL,
    SW_ANALYSIS_SMOOTHED_SPECTRUM_LABEL,
    SW_ANALYSIS_PHASE_LABEL,
    SW_ANALYSIS_UNWRAPPED_PHASE_LABEL,
    SW_ANALYSIS_GROUP_DELAY_LABEL,
#if 1
    SW_ANALYSIS_CEPSTRUM_LABEL,
#endif
    NULL,
};
static char *window_type_list[] =
{
    SW_WINDOW_RECTANGLE_LABEL,
    SW_WINDOW_HAMMING_LABEL,
    SW_WINDOW_HANNING_LABEL,
    SW_WINDOW_BLACKMAN_LABEL,
    SW_WINDOW_GAUSS_LABEL,
    NULL,
};
static char *fft_length_list[] =
{
    "128",
    "256",
    "512",
    "1024",
    "2048",
    "4096",
    "8192",
    "16384",
    NULL,
};
static char *max_fft_length_list[] =
{
    "2048",
    "4096",
    "8192",
    "16384",
    "16384",
    "32768",
    "65536",
    "131072",
    "262144",
    "524288",
    NULL,
};
static char *lifter_list[] =
{
    "0.5",
    "1.0",
    "1.5",
    "2.0",
    "2.5",
    "3.0",
    "4.0",
    "5.0",
    "6.0",
    NULL,
};

static swAnalysisDialog createAnalysisDialog(swConfig config)
{
    swAnalysisDialog analysis_dialog;
    spComponent tab_box;
    spComponent main_tab, attr_tab;
    char string[SP_MAX_LINE];
    
    analysis_dialog = xalloc(1, struct _swAnalysisDialog);
    memset(analysis_dialog, 0, sizeof(struct _swAnalysisDialog));
    analysis_dialog->config = config;
    
    analysis_dialog->window = spCreateDialogBox("analysisDialog",
						SppTitle, SW_ANALYSIS_DIALOG_TITLE,
						SppCallbackFunc, swPopdownAnalysisDialogCB,
						SppCallbackData, analysis_dialog,
						SppDialogBoxButtonType, SP_DB_OK_CANCEL_APPLY,
						SppPopupStyle, SP_MODELESS_POPUP,
						SppCloseStyle, SP_UNMAP_CLOSE,
						SppSpacingOn, SP_TRUE,
						NULL);
    
    /* create tab box */
    tab_box = spCreateTabBox(analysis_dialog->window, "analysisTabBox", 380,
			     NULL);
    
    /* create main tab */
    main_tab = spAddTabItem(tab_box, "analysisMainTab", -1,
			    SppTitle, SW_ANALYSIS_PARAMETER_LABEL,
			    NULL);
    
    /* create combo box to select analysis type */
    analysis_dialog->analysis_type_combo = spCreateParamField(main_tab, "analysisTypeComboBox", 60,
							      SppTitle, SW_ANALYSIS_TYPE_LABEL,
							      SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							      SppEditable, SP_FALSE,
							      SppFieldStrings, analysis_type_list,
							      SppFieldOffset, 140,
							      SppFieldSize, 160,
							      NULL);
    spSelectListIndex(analysis_dialog->analysis_type_combo, (int)config->analysis_type);

    /* create combo box to select window type */
    analysis_dialog->window_type_combo = spCreateParamField(main_tab, "windowTypeComboBox", 60,
							    SppTitle, SW_ANALYSIS_WINDOW_TYPE_LABEL,
							    SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							    SppEditable, SP_FALSE,
							    SppFieldStrings, window_type_list,
							    SppFieldOffset, 140,
							    SppFieldSize, 160,
							    NULL);
    spSelectListIndex(analysis_dialog->window_type_combo, (int)config->window_type);
    
    /* create combo box to input fft length */
    analysis_dialog->fft_length_combo = spCreateParamField(main_tab, "fftLengthComboBox", 60,
							   SppTitle, SW_ANALYSIS_FFT_LENGTH_LABEL,
							   SppDimension, "points",
							   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							   SppEditable, SP_TRUE,
							   SppFieldStrings, fft_length_list,
							   SppFieldOffset, 140,
							   SppFieldSize, 160,
							   NULL);
    sprintf(string, "%ld", config->fftl);
    spSetTextString(analysis_dialog->fft_length_combo, string);

    /* create combo box to input max fft length */
    analysis_dialog->max_fft_length_combo = spCreateParamField(main_tab, "maxFftLengthComboBox", 60,
							       SppTitle, SW_ANALYSIS_MAX_FFT_LENGTH_LABEL,
							       SppDimension, "points",
							       SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							       SppEditable, SP_TRUE,
							       SppFieldStrings, max_fft_length_list,
							       SppFieldOffset, 140,
							       SppFieldSize, 160,
							       NULL);
    sprintf(string, "%ld", config->max_fftl);
    spSetTextString(analysis_dialog->max_fft_length_combo, string);

    /* create combo box to input lifter */
    analysis_dialog->lifter_combo = spCreateParamField(main_tab, "lifterComboBox", 60,
						       SppTitle, SW_ANALYSIS_LIFTER_LABEL,
						       SppDimension, "ms",
						       SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						       SppEditable, SP_TRUE,
						       SppFieldStrings, lifter_list,
						       SppFieldOffset, 140,
						       SppFieldSize, 160,
						       NULL);
    sprintf(string, "%f", config->lifterm);
    spSetTextString(analysis_dialog->lifter_combo, string);

    /* create attribute tab */
    attr_tab = spAddTabItem(tab_box, "analysisAttributeTab", -1,
			    SppTitle, SW_ANALYSIS_ATTRIBUTE_LABEL,
			    NULL);
    

    /* create check box to set whether analysis always makes a new window */
    analysis_dialog->new_window_button = spCreateCheckBox(attr_tab, "analysisNewWindowButton",
							  SppTitle, SW_ANALYSIS_NEW_WINDOW_LABEL,
							  SppSet, analysis_dialog->config->analysis_new_window,
							  NULL);

    /* create check box for spectral analysis */
    analysis_dialog->linear_spectrum_button = spCreateCheckBox(attr_tab, "analysisLinearSpectrumButton",
					SppTitle, SW_ANALYSIS_LINEAR_SPECTRUM_LABEL,
					SppSet, analysis_dialog->config->wave_config->linear_spectrum,
					NULL);
    analysis_dialog->normalize_spectrum_button = spCreateCheckBox(attr_tab, "analysisNormalizeSpectrumButton",
					SppTitle, SW_ANALYSIS_NORMALIZE_SPECTRUM_LABEL,
					SppSet, analysis_dialog->config->wave_config->normalize_spectrum,
					NULL);

    return analysis_dialog;
}

swAnalysisDialog swCreateAnalysisDialog(swConfig config)
{
    static swAnalysisDialog sw_analysis_dialog = NULL;

    if (sw_analysis_dialog == NULL) {
	sw_analysis_dialog = createAnalysisDialog(config);
    }
    return sw_analysis_dialog;
}

void swPopupAnalysisDialogCB(spComponent component, swWindow window)
{
    swAnalysisDialog analysis_dialog;
    
    analysis_dialog = swCreateAnalysisDialog(window->config);

    /* popup dialog */
    spPopupWindow(analysis_dialog->window);

    return;
}

void swPopdownAnalysisDialogCB(spComponent component, swAnalysisDialog analysis_dialog)
{
    int index;
    char *string;
    spCallbackReason reason = SP_CR_NONE;

    if (analysis_dialog == NULL || spIsCreated(analysis_dialog->window) == SP_FALSE) return;

    reason = spGetCallbackReason(component);

    if (reason == SP_CR_OK || reason == SP_CR_CANCEL) {
	/* popdown analysis dialog */
	spPopdownWindow(analysis_dialog->window);
    }
    
    if (reason == SP_CR_OK || reason == SP_CR_APPLY) {
	if ((index = spGetSelectedListIndex(analysis_dialog->analysis_type_combo)) >= 0) {
	    analysis_dialog->config->analysis_type = index;
	    strcpy(analysis_dialog->config->analysis_type_string,
		   swGetAnalysisTypeParamLabel(analysis_dialog->config->analysis_type));
	}

	if ((index = spGetSelectedListIndex(analysis_dialog->window_type_combo)) >= 0) {
	    analysis_dialog->config->window_type = index;
	    strcpy(analysis_dialog->config->window_type_string,
		   swGetWindowTypeParamLabel(analysis_dialog->config->window_type));
	}
	
	if ((string = xspGetTextString(analysis_dialog->fft_length_combo)) != NULL) {
	    spDebug(10, "swPopdownAnalysisDialogCB", "fft length string = %s\n", string);
	    analysis_dialog->config->fftl = atol(string);
	    xfree(string);
	}
	if ((string = xspGetTextString(analysis_dialog->max_fft_length_combo)) != NULL) {
	    spDebug(10, "swPopdownAnalysisDialogCB", "max fft length string = %s\n", string);
	    analysis_dialog->config->max_fftl = atol(string);
	    xfree(string);
	}

	if ((string = xspGetTextString(analysis_dialog->lifter_combo)) != NULL) {
	    spDebug(10, "swPopdownAnalysisDialogCB", "lifter string = %s\n", string);
	    analysis_dialog->config->lifterm = atof(string);
	    xfree(string);
	}
	
	spGetToggleState(analysis_dialog->new_window_button, &analysis_dialog->config->analysis_new_window);
	spGetToggleState(analysis_dialog->linear_spectrum_button,
			 &analysis_dialog->config->wave_config->linear_spectrum);
	spGetToggleState(analysis_dialog->normalize_spectrum_button,
			 &analysis_dialog->config->wave_config->normalize_spectrum);

	swAnalysisAll(analysis_dialog->window);
    }
    
    return;
}
#endif
