#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>

#include "swWave.h"
#include "swWaveAudio.h"
#include "swWindow.h"
#include "swDraw.h"
#include "swEdit.h"
#include "swLabelDialog.h"
#include "swLabelList.h"
#include "swCursor.h"

int swSampToOverviewDisp(swWindow window, long samp)
{
    if (window == NULL || window->wave == NULL || samp < 0)
	return -1;

    if (window->wave->total_length <= 1) return 0;
    
    return (int)((double)samp * (double)window->overview_width
		 / (double)(window->wave->total_length - 1) + 0.5);
}

int swSampToDisp(swWindow window, long samp)
{
    if (window == NULL || window->wave == NULL || samp < 0)
	return -1;

    if (window->length <= 1) return 0;
    
    return (int)(((double)samp - (double)window->offset) * (double)swGetDrawWidth(window)
		 / (double)(window->length - 1) + 0.5);
}

long swDispToSamp(swWindow window, int disp)
{
    if (window == NULL || window->wave == NULL)
	return -1;
	
    return (long)((double)disp * (double)(window->length - 1) /
		  (double)swGetDrawWidth(window) + (double)window->offset + 0.5);
}

long swOverviewDispToSamp(swWindow window, int disp)
{
    if (window == NULL || window->wave == NULL)
	return -1;
	
    return (long)((double)disp * (double)(window->wave->total_length - 1) /
		  (double)window->overview_width + 0.5);
}

double swSampToDim(swWindow window, long samp)
{
    if (window == NULL || window->wave == NULL || samp < 0)
	return -1.0;

    
    if (window->data_type == SW_FREQ_DATA) {
	if (window->length <= 1) {
	    return 0.0;
	} else {
	    return ((double)samp * window->wave->samp_rate / 2.0
		    / (double)(window->wave->total_length - 1));
	}
    } else {
	return ((double)samp / (double)(window->wave->samp_rate));
    }
}

/* The return value is affected by current format setting. */
double swSampToCurrentDim(swWindow window, long samp)
{
    double dim = -1.0;
    
    if (window == NULL) return -1.0;

    if (window->data_type == SW_FREQ_DATA) {
	if (window->config->freq_format == SW_FREQ_FORMAT_KHZ) {
	    dim = (double)swSampToDim(window, samp) / 1000.0;
	} else if (window->config->freq_format == SW_FREQ_FORMAT_POINT) {
	    dim = (double)samp;
	} else {
	    dim = (double)swSampToDim(window, samp);
	}
    } else {
	if (window->config->time_format == SW_TIME_FORMAT_MSEC
	    || window->config->time_format == SW_TIME_FORMAT_FLOORED_MSEC) {
	    dim = (double)swSampToDim(window, samp) * 1000.0;
	} else if (window->config->time_format == SW_TIME_FORMAT_POINT) {
	    dim = (double)samp;
	} else {
	    dim = (double)swSampToDim(window, samp);
	}
    }

    return dim;
}

long swDimToSamp(swWindow window, double dim)
{
    if (window == NULL || window->wave == NULL || dim < 0)
	return -1;
    
    if (window->data_type == SW_FREQ_DATA) {
	return ((long)(((double)dim * (double)(window->wave->total_length - 1)
			/ (window->wave->samp_rate / 2.0)) + 0.5));
    } else {
	return ((long)((double)dim * (double)(window->wave->samp_rate) + 0.5));
    }
}

/* convert a dimension into an accurate display position */
int swDimToDisp(swWindow window, double dim)
{
    double samp;
    
    if (window == NULL || window->wave == NULL || dim < 0)
	return -1;
    
    if (window->data_type == SW_FREQ_DATA) {
	samp = (double)dim * (double)(window->wave->total_length - 1) / (window->wave->samp_rate / 2.0);
    } else {
	samp = (double)dim * (double)(window->wave->samp_rate);
    }

    return (int)((samp - (double)window->offset) * (double)swGetDrawWidth(window)
		 / (double)(window->length - 1) + 0.5);
}

int swLengthToDisp(swWindow window, long length)
{
    if (window == NULL || window->wave == NULL || length < 0)
	return -1;

    if (window->length <= 1) return 0;
    
    return (int)(((double)length) * (double)swGetDrawWidth(window)
		 / (double)(window->length - 1) + 0.5);
}

long swDispToLength(swWindow window, int disp)
{
    if (window == NULL || window->wave == NULL)
	return -1;
	
    return (long)((double)disp * (double)(window->length - 1) /
		  (double)swGetDrawWidth(window) + 0.5);
}

void swUpdatePoint(swWindow window, int point, spBool update_flag)
{
    swUpdateOverview(window, SP_TRUE, SP_FALSE, SP_FALSE);
    window->point = point;
    swUpdateOverview(window, SP_TRUE, SP_FALSE, update_flag);

    return;
}

void swSetCanvasCursor(spComponent canvas, spCursorType cursor_type)
{
    spCursor cursor = NULL;
    static spCursor wait_cursor = NULL;
    static spCursor move_cursor = NULL;
    static spCursor left_cursor = NULL;
    static spCursor right_cursor = NULL;
    static spCursor text_cursor = NULL;

    if (canvas == NULL) return;

    switch (cursor_type) {
      case SP_CURSOR_WAIT:
	if (wait_cursor == NULL) wait_cursor = spGetCursor(cursor_type);
	cursor = wait_cursor;
	break;
      case SP_CURSOR_MOVE:
	if (move_cursor == NULL) move_cursor = spGetCursor(cursor_type);
	cursor = move_cursor;
	break;
      case SP_CURSOR_W_RESIZE:
	if (left_cursor == NULL) left_cursor = spGetCursor(cursor_type);
	cursor = left_cursor;
	break;
      case SP_CURSOR_E_RESIZE:
	if (right_cursor == NULL) right_cursor = spGetCursor(cursor_type);
	cursor = right_cursor;
	break;
      case SP_CURSOR_TEXT:
	if (text_cursor == NULL) text_cursor = spGetCursor(cursor_type);
	cursor = text_cursor;
	break;
      default:
	break;
    }

    if (cursor != NULL) {
	spSetCanvasCursor(canvas, cursor);
    } else {
	spUnsetCanvasCursor(canvas);
    }

    return;
}

void swSetMouseCursor(swWindow window, spCursorType cursor_type)
{
    if (window == NULL) return;

    swSetCanvasCursor(window->canvas, cursor_type);
    
    return;
}

void swUnsetMouseCursor(swWindow window)
{
    if (window == NULL) return;
    
    spUnsetCanvasCursor(window->canvas);
    return;
}

static int swGetCursorChannel(swWindow window, int y)
{
    int i;
    int channel;
    int draw_min, draw_max;

    channel = -1;
    for (i = 0; i < window->wave->num_channel; i++) {
	swGetDrawRange(window, i, &draw_min, &draw_max);

	if (y >= draw_min && y < draw_max) {
	    channel = i;
	}
    }

    return channel;
}

static spBool swIsNearEdge(swWindow window, int x, int y, int *stdist, int *eddist)
{
    int std, edd;
    
    if (window == NULL || window->wave == NULL
	|| (window->sel_st_d < 0 && window->sel_ed_d < 0)
	|| (swIsWavePlaying(window->wave) == SP_FALSE
	    && swIsWaveProcessing(window->wave) == SP_TRUE)) {
	return SP_FALSE;
    }

    if (window->selecting == SP_FALSE
	&& window->wave->selected_channel >= 0
	&& window->wave->selected_channel != swGetCursorChannel(window, y)) {
	return SP_FALSE;
    }
    
    std = ABS(MAX(window->sel_st_d, 0) - x);
    edd = ABS(MAX(window->sel_ed_d, 0) - x);
    
    if (std <= SW_EDGE_MOTION_RANGE / 2 || edd <= SW_EDGE_MOTION_RANGE / 2) {
	if (window->sel_st_d < window->sel_ed_d) {
	    *stdist = std;
	    *eddist = edd;
	} else {
	    *stdist = edd;
	    *eddist = std;
	}
	return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool swIsInsideSelection(swWindow window, int x, int y)
{
    if (window == NULL || window->wave == NULL) return SP_FALSE;
    
    if (window->sel_st_d < x && x < window->sel_ed_d
	&& (window->wave->selected_channel == -1
	    || window->wave->selected_channel == swGetCursorChannel(window, y))) {
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool swIsNearSeparation(swWindow window, int x, int y)
{
    int i;
    int channel;
    int min, max;

    if (window == NULL || window->wave == NULL
	|| swIsWaveProcessing(window->wave) == SP_TRUE) return SP_FALSE;

    if (swIsInsideSelection(window, x, y) == SP_TRUE) {
	return SP_FALSE;
    }

    channel = -1;
    for (i = 1; i < window->wave->num_channel; i++) {
	swGetDrawRange(window, i, &min, &max);
	min -= SW_EDGE_MOTION_RANGE / 2;
	max = min + SW_EDGE_MOTION_RANGE;

	if (y >= min && y < max) {
	    return SP_TRUE;
	}
    }
    
    return SP_FALSE;
}

spBool swIsExtendKeyPressed(swWindow window)
{
    spModifierMask modmask;

    if (spGetModifierKeyMask(window->window, &modmask) == SP_TRUE
	&& (modmask & SP_EXTEND_MODIFIER_MASK)) {
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool swIsExtendRequired(swWindow window, int x, int y)
{
    if (/*window->sel_st_d >= 0
	  &&*/ (window->wave->selected_channel < 0
	    || window->wave->selected_channel == swGetCursorChannel(window, y))
	&& swIsExtendKeyPressed(window) == SP_TRUE) {
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

void swMoveAllCursor(swWindow window)
{
    long k;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL) return;

    if (swIsNoWave(window) == SP_TRUE) {
	return;
    } else {
	if (window->selecting == SP_TRUE
	    || swIsProcessing(window) == SP_FALSE) {
	    swLockWindowMutex(window);
	    window->point_d = swSampToDisp(window, window->point);
	    window->point_f = swSampToDim(window, window->point);
	
	    /* draw cursor */
	    swRefreshWindow(window, spIsFalse(swIsProcessing(window)));
	    swUnlockWindowMutex(window);
	}
    }
    
    if (swIsProcessing(window) == SP_TRUE) {
	return;
    }
    
    next = window->window;
    for (k = 0;; k++) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && next_window->wave != NULL) {
	    if (next_window->data_type == window->data_type) {
		swLockWindowMutex(next_window);
		if (swIsWaveProcessing(next_window->wave) == SP_FALSE) {
		    if (window->point_f >= 0.0) {
			next_window->point_f = window->point_f;
			swUpdatePoint(next_window,
				      swDimToSamp(next_window, next_window->point_f), SP_TRUE);
			next_window->point_d = swSampToDisp(next_window, next_window->point);
		    }
		    /* draw cursor */
		    swDrawCursor(next_window);
		} else {
		    next_window->prev_point_f = window->point_f;
		}
		swUnlockWindowMutex(next_window);
	    }
	}
    }
    
    return;
}

spBool swMoveCursor(swWindow window, long point, spBool update_flag)
{
    if (swIsNoWave(window) == SP_TRUE
	|| point < 0 || point >= window->wave->total_length) {
	return SP_FALSE;
    }
    
    if (window->selecting == SP_TRUE || swIsProcessing(window) == SP_FALSE) {
	swLockWindowMutex(window);
	swUpdatePoint(window, point, update_flag);
	swUnlockWindowMutex(window);
    } else if (update_flag == SP_TRUE && window->overview_canvas != NULL) {
	spRefreshCanvas(window->overview_canvas);
    }

    swMoveAllCursor(window);
    
    return SP_TRUE;
}

void swMoveCursorCB(spComponent component, swWindow window)
{
    int x, y;
    int stdist, eddist;
    int point_d;
    long point;
    double point_f;
    static spBool cursor_set = SP_FALSE;

    spDebug(50, "swMoveCursorCB", "in\n");
    
    if (swIsNoWave(window) == SP_TRUE) return;

    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	x = MIN(x, swGetDrawWidth(window));
	x = MAX(x, 0);
	
	window->target_channel = MAX(swGetCursorChannel(window, y), 0);
	    
	/* calculate new cursor position */
	point = swDispToSamp(window, x);
	point_d = swSampToDisp(window, point);
	point_f = swSampToDim(window, point);

	if (window->selecting == SP_FALSE && window->drag_region == SP_FALSE
	    && window->drag_label_type == SW_DRAG_NO_LABEL) {
	    if (swIsExtendRequired(window, x, y) == SP_TRUE) {
		if (point_d < (window->sel_st_d + window->sel_ed_d) / 2) {
		    swSetMouseCursor(window, SP_CURSOR_W_RESIZE);
		} else {
		    swSetMouseCursor(window, SP_CURSOR_E_RESIZE);
		}
	    } else if (window->active_label_index >= 0
		       && window->active_label_index == swFindNearLabelIndex(window, point_f, SP_FALSE, NULL)) {
		swSetMouseCursor(window, SP_CURSOR_MOVE);
		cursor_set = SP_TRUE;
	    } else if (swIsNearEdge(window, point_d, y, &stdist, &eddist) == SP_TRUE) {
		if (stdist < eddist) {
		    swSetMouseCursor(window, SP_CURSOR_W_RESIZE);
		} else {
		    swSetMouseCursor(window, SP_CURSOR_E_RESIZE);
		}
		cursor_set = SP_TRUE;
	    } else if (swIsNearSeparation(window, point_d, y) == SP_TRUE) {
		swSetMouseCursor(window, SP_CURSOR_TEXT);
		cursor_set = SP_TRUE;
	    } else if (cursor_set == SP_TRUE) {
		swUnsetMouseCursor(window);
		cursor_set = SP_FALSE;
	    }
	}

	if (swIsWaveProcessing(window->wave) == SP_FALSE) {
	    swLockWindowMutex(window);
	    swUpdatePoint(window, point, SP_TRUE);
	    window->point_d = point_d;
	    swUnlockWindowMutex(window);
	}
	    
	/* move all cursor */
	swMoveAllCursor(window);
    }
    
    return;
}

spBool swPlayRegion(swWindow window, long st, long ed)
{
    long offset;
    long length;
    spBool flag;

    if (swIsNoWave(window) == SP_TRUE
	|| window->data_type == SW_FREQ_DATA
	|| swIsProcessing(window) == SP_TRUE
	|| swIsWavePlayable(window->wave) == SP_FALSE
	|| window->num_blocked > 0) return SP_FALSE;
    
    spDebug(30, "swPlayRegion", "in\n");

    swLockWindowMutex(window);
    
    /* get edge */
    if (swGetEdge(window, st, ed, &offset, &length) != SP_TRUE) {
	offset = 0;
	length = window->wave->total_length;
	window->wave->selected_channel = -1;
    }

    spDebug(30, "swPlayRegion", "offset = %ld, length = %ld\n", offset, length);

    window->prev_point_f = window->point_f;
    window->wave->play_offset = offset;
    window->wave->play_length = length;
    swUnlockWindowMutex(window);
    
    flag = swPlayWave(window->wave, window->loop_play);

    if (flag == SP_FALSE) {
	spDisplayError(window->window, NULL, SW_PLAY_ERROR_MESSAGE);
    }
    
    return flag;
}

spBool swPlayStop(swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) return SP_FALSE;
    
    if (swIsWavePlaying(window->wave) == SP_TRUE) {
	swProcessStop(window->wave);
	return SP_TRUE;
    }
    return SP_FALSE;
}

void swPlayRegionCB(spComponent component, swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) return;
    
    swPlayRegion(window, window->sel_st, window->sel_ed);
    
    return;
}

void swPlayWindowCB(spComponent component, swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) return;
    
    swPlayRegion(window, window->offset, window->offset + window->length - 1);
    
    return;
}

void swPlayFileCB(spComponent component, swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) return;
    
    swPlayRegion(window, -1, -1);
    
    return;
}

void swPlayStopCB(spComponent component, swWindow window)
{
    spDialogResponse response = SP_DR_YES;
    
    if (swIsNoWave(window) == SP_FALSE) {
	spDebug(30, "swPlayStopCB", "in\n");
	
	swLockMutex(window->wave);

	if (swIsProcessing(window) == SP_TRUE
	    && swIsWaveProcessFinished(window->wave) == SP_FALSE) {
	    if (swIsWavePlaying(window->wave) == SP_FALSE) {
		response = spCreateMessageBox(window->window, NULL,
					      SW_STOP_EDIT_QUESTION_MESSAGE,
					      SppDialogType, SP_QUESTION_DIALOG,
					      SppMessageBoxButtonType, SP_MB_YES_NO,
					      NULL);
	    }
	
	    if (response == SP_DR_YES) {
		swProcessStop(window->wave);
	    }
	}
	
	swUnlockMutex(window->wave);
	
	spDebug(30, "swPlayStopCB", "done\n");
    }
    
    return;
}

void swLButtonPress(swWindow window, int x, int y, int *px, int *py)
{
    spBool flag;
    spBool extend_flag;
    long index;
    int stdist = -1, eddist = -1;

    spDebug(10, "swLButtonPress", "active label: %ld\n", window->active_label_index);
    
    if (window->active_label_index >= 0
	&& window->active_label_index == swFindNearLabelIndex(window, window->point_f, SP_FALSE, &flag)) {
	swSetMouseCursor(window, SP_CURSOR_MOVE);
	if (flag == SP_TRUE) {
	    window->drag_label_type = SW_DRAG_END_LABEL;
	} else {
	    window->drag_label_type = SW_DRAG_START_LABEL;
	}
	*px = x;
	*py = y;
    } else if ((extend_flag = swIsExtendRequired(window, x, y)) == SP_TRUE
	       || swIsNearEdge(window, x, y, &stdist, &eddist) == SP_TRUE) {
	if (swIsWavePlaying(window->wave) == SP_TRUE || swIsWaveProcessing(window->wave) == SP_FALSE) {
	    swLockWindowMutex(window);

	    if (extend_flag == SP_TRUE) {
		spDebug(10, "swLButtonPress", "extend: x = %d, sel_st_d = %d, sel_ed_d = %d\n",
			x, window->sel_st_d, window->sel_ed_d);
		if (x < (window->sel_st_d + window->sel_ed_d) / 2) {
		    swReverseRegion(window, window->wave->selected_channel,
				    x, window->sel_st_d);
		    swDrawCursor(window);
			
		    window->sel_st = swDispToSamp(window, x);
		    window->sel_st_d = swSampToDisp(window, window->sel_st);
		    stdist = 0; eddist = 1;
		} else {
		    swReverseRegion(window, window->wave->selected_channel,
				    window->sel_ed_d, x);
		    swDrawCursor(window);
		    
		    window->sel_ed = swDispToSamp(window, x);
		    window->sel_ed_d = swSampToDisp(window, window->sel_ed);
		    stdist = 1; eddist = 0;
		}
		spRefreshCanvas(window->overview_canvas);
	    }
	    spDebug(10, "swLButtonPress", "stdist = %d, eddist = %d\n", stdist, eddist);

	    if (stdist >= 0) {
		if (stdist < eddist) {
		    *px = window->sel_st_d;
		    window->sel_st_d = window->sel_ed_d;
		    window->sel_st = window->sel_ed;
		} else {
		    *px = window->sel_ed_d;
		}
		*py = y;
		window->selecting = SP_TRUE;
	    }
	    
	    swUnlockWindowMutex(window);
	}
    } else if ((index = swFindNearLabelIndex(window, window->point_f, SP_FALSE, NULL)) >= 0) {
	/* activate label */
	window->active_label_index = index;
	spDebug(10, "swLButtonPress", "label %ld is activated\n", index);
	swRedrawLabels(window);
	swSelectLabelList(window->label_list);
	*px = x;
	*py = y;
    } else if (swIsInsideSelection(window, x, y) == SP_TRUE) {
	spDebug(10, "swLButtonPress", "set canvas cursor: x = %d\n", x);
	
	if (swIsProcessing(window) == SP_FALSE && window->num_blocked <= 0) {
	    swSetMouseCursor(window, SP_CURSOR_MOVE);
	    window->drag_region = SP_TRUE;
	}
	
	*px = x;
	*py = -1;
    } else {
	if (swIsWavePlaying(window->wave) == SP_TRUE || swIsWaveProcessing(window->wave) == SP_FALSE) {
	    swLockWindowMutex(window);
	    
	    if (window->selecting == SP_TRUE) { /* could't catch button release */
		window->sel_st_d = -1;
		window->sel_ed_d = -1;
		window->sel_st = -1;
		window->sel_ed = -1;
		swRedrawLabels(window);
	    } else {
		swClearRegion(window);
	    }
	    window->selecting = SP_TRUE;
	    
	    window->sel_st = swDispToSamp(window, x);
	    spDebug(50, "swLButtonPress", "press: sel_st = %ld\n", window->sel_st);
	    window->sel_st_d = swSampToDisp(window, window->sel_st);

	    if (swIsWaveProcessing(window->wave) == SP_FALSE) {
		if (window->wave->num_channel > 1
		    && swIsNearSeparation(window, window->sel_st_d, y) == SP_FALSE) {
		    window->wave->selected_channel = swGetCursorChannel(window, y);
		} else {
		    window->wave->selected_channel = -1;
		}
	    }
	    
	    *px = window->sel_st_d;
	    *py = y;
	    spDebug(50, "swLButtonPress", "press: px = %d, py = %d\n", *px, *py);
	    spDebug(50, "swLButtonPress", "press: sel_st_d = %d\n", window->sel_st_d);
	    
	    swUnlockWindowMutex(window);
	}
    }
    spDebug(10, "swLButtonPress", "done\n");

    return;
}

void swLButtonRelease(swWindow window, int ox, int oy, int px, int py)
{
    if (py < 0) {
	if (swIsProcessing(window) == SP_FALSE && window->drag_region == SP_TRUE
	    && window->num_blocked <= 0) {
	    spDebug(10, "swLButtonRelease", "ox = %d, oy = %d\n", ox, oy);
	    swUnsetMouseCursor(window);
	    if (ox < 0 || oy < 0 || ox > window->width || oy > window->height) {
		int wx, wy;

		spGetWindowPosition(window->window, &wx, &wy);
		wx += ox; wy += oy;
		wx = MAX(wx - window->config->width / 2, 0);
		wy = MAX(wy - window->config->height / 2, 0);
	    
		if (window->config->autosave_by_drop == SP_TRUE) {
		    swExtractAutosaveWindowAt(window, wx, wy);
		} else {
		    swExtractWindowAt(window, window->sel_st, window->sel_ed, SP_FALSE, wx, wy);
		}
	    } else {
		if (px == ox) {
		    swSelectRegion(window, window->wave->selected_channel, -1, -1);
		}
	    }
	} else {
	    if ((swIsWavePlaying(window->wave) == SP_TRUE
		 || swIsWaveProcessing(window->wave) == SP_FALSE)
		&& !(ox < 0 || oy < 0 || ox > window->width || oy > window->height)
		&& px == ox) {
		swSelectRegion(window, window->wave->selected_channel, -1, -1);
		swSetPlayRegion(window->wave, 0, 0);
	    }
	}
	window->drag_region = SP_FALSE;
    } else if (window->drag_label_type != SW_DRAG_NO_LABEL) {
	swUnsetMouseCursor(window);

	if (px == ox) {
	    long index;
	    double time;
	    spBool start_flag, end_flag;

	    /* get end time of active label */
	    if (window->drag_label_type == SW_DRAG_END_LABEL
		&& window->active_label_index >= 0
		&& window->wave->labels->label[window->active_label_index].end_time >= 0.0) {
		time = window->wave->labels->label[window->active_label_index].end_time;
	    } else {
		time = -1.0;
	    }

	    spDebug(10, "swLButtonRelease", "px == ox: time = %f\n", time);
	    
	    index = -1;
	    window->drag_label_type = SW_DRAG_NO_LABEL;
	
	    if (time >= 0.0
		&& swFindIdenticalLabelIndex(window, time, SP_TRUE, &index,
					     &start_flag, &end_flag) == SP_TRUE
		&& index != window->active_label_index && start_flag == SP_TRUE) {
		/* if region label exists, select it. */
		spDebug(10, "swLButtonRelease", "found identical: %ld\n", index);
		window->active_label_index = index;
		swRedrawLabels(window);
		swSelectLabelList(window->label_list);
	    } else {
		/* deactivate label */
		spDebug(10, "swLButtonRelease", "label %ld is deactivated\n", window->active_label_index);
		window->active_label_index = -1;
		swRedrawLabels(window);
	    }
	} else {
	    window->drag_label_type = SW_DRAG_NO_LABEL;
	    swUpdateLabels(window);
	}
    } else if (window->selecting == SP_TRUE) {
	if (swIsWavePlaying(window->wave) == SP_TRUE
	    || swIsWaveProcessing(window->wave) == SP_FALSE) {
	    swLockWindowMutex(window);

	    spDebug(50, "swLButtonRelease", "selection finished\n");
	    
	    if (px == window->sel_st_d) {
		swResetRegion(window);
		swSetPlayRegion(window->wave, 0, 0);
	    } else {
		if (px < window->sel_st_d) {
		    window->sel_ed_d = window->sel_st_d;
		    window->sel_ed = window->sel_st;
		    window->sel_st_d = px;
		    window->sel_st = swDispToSamp(window, window->sel_st_d);
		} else {
		    window->sel_ed_d = px;
		    window->sel_ed = swDispToSamp(window, window->sel_ed_d);
		}
		spDebug(50, "swLButtonRelease",
			"release: sel_st = %d, sel_ed = %d\n", window->sel_st_d, 
			window->sel_ed_d);

		if (swIsWaveProcessing(window->wave) == SP_FALSE) {
		    if (window->wave->selected_channel != swGetCursorChannel(window, oy)) {
			window->wave->selected_channel = -1;
		    }
		}
		swSetSelectSenseLevel(window, SP_TRUE);
		swSetPlayRegion(window->wave, window->sel_st, window->sel_ed - window->sel_st + 1);
	    }

	    spRefreshCanvas(window->overview_canvas);
	    
	    window->selecting = SP_FALSE;
	    swUnlockWindowMutex(window);
	}
    }
	
    return;
}

void swLButtonMotion(swWindow window, int x, int y, int *px, int *py)
{
    long x_s;
    double x_f;

    if (window->selecting == SP_TRUE || window->drag_label_type != SW_DRAG_NO_LABEL) {
	swLockWindowMutex(window);
	x_s = swDispToSamp(window, x);
	if (x_s != swDispToSamp(window, *px)) {
	    x = swSampToDisp(window, x_s);
	    
	    spDebug(70, "swLButtonMotion",  "(%d, %d)\n", *px, x);

	    if (window->drag_label_type != SW_DRAG_NO_LABEL) {
		if (window->wave->labels != NULL && window->active_label_index >= 0) {
		    x_f = swSampToDim(window, x_s);
		    if (window->drag_label_type == SW_DRAG_END_LABEL
			&& window->wave->labels->label[window->active_label_index].end_time >= 0.0) {
			/* dragging end label */
			spDebug(100, "swLButtonMotion",  "label %ld: %f --> %f\n",
				window->active_label_index,
				window->wave->labels->label[window->active_label_index].end_time, x_f);

			swReplaceIdenticalLabelIndex(window,
					window->wave->labels->label[window->active_label_index].end_time, x_f, SP_TRUE);

			if (x_f < window->wave->labels->label[window->active_label_index].time) {
			    /* swap end for start */
			    window->wave->labels->label[window->active_label_index].end_time
				= window->wave->labels->label[window->active_label_index].time;
			    window->wave->labels->label[window->active_label_index].time = x_f;
			    
			    /* dragging label is start label */
			    window->drag_label_type = SW_DRAG_START_LABEL;
			} else {
			    window->wave->labels->label[window->active_label_index].end_time = x_f;
			}
		    } else {
			/* dragging start label */
			spDebug(100, "swLButtonMotion",  "label %ld: %f --> %f\n",
				window->active_label_index,
				window->wave->labels->label[window->active_label_index].time, x_f);

			swReplaceIdenticalLabelIndex(window,
					window->wave->labels->label[window->active_label_index].time, x_f,
					(window->wave->labels->label[window->active_label_index].end_time >= 0.0
					 ? SP_TRUE: SP_FALSE));
			
			if (window->wave->labels->label[window->active_label_index].end_time >= 0.0
			    && x_f > window->wave->labels->label[window->active_label_index].end_time) {
			    /* swap start for end */
			    window->wave->labels->label[window->active_label_index].time
				= window->wave->labels->label[window->active_label_index].end_time;
			    window->wave->labels->label[window->active_label_index].end_time = x_f;
			    
			    /* dragging label is end label */
			    window->drag_label_type = SW_DRAG_END_LABEL;
			} else {
			    window->wave->labels->label[window->active_label_index].time = x_f;
			}
		    }
		    swRedrawLabels(window);
		}
	    } else {
		if (swIsWaveProcessing(window->wave) == SP_TRUE) {
		    swReverseRegion(window, window->wave->selected_channel, *px, x);
		} else {
		    if (window->wave->selected_channel >= 0
			&& window->wave->selected_channel != swGetCursorChannel(window, y)) {
			if (window->wave->selected_channel != swGetCursorChannel(window, *py)) {
			    /* multiple channel --> multiple channel */
			    swReverseRegion(window, -1, *px, x);
			} else {
			    /* single channel --> multiple channel */
			    swReverseRegion(window, window->wave->selected_channel,
					    window->sel_st_d, *px);
			    swReverseRegion(window, -1, window->sel_st_d, x);
			}
		    } else {
			if (window->wave->selected_channel >= 0
			    && window->wave->selected_channel != swGetCursorChannel(window, *py)) {
			    /* multiple channel --> single channel */
			    swReverseRegion(window, -1, window->sel_st_d, *px);
			    swReverseRegion(window, window->wave->selected_channel,
					    window->sel_st_d, x);
			} else {
			    /* single channel --> single channel */
			    /* or multiple channel --> multiple channel */
			    swReverseRegion(window, window->wave->selected_channel, *px, x);
			}
		    }
		}
		window->sel_ed_d = x;
		window->sel_ed = swDispToSamp(window, window->sel_ed_d);
		*px = x;
		*py = y;
	    }
	}
	swUnlockWindowMutex(window);
    }
    
    return;
}

void swLButtonScroll(swWindow window, int ox, int oy, int *px, int *py)
{
    long value;
    long offset;

    if (swIsProcessing(window) == SP_TRUE
	|| window->num_blocked > 0
	|| (swIsWavePeakAvailable(window->wave) == SP_FALSE
	    && window->length > window->config->wave_config->read_callback_length)) return;
	
    if (*py >= 0) {
	swLockWindowMutex(window);
	
	value = MAX(window->length / 16, 1);
	if (ox < 0) {
	    value *= -1;
	}

	offset = window->offset + value;
	offset = MIN(offset, window->wave->total_length - window->length);
	offset = MAX(offset, 0);
	value = offset - window->offset;

	if (window->selecting == SP_TRUE) {
	    if (window->wave->selected_channel >= 0
		&& window->wave->selected_channel != swGetCursorChannel(window, *py)) {
		window->wave->selected_channel = -1;
	    }
	    swReverseOverviewRegion(window, window->wave->selected_channel,
				    swSampToOverviewDisp(window, window->sel_st),
				    swSampToOverviewDisp(window, window->sel_ed));
	
	    window->sel_ed = swDispToSamp(window, *px) + value;
	}
	
	spDebug(60, "swLButtonScroll", "(%ld, %ld)\n", window->sel_st, window->sel_ed);
	
	swUnlockWindowMutex(window);
	
	swScrollWindow(window, offset, SP_FALSE, SP_FALSE);
	swReverseOverviewRegion(window, window->wave->selected_channel,
				swSampToOverviewDisp(window, window->sel_st),
				swSampToOverviewDisp(window, window->sel_ed));
    }
    
    return;
}

void swMButtonPress(swWindow window)
{
    spBool flag = SP_FALSE;
    
    swLockMutex(window->wave);
    if (swIsProcessing(window) == SP_TRUE) {
	if (swIsWavePlaying(window->wave) == SP_TRUE
	    && swIsWaveProcessFinished(window->wave) == SP_FALSE) {
	    swProcessStop(window->wave);
	}
	flag = SP_TRUE;
    }
    swUnlockMutex(window->wave);
    
    if (flag == SP_FALSE && swIsPlayable(window) == SP_TRUE) {
	spDebug(50, "swMButtonPress", "mbutton press: sel_st = %d sel_ed = %d\n",
		window->sel_st_d, window->sel_ed_d);
	
	swPlayRegion(window, window->sel_st, window->sel_ed);
    }
    
    return;
}

void swButtonMotionCB(spComponent component, swWindow window)
{
    int ox, oy;
    int x, y;
    int window_width;
    spCallbackReason reason;
    static int prev_x = 0, prev_y = 0;
    
    if (swIsNoWave(window) == SP_TRUE) return;

    if (spGetCallbackMousePosition(component, &ox, &oy) == SP_FALSE) {
	return;
    }

    spDebug(30, "swButtonMotionCB", "x = %d, y = %d\n", ox, oy);
    
    x = MIN(ox, swGetDrawWidth(window));
    y = MIN(oy, window->height);
    x = MAX(ox, 0);
    y = MAX(oy, 0);

    reason = spGetCallbackReason(component);

    if (swIsWavePlaying(window->wave) == SP_TRUE && window->sync_play == SP_TRUE
	&& reason != SP_CR_MBUTTON_PRESS) {
        return;
    }
    
    switch (reason) {
      case SP_CR_LBUTTON_PRESS:
	swLButtonPress(window, x, y, &prev_x, &prev_y);
	break;
      case SP_CR_LBUTTON_RELEASE:
	swLButtonRelease(window, ox, oy, prev_x, prev_y);
	break;
      case SP_CR_LBUTTON_MOTION:
	window_width = swGetDrawWidth(window);
	if (ox > window_width || ox < 0) {
	    swLButtonScroll(window, ox, oy, &prev_x, &prev_y);
	} else {
	    swLButtonMotion(window, x, y, &prev_x, &prev_y);
	}
	break;
      case SP_CR_MBUTTON_PRESS:
	swMButtonPress(window);
	break;
      case SP_CR_RBUTTON_PRESS:
	if (swGetNumLabel(window->wave) > 0) {
	    long index;
	    double point_f;

	    point_f = swSampToDim(window, window->point);
	    spDebug(10, "swButtonMotionCB", "right button pressed: point = %f\n", point_f);

	    if ((index = swFindNearLabelIndex(window, point_f, SP_FALSE, NULL)) >= 0) {
		if (window->wave->labels->label[index].end_time >= 0.0) {
		    spSetWindowSenseLevel(window->window, SW_REGION_LABEL_GROUP_ID,
					  SW_STATE_EXIST_LABEL_HERE);
		}
		spSetWindowSenseLevel(window->window, SW_LABEL_GROUP_ID,
				      SW_STATE_EXIST_LABEL_HERE);
	    } else {
		spSetWindowSenseLevel(window->window, SW_REGION_LABEL_GROUP_ID,
				      SW_STATE_EXIST_LABEL);
		spSetWindowSenseLevel(window->window, SW_LABEL_GROUP_ID,
				      SW_STATE_EXIST_LABEL);
	    }
	}
	break;
      default:
	break;
    }

    return;
}

void swKeyPressCB(spComponent component, swWindow window)
{
    spKeySym key_sym;

    if (window == NULL || window->wave == NULL) return;
    
    spDebug(50, "swKeyPressCB", "in\n");
    
    if (spGetCallbackKeySym(component, &key_sym) == SP_TRUE) {
	spDebug(50, "swKeyPressCB", "%ld\n", key_sym);
	
	if (key_sym == SPK_Character) {
	    int len;
	    char buf[SP_MAX_LINE];
	    spBool overflow;
	    
	    if ((len = spGetCallbackKeyString(component, buf, sizeof(buf), &overflow)) >= 0) {
		spDebug(50, "swKeyPressCB", "len = %d, buf = %s\n", len, buf);
		if (streq(buf, " ")) {
		    if (swPlayStop(window) == SP_FALSE) {
			swPlayRegion(window, window->sel_st, window->sel_ed);
		    }
		}
	    }
	} else {
	    long delta = 0;
	    
	    switch (key_sym) {
	      case SPK_Prior:
		delta = MAX(window->length / 4, 1);
		break;
	      case SPK_Next:
		delta = -MAX(window->length / 4, 1);
		break;
	
	      case SPK_Left:
		delta = -MAX(window->length / 8, 1);
		break;
	      case SPK_Right:
		delta = MAX(window->length / 8, 1);
		break;
	    
	      case SPK_Home:
		delta = -window->offset;
		break;
	      case SPK_End:
		delta = -window->offset + window->wave->total_length - window->length;
		break;

	      default:
		return;
	    }
	    
	    if (swIsExtendKeyPressed(window) == SP_TRUE) {
		int channel = -1;
		long sel_st, sel_ed;
		long point;
	    
		if (key_sym == SPK_Home) {
		    if (window->sel_st >= 0) {
			channel = window->wave->selected_channel;
			sel_ed = window->sel_ed;
		    } else {
			sel_ed = window->point;
		    }
		    sel_st = 0;
		    point = window->point;
		} else if (key_sym == SPK_End) {
		    if (window->sel_st >= 0) {
			channel = window->wave->selected_channel;
			sel_st = window->sel_st;
		    } else {
			sel_st = window->point;
		    }
		    sel_ed = window->wave->total_length - 1;
		    point = window->point;
		} else if (window->sel_st >= 0) {
		    sel_st = window->sel_st;
		    sel_ed = window->sel_ed;
		    
		    if (window->point < (window->sel_st + window->sel_ed) / 2) {
			sel_st += delta;
			point = sel_st;
		    } else {
			sel_ed += delta;
			point = sel_ed;
		    }
		    channel = window->wave->selected_channel;
		} else {
		    sel_st = sel_ed = window->point;
		    
		    if (delta < 0) {
			sel_st += delta;
			point = sel_st;
		    } else {
			sel_ed += delta;
			point = sel_ed;
		    }
		}
		swSelectRegion(window, channel, sel_st, sel_ed);
		swUpdatePoint(window, point, SP_TRUE);
		swMoveAllCursor(window);
	    } else {
		swScrollWindow(window, window->offset + delta, SP_TRUE, SP_FALSE);
	    }
	}
    }
    
    return;
}
