/*
 *	swDialog.c
 *
 *	Last modified: <2002-03-16 19:54:33 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spAudioLib.h>
#include <sp/spComponentLib.h>

#include <sp/sp.h>

#include "swWindow.h"
#include "swDraw.h"
#include "swEdit.h"
#include "swCursor.h"
#include "swLabel.h"
#include "swDialog.h"

static char *format_list_strings[] =
{
    SP_WAVE_FORMAT_RAW_LABEL,
    SP_WAVE_FORMAT_SWAP_LABEL,
    SP_WAVE_FORMAT_LITTLE_LABEL,
    SP_WAVE_FORMAT_BIG_LABEL,
    SP_WAVE_FORMAT_ULAW_LABEL,
    SP_WAVE_FORMAT_ALAW_LABEL,
    SP_WAVE_FORMAT_TEXT_LABEL,
    SP_WAVE_FORMAT_TEXT_TIME_LABEL,
    SP_WAVE_FORMAT_TEXT_FREQ_LABEL,
    NULL,
};

static char *format_list_short_strings[] =
{
    "raw",
    "swap",
    "little",
    "big",
    "ulaw",
    "alaw",
    "text",
    "time",
    "freq",
    NULL,
};

static char *bit_list_strings[] =
{
    "8",
    "16",
    "24",
    "32",
    SW_FLOAT_BIT_STRING,
    "64",
    NULL,
};

static char *samp_rate_list_strings[] =
{
    "8000",
    "10000",
    "11025",
    "12000",
    "16000",
    "22050",
    "24000",
    "32000",
    "44100",
    "48000",
    "96000",
    NULL,
};

static char *channel_list_strings[] =
{
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "8",
    "10",
    "12",
    "16",
    "24",
    "32",
    NULL,
};

int swGetSampleBit(char *string)
{
    if (string == NULL) return 0;
    
    if (streq(string, SW_FLOAT_BIT_STRING)) {
	return 33;
    } else {
	return atoi(string);
    }
}

spBool swGetSampleBitString(char *string, int samp_bit)
{
    if (string == NULL) return SP_FALSE;
    
    if (samp_bit == 33) {
	strcpy(string, SW_FLOAT_BIT_STRING);
    } else {
	sprintf(string, "%d", samp_bit);
    }
    
    return SP_TRUE;
}

#ifdef SW_USE_PREFERENCE_DIALOG

static swLook sw_looks[] = {
    {SW_MY_LOOK_INDEX, SW_MY_LOOK_LABEL, "", "", "", "", "", ""},
    {SW_ORIGINAL_LOOK_INDEX, SW_ORIGINAL_LOOK_LABEL,
	 "black", "white", "red", "blue", "red", "grey50"},
    {SW_OSCILLO_LOOK_INDEX, SW_OSCILLO_LOOK_LABEL,
	 "aquamarine", "black", "orange", "white", "orange", "grey60"}, 
    {SW_MONOTONE_LOOK_INDEX, SW_MONOTONE_LOOK_LABEL,
	 "white", "black", "grey70", "grey80", "grey70", "grey60"},
    {SW_SKY_LOOK_INDEX, SW_SKY_LOOK_LABEL,
	 "white", "MediumBlue", "yellow", "SkyBlue", "yellow", "LightBlue1"},
    {SW_FOREST_LOOK_INDEX, SW_FOREST_LOOK_LABEL,
	 "DarkGreen", "LimeGreen", "yellow", "green", "yellow", "GreenYellow"},
    {SW_TOMATO_LOOK_INDEX, SW_TOMATO_LOOK_LABEL,
	 "white", "red", "green3", "orange", "green3", "pink"},
    {SW_MATLAB_LOOK_INDEX, SW_MATLAB_LOOK_LABEL,
	 "yellow", "black", "red", "blue", "red", "white"},
    {SW_WAVES_LOOK_INDEX, SW_WAVES_LOOK_LABEL,
	 "black", "#00afaf", "yellow", "white", "#9400d3", "#0000d2"},
};
static int sw_num_looks = 0;

#ifdef SW_USE_TOOL_BAR
void swShowAllToolBar(swWindow window, spBool flag)
{
    spComponent next = NULL;
    swWindow next_window = NULL;
    
    if (window == NULL) return;

    spDebug(10, "swShowAllToolBar", "in\n");

    if (window->tool_bar != NULL) {
	if (flag == SP_TRUE) {
	    spMapComponent(window->tool_bar);
	} else {
	    spUnmapComponent(window->tool_bar);
	}
    }
    
    next = window->window;
    while (1) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}

	if ((next_window = (swWindow)spGetUserData(next)) != NULL) {
	    if (next_window->tool_bar != NULL) {
		if (flag == SP_TRUE) {
		    spMapComponent(next_window->tool_bar);
		} else {
		    spUnmapComponent(next_window->tool_bar);
		}
	    }
	}
    }

    return;
}
#endif

void swShowAllToolBarCB(spComponent component, swWindow window)
{
#ifdef SW_USE_TOOL_BAR
    if (window->config->no_tool_bar == SP_FALSE) {
	swShowAllToolBar(window, SP_FALSE);
	window->config->no_tool_bar = SP_TRUE;
    } else {
	swShowAllToolBar(window, SP_TRUE);
	window->config->no_tool_bar = SP_FALSE;
    }
#endif
    
    return;
}

void swPopdownPreferenceDialogCB(spComponent component, swPrefDialog pref_dialog)
{
    int i;
    spBool flag = SP_FALSE;
    spBool use_tool_bar;
    int index;
    char *string;
    spCallbackReason reason = SP_CR_NONE;

    if (pref_dialog == NULL || spIsCreated(pref_dialog->window) == SP_FALSE) return;

    reason = spGetCallbackReason(component);

    if (reason == SP_CR_OK || reason == SP_CR_CANCEL) {
	/* popdown preference dialog */
	spPopdownWindow(pref_dialog->window);
    }
    
    if (reason == SP_CR_OK || reason == SP_CR_APPLY) {
	spGetToggleState(pref_dialog->default_format_button, &pref_dialog->config->use_def_format);
	
	if ((string = xspGetTextString(pref_dialog->format_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "format string = %s\n", string);
	    spStrCopy(pref_dialog->config->def_format_string, SP_MAX_SETUP_VALUE, string);
	    xfree(string);
	}

	if ((string = xspGetTextString(pref_dialog->bit_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "bit string = %s\n", string);
	    pref_dialog->config->def_samp_bit = swGetSampleBit(string);
	    xfree(string);
	}
	
	if ((string = xspGetTextString(pref_dialog->samp_rate_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "samp_rate string = %s\n", string);
	    pref_dialog->config->def_samp_rate = atof(string);
	    xfree(string);
	}
	
	if ((string = xspGetTextString(pref_dialog->channel_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "channel string = %s\n", string);
	    pref_dialog->config->def_num_channel = atoi(string);
	    xfree(string);
	}

	if (spGetToggleState(pref_dialog->use_play_command_button,
			     &pref_dialog->config->use_play_command) == SP_TRUE) {
	    pref_dialog->config->wave_config->play_use_audio = 
		(pref_dialog->config->use_play_command == SP_TRUE ? SP_FALSE : SP_TRUE);
	}
	
	if ((string = xspGetTextString(pref_dialog->play_command_text)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "play command string = %s\n", string);
	    spStrCopy(pref_dialog->config->play_command, SP_MAX_SETUP_VALUE, string);
	    spSetPlayCommand(pref_dialog->config->play_command);
	    xfree(string);
	}

#ifdef SP_SUPPORT_AUDIO
	if ((string = xspGetTextString(pref_dialog->buffer_size_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "buffer size = %s\n", string);
	    pref_dialog->config->audio_buffer_size = atoi(string);
	    pref_dialog->config->wave_config->audio_buffer_size = pref_dialog->config->audio_buffer_size;
	    xfree(string);
	}
	
	spGetToggleState(pref_dialog->use_loop_button, &pref_dialog->config->use_loop_play);
	spGetToggleState(pref_dialog->use_sync_button, &pref_dialog->config->use_sync_play);
	spGetToggleState(pref_dialog->play_draw_button, &pref_dialog->config->play_draw);
#endif
	
#ifdef SW_USE_THREAD
	if (pref_dialog->config->wave_config->thread_safe == SP_TRUE) {
	    spGetToggleState(pref_dialog->use_thread_button,
			     &pref_dialog->config->wave_config->process_use_thread);
	    
	    if (spIsThreadLevelVariable() == SP_TRUE) {
		spGetToggleState(pref_dialog->use_lowlevel_thread_button,
				 &pref_dialog->config->use_lowlevel_thread);
	    }
	}
#endif

	spGetToggleState(pref_dialog->display_info_area_button,
			 &pref_dialog->config->display_info_area);
	spGetToggleState(pref_dialog->display_freq_info_area_button,
			 &pref_dialog->config->display_freq_info_area);
	spGetToggleState(pref_dialog->link_region_label_button,
			 &pref_dialog->config->link_region_label);
	spGetToggleState(pref_dialog->erase_label_prompt_button,
			 &pref_dialog->config->erase_label_prompt);
	
#ifdef SW_SUPPORT_AUTOSAVE
	if ((string = xspGetTextString(pref_dialog->autosave_dir_text)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "autosave_dir = %s\n", string);
	    spStrCopy(pref_dialog->config->autosave_dir, SP_MAX_SETUP_VALUE, string);
	    xfree(string);
	}
	if ((index = spGetSelectedListIndex(pref_dialog->autosave_id_type_combo)) >= 0) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "autosave ID = %d\n", index);
	    pref_dialog->config->autosave_id_type = index;
	}
	if ((string = xspGetTextString(pref_dialog->autosave_id_prefix_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "autosave_id_prefix = %s\n", string);
	    spStrCopy(pref_dialog->config->autosave_id_prefix, SP_MAX_SETUP_VALUE, string);
	    xfree(string);
	}
	if ((string = xspGetTextString(pref_dialog->autosave_label_suffix_combo)) != NULL) {
	    spDebug(10, "swPopdownPreferenceDialogCB", "autosave_label_suffix = %s\n", string);
	    spStrCopy(pref_dialog->config->autosave_label_suffix, SP_MAX_SETUP_VALUE, string);
	    xfree(string);
	}
	spGetToggleState(pref_dialog->autosave_label_fullpath_button,
			 &pref_dialog->config->autosave_label_fullpath);
	spGetToggleState(pref_dialog->autosave_by_drop_button,
			 &pref_dialog->config->autosave_by_drop);
#endif
	
	if ((string = xspGetTextString(pref_dialog->size_text)) != NULL) {
	    int width = 0, height = 0;
	    spDebug(10, "swPopdownPreferenceDialogCB", "size string = %s\n", string);
	    if (sscanf(string, "%dx%d", &width, &height) == 2) {
		flag = SP_TRUE;
	    } else if (sscanf(string, "%d %d", &width, &height) == 2) {
		flag = SP_TRUE;
	    } else if (sscanf(string, "%d,%d", &width, &height) == 2) {
		flag = SP_TRUE;
	    }
	    if (width > 0) pref_dialog->config->width = width;
	    if (height > 0) pref_dialog->config->height = height;
	    
	    xfree(string);
	}

	spGetToggleState(pref_dialog->alt_ctrl_swap_button,
			 &pref_dialog->config->alt_ctrl_swap);
	
#ifdef SW_USE_TOOL_BAR
	if (spGetToggleState(pref_dialog->use_tool_bar_button, &use_tool_bar) == SP_TRUE) {
	    if (use_tool_bar == SP_TRUE) {
		pref_dialog->config->no_tool_bar = SP_FALSE;
	    } else {
		pref_dialog->config->no_tool_bar = SP_TRUE;
	    }
	    swShowAllToolBar(pref_dialog->config->toplevel->current_window, use_tool_bar);
	    /*flag = SP_TRUE;*/
	}
#endif
	
	if (spGetToggleState(pref_dialog->display_meter_button,
			     &pref_dialog->config->display_meter) == SP_TRUE) {
	    flag = SP_TRUE;
	}
	if (spGetToggleState(pref_dialog->draw_selection_times_button,
			     &pref_dialog->config->draw_selection_times) == SP_TRUE) {
	    flag = SP_TRUE;
	}
	if (spGetToggleState(pref_dialog->draw_selection_length_button,
			     &pref_dialog->config->draw_selection_length) == SP_TRUE) {
	    flag = SP_TRUE;
	}
	if (spGetToggleState(pref_dialog->draw_scale_button, &pref_dialog->config->scale_flag) == SP_TRUE) {
	    flag = SP_TRUE;
	}
	if (spGetToggleState(pref_dialog->draw_grid_button, &pref_dialog->config->grid_flag) == SP_TRUE) {
	    flag = SP_TRUE;
	}
	
	if ((index = spGetSelectedListIndex(pref_dialog->look_like_combo)) >= 0) {
	    if (sw_num_looks <= 0) {
		sw_num_looks = spArraySize(sw_looks);
	    }

	    for (i = 1; i < sw_num_looks; i++) {
		if (index == sw_looks[i].id) {
		    strcpy(pref_dialog->config->wave_fg, sw_looks[i].wave_fg);
		    strcpy(pref_dialog->config->wave_bg, sw_looks[i].wave_bg);
		    strcpy(pref_dialog->config->pointer_color, sw_looks[i].pointer_color);
		    strcpy(pref_dialog->config->label_color, sw_looks[i].label_color);
		    strcpy(pref_dialog->config->string_color, sw_looks[i].string_color);
		    strcpy(pref_dialog->config->scale_color, sw_looks[i].scale_color);
		    flag = SP_TRUE;
		    break;
		}
	    }

	    if (flag == SP_TRUE) {
		swSetColor(pref_dialog->config);
	    }
	}

	if (flag == SP_TRUE) {
	    swDrawAllWave(pref_dialog->config->toplevel->current_window);
	}
    }
    
    return;
}

static int selectFormatItem(spComponent combo, char *format_string)
{
    int i;
    int index = -1;
    
    for (i = 0; format_list_strings[i] != NULL; i++) {
	if (strcaseeq(format_string, format_list_strings[i])
	    || strcaseeq(format_string, format_list_short_strings[i])) {
	    spSelectListIndex(combo, i);
	    index = i;
	    break;
	}
    }

    return index;
}

static void createPreferenceFileTab(swPrefDialog pref_dialog, spComponent tab_box)
{
    char string[SP_MAX_LINE];
    
    pref_dialog->file_tab = spAddTabItem(tab_box, "fileTab", -1,
					 SppTitle, SW_FILE_TAB_LABEL,
					 NULL);
    
    /* create check box to set whether you use the default format */
    pref_dialog->default_format_button = spCreateCheckBox(pref_dialog->file_tab, "useDefaultFormatButton",
							  SppTitle, SW_USE_DEFAULT_FORMAT_LABEL,
							  SppSet, pref_dialog->config->use_def_format,
							  NULL);

    /* create combo box to input the wave format */
    pref_dialog->format_combo = spCreateParamField(pref_dialog->file_tab, "formatComboBox", 60,
						   SppTitle, SW_FORMAT_LABEL,
						   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						   SppEditable, SP_FALSE,
						   SppFieldStrings, format_list_strings,
						   SppFieldOffset, 140,
						   SppFieldSize, 160,
						   NULL);
    selectFormatItem(pref_dialog->format_combo, pref_dialog->config->def_format_string);
    
    /* create combo box to input bits per sample */
    pref_dialog->bit_combo = spCreateParamField(pref_dialog->file_tab, "bitComboBox", 60,
						SppTitle, SW_BIT_LABEL,
						SppDimension, "bits/sample",
						SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						SppEditable, SP_FALSE,
						SppFieldStrings, bit_list_strings,
						SppFieldOffset, 140,
						SppFieldSize, 160,
						NULL);
    swGetSampleBitString(string, pref_dialog->config->def_samp_bit);
    spSelectListItem(pref_dialog->bit_combo, string);
    
    /* create combo box to input sampling frequency */
    pref_dialog->samp_rate_combo = spCreateParamField(pref_dialog->file_tab, "sampFreqComboBox", 60,
						      SppTitle, SW_SAMP_RATE_LABEL,
						      SppDimension, "Hz",
						      SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						      SppEditable, SP_TRUE,
						      SppFieldStrings, samp_rate_list_strings,
						      SppFieldOffset, 140,
						      SppFieldSize, 160,
						      NULL);
    sprintf(string, "%.0f", pref_dialog->config->def_samp_rate);
    spSetTextString(pref_dialog->samp_rate_combo, string);
    
    /* create combo box to input a number of channel */
    pref_dialog->channel_combo = spCreateParamField(pref_dialog->file_tab, "channelComboBox", 60,
						    SppTitle, SW_CHANNEL_LABEL,
						    SppDimension, "channels",
						    SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						    SppEditable, SP_TRUE,
						    SppFieldStrings, channel_list_strings,
						    SppFieldOffset, 140,
						    SppFieldSize, 160,
						    NULL);
    sprintf(string, "%d", pref_dialog->config->def_num_channel);
    spSetTextString(pref_dialog->channel_combo, string);

    return;
}

static void displayRestartWarningCB(spComponent component, void *data)
{
    spDisplayWarning(component, NULL, SW_NEED_RESTART_WARNING_MESSAGE);
    return;
}

static void createPreferenceLookTab(swPrefDialog pref_dialog, spComponent tab_box)
{
    int i;
    char string[SP_MAX_LINE];
    static char **list_strings = NULL;

    if (list_strings == NULL) {
	if (sw_num_looks <= 0) {
	    sw_num_looks = spArraySize(sw_looks);
	}
	list_strings = xalloc(sw_num_looks + 1, char *);
	for (i = 0; i < sw_num_looks; i++) {
	    list_strings[i] = sw_looks[i].label;
	}
	list_strings[sw_num_looks] = NULL;
    }
    
    pref_dialog->look_tab = spAddTabItem(tab_box, "lookTab", -1,
					 SppTitle, SW_LOOK_TAB_LABEL,
					 NULL);
    
    /* create combo box to input the look name */
    pref_dialog->look_like_combo = spCreateParamField(pref_dialog->look_tab, "lookComboBox", 60,
						      SppTitle, SW_LOOK_LIKE_LABEL,
						      SppDimension, SW_LOOK_LIKE_DIMENSION,
						      SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						      SppEditable, SP_FALSE,
						      SppFieldStrings, list_strings,
						      SppFieldOffset, 140,
						      SppFieldSize, 160,
						      NULL);
    spSelectListIndex(pref_dialog->look_like_combo, 0);
    
    /* create text field to input the size */
    pref_dialog->size_text = spCreateParamField(pref_dialog->look_tab, "sizeText", 60,
						SppTitle, SW_WINDOW_SIZE_LABEL,
						SppFieldType, SP_FIELD_TYPE_TEXT,
						SppEditable, SP_TRUE,
						SppFieldOffset, 140,
						SppFieldSize, 160,
						NULL);
    sprintf(string, "%dx%d", pref_dialog->config->width, pref_dialog->config->height);
    spSetTextString(pref_dialog->size_text, string);

    pref_dialog->alt_ctrl_swap_button = spCreateCheckBox(pref_dialog->look_tab, "altCtrlSwapButton",
							 SppTitle, SW_ALT_CTRL_SWAP_LABEL,
							 SppSet, pref_dialog->config->alt_ctrl_swap,
							 SppCallbackFunc, displayRestartWarningCB,
							 NULL);
    
#ifdef SW_USE_TOOL_BAR
    /* create check box to set whether you use the tool bar */
    pref_dialog->use_tool_bar_button = spCreateCheckBox(pref_dialog->look_tab, "useToolBarButton",
							SppTitle, SW_USE_TOOL_BAR_LABEL,
							SppSet, (pref_dialog->config->no_tool_bar == SP_TRUE ? SP_FALSE : SP_TRUE),
							NULL);
#endif
    
    /* create check box to set whether you draw the following */
    pref_dialog->display_meter_button = spCreateCheckBox(pref_dialog->look_tab, "displayMeterButton",
							 SppTitle, SW_DISPLAY_METER_LABEL,
							 SppSet, pref_dialog->config->display_meter,
							 NULL);
    pref_dialog->draw_selection_times_button = spCreateCheckBox(pref_dialog->look_tab, "drawSelectionTimesButton",
								SppTitle, SW_DRAW_SELECTION_TIMES_LABEL,
								SppSet, pref_dialog->config->draw_selection_times,
								NULL);
    pref_dialog->draw_selection_length_button = spCreateCheckBox(pref_dialog->look_tab, "drawSelectionLengthButton",
								 SppTitle, SW_DRAW_SELECTION_LENGTH_LABEL,
								 SppSet, pref_dialog->config->draw_selection_length,
								 NULL);
    pref_dialog->draw_scale_button = spCreateCheckBox(pref_dialog->look_tab, "drawScaleButton",
						      SppTitle, SW_DRAW_SCALE_LABEL,
						      SppSet, pref_dialog->config->scale_flag,
						      NULL);
    pref_dialog->draw_grid_button = spCreateCheckBox(pref_dialog->look_tab, "drawGridButton",
						     SppTitle, SW_DRAW_GRID_LABEL,
						     SppSet, pref_dialog->config->grid_flag,
						     NULL);
    
    return;
}

static void createPreferenceSoundTab(swPrefDialog pref_dialog, spComponent tab_box)
{
    char string[SP_MAX_SETUP_VALUE];
    static char *buffer_list_strings[] =
    {
	"256",
	"512",
	"1024",
	"2048",
	"4096",
	"8192",
	"16384",
	"32768",
	"65536",
	NULL,
    };

    pref_dialog->sound_tab = spAddTabItem(tab_box, "soundTab", -1,
					  SppTitle, SW_SOUND_TAB_LABEL,
					  NULL);
    
    /* create check box to set whether you use the play command */
    pref_dialog->use_play_command_button = spCreateCheckBox(pref_dialog->sound_tab, "usePlayCommandButton",
							    SppTitle, SW_USE_PLAY_COMMAND_LABEL,
							    SppSet, pref_dialog->config->use_play_command,
							    NULL);
    
    /* create text field to input play command */
    pref_dialog->play_command_text = spCreateParamField(pref_dialog->sound_tab, "playCommandText", 60,
							SppTitle, SW_PLAY_COMMAND_LABEL,
							SppFieldType, SP_FIELD_TYPE_TEXT,
							SppEditable, SP_TRUE,
							SppFieldOffset, 140,
							SppFieldSize, 240,
							NULL);
    spStrCopy(string, SP_MAX_SETUP_VALUE, pref_dialog->config->play_command);
    spSetTextString(pref_dialog->play_command_text, string);
    
#ifdef SP_SUPPORT_AUDIO
    /* create combo box to input buffer size */
    pref_dialog->buffer_size_combo = spCreateParamField(pref_dialog->sound_tab, "bufferSizeComboBox", 60,
							SppTitle, SW_BUFFER_SIZE_LABEL,
							SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							SppEditable, SP_TRUE,
							SppFieldStrings, buffer_list_strings,
							SppFieldOffset, 140,
							SppFieldSize, 160,
							NULL);
    sprintf(string, "%d", pref_dialog->config->audio_buffer_size);
    spSetTextString(pref_dialog->buffer_size_combo, string);

    pref_dialog->use_loop_button = spCreateCheckBox(pref_dialog->sound_tab, "useLoopButton",
						    SppTitle, SW_USE_LOOP_PLAY_LABEL,
						    SppSet, pref_dialog->config->use_loop_play,
						    NULL);
    pref_dialog->use_sync_button = spCreateCheckBox(pref_dialog->sound_tab, "useSyncButton",
						    SppTitle, SW_USE_SYNC_PLAY_LABEL,
						    SppSet, pref_dialog->config->use_sync_play,
						    NULL);
    
    pref_dialog->play_draw_button = spCreateCheckBox(pref_dialog->sound_tab, "playDrawButton",
						     SppTitle, SW_PLAY_DRAW_LABEL,
						     SppSet, pref_dialog->config->play_draw,
						     NULL);
#endif

#ifdef SW_USE_THREAD
    if (pref_dialog->config->wave_config->thread_safe == SP_TRUE) {
	pref_dialog->use_thread_button = spCreateCheckBox(pref_dialog->sound_tab, "useThreadButton",
							  SppTitle, SW_USE_THREAD_LABEL,
							  SppSet, pref_dialog->config->wave_config->process_use_thread,
							  NULL);
	if (spIsThreadLevelVariable() == SP_TRUE) {
	    pref_dialog->use_lowlevel_thread_button = spCreateCheckBox(pref_dialog->sound_tab, "useLowLevelThreadButton",
						       SppTitle, SW_USE_LOWLEVEL_THREAD_LABEL,
						       SppSet, pref_dialog->config->use_lowlevel_thread,
						       SppCallbackFunc, displayRestartWarningCB,
						       NULL);
	}
    }
#endif
    
    return;
}

static void createPreferenceLabelTab(swPrefDialog pref_dialog, spComponent tab_box)
{
    pref_dialog->label_tab = spAddTabItem(tab_box, "labelTab", -1,
					  SppTitle, SW_LABEL_TAB_LABEL,
					  NULL);
    
    pref_dialog->display_info_area_button = spCreateCheckBox(pref_dialog->label_tab, "displayInfoAreaButton",
							     SppTitle, SW_DISPLAY_INFO_AREA_LABEL,
							     SppSet, pref_dialog->config->display_info_area,
							     NULL);

    pref_dialog->display_freq_info_area_button = spCreateCheckBox(pref_dialog->label_tab, "displayFreqInfoAreaButton",
								  SppTitle, SW_DISPLAY_FREQ_INFO_AREA_LABEL,
								  SppSet, pref_dialog->config->display_freq_info_area,
								  NULL);

    pref_dialog->link_region_label_button = spCreateCheckBox(pref_dialog->label_tab, "linkRegionLabelButton",
							     SppTitle, SW_LINK_REGION_LABEL_LABEL,
							     SppSet, pref_dialog->config->link_region_label,
							     NULL);
    
    pref_dialog->erase_label_prompt_button = spCreateCheckBox(pref_dialog->label_tab, "eraseLabelPromptButton",
							       SppTitle, SW_ERASE_LABEL_PROMPT_LABEL,
							       SppSet, pref_dialog->config->erase_label_prompt,
							       NULL);
    
    return;
}

static void createPreferenceAutosaveTab(swPrefDialog pref_dialog, spComponent tab_box)
{
#ifdef SW_SUPPORT_AUTOSAVE
    static char *id_prefix_list[] =
    {
	"",
	"-",
	"_",
	NULL,
    };
    static char *id_type_list[] =
    {
	"2-Digit Serial Number",
	"3-Digit Serial Number",
	"4-Digit Serial Number",
	"Serial Number",
	"Edge Points",
	NULL,
    };
    static char *label_suffix_list[] =
    {
	"_cut.txt",
	".cut",
	".lab",
	NULL,
    };

    pref_dialog->autosave_tab = spAddTabItem(tab_box, "autosaveTab", -1,
					     SppTitle, SW_AUTOSAVE_TAB_LABEL,
					     NULL);
    
    /* create text field to input directory */
    pref_dialog->autosave_dir_text =
	spCreateParamField(pref_dialog->autosave_tab, "autosaveDirText", 60,
			   SppTitle, SW_AUTOSAVE_DIR_LABEL,
			   SppFieldType, SP_FIELD_TYPE_DIR_TEXT,
			   SppEditable, SP_TRUE,
			   SppFieldOffset, 140,
			   SppFieldSize, 180,
			   SppDimensionSize, 60,
			   NULL);
    spSetTextString(pref_dialog->autosave_dir_text,
		    pref_dialog->config->autosave_dir);

    /* create combo box to input the ID type */
    pref_dialog->autosave_id_type_combo =
	spCreateParamField(pref_dialog->autosave_tab, "autosaveIdTypeCombo", 60,
			   SppTitle, SW_AUTOSAVE_ID_TYPE_LABEL,
			   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
			   SppEditable, SP_FALSE,
			   SppFieldStrings, id_type_list,
			   SppFieldOffset, 140,
			   SppFieldSize, 180,
			   NULL);
    spSelectListIndex(pref_dialog->autosave_id_type_combo,
		      pref_dialog->config->autosave_id_type);
    
    /* create combo box to input the ID prefix */
    pref_dialog->autosave_id_prefix_combo =
	spCreateParamField(pref_dialog->autosave_tab, "autosaveIdPrefixCombo", 60,
			   SppTitle, SW_AUTOSAVE_ID_PREFIX_LABEL,
			   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
			   SppEditable, SP_TRUE,
			   SppFieldStrings, id_prefix_list,
			   SppFieldOffset, 140,
			   SppFieldSize, 160,
			   NULL);
    spSetTextString(pref_dialog->autosave_id_prefix_combo,
		    pref_dialog->config->autosave_id_prefix);

    /* create combo box to input the suffix of a label file */
    pref_dialog->autosave_label_suffix_combo =
	spCreateParamField(pref_dialog->autosave_tab, "autosaveLabelSuffixCombo", 60,
			   SppTitle, SW_AUTOSAVE_LABEL_SUFFIX_LABEL,
			   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
			   SppEditable, SP_TRUE,
			   SppFieldStrings, label_suffix_list,
			   SppFieldOffset, 140,
			   SppFieldSize, 160,
			   NULL);
    spSetTextString(pref_dialog->autosave_label_suffix_combo,
		    pref_dialog->config->autosave_label_suffix);
    
    /* create check box to set whether you write the fullpath into a label file */
    pref_dialog->autosave_label_fullpath_button =
	spCreateCheckBox(pref_dialog->autosave_tab, "autosaveLabelFullpathButton",
			 SppTitle, SW_AUTOSAVE_LABEL_FULLPATH_LABEL,
			 SppSet, pref_dialog->config->autosave_label_fullpath,
			 NULL);

    /* create check box */
    pref_dialog->autosave_by_drop_button =
	spCreateCheckBox(pref_dialog->autosave_tab, "autosaveByDropButton",
			 SppTitle, SW_AUTOSAVE_DROP_LABEL,
			 SppSet, pref_dialog->config->autosave_by_drop,
			 NULL);
#endif
    
    return;
}

static swPrefDialog createPreferenceDialog(swConfig config)
{
    swPrefDialog pref_dialog;

    pref_dialog = xalloc(1, struct _swPrefDialog);
    memset(pref_dialog, 0, sizeof(struct _swPrefDialog));
    pref_dialog->config = config;

    pref_dialog->window = spCreateDialogBox("preferenceDialog",
					    SppTitle, SW_PREFERENCE_TITLE,
					    SppCallbackFunc, swPopdownPreferenceDialogCB,
					    SppCallbackData, pref_dialog,
					    SppDialogBoxButtonType, SP_DB_OK_CANCEL_APPLY,
					    SppCloseStyle, SP_UNMAP_CLOSE,
					    SppSpacingOn, SP_TRUE,
					    NULL);

    /* create tab box */
    pref_dialog->tab_box = spCreateTabBox(pref_dialog->window, "preferenceTabBox", 430,
					  NULL);
    
    /* create file tab */
    createPreferenceFileTab(pref_dialog, pref_dialog->tab_box);
    
    /* create look tab */
    createPreferenceLookTab(pref_dialog, pref_dialog->tab_box);
    
    /* create sound tab */
    createPreferenceSoundTab(pref_dialog, pref_dialog->tab_box);
    
    /* create label tab */
    createPreferenceLabelTab(pref_dialog, pref_dialog->tab_box);
    
    /* create autosave tab */
    createPreferenceAutosaveTab(pref_dialog, pref_dialog->tab_box);
    
    return pref_dialog;
}

swPrefDialog swCreatePreferenceDialog(swConfig config)
{
    static swPrefDialog sw_pref_dialog = NULL;

    if (sw_pref_dialog == NULL) {
	sw_pref_dialog = createPreferenceDialog(config);
    }
    
    return sw_pref_dialog;
}

spBool swPopupPreferenceDialogCB(spComponent component, swWindow window)
{
    swPrefDialog pref_dialog;
    
    window->config->toplevel->current_window = window;

    pref_dialog = swCreatePreferenceDialog(window->config);

    /* popup dialog */
    spPopupWindow(pref_dialog->window);

    return SP_TRUE;
}
#endif /* SW_USE_PREFERENCE_DIALOG */

/*
 *	functions for format dialog
 */
void swPopdownFormatDialogCB(spComponent component, swFormatDialog format_dialog)
{
    char *string;

    if (spIsCreated(format_dialog->window) == SP_FALSE) return;

    /* popdown format dialog */
    spPopdownWindow(format_dialog->window);
    
    format_dialog->call_reason = spGetCallbackReason(component);

    if (format_dialog->call_reason == SP_CR_OK) {
	if ((string = xspGetTextString(format_dialog->format_combo)) != NULL) {
	    spDebug(10, "swPopdownFormatDialogCB", "format string = %s\n", string);
	    spStrCopy(format_dialog->config->format_string, SP_MAX_SETUP_VALUE, string);
	    xfree(string);
	}

	if ((string = xspGetTextString(format_dialog->bit_combo)) != NULL) {
	    spDebug(10, "swPopdownFormatDialogCB", "bit string = %s\n", string);
	    format_dialog->config->samp_bit = swGetSampleBit(string);
	    xfree(string);
	}
	
	if ((string = xspGetTextString(format_dialog->samp_rate_combo)) != NULL) {
	    spDebug(10, "swPopdownFormatDialogCB", "samp_rate string = %s\n", string);
	    format_dialog->config->samp_rate = atof(string);
	    xfree(string);
	}
	
	if ((string = xspGetTextString(format_dialog->channel_combo)) != NULL) {
	    spDebug(10, "swPopdownFormatDialogCB", "channel string = %s\n", string);
	    format_dialog->config->num_channel = atoi(string);
	    xfree(string);
	}
    }
    
    return;
}

static swFormatDialog createFormatDialog(swConfig config)
{
    swFormatDialog format_dialog;
    char string[SP_MAX_LINE];
    
    format_dialog = xalloc(1, struct _swFormatDialog);
    memset(format_dialog, 0, sizeof(struct _swFormatDialog));
    format_dialog->call_reason = SP_CR_NONE;
    format_dialog->config = config;

    format_dialog->window = spCreateDialogBox("formatDialog",
					      /*SppTitle, SW_FORMAT_TITLE,*/
					      SppCallbackFunc, swPopdownFormatDialogCB,
					      SppCallbackData, format_dialog,
					      SppCloseStyle, SP_UNMAP_CLOSE,
					      NULL);
    
    /* create combo box to input wave format */
    format_dialog->format_combo = spCreateParamField(format_dialog->window, "formatComboBox", 60,
						     SppTitle, SW_FORMAT_LABEL,
						     SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						     SppEditable, SP_FALSE,
						     SppFieldStrings, format_list_strings,
						     SppFieldOffset, 140,
						     SppFieldSize, 160,
						     NULL);
    selectFormatItem(format_dialog->format_combo, config->format_string);
    /*spSelectListItem(format_dialog->format_combo, config->format_string);*/
    
    /* create combo box to input bits per sample */
    format_dialog->bit_combo = spCreateParamField(format_dialog->window, "bitComboBox", 60,
						  SppTitle, SW_BIT_LABEL,
						  SppDimension, "bits/sample",
						  SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						  SppEditable, SP_FALSE,
						  SppFieldStrings, bit_list_strings,
						  SppFieldOffset, 140,
						  SppFieldSize, 160,
						  NULL);
    swGetSampleBitString(string, config->samp_bit);
    spSelectListItem(format_dialog->bit_combo, string);
    
    /* create combo box to input sampling frequency */
    format_dialog->samp_rate_combo = spCreateParamField(format_dialog->window, "sampFreqComboBox", 60,
							SppTitle, SW_SAMP_RATE_LABEL,
							SppDimension, "Hz",
							SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							SppEditable, SP_TRUE,
							SppFieldStrings, samp_rate_list_strings,
							SppFieldOffset, 140,
							SppFieldSize, 160,
							NULL);
    sprintf(string, "%.0f", config->samp_rate);
    spSetTextString(format_dialog->samp_rate_combo, string);
    
    /* create combo box to input number of channel */
    format_dialog->channel_combo = spCreateParamField(format_dialog->window, "channelComboBox", 60,
						      SppTitle, SW_CHANNEL_LABEL,
						      SppDimension, "channels",
						      SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						      SppEditable, SP_TRUE,
						      SppFieldStrings, channel_list_strings,
						      SppFieldOffset, 140,
						      SppFieldSize, 160,
						      NULL);
    sprintf(string, "%d", config->num_channel);
    spSetTextString(format_dialog->channel_combo, string);
    
    return format_dialog;
}

static swFormatDialog sw_format_dialog = NULL;

swFormatDialog swCreateFormatDialog(swConfig config)
{
    if (sw_format_dialog == NULL) {
	sw_format_dialog = createFormatDialog(config);
    }
    sw_format_dialog->call_reason = SP_CR_NONE;
    
    return sw_format_dialog;
}

spComponent swGetFormatDialogWindow(void)
{
    if (sw_format_dialog == NULL) {
	return NULL;
    }

    return sw_format_dialog->window;
}

spBool swPopupFormatDialog(swConfig config, char *filename)
{
    swFormatDialog format_dialog;
    char title[SP_MAX_MESSAGE];
    
    format_dialog = swCreateFormatDialog(config);

    /* set window title */
    if (!strnone(filename)) {
	spStrCopy(title, SP_MAX_MESSAGE, SW_FORMAT_TITLE);
	spStrCat(title, SP_MAX_MESSAGE, ": ");
	spStrCat(title, SP_MAX_MESSAGE, filename);
	
	spSetParams(format_dialog->window, SppTitle, title, NULL);
    } else {
	spSetParams(format_dialog->window, SppTitle, SW_FORMAT_TITLE, NULL);
    }
    
    /* popup dialog */
    spPopupWindow(format_dialog->window);

    if (format_dialog->call_reason == SP_CR_OK) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

/*
 *	functions for bit conversion
 */
typedef struct _swBitConvDialog {
    spComponent window;
    spComponent bit_combo;
    spComponent float_normalized_button;
    int samp_bit;
    spBool *float_normalized;
} *swBitConvDialog;

static void popdownBitConvDialogCB(spComponent component, swBitConvDialog dialog)
{
    char *string;

    spPopdownWindow(component);
    
    if (dialog == NULL) return;
    
    if (spGetCallbackReason(component) == SP_CR_OK) {
	if ((string = xspGetTextString(dialog->bit_combo)) != NULL) {
	    spDebug(10, "popdownBitConvDialogCB", "bit string = %s\n", string);
	    dialog->samp_bit = swGetSampleBit(string);
	    xfree(string);
	}
	spGetToggleState(dialog->float_normalized_button, dialog->float_normalized);
    }

    return;
}

void swPopupBitConvDialogCB(spComponent component, swWindow window)
{
    char string[SP_MAX_LINE];
    static swBitConvDialog dialog = NULL;

    if (window == NULL || window->wave == NULL) return;

    if (dialog == NULL) {
	dialog = xalloc(1, struct _swBitConvDialog);
	memset(dialog, 0, sizeof(struct _swBitConvDialog));
	
	dialog->window = spCreateDialogBox("bitConvDialog",
					   SppTitle, SW_BIT_CONV_DIALOG_TITLE,
					   SppCallbackFunc, popdownBitConvDialogCB,
					   SppCallbackData, dialog,
					   SppDialogBoxButtonType, SP_DB_OK_CANCEL,
					   SppCloseStyle, SP_UNMAP_CLOSE,
					   NULL);
	
	/* create combo box to input bits per sample */
	dialog->bit_combo = spCreateParamField(dialog->window, "bitComboBox", 0,
					       SppTitle, SW_BIT_LABEL,
					       SppDimension, "bits/sample",
					       SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					       SppEditable, SP_FALSE,
					       SppFieldStrings, bit_list_strings,
					       SppFieldOffset, 120,
					       SppFieldSize, 120,
					       NULL);
	
	dialog->float_normalized_button = spCreateCheckBox(dialog->window, "floatNormalizedButton",
							   SppTitle, SW_FLOAT_NORMALIZED_LABEL,
							   NULL);
    }
    swGetSampleBitString(string, window->wave->samp_bit);
    spSelectListItem(dialog->bit_combo, string);
    dialog->samp_bit = window->wave->samp_bit;

    spSetToggleState(dialog->float_normalized_button, window->config->wave_config->float_normalized);
    dialog->float_normalized = &window->config->wave_config->float_normalized;
    
    /* popup dialog */
    spPopupWindow(dialog->window);

    if (dialog->samp_bit != window->wave->samp_bit) {
	swConvertBit(window, dialog->samp_bit);
    }

    return;
}

#define SW_MIN_AMPLIFY_RATE 0
#define SW_MAX_AMPLIFY_RATE 300

void swPopupAmplifyDialogCB(spComponent component, swWindow window)
{
    double rate;
    static swRateChangeDialog dialog = NULL;

    if (window == NULL || window->wave == NULL) return;

    spDebug(60, "swPopupAmplifyDialogCB", "in\n");

    if (dialog == NULL) {
	static char *amplify_rate_strings[] =
	{
	    "50",
	    "80",
	    "90",
	    "100",
	    "110",
	    "120",
	    "150",
	    "200",
	    NULL,
	};
	static char *amplify_dB_strings[] =
	{
	    "-6.0",
	    "-3.0",
	    "-2.0",
	    "-1.0",
	    "0.0",
	    "1.0",
	    "2.0",
	    "3.0",
	    "6.0",
	    NULL,
	};
	
	dialog = swCreateRateChangeDialog(SW_AMPLIFY_DIALOG_TITLE,
					  SW_AMPLIFY_RATE_LABEL, amplify_rate_strings,
					  SW_AMPLIFY_DB_LABEL, amplify_dB_strings,
					  SW_MIN_AMPLIFY_RATE, SW_MAX_AMPLIFY_RATE);
    }

    rate = 100.0;
    
    /* popup dialog */
    if (swPopupRateChangeDialog(dialog, &rate) == SP_TRUE) {
	swEditWindow(window, SW_EDIT_AMPLIFY, window->sel_st, window->sel_ed, rate / 100.0);
    }

    return;
}

void swPopupMaximizeDialogCB(spComponent component, swWindow window)
{
    double min, max;
    double absmax;
    double rate;
    static swRateChangeDialog dialog = NULL;

    if (window == NULL || window->wave == NULL) return;

    spDebug(60, "swPopupMaximizeDialogCB", "in\n");

    min = swGetWaveMin(window->wave);
    max = swGetWaveMax(window->wave);
    absmax = MAX(FABS(min), FABS(max));
    spDebug(60, "swPopupMaximizeDialogCB", "min = %f, max = %f, absmax = %f\n",
	    min, max, absmax);

    if (absmax <= 0.0) {
	/* maximum value is too small. */
	spDisplayError(window->window, NULL, SW_MAXIMUM_VALUE_TOO_SMALL_ERROR_MESSAGE);
	return;
    }
    
    if (swIsWaveFloat(window->wave) == SP_TRUE
	&& window->config->wave_config->float_normalized == SP_FALSE) {
	if (spCreateMessageBox(window->window, NULL,
			       SW_ASSUME_FLOAT_NORMALIZED_WARNING_MESSAGE,
			       SppDialogType, SP_WARNING_DIALOG,
			       SppMessageBoxButtonType, SP_MB_YES_NO,
			       NULL) != SP_DR_YES) {
	    return;
	}
	window->config->wave_config->float_normalized = SP_TRUE;
    }

    if (dialog == NULL) {
	static char *maximize_rate_strings[] =
	{
	    "50",
	    "60",
	    "70",
	    "80",
	    "90",
	    "100",
	    NULL,
	};
	static char *maximize_dB_strings[] =
	{
	    "-6.0",
	    "-3.0",
	    "-2.0",
	    "-1.0",
	    "-0.5",
	    "0.0",
	    NULL,
	};
	
	dialog = swCreateRateChangeDialog(SW_MAXIMIZE_DIALOG_TITLE,
					  SW_MAXIMIZE_RATE_LABEL, maximize_rate_strings,
					  SW_MAXIMIZE_DB_LABEL, maximize_dB_strings,
					  0, 100);
    }

    rate = 100.0 * absmax / swGetLimitValue(window->wave->samp_bit);
    spDebug(60, "swPopupMaximizeDialogCB", "initial rate = %f\n", rate);
	
    /* popup dialog */
    if (swPopupRateChangeDialog(dialog, &rate) == SP_TRUE) {
	spDebug(60, "swPopupMaximizeDialogCB", "rate = %f\n", rate);
	swEditWindow(window, SW_EDIT_AMPLIFY, 0, window->wave->total_length,
		     swGetLimitValue(window->wave->samp_bit) * rate / 100.0 / absmax);
    }

   return;
}

/*
 *	functions for dialog for sampling frequency conversion
 */
typedef struct _swSampFreqConvDialog {
    spComponent window;
    spComponent samp_rate_field;
    spComponent cutoff_field;
    spComponent transition_field;
    spComponent gain_field;
    spComponent sidelobe_field;
    spComponent tolerance_field;
    spComponent buffer_length_field;
    
    double samp_rate;
    swConfig config;
} *swSampFreqConvDialog;

static void popdownSampFreqConvDialogCB(spComponent component, swSampFreqConvDialog dialog)
{
    char *string;
    double value;

    spPopdownWindow(component);
    
    if (dialog == NULL) return;
    
    if (spGetCallbackReason(component) == SP_CR_OK) {
	if ((string = xspGetTextString(dialog->samp_rate_field)) != NULL) {
	    spDebug(10, "popdownSampFreqConvDialogCB", "string = %s\n", string);
	    if ((value = atof(string)) <= 0.0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SAMP_RATE_LABEL);
	    } else {
		dialog->samp_rate = value;
	    }
	    xfree(string);
	}
	
	if ((string = xspGetTextString(dialog->cutoff_field)) != NULL) {
	    if ((value = atof(string)) <= 0.0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SFC_CUTOFF_LABEL);
	    } else {
		dialog->config->wave_config->sfc_cutoff = value;
	    }
	    xfree(string);
	}
	if ((string = xspGetTextString(dialog->transition_field)) != NULL) {
	    if ((value = atof(string)) <= 0.0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SFC_TRANSITION_LABEL);
	    } else {
		dialog->config->wave_config->sfc_transition = value;
	    }
	    xfree(string);
	}
	if ((string = xspGetTextString(dialog->sidelobe_field)) != NULL) {
	    if ((value = atof(string)) <= 0.0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SFC_SIDELOBE_LABEL);
	    } else {
		dialog->config->wave_config->sfc_sidelobe = value;
	    }
	    xfree(string);
	}
	if ((string = xspGetTextString(dialog->gain_field)) != NULL) {
	    if ((value = atof(string)) <= 0.0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SFC_GAIN_LABEL);
	    } else {
		dialog->config->wave_config->sfc_gain = value;
	    }
	    xfree(string);
	}
	if ((string = xspGetTextString(dialog->tolerance_field)) != NULL) {
	    if ((value = atof(string)) < 0.0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SFC_TOLERANCE_LABEL);
	    } else {
		dialog->config->wave_config->sfc_tolerance = value;
	    }
	    xfree(string);
	}
	if ((string = xspGetTextString(dialog->buffer_length_field)) != NULL) {
	    long buffer_length;
	    
	    if ((buffer_length = atol(string)) <= 0) {
		spDisplayError(component, NULL, SW_PARAMETER_ERROR_MESSAGE, SW_SFC_BUFFER_LENGTH_LABEL);
	    } else {
		dialog->config->wave_config->sfc_buffer_length = buffer_length;
	    }
	    xfree(string);
	}
    }

    return;
}

void swPopupSampFreqConvDialogCB(spComponent component, swWindow window)
{
    char string[SP_MAX_LINE];
    static swSampFreqConvDialog dialog = NULL;
    static char *cutoff_list_strings[] =
    {
	"0.90",
	"0.92",
	"0.93",
	"0.94",
	"0.95",
	"0.96",
	"0.97",
	"0.98",
	NULL,
    };
    static char *transition_list_strings[] =
    {
	"0.02",
	"0.03",
	"0.04",
	"0.05",
	"0.06",
	"0.07",
	"0.08",
	"0.09",
	"0.10",
	NULL,
    };
    static char *sidelobe_list_strings[] =
    {
	"40.0",
	"50.0",
	"60.0",
	"70.0",
	"80.0",
	"90.0",
	"100.0",
	NULL,
    };
    static char *tolerance_list_strings[] =
    {
	"0.2",
	"0.5",
	"1.0",
	"1.5",
	"2.0",
	"2.5",
	"3.0",
	NULL,
    };
    static char *buffer_length_list_strings[] =
    {
	"8192",
	"16384",
	"32768",
	"65536",
	"131072",
	"262144",
	NULL,
    };
    static char *gain_list_strings[] =
    {
	"0.5",
	"0.8",
	"0.9",
	"1.0",
	"1.1",
	"1.2",
	"1.5",
	"2.0",
	NULL,
    };

    if (window == NULL || window->wave == NULL) return;

    if (dialog == NULL) {
	dialog = xalloc(1, struct _swSampFreqConvDialog);
	memset(dialog, 0, sizeof(struct _swSampFreqConvDialog));
	dialog->config = window->config;
	
	dialog->window = spCreateDialogBox("sampFreqConvDialog",
					   SppTitle, SW_SAMP_RATE_CONV_DIALOG_TITLE,
					   SppCallbackFunc, popdownSampFreqConvDialogCB,
					   SppCallbackData, dialog,
					   SppDialogBoxButtonType, SP_DB_OK_CANCEL,
					   SppCloseStyle, SP_UNMAP_CLOSE,
					   NULL);
	
	/* create combo box to input sampling frequency */
	dialog->samp_rate_field = spCreateParamField(dialog->window, "sampFreqField", 60,
						     SppTitle, SW_SAMP_RATE_LABEL,
						     SppDimension, "Hz",
						     SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						     SppEditable, SP_TRUE,
						     SppFieldStrings, samp_rate_list_strings,
						     SppFieldOffset, 140,
						     SppFieldSize, 160,
						     NULL);
	
	sprintf(string, "%.2f", dialog->config->wave_config->sfc_cutoff);
	dialog->cutoff_field = spCreateParamField(dialog->window, "cutoffField", 60,
						  SppTitle, SW_SFC_CUTOFF_LABEL,
						  SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						  SppEditable, SP_TRUE,
						  SppFieldString, string,
						  SppFieldStrings, cutoff_list_strings,
						  SppFieldOffset, 140,
						  SppFieldSize, 160,
						  NULL);
	
	sprintf(string, "%.2f", dialog->config->wave_config->sfc_transition);
	dialog->transition_field = spCreateParamField(dialog->window, "transitionField", 60,
						      SppTitle, SW_SFC_TRANSITION_LABEL,
						      SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						      SppEditable, SP_TRUE,
						      SppFieldString, string,
						      SppFieldStrings, transition_list_strings,
						      SppFieldOffset, 140,
						      SppFieldSize, 160,
						      NULL);
	
	sprintf(string, "%.1f", dialog->config->wave_config->sfc_sidelobe);
	dialog->sidelobe_field = spCreateParamField(dialog->window, "sidelobeField", 60,
						    SppTitle, SW_SFC_SIDELOBE_LABEL,
						    SppDimension, "dB",
						    SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						    SppEditable, SP_TRUE,
						    SppFieldString, string,
						    SppFieldStrings, sidelobe_list_strings,
						    SppFieldOffset, 140,
						    SppFieldSize, 160,
						    NULL);
	
	sprintf(string, "%.1f", dialog->config->wave_config->sfc_gain);
	dialog->gain_field = spCreateParamField(dialog->window, "gainField", 60,
						SppTitle, SW_SFC_GAIN_LABEL,
						SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						SppEditable, SP_TRUE,
						SppFieldString, string,
						SppFieldStrings, gain_list_strings,
						SppFieldOffset, 140,
						SppFieldSize, 160,
						NULL);
	
	sprintf(string, "%.1f", dialog->config->wave_config->sfc_tolerance);
	dialog->tolerance_field = spCreateParamField(dialog->window, "toleranceField", 60,
						     SppTitle, SW_SFC_TOLERANCE_LABEL,
						     SppDimension, "%",
						     SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
						     SppEditable, SP_TRUE,
						     SppFieldString, string,
						     SppFieldStrings, tolerance_list_strings,
						     SppFieldOffset, 140,
						     SppFieldSize, 160,
						     NULL);

	sprintf(string, "%ld", dialog->config->wave_config->sfc_buffer_length);
	dialog->buffer_length_field = spCreateParamField(dialog->window, "bufferLengthField", 60,
							 SppTitle, SW_SFC_BUFFER_LENGTH_LABEL,
							 SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
							 SppEditable, SP_TRUE,
							 SppFieldString, string,
							 SppFieldStrings, buffer_length_list_strings,
							 SppFieldOffset, 140,
							 SppFieldSize, 160,
							 NULL);
    }
    sprintf(string, "%.0f", window->wave->samp_rate);
    spSetTextString(dialog->samp_rate_field, string);
    
    dialog->samp_rate = window->wave->samp_rate;
    
    /* popup dialog */
    spPopupWindow(dialog->window);

    if (dialog->samp_rate != window->wave->samp_rate) {
	swConvertSampFreq(window, dialog->samp_rate);
    }

    return;
}

void swPopupValueChangeDialogCB(spComponent component, swWindow window)
{
    int channel;
    long point;
    long data_offset;
    double new_value;
    double value = 0.0;
    char *string;
    char buf[SP_MAX_MESSAGE];
    static swStringChangeDialog dialog = NULL;

    if (swIsNoWave(window) == SP_TRUE) return;

    if (dialog == NULL) {
	dialog = swCreateStringChangeDialog(SW_CHANGE_VALUE_DIALOG_TITLE, SW_VALUE_LABEL, NULL,
					    NULL, NULL, 80, 120, SP_TRUE);
    }

    channel = window->target_channel;
    point = window->point;
    data_offset = (window->point - window->offset) / swGetWaveThinLength(window->wave);
    swGetWaveData(window->wave, window->target_channel, data_offset, &value);
    if (swIsWaveFloat(window->wave) == SP_TRUE) {
	sprintf(buf, "%f", value);
    } else {
	sprintf(buf, "%ld", (long)value);
    }

    if ((string = xswPopupStringChangeDialog(dialog, NULL, buf)) != NULL) {
	new_value = atof(string);
	if (new_value != value) {
	    swChangeWindowValue(window, channel, point, new_value);
	}
	xfree(string);
    }
    
    return;
}

void swPopupInsertPauseDialogCB(spComponent component, swWindow window)
{
    char *string;
    char *dim_string;
    static swStringChangeDialog dialog = NULL;

    if (swIsNoWave(window) == SP_TRUE) return;

    if (dialog == NULL) {
	dialog = swCreateStringChangeDialog(SW_INSERT_PAUSE_DIALOG_TITLE,
					    SW_PAUSE_LABEL, "points",
					    NULL, NULL, 100, 140, SP_TRUE);
    }

    if (window->config->time_format == SW_TIME_FORMAT_MSEC
	|| window->config->time_format == SW_TIME_FORMAT_FLOORED_MSEC) {
	dim_string = "msec";
    } else {
	if (window->config->time_format == SW_TIME_FORMAT_POINT) {
	    dim_string = "points";
	} else {
	    dim_string = "sec";
	}
    }

    if ((string = xswPopupStringChangeDialog(dialog, dim_string, "0")) != NULL) {
	double sec;
	long pause_length;

	spDebug(50, "swInsertPauseDialogCB", "string = %s\n", string);
	
	if (swConvertTimeString(string, window->config->time_format, &sec) == SP_TRUE) {
	    pause_length = swDimToSamp(window, sec);
	    spDebug(50, "swInsertPauseDialogCB", "sec = %f, pause_length = %ld\n",
		    sec, pause_length);
	    
	    if (pause_length > 0) {
		swEditWindow(window, SW_EDIT_INSERT_PAUSE, window->point, pause_length, 0.0);
	    }
	}

	xfree(string);
    }
    
    return;
}

/*
 *	functions for dialog to change string
 */
struct _swStringChangeDialog {
    spComponent window;
    spComponent field;
    spCallbackReason reason;
};

static void popdownStringChangeCB(spComponent component, swStringChangeDialog dialog)
{
    if (dialog != NULL) {
	dialog->reason = spGetCallbackReason(component);
    }

    spPopdownWindow(component);
    
    return;
}

swStringChangeDialog swCreateStringChangeDialog(char *title, char *label, char *dimension,
						char **list, char *default_string, 
						int offset, int size, spBool editable)
{
    swStringChangeDialog dialog;

    dialog = xalloc(1, struct _swStringChangeDialog);
    memset(dialog, 0, sizeof(struct _swStringChangeDialog));
    dialog->reason = SP_CR_NONE;
    
    dialog->window = spCreateDialogBox("stringChangeDialog",
				       SppTitle, title,
				       SppCallbackFunc, popdownStringChangeCB,
				       SppCallbackData, dialog,
				       SppDialogBoxButtonType, SP_DB_OK_CANCEL,
				       SppCloseStyle, SP_UNMAP_CLOSE,
				       NULL);

    if (list != NULL) {
	/* create combo box */
	dialog->field = spCreateParamField(dialog->window, "stringChangeCombo", 60,
					   SppTitle, label,
					   SppDimension, dimension,
					   SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					   SppEditable, editable,
					   SppFieldStrings, list,
					   SppFieldOffset, offset,
					   SppFieldSize, size,
					   NULL);

	if (editable == SP_TRUE) {
	    spSetTextString(dialog->field, default_string);
	} else {
	    spSelectListItem(dialog->field, default_string);
	}
    } else {
	/* create text field */
	dialog->field = spCreateParamField(dialog->window, "stringChangeText", 60,
					   SppTitle, label,
					   SppDimension, dimension,
					   SppFieldType, SP_FIELD_TYPE_TEXT,
					   SppEditable, editable,
					   SppTextString, default_string,
					   SppFieldOffset, offset,
					   SppFieldSize, size,
					   NULL);
    }

    return dialog;
}

char *xswPopupStringChangeDialog(swStringChangeDialog dialog, char *dimension, char *string)
{
    dialog->reason = SP_CR_NONE;

    if (!strnone(dimension)) {
	spSetParams(dialog->field,
		    SppDimension, dimension,
		    NULL);
    }
    if (!strnone(string)) {
	spSetTextString(dialog->field, string);
    }

    spPopupWindow(dialog->window);

    if (dialog->reason == SP_CR_OK) {
	return xspGetTextString(dialog->field);
    } else {
	return NULL;
    }
}

/*
 *	functions for dialog to change rate in dB
 */
struct _swRateChangeDialog {
    spComponent window;

    spComponent track_bar;
    spComponent rate_combo;
    spComponent db_combo;

    int min_rate;
    int max_rate;
    
    double rate;
    spCallbackReason reason;
};

static void popdownRateChangeDialogCB(spComponent component, swRateChangeDialog dialog)
{
    char *string;

    spPopdownWindow(component);
    
    if (dialog == NULL) return;

    dialog->reason = spGetCallbackReason(component);
    if (dialog->reason == SP_CR_OK) {
	if ((string = xspGetTextString(dialog->rate_combo)) != NULL) {
	    spDebug(10, "popdownRateChangeDialogCB", "rate string = %s\n", string);
	    dialog->rate = atof(string);
	    xfree(string);
	}
    }

    return;
}

static void updateRateChangeDialog(spComponent component, swRateChangeDialog dialog, double rate)
{
    char buf[SP_MAX_LINE];
    static spBool flag = SP_FALSE; /* flag to prevent endless loop */

    if (flag == SP_TRUE
	|| dialog == NULL || dialog->track_bar == NULL
	|| dialog->rate_combo == NULL || dialog->db_combo == NULL) return;

    flag = SP_TRUE;
    
    rate = MAX(rate, dialog->min_rate);
    rate = MIN(rate, dialog->max_rate);
    
    spDebug(10, "updateRateChangeDialog", "rate = %f\n", rate);

    if (component != dialog->db_combo) {
	if (rate <= 0.0) {
	    sprintf(buf, "-Inf");
	} else {
	    sprintf(buf, "%.3f", dB(rate/100.0));
	}
	spSetTextString(dialog->db_combo, buf);
    }
    
    if (component != dialog->rate_combo) {
	sprintf(buf, "%.1f", rate);
	spSetTextString(dialog->rate_combo, buf);
    }

    if (component != dialog->track_bar) {
	spSetSliderValue(dialog->track_bar, (int)rate);
    }

    flag = SP_FALSE;
    
    return;
}

static void rateChangeTrackBarCB(spComponent component, swRateChangeDialog dialog)
{
    int value;

    if (spGetSliderValue(component, &value) == SP_TRUE) {
	spDebug(10, "rateChangeTrackBarCB", "value = %d\n", value);
	updateRateChangeDialog(dialog->track_bar, dialog, (double)value);
    }
    
    return;
}

static void rateChangeEditRateCB(spComponent component, swRateChangeDialog dialog)
{
    double rate;
    char *string;
    
    if ((string = xspGetTextString(component)) != NULL) {
	rate = atof(string);
	updateRateChangeDialog(dialog->rate_combo, dialog, rate);
	xfree(string);
    }
    
    return;
}

static void rateChangeEditdBCB(spComponent component, swRateChangeDialog dialog)
{
    double rate;
    char *string;
    
    if ((string = xspGetTextString(component)) != NULL) {
	if (strcaseeq(string, "-Inf")) {
	    rate = dialog->min_rate;
	} else {
	    rate = MIN(100.0 * pow(10.0, atof(string) / 20.0), dialog->max_rate);
	}
	updateRateChangeDialog(dialog->db_combo, dialog, rate);
	xfree(string);
    }
    
    return;
}

swRateChangeDialog swCreateRateChangeDialog(char *title,
					    char *rate_label, char **rate_strings,
					    char *db_label, char **dB_strings,
					    int min_rate, int max_rate)
{
    swRateChangeDialog dialog = NULL;

    dialog = xalloc(1, struct _swRateChangeDialog);
    memset(dialog, 0, sizeof(struct _swRateChangeDialog));
    dialog->min_rate = min_rate;
    dialog->max_rate = max_rate;
    dialog->rate = 100.0;
    dialog->reason = SP_CR_NONE;
    
    dialog->window = spCreateDialogBox("rateChangeDialog",
				       SppTitle, title,
				       SppCallbackFunc, popdownRateChangeDialogCB,
				       SppCallbackData, dialog,
				       SppDialogBoxButtonType, SP_DB_OK_CANCEL,
				       SppCloseStyle, SP_UNMAP_CLOSE,
				       NULL);
    
    dialog->track_bar = spCreateTrackBar(dialog->window, "rageChangeTrackBar",
					 SppCallbackFunc, rateChangeTrackBarCB,
					 SppCallbackData, dialog,
					 SppTrackCallbackOn, SP_TRUE,
					 SppShowScale, SP_TRUE,
					 SppShowValue, SP_TRUE,
					 SppValue, (int)dialog->rate,
					 SppMinimum, min_rate,
					 SppMaximum, max_rate,
					 NULL);
    
    /* create combo box to input rate */
    dialog->rate_combo = spCreateParamField(dialog->window, "rateChangeRateCombo", 60,
					    SppTitle, rate_label,
					    SppDimension, "%",
					    SppCallbackFunc, rateChangeEditRateCB,
					    SppCallbackData, dialog,
					    SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					    SppEditable, SP_TRUE,
					    SppFieldStrings, rate_strings,
					    SppFieldOffset, 120,
					    SppFieldSize, 120,
					    NULL);
    
    /* create combo box to input factor in dB */
    dialog->db_combo = spCreateParamField(dialog->window, "rateChangedBCombo", 60,
					  SppTitle, db_label,
					  SppDimension, "dB",
					  SppCallbackFunc, rateChangeEditdBCB,
					  SppCallbackData, dialog,
					  SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					  SppEditable, SP_TRUE,
					  SppFieldStrings, dB_strings,
					  SppFieldOffset, 120,
					  SppFieldSize, 120,
					  NULL);

   return dialog;
}

spBool swPopupRateChangeDialog(swRateChangeDialog dialog, double *rate)
{
    dialog->reason = SP_CR_NONE;
    dialog->rate = *rate;

    updateRateChangeDialog(NULL, dialog, dialog->rate);
    
    spPopupWindow(dialog->window);

    if (dialog->reason == SP_CR_OK) {
	*rate = dialog->rate;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}
