/*
 *	swDraw.c
 *
 *	Last modified: <2002-03-16 23:34:18 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spAudioLib.h>
#include <sp/spLib.h>
#include <sp/spComponentLib.h>

#include "swWaveAudio.h"

#include "swWindow.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swEdit.h"
#include "swDialog.h"
#include "swLabel.h"
#include "swLabelDialog.h"
#include "swLabelList.h"
#include "swAnalysis.h"

#define SW_DRAW_LABEL_CANVAS	/* cursor will be sometimes slow */
/*#undef SW_DRAW_LABEL_CANVAS*/ /* dragging a label will be slow */

#define SW_METER_FONT "-*-*-medium-r-normal--8-*-*-*-*-*-*-*"

static swGraphics sw_graphics = NULL;

void swSetColor(swConfig config)
{
    if (sw_graphics == NULL) {
	sw_graphics = xalloc(1, struct _swGraphics);
	sw_graphics->gx_fg = NULL;
	sw_graphics->gx_bg = NULL;
	sw_graphics->gx_xor = NULL;
	sw_graphics->gx_pointer = NULL;
	sw_graphics->gx_string = NULL;
	sw_graphics->gx_label = NULL;
	sw_graphics->gx_region = NULL;
	sw_graphics->gx_region_bg = NULL;
	sw_graphics->gx_scale = NULL;
	sw_graphics->gx_meter = NULL;
	sw_graphics->gx_overview_xor = NULL;
	
	sw_graphics->region_pixel = 0L;
	sw_graphics->current_region_pixel = 0L;
	sw_graphics->region_bg_pixel = 0L;
	sw_graphics->region_pixel_incr = 0;
	sw_graphics->mutex = spCreateMutex(NULL);
    }
    
    spLockMutex(sw_graphics->mutex);
    
    spDebug(40, "swCreateGraphics", "%s %s %s %s\n",
	    config->wave_fg, config->wave_bg, config->pointer_color, config->label_color);
    
    if (sw_graphics->gx_fg == NULL) {
	sw_graphics->gx_fg = spCreateGraphics("Foreground",
					      SppForeground, config->wave_fg,
					      NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_fg,
			    SppForeground, config->wave_fg,
			    NULL);
    }
    
    if (sw_graphics->gx_bg == NULL) {
	sw_graphics->gx_bg = spCreateGraphics("Background",
					      SppForeground, config->wave_bg,
					      NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_bg,
			    SppForeground, config->wave_bg,
			    NULL);
    }
    
    if (sw_graphics->gx_xor == NULL) {
	sw_graphics->gx_xor = spCreateGraphics("XOR",
					       SppForeground, "max",
					       SppGraphicsMode, SP_GM_XOR,
					       NULL);
    }
    
    if (sw_graphics->gx_pointer == NULL) {
	sw_graphics->gx_pointer = spCreateGraphics("Pointer",
						   SppForeground, config->pointer_color,
						   NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_pointer,
			    SppForeground, config->pointer_color,
			    NULL);
    }
    
    if (sw_graphics->gx_string == NULL) {
	sw_graphics->gx_string = spCreateGraphics("String",
						  SppForeground, config->string_color,
						  NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_string,
			    SppForeground, config->string_color,
			    NULL);
    }
    
    if (sw_graphics->gx_label == NULL) {
	sw_graphics->gx_label = spCreateGraphics("Label",
						 SppForeground, config->label_color,
						 NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_label,
			    SppForeground, config->label_color,
			    NULL);
    }
    {
	int i;
	spPixel pixel;
	int r, g, b;
	
	pixel = spGetColorPixel(config->wave_bg);
	r = spGetRValue(pixel);
	g = spGetGValue(pixel);
	b = spGetBValue(pixel);
	if (r + g + b >= 386) {
	    sw_graphics->region_pixel_incr = SW_REGION_PIXEL_NEGA_INCR;
	} else {
	    sw_graphics->region_pixel_incr = SW_REGION_PIXEL_POSI_INCR;
	}

	for (i = 0;; i++) {
	    r = MAX(r + 2 * (i + 1) * sw_graphics->region_pixel_incr, 0); r = MIN(r, 255);
	    g = MAX(g + 2 * (i + 1) * sw_graphics->region_pixel_incr, 0); g = MIN(g, 255);
	    b = MAX(b + 2 * (i + 1) * sw_graphics->region_pixel_incr, 0); b = MIN(b, 255);
	    pixel = spRGB(r, g, b);

	    if (i == 0) {
		sw_graphics->region_pixel = pixel;
		spGetColorName(pixel, sw_graphics->region_color);

#if 0
		sw_graphics->region_bg_pixel = spRGB(255 - r, 255 - g, 255 - b);
#else
		sw_graphics->region_bg_pixel = pixel;
#endif
		spGetColorName(sw_graphics->region_bg_pixel, sw_graphics->region_bg_color);
	    } else if (i == 1) {
		spGetColorName(pixel, sw_graphics->region_line_color);
	    } else {
		spGetColorName(pixel, sw_graphics->region_label_color);
		break;
	    }
	}
	
	if (sw_graphics->gx_region == NULL) {
	    sw_graphics->gx_region = spCreateGraphics("Region",
						      SppForeground, sw_graphics->region_color,
						      NULL);
	} else {
	    spSetGraphicsParams(sw_graphics->gx_region,
				SppForeground, sw_graphics->region_color,
				NULL);
	}
	if (sw_graphics->gx_region_bg == NULL) {
	    sw_graphics->gx_region_bg = spCreateGraphics("RegionBG",
							 SppGraphicsMode, /*SP_GM_XOR*/SP_GM_AND,
							 SppForeground, sw_graphics->region_bg_color,
							 NULL);
	} else {
	    spSetGraphicsParams(sw_graphics->gx_region_bg,
				SppForeground, sw_graphics->region_bg_color,
				NULL);
	}
    }
    
    if (sw_graphics->gx_scale == NULL) {
	sw_graphics->gx_scale = spCreateGraphics("Scale",
						 SppForeground, config->scale_color,
						 NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_scale,
			    SppForeground, config->scale_color,
			    NULL);
    }
    
    if (sw_graphics->gx_meter == NULL) {
	sw_graphics->gx_meter = spCreateGraphics("Meter",
						 SppFontName, SW_METER_FONT,
						 NULL);
    }
    
    if (sw_graphics->gx_overview_xor == NULL) {
	sw_graphics->gx_overview_xor = spCreateGraphics("OverviewXOR",
							SppForeground, "max",
							SppGraphicsMode, SP_GM_XOR,
							NULL);
    }
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

static void getDrawRange(swWindow window, double draw_height,
			 int channel, int *draw_min, int *draw_max)
{
    int dmin, dmax;

    dmin = (int)((double)channel * draw_height);
    dmax = dmin + (int)draw_height;

    if (draw_min != NULL) *draw_min = dmin;
    if (draw_max != NULL) *draw_max = dmax;

    return;
}

void swGetDrawRange(swWindow window, int channel, int *draw_min, int *draw_max)
{
    getDrawRange(window, window->draw_height, channel, draw_min, draw_max);
    return;
}

int swGetDrawWidth(swWindow window)
{
#ifdef SW_SUPPORT_METER
    return MAX(window->width - window->meter_width, SW_MIN_DRAW_WIDTH);
#else
    return MAX(window->width, SW_MIN_DRAW_WIDTH);
#endif
}

#ifdef SW_SUPPORT_METER
spBool swIsMeterVisible(swWindow window)
{
    if (window == NULL || window->draw_height <= 0.0
	|| window->meter_width <= 0) {
	return SP_FALSE;
    }

    return SP_TRUE;
}

#define SW_MIN_METER_VALUE -300.0
#define SW_METER_DRAW_HEIGHT_THRESHOLD 150
#define SW_METER_TICK_STRING_LEFT_OFFSET 5
#define SW_METER_TICK_STRING_BOTTOM_OFFSET 5

void swDrawMeterBar(spComponent component, swWindow window, 
		    double draw_height, int left_offset, int top_offset, double value)
{
    int i;
    int y, py;
    int height;
    int meter_range;
    int meter_min, meter_max;
    double y_factor;
    unsigned char green, red;
    char buf[SP_MAX_LINE];

    if (swIsMeterVisible(window) == SP_FALSE) return;

    meter_min = -window->config->meter_range;
    meter_max = 6;
    meter_range = meter_max - meter_min;

    if (meter_range <= 0) return;
    
    spLockMutex(sw_graphics->mutex);
    
    y_factor = draw_height / (double)meter_range;
    spDebug(60, "swDrawMeterBar", "y_factor = %f, draw_height = %f, left_offset = %d, top_offset = %d\n",
	    y_factor, draw_height, left_offset, top_offset);

    py = top_offset + (int)draw_height;
    for (i = meter_min; i <= meter_max; i++) {
	y = top_offset + (int)draw_height - (int)spRound(y_factor * (double)(i - meter_min));
	height = py - y;

	if (height >= 1) {
	    if (i < -24) {
		green = (unsigned char)MAX(256 + 3 * (24 + i), 0);
		red = 0;
	    } else if (i > -6) {
		green = MAX(-128 * i / 3, 0);
		red = 255;
	    } else if (i > -18) {
		green = 255;
		red = MIN(64 * i / 3 + 384, 255);
	    } else {
		green = 255;
		red = 0;
	    }
	    
	    if ((double)i > value) {
		green /= 2;
		red /= 2;
	    }
	    
	    spDebug(120, "swDrawMeterBar", "i = %d, green = %d, red = %d, y = %d, height = %d\n",
		    i, green, red, y, height);
	    spSetForegroundPixel(sw_graphics->gx_meter, spRGB(red, green, 0));
	    
	    spFillRectangle(component, sw_graphics->gx_meter,
			    left_offset, y,
			    window->meter_width, height);
	    
	    py = y;
	}
    }
    
    py = top_offset + (int)draw_height;
    for (i = meter_min; i < meter_max; i++) {
	y = top_offset + (int)draw_height - (int)spRound(y_factor * (double)(i - meter_min));
	if ((draw_height > SW_METER_DRAW_HEIGHT_THRESHOLD && i >= -6 && i % 3 == 0)
	    || (i >= -12 && i % 6 == 0)
	    || (i % 24 == 0)) {
	    /* draw tick */
	    spSetForegroundPixel(sw_graphics->gx_meter, spRGB(255, 255, 255));
	    
	    spDrawLine(component, sw_graphics->gx_meter,
		       left_offset, y, left_offset + 2, y);
	    spDrawLine(component, sw_graphics->gx_meter,
		       left_offset + window->meter_width,
		       y, left_offset + window->meter_width - 3, y);

	    sprintf(buf, "%3d", i);
	    spDrawString(component, sw_graphics->gx_meter,
			 left_offset + SW_METER_TICK_STRING_LEFT_OFFSET,
			 y + SW_METER_TICK_STRING_BOTTOM_OFFSET, buf);
	}
    }
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

void swDrawMeter(spComponent component, swWindow window,
		 double draw_height, int left_offset, int channel, double value)
{
    int samp_bit;
    double min, max;
    double dbvalue;
    
    if (swIsMeterVisible(window) == SP_FALSE) return;

    if (swIsWaveFloat(window->wave) == SP_TRUE) {
	min = swGetWaveMin(window->wave);
	max = swGetWaveMax(window->wave);
	max = MAX(FABS(min), FABS(max));
	if (max < 1.0) {
	    max = 1.0;
	}
    } else {
	samp_bit = swGetWaveSampleBit(window->wave);
	if (samp_bit >= 32) {
	    max = 2147483648.0;
	} else if (samp_bit >= 24) {
	    max = 8388608.0;
	} else {
	    max = 32768.0;
	}
    }

    value = FABS(value);

    if (value <= 0.0) {
	dbvalue = -300;
    } else {
	dbvalue = dB(value/max);
    }
    
    spDebug(80, "swDrawMeter", "channel = %d, value = %f, dbvalue = %f\n",
	    channel, value, dbvalue);
    swDrawMeterBar(component, window, draw_height, left_offset,
		   (int)spRound((double)channel * draw_height),
		   dbvalue);
    
    return;
}
#endif

double swGetDataHeight(swWindow window, double *data_min, double *data_max)
{
    double dmin, dmax;
    
    dmin = swGetWaveMin(window->wave);
    dmax = swGetWaveMax(window->wave);
    
    if (dmin >= dmax) {
	dmin = 0.0; dmax = 1.0;
    }
    if (data_min != NULL) *data_min = dmin;
    if (data_max != NULL) *data_max = dmax;

    return dmax - dmin;
}

#define SW_MIN_HSCALE_DIV 80
#define SW_MIN_VSCALE_DIV /*30*/25
#define SW_TICK_LENGTH 5
#define SW_HSCALE_LEFT_SPACING 65
#define SW_HSCALE_BOTTOM_SPACING 25
#define SW_HSCALE_STRING_LEFT_OFFSET 2
#define SW_HSCALE_STRING_BOTTOM_OFFSET 3
#define SW_VSCALE_STRING_LEFT_OFFSET 5
#define SW_VSCALE_STRING_TOP_OFFSET 2
#define SW_VSCALE_STRING_BOTTOM_OFFSET 2

#define SW_TIME_STRING_HEIGHT 16

static double swGetScaleTick(int width, double data_width, double *pw, spBool v_flag)
{
    int idx;
    double ndiv;
    double pn, fract;
    double tick, ntick;
    /* table for drawing scale */
    static double scale_table[] = {1.0, 2.0, 5.0, 10.0};
    static double scale_log_table[] = {0.0, 0.301, 0.699, 1.0};

    if (data_width <= 0.0) {
	fract = 0.0;
	pn = 0.0;
    } else {
	if (v_flag == SP_TRUE) {
	    ndiv = MAX(floor((double)width / SW_MIN_VSCALE_DIV), 2.0);
	} else {
	    ndiv = MAX(floor((double)width / SW_MIN_HSCALE_DIV), 2.0);
	}
	fract = modf(log10(data_width / ndiv), &pn);
    }

    if (pn < 0.0 || fract < 0.0) {
	pn = pn - 1.0;
	fract = 1.0 + fract;
    }

    spDebug(40, "swGetScaleTick", "pn = %f, fract = %f\n", pn, fract);

    if (fract > scale_log_table[0] && fract <= scale_log_table[1]) {
	idx = 1;
    } else if (fract > scale_log_table[1] && fract <= scale_log_table[2]) {
	idx = 2;
    } else if (fract > scale_log_table[2] && fract <= scale_log_table[3]) {
	idx = 0;
	pn++;
    } else {
	idx = 0;
    }
    tick = scale_table[idx];

    ntick =  tick * pow(10.0, pn);
    ndiv = floor(data_width / ntick);
    if (ndiv * tick >= 10.0) {
	tick /= 10.0;
	pn++;
    }

    if (pw != NULL) {
	*pw = pn;
    }

    return tick;
}

int swGetScaleString(char *string, double tick, double pn, double value)
{
    int flag = 0;
    int pni;
    double n, ntick;
    char exp_string[SP_MAX_MESSAGE];

    n = pow(10.0, pn);
    ntick = n * tick;
    
    if (value == 0.0) {
	sprintf(string, "0");
    } else if (pn < 0) {
	pni = (int)spRound(pn);
	if (tick < 1.0) {
	    pni--;
	}
	switch (pni) {
	  case -1:
	    sprintf(string, "%.1f", value);
	    break;
	  case -2:
	    sprintf(string, "%.2f", value);
	    break;
	  case -3:
	    sprintf(string, "%.3f", value);
	    break;
	  case -4:
	    sprintf(string, "%.4f", value);
	    break;
	  case -5:
	    sprintf(string, "%.5f", value);
	    break;
	  default:
	    if (tick < 1.0) {
		sprintf(string, "%.1f", value / n);
	    } else {
		sprintf(string, "%.0f", value / n);
	    }
	    flag = 1;
	}
    } else {
	if (pn >= 6.0) {
	    if (tick < 1.0) {
		sprintf(string, "%.1f", value / n);
	    } else {
		sprintf(string, "%.0f", value / n);
	    }
	    flag = 1;
	} else {
	    if (ntick < 1.0) {
		sprintf(string, "%.1f", value);
	    } else {
		sprintf(string, "%.0f", value);
	    }
	}
    }
		
    if (flag) {
	sprintf(exp_string, "x10e%.0f", pn);
	strcat(string, exp_string);
    }

    return flag;
}

static void swClipZeroChars(char *string)
{
    int i;
    
    for (i = strlen(string)-1; i >= 0; i--) {
	if (string[i] == '0') {
	    string[i] = NUL;
	} else {
	    if (i >= 1 && string[i] == '.') {
		string[i] = NUL;
	    }
	    break;
	}
    }

    return;
}

void swDrawHScale(spComponent component, swWindow window,
		  double draw_width, double draw_height,
		  int channel, long offset, long length)
{
    int w, h;
    int x;
    int direct;
    int x_width;
    int draw_min, draw_max;
    long k;
    long ndmax, ndmin;
    double tick, ntick;
    double n, pn;
    double x_data_width;
    double x_max, x_min;
    double x_zero, x_fact;
    double sx;
    char string[SP_MAX_MESSAGE];

    getDrawRange(window, draw_height, channel, &draw_min, &draw_max);
    x_width = (int)draw_width;

    direct = 1;
    x_min = swSampToCurrentDim(window, offset);
    x_max = swSampToCurrentDim(window, offset + length - 1);
    x_data_width = x_max - x_min;

    tick = swGetScaleTick(x_width, x_data_width, &pn, SP_FALSE);
    n = pow(10.0, pn);
    ntick = n * tick;

    ndmin = (long)floor(x_min / ntick);
    ndmax = (long)ceil(x_max / ntick);

    x_fact = (double)(direct * x_width) / x_data_width;
    x_zero = x_fact * (direct > 0 ? -x_min : x_max);

    if (window->config->grid_flag == SP_TRUE) {
	spLockMutex(sw_graphics->mutex);
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_DOT, NULL);
	for (k = ndmin; k <= ndmax; k++) {
	    sx = (double)k * ntick;
	    x = (int)spRound(x_zero + sx * x_fact);

	    if (x >= 0 && x <= x_width) {
		if (x == x_width) x -= 1;
		
		spDrawLine(component, sw_graphics->gx_scale, x, draw_min, x, draw_max);
	    }
	}
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_SOLID, NULL);
	spUnlockMutex(sw_graphics->mutex);
    }

    if (draw_max <= window->height) {
	spDrawLine(component, sw_graphics->gx_scale, 0, draw_max, /*x_width*/window->width, draw_max);
    }
    
    if (window->config->scale_flag == SP_FALSE) {
	return;
    }

    for (k = ndmin; k <= ndmax; k++) {
	sx = (double)k * ntick;
	x = (int)spRound(x_zero + sx * x_fact);

	if (x >= 0 && x <= x_width) {
	    if (x == x_width) x -= 1;
	    
	    spDrawLine(component, sw_graphics->gx_scale, x, draw_min, x, draw_min + SW_TICK_LENGTH);
	    spDrawLine(component, sw_graphics->gx_scale, x, draw_max, x, draw_max - SW_TICK_LENGTH);

	    if (x >= SW_HSCALE_LEFT_SPACING
		&& channel == window->wave->num_channel - 1) {
		if (window->config->time_format == SW_TIME_FORMAT_SEPARATED_SEC) {
		    spGetTimeString(sx, SP_TIME_FORMAT_SEPARATED_SEC, string);
		    swClipZeroChars(string);
		} else {
		    swGetScaleString(string, tick, pn, sx);
		}
		
		if (spGetStringExtent(sw_graphics->gx_scale, string, NULL, NULL, &w, &h) == SP_TRUE) {
		    spDrawString(component, sw_graphics->gx_scale,
				 MAX(x - w / 2, SW_HSCALE_STRING_LEFT_OFFSET),
				 draw_max - SW_TICK_LENGTH - SW_HSCALE_STRING_BOTTOM_OFFSET,
				 string);
		}
	    }
	}
    }

    return;
}

void swDrawVScale(spComponent component, swWindow window,
		  double draw_width, double draw_height,
		  int channel, double data_min, double data_height,
		  double *y_factor, double *y_zero_offset, spBool draw_flag)
{
    int flag;
    int y;
    int direct;
    int x_width;
    int y_width;
    int draw_min, draw_max;
    int w, h;
    long k;
    long ndmax, ndmin;
    double data_max;
    double tick, ntick;
    double n, pn;
    double y_fact, y_zero;
    double sy;
    char string[SP_MAX_MESSAGE];

    spDebug(40, "swDrawVScale", "data_min = %f, data_height = %f\n",
	    data_min, data_height);
    
    getDrawRange(window, draw_height, channel, &draw_min, &draw_max);
    x_width = (int)draw_width;
    y_width = (int)draw_height;

    flag = 0;
    direct = -1;
    data_max = data_min + data_height;
    if (data_min == data_max) {
	flag = 1;
    }

    tick = swGetScaleTick(y_width, data_height, &pn, SP_TRUE);
    n = pow(10.0, pn);
    ntick = n * tick;

    ndmax = (long)ceil(data_max / ntick);
    ndmin = (long)floor(data_min / ntick);
    if (flag) {
	data_min = ntick * (double)ndmin;
	data_max = ntick * (double)ndmax;

	if (data_max == data_min) {
	    data_min -= 1.0;
	    data_max += 1.0;
	}
	data_height = data_max - data_min;
    }

    spDebug(40, "swDrawVScale", "data_min = %f, data_max = %f\n",
	    data_min, data_max);
    
    y_fact = (double)(window->direction * (y_width)) / data_height;
    y_zero = y_fact * ((window->direction > 0) ? data_max : -data_min);

    if (y_factor != NULL) *y_factor = y_fact;
    if (y_zero_offset != NULL) *y_zero_offset = y_zero;

    if (draw_flag == SP_FALSE) return;
    
    y_zero += (double)draw_min;
    
    if (window->config->grid_flag == SP_TRUE || window->config->zero_flag == SP_TRUE) {
	spLockMutex(sw_graphics->mutex);
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_DOT, NULL);
	if (window->config->grid_flag == SP_TRUE) {
	    for (k = ndmin; k <= ndmax; k++) {
		sy = (double)k * ntick;
		y = (int)spRound(y_zero - sy * y_fact);

		if (y >= draw_min && y <= draw_max) {
		    spDrawLine(component, sw_graphics->gx_scale, 0, y, x_width, y);
		}
	    }
	} else {
	    y = (int)spRound(y_zero);
	    if (y >= draw_min && y <= draw_max) {
		spDrawLine(component, sw_graphics->gx_scale, 0, y, x_width, y);
	    }
	}
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_SOLID, NULL);
	spUnlockMutex(sw_graphics->mutex);
    }

    if (window->config->scale_flag == SP_TRUE) {
	for (k = ndmin; k <= ndmax; k++) {
	    sy = (double)k * ntick;
	    y = (int)spRound(y_zero - sy * y_fact);

	    if (y >= draw_min && y <= draw_max) {
		spDrawLine(component, sw_graphics->gx_scale, 0, y, SW_TICK_LENGTH, y);
		spDrawLine(component, sw_graphics->gx_scale,
			   x_width - SW_TICK_LENGTH, y, x_width, y);
	    
		swGetScaleString(string, tick, pn, sy);
	    
		if (spGetStringExtent(sw_graphics->gx_scale, string, NULL, NULL, &w, &h) == SP_TRUE) {
		    y = MAX(y -  direct * (h / 2 - SW_VSCALE_STRING_BOTTOM_OFFSET),
			    (draw_min + SW_VSCALE_STRING_TOP_OFFSET - direct * h));
		    y = MIN(y, draw_max + direct * SW_VSCALE_STRING_BOTTOM_OFFSET);
		    spDrawString(component, sw_graphics->gx_scale,
				 SW_TICK_LENGTH + SW_VSCALE_STRING_LEFT_OFFSET, y,
				 string);
		}
	    }
	}
    }

    return;
}

void swDrawWaveform(spComponent component, swWindow window,
		    double draw_height, double zero_offset, double x_factor, double y_factor,
		    int channel, long draw_offset, long draw_length, spBool peak_flag)
{
    int x, y;
    int px, py;
    int min, max;
    int draw_min, draw_max;
    long k, l;
    long thin_length;
    double value;
    double zero_height;

    getDrawRange(window, draw_height, channel, &draw_min, &draw_max);
    zero_height = (double)draw_min + zero_offset;

    if (peak_flag == SP_TRUE) {
	thin_length = window->wave->peak_buf_thin_length;
	swGetPeakData(window->wave, channel, draw_offset * window->wave->peak_buf_thin_length, &value);
    } else {
	thin_length = window->wave->thin_length;
	swGetWaveData(window->wave, channel, draw_offset, &value);
    }
    
    spDebug(50, "swDrawWaveform", "x_factor = %f, thin_length = %ld, draw_length = %ld\n",
	    x_factor, thin_length, draw_length);
    
    /*px = 0;*/
    px = (int)((double)draw_offset * x_factor);
    py = (int)(zero_height - (value * y_factor));
    if (x_factor > 0.2 && thin_length < 16) {
	for (k = 1; k < draw_length; k++) {
	    l = draw_offset + k;
	    if (peak_flag == SP_TRUE) {
		swGetPeakData(window->wave, channel, l * thin_length, &value);
	    } else {
		swGetWaveData(window->wave, channel, l, &value);
	    }
	    x = (int)((double)l * x_factor);
	    y = (int)(zero_height - (value * y_factor));

	    if (y >= draw_min && y <= draw_max) {
		if (py >= 0) {
		    spDrawLine(component, sw_graphics->gx_fg, px, py, x, y);
		}
		py = y;
	    } else {
		py = -1;
	    }
	    px = x;
	}
    } else {
	x = px;
	min = max = py;
	for (k = 1; k < draw_length; k++) {
	    l = draw_offset + k;
	    if (peak_flag == SP_TRUE) {
		swGetPeakData(window->wave, channel, l * thin_length, &value);
	    } else {
		swGetWaveData(window->wave, channel, l, &value);
	    }
	    x = (int)((double)l * x_factor);
	    y = (int)(zero_height - (value * y_factor));
		
	    y = MAX(y, draw_min); y = MIN(y, draw_max);
		
	    if (x == px || l % 2 != 0) {
		if (py < min) min = py;
		if (py > max) max = py;
	    } else {
		if (py < min) {
		    min = MAX(py, draw_min);
		}
		if (py > max) {
		    max = MIN(py, draw_max);
		}
		if (thin_length >= 16) {
		    spFillRectangle(component, sw_graphics->gx_fg, px, min, x - px,  MAX(max - min, 1));
		} else {
		    spDrawLine(component, sw_graphics->gx_fg, px, min, px, max);
		}
		
		px = x;

#if 0
		if (thin_length >= 16) {
		    if (y > zero_height) {
			min = (int)zero_height;
			max = y;
		    } else {
			min = y;
			max = (int)zero_height;
		    }
		} else {
		    min = max = y;
		}
#else
		min = max = y;
#endif
	    }
	    py = y;
	}
	if (x > px) {
	    if (thin_length >= 16) {
		spFillRectangle(component, sw_graphics->gx_fg, px, min, x - px,  MAX(max - min, 1));
	    } else {
		spDrawLine(component, sw_graphics->gx_fg, px, min, px, max);
	    }
	}
    }
    
    return;
}

void swFillRegion(spComponent component, spGraphics graphics, swWindow window,
		  double draw_width, double draw_height, int channel, int start_d, int end_d)
{
    int st, ed;
    int width;
    int window_width;

    if (window == NULL || (start_d < 0 && end_d < 0) || start_d == end_d)
	return;

    /*spLockMutex(sw_graphics->mutex);*/
    
    window_width = (int)draw_width;
    if (start_d < end_d) {
	st = MAX(start_d, 0);
	ed = MIN(end_d, window_width);
    } else {
	st = MAX(end_d, 0);
	ed = MIN(start_d, window_width);
    }

    width = ed - st;

    if (width > 0) {
	spDebug(50, "swFillRegion", "st = %d, width = %d\n", st, width);

	if (channel >= 0) {
	    spFillRectangle(component, graphics, st,
			    (int)((double)channel * draw_height),
			    width, (int)draw_height);
	} else {
#if 0
	    if (draw_height <= 0.0) {
		spFillRectangle(component, graphics, st, 0, width, window->height);
	    } else {
		spFillRectangle(component, graphics, st, 0, width, (int)draw_height);
	    }
#else
	    spFillRectangle(component, graphics, st, 0, width, window->height);
#endif
	}
    }

    /*spUnlockMutex(sw_graphics->mutex);*/
    
    return;
}

void swDrawRegionLabels(spComponent component, swWindow window)
{
    long k;
    int st_d, ed_d;

    if (swIsNoLabel(window) == SP_TRUE || window->draw_label == SP_FALSE)
	return;


#if 0
    if (window->drag_label_type == SW_DRAG_NO_LABEL) {
	swSortLabels(window->wave->labels, &window->active_label_index);
	spDebug(30, "swDrawRegionLabels", "sorted active label %ld\n", window->active_label_index);
    }
#endif
    
    spDebug(30, "swDrawRegionLabels", "num_buffer = %ld\n", window->wave->labels->num_buffer);

    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0
	    && window->wave->labels->label[k].end_time >= 0.0) {
#if 0
	    long st, ed;
	    st = swDimToSamp(window, window->wave->labels->label[k].time);
	    ed = swDimToSamp(window, window->wave->labels->label[k].end_time);
	    st_d = swSampToDisp(window, st);
	    ed_d = swSampToDisp(window, ed);
#else
	    st_d = swDimToDisp(window, window->wave->labels->label[k].time);
	    ed_d = swDimToDisp(window, window->wave->labels->label[k].end_time);
#endif

	    swFillRegion(component, sw_graphics->gx_region_bg, window,
			 (double)swGetDrawWidth(window), window->draw_height, -1, st_d, ed_d);
	}
    }
    
    spDebug(80, "swDrawRegionLabels", "done\n");
    
    return;
}

void swDrawBackground(spComponent component, swWindow window)
{
    if (window == NULL)
	return;

    spFillRectangle(component, sw_graphics->gx_bg, 0, 0, window->width, window->height);
    spDebug(40, "swDrawBackground", "fill rectangle done\n");

#ifndef SW_DRAW_LABEL_CANVAS
    swDrawRegionLabels(component, window);
#endif

    return;
}

void swDrawWaveImage(spComponent component, swWindow window, long draw_pos)
{
    int n;
    long offset = 0, length = 0;
    long draw_offset;
    double data_min, data_max;
    double data_height;
    double zero_offset;
    double x_factor, y_factor;

    spDebug(40, "swDrawWaveImage", "in: created = %d, draw_pos = %ld\n",
	    spIsCreated(component), draw_pos);
    
    if (component == NULL || window == NULL)
	return;

    if (swIsNoWave(window) == SP_TRUE || window->wave->length <= 1) {
	swDrawBackground(component, window);
    } else {
	if (draw_pos < 0) {
	    draw_offset = 0;
	} else {
	    spDebug(40, "swDrawWaveImage", "window->drawn_pos = %ld, wave->length = %ld\n",
		    window->drawn_pos, window->wave->length);
	    draw_offset = MAX(window->drawn_pos, 0);
	    if (draw_offset >= window->wave->length) {
		draw_offset = 0;
		draw_pos = window->wave->length;
	    } else {
		if (swIsWaveRangeAvailable(window->wave) == SP_FALSE) {
		    draw_offset = 0;
		} else if (draw_pos <= 0){
		    draw_pos = window->wave->length - draw_offset;
		}
	    }
	}
	
	spDebug(40, "swDrawWaveImage", "draw_offset = %ld, draw_pos = %ld\n",
		draw_offset, draw_pos);

	data_height = swGetDataHeight(window, &data_min, &data_max);
	
	spDebug(40, "swDrawWaveImage", "data_min = %f, data_max = %f\n",
		data_min, data_max);
	
	if (draw_offset == 0) {
	    swDrawBackground(component, window);
	
	    offset = MIN(window->offset, window->wave->total_length - 2);
	    length = MIN(window->length, window->wave->total_length - offset);
	    length = MAX(length, 2);
	    spDebug(40, "swDrawWaveImage", "offset = %ld, length = %ld\n",
		    offset, length);
	
	    window->draw_width = (double)swGetDrawWidth(window);
	    window->draw_height = (double)window->height / (double)window->wave->num_channel;
	    window->draw_height = MAX(window->draw_height, 2);
	    window->draw_height = MAX(window->draw_height, SW_MIN_DRAW_HEIGHT);
	} else {
	    if (draw_pos > draw_offset) {
		/* unselect region */
		swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d);
	    }
	}
	
	x_factor = window->draw_width / (double)(MAX(window->wave->length - 1, 1));
	
	spDebug(30, "swDrawWaveImage", "width = %f, height = %f\n",
		window->draw_width, window->draw_height);
	spDebug(30, "swDrawWaveImage", "data_height = %f, x_factor = %f\n",
		data_height, x_factor);

	for (n = 0; n < window->wave->num_channel; n++) {
	    if (draw_offset == 0) {
		swDrawHScale(component, window, window->draw_width, window->draw_height,
			     n, offset, length);
		swDrawVScale(component, window, window->draw_width, window->draw_height,
			     n, data_min, data_height, &y_factor, &zero_offset, SP_TRUE);
		spDebug(30, "swDrawWaveImage", "y_factor = %f, zero_offset = %f\n",
			y_factor, zero_offset);
	    } else {
		swDrawVScale(component, window, window->draw_width, window->draw_height,
			     n, data_min, data_height, &y_factor, &zero_offset, SP_FALSE);
	    }
#ifdef SW_SUPPORT_METER
	    swDrawMeter(component, window, window->draw_height, (int)window->draw_width, n, 0.0);
#endif

	    if (draw_pos > draw_offset) {
		swDrawWaveform(component, window, window->draw_height, zero_offset,
			       x_factor, y_factor, n, draw_offset, draw_pos - draw_offset, SP_FALSE);
	    }
	}
	
	if (draw_pos > draw_offset) {
	    window->drawn_pos = draw_pos;

#ifndef SW_DRAW_LABEL_CANVAS
	    swDrawLabels(component, window);
#endif
	
	    /* select region */
	    swRedrawRegion(window);
	}
    }
    
    spDebug(30, "swDrawWaveImage", "done\n");
    
    return;
}

void swDrawWaveImageCB(spComponent component, swWindow window)
{
    if (window == NULL) return;

    spDebug(30, "swDrawWaveImageCB", "in\n");
    
    if (window->image == NULL) {
	spDebug(30, "swDrawWaveImageCB", "initial callback\n");
	window->image = component;
	if (window->wave != NULL) {
	    /* drawing will be performed in the callback */
	    swReadTotalWave(window->wave, SP_FALSE);

	    swDrawWaveImage(window->image, window, 0);
	    swDrawOverview(window);
	} else {
	    swDrawBackground(window->image, window);
	}
	swDrawCursor(window);
    } else if (window->wave == NULL
	       || swIsWaveRangeAvailable(window->wave) == SP_TRUE) {
	spDebug(30, "swDrawWaveImageCB", "drawn_pos = %ld\n", window->drawn_pos);
	if (window->drawn_pos <= 0
	    || window->drawn_pos >= window->wave->length) { /* if drawing has been completed */
	    window->drawn_pos = 0;
	    swDrawWaveImage(component, window, 0);
	}
    }

    spDebug(30, "swDrawWaveImageCB", "done\n");
    
    return;
}

void swCopyImage(swWindow window)
{
    if (window == NULL) return;
    
    /* copy wave image */
    spCopyImage(window->image, window->canvas, 0, 0,
		window->width, window->height, 0, 0);
    
    return;
}

spBool swGetDrawTimeString(swWindow window, double point_f, char *string, spBool align_left)
{
    long point;
    char *format1;
    char format[SP_MAX_MESSAGE];
    static char format_left[] = "%-9";
    static char format_right[] = "%9";
    
    if (window == NULL || swIsNoWave(window) == SP_TRUE) return SP_FALSE;
    
    if (align_left == SP_TRUE) {
	format1 = format_left;
    } else {
	format1 = format_right;
    }
    
    if (window->data_type == SW_FREQ_DATA) {
	if (window->config->freq_format == SW_FREQ_FORMAT_KHZ) {
	    sprintf(format, "%s.6f", format1);
	    sprintf(string, format, point_f / 1000.0);
	} else if (window->config->freq_format == SW_FREQ_FORMAT_POINT) {
	    point = swDimToSamp(window, point_f);
	    sprintf(format, "%sld", format1);
	    sprintf(string, format, point);
	} else {
	    sprintf(format, "%s.3f", format1);
	    sprintf(string, format, point_f);
	}
    } else {
	if (window->config->time_format == SW_TIME_FORMAT_POINT) {
	    point = swDimToSamp(window, point_f);
	    sprintf(format, "%sld", format1);
	    sprintf(string, format, point);
	} else {
	    char buf[SP_MAX_MESSAGE];
	    
	    swGetTimeString(point_f, window->config->time_format, buf);
	    sprintf(format, "%ss", format1);
	    sprintf(string, format, buf);
	}
    }

    return SP_TRUE;
}

void swDrawString(swWindow window)
{
    spBool zero_flag = SP_FALSE;
    int n;
    int sx, sy, swidth, sheight;
    int left_offset;
    long point;
    double top_offset;
    double value;
    double point_f;
    long data_offset;
    char string[SP_MAX_MESSAGE];
    char label[SP_MAX_MESSAGE];

    if (window == NULL || swIsNoWave(window) == SP_TRUE) return;

    top_offset = SW_TIME_STRING_TOP_OFFSET;
    left_offset = SW_TIME_STRING_LEFT_OFFSET;
    if (top_offset < 0) {
	top_offset += window->height;
    }
    if (left_offset < 0) {
	/*left_offset += window->width;*/
	left_offset += (int)window->draw_width;
    }
    if (window->config->scale_flag == SP_TRUE) {
	top_offset = MIN(top_offset, window->height - SW_HSCALE_BOTTOM_SPACING);
	left_offset = MAX(left_offset, SW_HSCALE_LEFT_SPACING);
    }

    if (window->point < 0 || window->point >= window->wave->total_length) {
	point_f = 0.0;
	point = 0;
	zero_flag = SP_TRUE;
    } else {
	point_f = window->point_f;
	point = window->point;
    }

    {
	char title_string[SP_MAX_LINE];
	char time_string[SP_MAX_LINE];
	char dim_string[SP_MAX_LINE];

	swGetDrawTimeString(window, point_f, time_string, SP_FALSE);
	
	if (window->data_type == SW_FREQ_DATA) {
	    strcpy(title_string, "Frequency");
	    if (window->config->freq_format == SW_FREQ_FORMAT_KHZ) {
		strcpy(dim_string, "[kHz]");
	    } else if (window->config->freq_format == SW_FREQ_FORMAT_POINT) {
		strcpy(dim_string, "[points]");
	    } else {
		strcpy(dim_string, "[Hz]");
	    }
	} else {
	    strcpy(title_string, "Time");
	    if (window->config->time_format == SW_TIME_FORMAT_MSEC
		|| window->config->time_format == SW_TIME_FORMAT_FLOORED_MSEC) {
		strcpy(dim_string, "[ms]");
	    } else if (window->config->time_format == SW_TIME_FORMAT_POINT) {
		strcpy(dim_string, "[points]");
	    } else {
		strcpy(dim_string, "[s]");
	    }
	}
	sprintf(string, "%s: %s %s", title_string, time_string, dim_string);
	spDebug(100, "swDrawString", "string = %s\n", string);

	/* draw time string */
	spDrawString(window->canvas, sw_graphics->gx_string,
		     left_offset, (int)top_offset, string);

	if (window->config->draw_selection_times == SP_TRUE) {
	    char end_time_string[SP_MAX_LINE];

	    if (window->sel_st >= 0 && window->sel_ed >= 0) {
		double st_f, ed_f;

		st_f = swSampToDim(window, window->sel_st);
		ed_f = swSampToDim(window, window->sel_ed);
		
		swGetDrawTimeString(window, st_f, time_string, SP_FALSE);
		swGetDrawTimeString(window, ed_f, end_time_string, SP_TRUE);
	    } else {
		swGetDrawTimeString(window, 0.0, time_string, SP_FALSE);
		swGetDrawTimeString(window, 0.0, end_time_string, SP_TRUE);
	    }
	    
	    sprintf(string, "Selection: %s - %s %s", time_string, end_time_string, dim_string);
	    
	    if (spGetStringExtent(sw_graphics->gx_string, string, &sx, &sy, &swidth, &sheight) == SP_TRUE) {
		top_offset -= (-sy + 2);
	    } else {
		top_offset -= SW_TIME_STRING_HEIGHT;
	    }

	    /* draw selection string */
	    spDrawString(window->canvas, sw_graphics->gx_string,
			 left_offset, (int)top_offset, string);
	}

	if (window->config->draw_selection_length == SP_TRUE) {
	    long len = 0;
	    double len_f = 0.0;

	    if (window->sel_st >= 0 && window->sel_ed >= 0) {
		if (window->sel_st < window->sel_ed) {
		    len = window->sel_ed - window->sel_st;
		} else {
		    len = window->sel_st - window->sel_ed;
		}
		len_f = swSampToDim(window, len);
		
		swGetDrawTimeString(window, len_f, time_string, SP_FALSE);
	    } else {
		swGetDrawTimeString(window, 0.0, time_string, SP_FALSE);
	    }
	    
	    if (window->data_type == SW_FREQ_DATA) {
		sprintf(string, "Length: %s %s", time_string, dim_string);
	    } else {
		char freq_string[SP_MAX_LINE];

		if (len_f > 0.0) {
		    sprintf(freq_string, "%.4f", 1.0/len_f);
		} else {
		    sprintf(freq_string, "oo");
		}
		sprintf(string, "Length: %s %s (%9s [Hz])", time_string, dim_string, freq_string);
	    }
	    
	    if (spGetStringExtent(sw_graphics->gx_string, string, &sx, &sy, &swidth, &sheight) == SP_TRUE) {
		top_offset -= (-sy + 2);
	    } else {
		top_offset -= SW_TIME_STRING_HEIGHT;
	    }

	    /* draw selection string */
	    spDrawString(window->canvas, sw_graphics->gx_string,
			 left_offset, (int)top_offset, string);
	}
    }

#ifdef SW_SUPPORT_METER
    for (n = 0; n < window->wave->num_channel; n++) {
	if (swGetPeakData(window->wave, n, window->point, &value) == SP_FALSE) {
	    value = 0.0;
	}
	swDrawMeter(window->canvas, window, window->draw_height,
		    (int)window->draw_width, n, value);
    }
#endif
    
    if (zero_flag == SP_TRUE || swIsWaveProcessing(window->wave) == SP_TRUE) return;

    top_offset = SW_AMP_STRING_TOP_OFFSET;
    left_offset = SW_AMP_STRING_LEFT_OFFSET;
    if (top_offset < 0) {
	top_offset += window->draw_height;
    }
    if (left_offset < 0) {
	left_offset += (int)window->draw_width;
    }
    if (window->config->scale_flag == SP_TRUE) {
	top_offset = MIN(top_offset, window->height - SW_HSCALE_BOTTOM_SPACING);
	left_offset = MAX(left_offset, SW_HSCALE_LEFT_SPACING);
    }

    data_offset = (window->point - window->offset) / window->wave->thin_length;
    
    spDebug(40, "swDrawString", "data_offset = %ld, point = %ld, offset = %ld\n",
	    data_offset, window->point, window->offset);
    spDebug(40, "swDrawString", "length = %ld, total_length = %ld, num_channel = %d\n",
	    window->wave->length, window->wave->total_length, window->wave->num_channel);

    for (n = 0; n < window->wave->num_channel; n++) {
	if (swGetWaveData(window->wave, n, data_offset, &value) == SP_FALSE) {
	    break;
	}

	if (!strnone(window->data_label)) {
	    strcpy(label, window->data_label);
	} else {
	    strcpy(label, "Amplitude");
	}
	
	if (swIsWaveFloat(window->wave) == SP_TRUE) {
	    sprintf(string, "%s:%10.4f", label, value);
	} else if (swIsWaveLong(window->wave) == SP_TRUE) {
	    sprintf(string, "%s:%10.0f", label, value);
	} else {
	    sprintf(string, "%s:%7.0f", label, value);
	}

	if (!strnone(window->data_dim)) {
	    strcat(string, " ");
	    strcat(string, window->data_dim);
	}

	spDebug(50, "swDrawString", "string = %s\n", string);

	/* draw amplitude string */
	spDrawString(window->canvas, sw_graphics->gx_string,
		     left_offset, (int)top_offset, string);

	top_offset += window->draw_height;
    }
    
    return;
}

static int sw_label_string_px = 0;
static int sw_label_string_py = 0;

static int sw_region_label_px = 0;
static int sw_region_label_height = 0;

#define SW_LABEL_STRING_OFFSET 5
/*#define SW_LABEL_STRING_TOP_OFFSET 25*/
#define SW_LABEL_STRING_TOP_OFFSET 10
#define SW_REGION_LABEL_OFFSET 8
#define SW_REGION_LABEL_BOTTOM_OFFSET 23

#define SW_ACTIVE_LABEL_RECT_SIZE 6

void swDrawLabel(spComponent component, int draw_width, int window_height,
		 int st_d, int ed_d, char *string, spBool active_flag)
{
    int cx, cy;
    int x, y;
    int width, height;
    int region_height;
    spGraphics graphics;
    
    spLockMutex(sw_graphics->mutex);
    
    if (ed_d >= 0) {
	graphics = sw_graphics->gx_region;

	spSetGraphicsParams(graphics,
			    SppForeground, sw_graphics->region_line_color,
			    NULL);

	/* draw arrow */
	if (st_d >= 0 && sw_region_label_px > st_d) {
	    region_height = sw_region_label_height + SW_REGION_LABEL_OFFSET;
	    if (region_height > window_height) {
		region_height = SW_REGION_LABEL_BOTTOM_OFFSET;
	    }
	} else {
	    region_height = SW_REGION_LABEL_BOTTOM_OFFSET;
	}
	    
	y = window_height - region_height;
	
	if (ed_d <= draw_width) {
	    spDrawLine(component, graphics, ed_d, 0, ed_d, window_height);
	    
	    if (active_flag == SP_TRUE) {
		spFillRectangle(component, graphics, ed_d - SW_ACTIVE_LABEL_RECT_SIZE/2, 0,
				SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
		spFillRectangle(component, graphics, ed_d - SW_ACTIVE_LABEL_RECT_SIZE/2,
				window_height - SW_ACTIVE_LABEL_RECT_SIZE,
				SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
	    }
	    
	    spSetGraphicsParams(graphics,
				SppForeground, sw_graphics->region_label_color,
				NULL);

	    /* arrow's head */
	    spDrawLine(component, graphics, ed_d, y, ed_d - 9, y - 3);
	    spDrawLine(component, graphics, ed_d, y, ed_d - 9, y + 3);
	    spDrawLine(component, graphics, MAX(st_d, 0), y, ed_d, y);

	    sw_region_label_px = MAX(ed_d, sw_region_label_px);
	    sw_region_label_height = region_height;
	} else {
	    spSetGraphicsParams(graphics,
				SppForeground, sw_graphics->region_label_color,
				NULL);

	    if (st_d <= draw_width) {
		/* arrow's body */
		spDrawLine(component, graphics, MAX(st_d, 0), y, MIN(ed_d, draw_width), y);
	    }
	}
    } else {
	graphics = sw_graphics->gx_label;
    }
    
    if (st_d >= 0 && st_d <= draw_width) {
	spDrawLine(component, graphics, st_d, 0, st_d, window_height);
	
	if (active_flag == SP_TRUE) {
	    spFillRectangle(component, graphics, st_d - SW_ACTIVE_LABEL_RECT_SIZE/2, 0,
			    SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
	    spFillRectangle(component, graphics, st_d - SW_ACTIVE_LABEL_RECT_SIZE/2,
			    window_height - SW_ACTIVE_LABEL_RECT_SIZE,
			    SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
	}
	
	if (!strnone(string)) {
	    spGetStringExtent(graphics, string, &x, &y, &width, &height);
	    
	    cx = st_d;
	    if (sw_label_string_px >= cx) {
		cy = sw_label_string_py + height + SW_LABEL_STRING_OFFSET;
		if (cy > window_height) {
		    cy = height + SW_LABEL_STRING_TOP_OFFSET;
		}
	    } else {
		cy = height + SW_LABEL_STRING_TOP_OFFSET;
	    }
	    
	    sw_label_string_px = cx + width;
	    sw_label_string_py = cy;
	    
	    spDrawLine(component, graphics, cx, cy, sw_label_string_px, cy);
	    spDrawString(component, graphics, st_d, cy - 1, string);
	}
    }

    if (ed_d >= 0) {
	spSetGraphicsParams(sw_graphics->gx_region,
			    SppForeground, sw_graphics->region_color,
			    NULL);
    }
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

void swDrawLabels(spComponent component, swWindow window)
{
    long k;
    int st_d, ed_d;
    spBool active_flag;

    if (swIsNoLabel(window) == SP_TRUE
	|| window->draw_label == SP_FALSE || component == NULL)
	return;

    spDebug(80, "swDrawLabels", "in\n");
    
    sw_label_string_px = 0;
    sw_label_string_py = SW_LABEL_STRING_TOP_OFFSET;
    sw_region_label_px = 0;
    sw_region_label_height = SW_REGION_LABEL_BOTTOM_OFFSET;

    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0) {
	    st_d = swDimToDisp(window, window->wave->labels->label[k].time);

	    if (k == window->active_label_index) {
		active_flag = SP_TRUE;
	    } else {
		active_flag = SP_FALSE;
	    }

#if 0
	    if (window->wave->labels->label[k].end_time >= 0.0) {
		ed_d = swDimToDisp(window, window->wave->labels->label[k].end_time);
		swDrawLabel(component, window->draw_width, window->height,
			    st_d, ed_d, window->wave->labels->label[k].data, active_flag);
	    } else {
		swDrawLabel(component, window->draw_width, window->height,
			    st_d, -1, window->wave->labels->label[k].string, active_flag);
	    }
#else
	    if (window->wave->labels->label[k].end_time >= 0.0) {
		ed_d = swDimToDisp(window, window->wave->labels->label[k].end_time);
	    } else {
		ed_d = -1;
	    }
	    swDrawLabel(component, (int)window->draw_width, window->height,
			st_d, ed_d, window->wave->labels->label[k].string, active_flag);
#endif
	}
    }

    spDebug(80, "swDrawLabels", "done\n");
    
    return;
}

void swRedrawLabels(swWindow window)
{
#ifndef SW_DRAW_LABEL_CANVAS
    swRedrawWave(window);
#else
    swDrawCursor(window);
#endif
    
    return;
}

void swUpdateLabels(swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) return;

    if (swIsWaveLabelNone(window->wave) == SP_TRUE) {
	window->active_label_index = -1;
    } else {
	if (window->drag_label_type == SW_DRAG_NO_LABEL) {
	    swSortLabels(window->wave->labels, &window->active_label_index);
	    spDebug(50, "swUpdateLabels", "sorted active label %ld\n", window->active_label_index);
	}
	
	if (window->active_label_index >= 0
	    && window->wave->labels->label[window->active_label_index].time < 0.0) {
	    window->active_label_index = -1;
	}
    }
    
    swRedrawLabels(window);
    swUpdateLabelList(window->label_list);

    return;
}

void swResetRegion(swWindow window)
{
    if (window == NULL) return;

    window->sel_st = -1;
    window->sel_ed = -1;
    window->sel_st_d = -1;
    window->sel_ed_d = -1;
	
    swSetSelectSenseLevel(window, SP_FALSE);
    
    return;
}

void swSelectRegion(swWindow window, int channel, long start, long end)
{
    if (window == NULL || window->wave == NULL)	return;

    spDebug(50, "swSelectRegion", "in\n");
    
    swLockWindowMutex(window);

    swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d);
    spDebug(50, "swSelectRegion", "reverse done\n");

    if ((start < 0 && end < 0) || 
	(start >= window->wave->total_length && end >= window->wave->total_length)) {
	swResetRegion(window);
    } else {
	if (start < end) {
	    window->sel_st = MAX(start, 0);
	    window->sel_ed = MIN(end, window->wave->total_length - 1);
	} else {
	    window->sel_st = MAX(end, 0);
	    window->sel_ed = MIN(start, window->wave->total_length - 1);
	}
	
	if (window->sel_st == window->sel_ed) {
	    swResetRegion(window);
	} else {
	    window->sel_st_d = swSampToDisp(window, window->sel_st);
	    window->sel_ed_d = swSampToDisp(window, window->sel_ed);
	    
	    if (window->sel_st != window->sel_ed && 
		window->sel_st_d == window->sel_ed_d) {
		window->sel_ed_d += 1;
	    }
	    
	    window->wave->selected_channel = channel;
	    
	    /* redraw region */
	    swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d);

	    swSetSelectSenseLevel(window, SP_TRUE);
	}
    }

    /* draw cursor */
    swDrawCursor(window);
    spRefreshCanvas(window->overview_canvas);
    
    swUnlockWindowMutex(window);
    
    spDebug(50, "swSelectRegion", "done\n");
    
    return;
}

void swSelectRegionCB(spComponent component, swWindow window)
{
    long start, end;

    if (window == NULL || window->wave == NULL) return;
    
    start = window->offset;
    end = start + window->length - 1;
    
    if (start == window->sel_st && end == window->sel_ed) {
	start = -1;
	end = -1;
    }
    
    swSelectRegion(window, -1, start, end);

    return;
}

void swSelectFromHereCB(spComponent component, swWindow window)
{
    int channel;
    long start, end;

    if (window == NULL || window->wave == NULL) return;
    
    start = window->point;
    if (window->sel_ed > start) {
	end = window->sel_ed;
	channel = window->wave->selected_channel;
    } else {
	end = window->wave->total_length - 1;
	channel = -1;
    }
    
    swSelectRegion(window, channel, start, end);

    return;
}

void swSelectToHereCB(spComponent component, swWindow window)
{
    int channel;
    long start, end;

    if (window == NULL || window->wave == NULL) return;
    
    end = window->point;
    if (window->sel_st >= 0 && window->sel_st < end) {
	start = window->sel_st;
	channel = window->wave->selected_channel;
    } else {
	start = 0;
	channel = -1;
    }
    
    swSelectRegion(window, channel, start, end);

    return;
}

/*
 *	select region of other window
 */
void swSelectAllRegionCB(spComponent component, swWindow window)
{
    long k;
    long sel_st, sel_ed;
    double sel_st_f, sel_ed_f;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL || window->wave == NULL
	|| (window->sel_st < 0 && window->sel_ed < 0))	return;

    sel_st_f = swSampToDim(window, window->sel_st);
    sel_ed_f = swSampToDim(window, window->sel_ed);

    /* draw cursor */
    swDrawCursor(window);

    next = window->window;
    for (k = 0;; k++) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && next_window->wave != NULL) {
	    if (next_window->data_type == window->data_type
		&& swIsWaveProcessing(next_window->wave) == SP_FALSE) {
		sel_st = swDimToSamp(next_window, sel_st_f);
		sel_ed = swDimToSamp(next_window, sel_ed_f);

		/* select region */
		swSelectRegion(next_window, -1, sel_st, sel_ed);
	    }
	}
    }

    return;
}

void swSelectAllCB(spComponent component, swWindow window)
{
    long start, end;

    if (window == NULL || window->wave == NULL) return;
    
    start = 0;
    end = window->wave->total_length - 1;
    
    swSelectRegion(window, -1, start, end);

    return;
}

void swSelectNextChannelCB(spComponent component, swWindow window)
{
    int channel;

    if (window == NULL || window->wave == NULL) return;

    
    if ( window->wave->selected_channel >= window->wave->num_channel - 1) {
	channel = -1;
    } else if (window->wave->selected_channel < 0) {
	channel = 0;
    } else {
	channel = window->wave->selected_channel + 1;
    }
    
    swSelectRegion(window, channel, window->sel_st, window->sel_ed);

    return;
}

void swReverseMainRegion(swWindow window, int channel, int start_d, int end_d)
{
    if (window == NULL) return;

    swFillRegion(window->image, sw_graphics->gx_xor, window,
		 (double)swGetDrawWidth(window), window->draw_height, channel, start_d, end_d);
    
    return;
}

void swReverseRegion(swWindow window, int channel, int start_d, int end_d)
{
    if (window == NULL) return;

    swReverseMainRegion(window, channel, start_d, end_d);

    if (window->overview_canvas != NULL) {
	long start, end;
    
	start = swDispToSamp(window, start_d);
	end = swDispToSamp(window, end_d);
	swReverseOverviewRegion(window, channel,
				swSampToOverviewDisp(window, start),
				swSampToOverviewDisp(window, end));
    }
    
    return;
}

void swRedrawRegion(swWindow window)
{
    if (window == NULL || (window->sel_st < 0 && window->sel_ed < 0))
	return;
    
    spDebug(30, "swRedrawRegion", "set_st = %ld, sel_ed = %ld\n",
	    window->sel_st, window->sel_ed);
    
    window->sel_st_d = swSampToDisp(window, window->sel_st);
    window->sel_ed_d = swSampToDisp(window, window->sel_ed);
    
    /* redraw region */
    swReverseMainRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d);
    
    return;
}

void swClearRegion(swWindow window)
{
    if (window == NULL) return;
    
    spDebug(30, "swClearRegion", "sel_st_d = %d, sel_ed_d = %d\n",
	    window->sel_st_d, window->sel_ed_d);

    /* clear region */
    swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d);
    
    window->sel_st_d = -1;
    window->sel_ed_d = -1;
    window->sel_st = -1;
    window->sel_ed = -1;

    /* refresh window */
    swDrawCursor(window);

    if (window->overview_canvas != NULL) {
	spRefreshCanvas(window->overview_canvas);
    }
    
    return;
}

static void drawHCursor(swWindow window, spComponent canvas,
			spGraphics graphics, int point_d)
{
    if (point_d >= 0 && point_d < swGetDrawWidth(window)) {
	/* draw cursor line */
	spDrawLine(canvas, graphics,
		   point_d, 0, point_d, window->height);
    }
    
    return;
}

void swDrawHCursor(swWindow window)
{
    drawHCursor(window, window->canvas,
		sw_graphics->gx_pointer, window->point_d);
    
    return;
}

void swRefreshWindow(swWindow window, spBool draw_cursor)
{
    /* copy wave image */
    swCopyImage(window);

#ifdef SW_DRAW_LABEL_CANVAS
    swDrawRegionLabels(window->canvas, window);
    swDrawLabels(window->canvas, window);
#endif

    if (draw_cursor == SP_TRUE) {
	/* draw horizontal cursor */
	swDrawHCursor(window);

	/* draw string */
	swDrawString(window);
    }
    
    /* refresh window */
    spRefreshCanvas(window->canvas);

    spDebug(100, "swRefreshWindow", "done\n");
    
    return;
}

void swDrawCursor(swWindow window)
{
    swRefreshWindow(window, spIsFalse(swIsProcessing(window)));
    return;
}


static spBool swGetOverviewEdgeDiff(swWindow window, int x, int *stdiff, int *eddiff,
				    spBool *thumb_flag)
{
    int st_d, ed_d;
    
    if (window == NULL || window->wave == NULL
	|| (swIsWaveProcessing(window->wave) == SP_TRUE
	    && (swIsWavePlaying(window->wave) == SP_FALSE
		|| window->sync_play == SP_TRUE
		/*|| swIsWaveThreadSafe(window->wave) == SP_FALSE*/))) {
	return SP_FALSE;
    }

    st_d = swSampToOverviewDisp(window, window->offset);
    ed_d = swSampToOverviewDisp(window, window->offset + window->length - 1);

    *stdiff = x - st_d;
    *eddiff = x - ed_d;

    if (ed_d - st_d <= SW_EDGE_MOTION_RANGE * 2
	&& (x > st_d && x < ed_d)) {
	*thumb_flag = SP_TRUE;
    } else {
	*thumb_flag = SP_FALSE;
    }

    return SP_TRUE;
}

static swOverviewDragType swGetOverviewDragType(swWindow window,
						int stdiff, int eddiff, spBool thumb_flag)
{
    swOverviewDragType drag_type = SW_OVERVIEW_DRAG_NONE;

    if (thumb_flag == SP_FALSE && ABS(stdiff) <= SW_EDGE_MOTION_RANGE / 2
	&& ABS(stdiff) < ABS(eddiff)) {
	drag_type = SW_OVERVIEW_DRAG_START;
    } else if (thumb_flag == SP_FALSE && ABS(eddiff) <= SW_EDGE_MOTION_RANGE / 2) {
	drag_type = SW_OVERVIEW_DRAG_END;
    } else if (stdiff >= 0 && eddiff <= 0) {
	drag_type = SW_OVERVIEW_DRAG_THUMB;
    }
	
    return drag_type;
}

void swOverviewPointerMotionCB(spComponent component, swWindow window)
{
    int x, y;
    int stdiff, eddiff;
    spBool thumb_flag;
    swOverviewDragType drag_type = SW_OVERVIEW_DRAG_NONE;
    static swOverviewDragType prev_drag_type = SW_OVERVIEW_DRAG_NONE;
    
    if (swIsNoWave(window) == SP_TRUE || window->overview_canvas == NULL) return;

    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	spDebug(100, "swOverviewPointerMotionCB",
		"component = %ld, overview_canvas = %ld, canvs = %ld\n",
		component, window->overview_canvas, window->canvas);
	x = MAX(x, 0);
	x = MIN(x, window->overview_width);
	if (swGetOverviewEdgeDiff(window, x, &stdiff, &eddiff, &thumb_flag) == SP_TRUE) {
	    drag_type = swGetOverviewDragType(window, stdiff, eddiff, thumb_flag);
	    if (drag_type == SW_OVERVIEW_DRAG_START) {
		swSetCanvasCursor(component, SP_CURSOR_W_RESIZE);
	    } else if (drag_type == SW_OVERVIEW_DRAG_END) {
		swSetCanvasCursor(component, SP_CURSOR_E_RESIZE);
	    } else if (drag_type == SW_OVERVIEW_DRAG_THUMB) {
		swSetCanvasCursor(component, SP_CURSOR_MOVE);
	    } else {
		drag_type = SW_OVERVIEW_DRAG_NONE;
	    }
	}
    }

    if (drag_type == SW_OVERVIEW_DRAG_NONE && prev_drag_type != SW_OVERVIEW_DRAG_NONE) {
	spUnsetCanvasCursor(component);
    }
    prev_drag_type = drag_type;
    
    return;
}

void swOverviewButtonCB(spComponent component, swWindow window)
{
    int x, y;
    int stdiff, eddiff;
    long offset, length;
    spBool thumb_flag;
    spCallbackReason reason;
    spModifierMask modmask;
    static swOverviewDragType drag_type = SW_OVERVIEW_DRAG_NONE;
    static long offset_diff = 0;
    
    if (swIsNoWave(window) == SP_TRUE || window->overview_canvas == NULL) return;

    spDebug(50, "swOverviewButton", "in\n");
    
    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	x = MAX(x, 0);
	x = MIN(x, window->overview_width);
	
	reason = spGetCallbackReason(component);
	spDebug(50, "swOverviewButton", "reason = %d\n", reason);
	
	if (swGetOverviewEdgeDiff(window, x, &stdiff, &eddiff, &thumb_flag) == SP_TRUE) {
	    switch (reason) {
	      case SP_CR_LBUTTON_PRESS:
		modmask = 0;
		spGetModifierKeyMask(window->window, &modmask);
		
		drag_type = swGetOverviewDragType(window, stdiff, eddiff, thumb_flag);
		if (modmask & SP_COMMAND_MODIFIER_MASK) {
		    offset = swOverviewDispToSamp(window, x);
		    swScrollWindow(window, MAX(offset - window->length / 2, 0), SP_TRUE, SP_FALSE);
		} else if (modmask & SP_EXTEND_MODIFIER_MASK) {
		    if (stdiff < 0 || (eddiff <= 0 && ABS(eddiff) > ABS(stdiff))) {
			offset = swOverviewDispToSamp(window, x);
			length = window->length - (offset - window->offset);
			swZoomRegion(window, offset, length, SP_FALSE);
			drag_type = SW_OVERVIEW_DRAG_START;
		    } else {
			offset = window->offset;
			length = swOverviewDispToSamp(window, x) - offset + 1;
			swZoomRegion(window, offset, length, SP_FALSE);
			drag_type = SW_OVERVIEW_DRAG_END;
		    }
		} else {
		    if (drag_type == SW_OVERVIEW_DRAG_NONE) {
			if (stdiff < 0) {
			    /* backward scroll */
			    swScrollWindow(window, window->offset - MAX(window->length, 1), SP_TRUE, SP_FALSE);
			} else if (eddiff > 0) {
			    /* forward scroll */
			    swScrollWindow(window, window->offset + MAX(window->length, 1), SP_TRUE, SP_FALSE);
			}
		    } else if (drag_type == SW_OVERVIEW_DRAG_THUMB) {
			offset_diff = swOverviewDispToSamp(window, x) - window->offset;
		    }
		}
		break;
		
	      case SP_CR_LBUTTON_MOTION:
		if (drag_type == SW_OVERVIEW_DRAG_START) {
		    offset = swOverviewDispToSamp(window, x);
		    length = window->length - (offset - window->offset);
		    swZoomRegion(window, offset, length, SP_FALSE);
		} else if (drag_type == SW_OVERVIEW_DRAG_END) {
		    offset = window->offset;
		    length = swOverviewDispToSamp(window, x) - offset + 1;
		    swZoomRegion(window, offset, length, SP_FALSE);
		} else if (drag_type == SW_OVERVIEW_DRAG_THUMB) {
		    offset = swOverviewDispToSamp(window, x) - offset_diff;
#if 0
		    if (window->sync_play == SP_FALSE
			|| (offset >= window->offset && offset < window->offset + window->length
			    || offset < window->offset && offset + window->length < window->offset + window->length)) {
			swScrollWindow(window, offset, SP_TRUE, SP_FALSE);
		    }
#else
		    swScrollWindow(window, offset, SP_TRUE, SP_FALSE);
#endif
		}
		break;
		
	      default:
		break;
	    }
	}
	    
	switch (reason) {
	  case SP_CR_LBUTTON_RELEASE:
	    if (drag_type == SW_OVERVIEW_DRAG_AUDIO) {
		swPlayStop(window);
	    }
	    drag_type = SW_OVERVIEW_DRAG_NONE;
	    break;
	    
	  case SP_CR_MBUTTON_PRESS:
	  case SP_CR_RBUTTON_PRESS:
	    offset = swOverviewDispToSamp(window, x);
	    if (swIsWavePlaying(window->wave) == SP_TRUE) {
		swSetPlayRegion(window->wave, offset, window->wave->total_length - offset);
	    } else {
		swPlayRegion(window, offset, window->wave->total_length - 1);
	    }
	    drag_type = SW_OVERVIEW_DRAG_AUDIO;
	    break;
		
	  case SP_CR_MBUTTON_RELEASE:
	  case SP_CR_RBUTTON_RELEASE:
	    swPlayStop(window);
	    drag_type = SW_OVERVIEW_DRAG_NONE;
	    break;
		
	  case SP_CR_MBUTTON_MOTION:
	  case SP_CR_RBUTTON_MOTION:
	    if (drag_type == SW_OVERVIEW_DRAG_AUDIO) {
		offset = swOverviewDispToSamp(window, x);
		spDebug(50, "swOverviewButton", "drag audio: offset = %ld\n", offset);
		swSetPlayRegion(window->wave, offset, window->wave->total_length - offset);
	    }
	    break;
	  default:
	    return;
	}
    }
    
    spDebug(50, "swOverviewButton", "done\n");
    
    return;
}

void swReverseOverviewRegion(swWindow window, int channel, int start_d, int end_d)
{
    if (window == NULL || window->overview_canvas == NULL) return;

    swFillRegion(window->overview_canvas, sw_graphics->gx_xor, window,
		 (double)window->overview_width,
		 (double)window->overview_height / (double)window->wave->num_channel,
		 channel, start_d, end_d);
    
    return;
}

void swUpdateOverview(swWindow window, spBool point_flag,
		      spBool region_flag, spBool refresh_flag)
{
    int point_d = 0;
    int offset_d = 0;
    int length_d = 0;
    
    if (window == NULL || window->overview_canvas == NULL) return;

    /*spLockMutex(sw_graphics->mutex);*/
    
    if (point_flag == SP_TRUE && window->point >= 0) {
	point_d = swSampToOverviewDisp(window, window->point) + 1;
    }
    if (region_flag == SP_TRUE) {
	offset_d = swSampToOverviewDisp(window, window->offset) + 1;
	length_d = MAX(swSampToOverviewDisp(window, window->length) - 2, 1);
    }
    
    if (region_flag == SP_TRUE) {
	spSetGraphicsParams(sw_graphics->gx_overview_xor,
			    SppForeground, "#c0c0c0",
			    SppLineWidth, 2,
			    NULL);
	spDrawRectangle(window->overview_canvas, sw_graphics->gx_overview_xor,
			offset_d, 1, length_d, window->overview_height - 2);
    }
    
    if (point_flag == SP_TRUE && window->point >= 0) {
	spPixel pixel;
	
	pixel = spGetForegroundPixel(sw_graphics->gx_bg) ^ spGetForegroundPixel(sw_graphics->gx_pointer);
	spSetGraphicsParams(sw_graphics->gx_overview_xor,
			    SppForegroundPixel, pixel,
			    SppLineWidth, 1,
			    NULL);
	spDrawLine(window->overview_canvas, sw_graphics->gx_overview_xor,
		   point_d, 0, point_d, window->overview_height);
    }

    spSetGraphicsParams(sw_graphics->gx_overview_xor,
			SppForeground, "max",
			SppLineWidth, 1,
			NULL);
    
    /*spUnlockMutex(sw_graphics->mutex);*/
    
    if (refresh_flag == SP_TRUE) {
	spRefreshCanvas(window->overview_canvas);
    }

    return;
}

void swDrawOverview(swWindow window)
{
    int n;
    double zero_offset;
    double x_factor, y_factor;
    double data_min, data_max;
    double data_height;
    double draw_width, draw_height;
    
    if (window == NULL || window->overview_canvas == NULL) return;

    spGetSize(window->overview_canvas, &(window->overview_width), &(window->overview_height));

    spFillRectangle(window->overview_canvas, sw_graphics->gx_bg,
		    0, 0, window->overview_width, window->overview_height);

    if (swIsWavePeakAvailable(window->wave) == SP_FALSE) {
	return;
    }
    
    draw_width = (double)window->overview_width;
    draw_height = (double)window->overview_height / (double)window->wave->num_channel;
    draw_height = MAX(draw_height, 2);

    spDebug(30, "swDrawOverview", "peak_buf_length = %ld, draw_width = %f, draw_height = %f\n",
	    window->wave->peak_buf_length, draw_width, draw_height);
    
    data_height = swGetDataHeight(window, &data_min, &data_max);

    x_factor = draw_width / (double)(MAX(window->wave->peak_buf_length - 1, 1));
    spDebug(30, "swDrawOverview", "data_min = %f, data_max = %f, data_height = %f, x_factor = %f\n",
	    data_min, data_max, data_height, x_factor);
    
    for (n = 0; n < window->wave->num_channel; n++) {
#if 1
	swDrawHScale(window->overview_canvas, window, draw_width, draw_height,
		     n, 0, window->wave->total_length);
#endif
	swDrawVScale(window->overview_canvas, window, draw_width, draw_height,
		     n, data_min, data_height, &y_factor, &zero_offset, SP_FALSE);
	
	spDebug(30, "swDrawOverview", "n = %d, y_factor = %f, zero_offset = %f\n",
		n, y_factor, zero_offset);
	
	swDrawWaveform(window->overview_canvas, window, draw_height, zero_offset,
		       x_factor, y_factor, n, 0, window->wave->peak_buf_length, SP_TRUE);
    }

    swUpdateOverview(window, SP_TRUE, SP_TRUE, SP_TRUE);

    swReverseOverviewRegion(window, window->wave->selected_channel,
			    swSampToOverviewDisp(window, window->sel_st),
			    swSampToOverviewDisp(window, window->sel_ed));
    
    return;
}

void swDrawWave(swWindow window)
{
    if (window == NULL || spIsCreated(window->canvas) == SP_FALSE)
	return;

    window->drawn_pos = 0;
    
    spGetSize(window->canvas, &(window->width), &(window->height));

    spDebug(30, "swDrawWave", "%s: width = %d, height = %d\n",
	    window->name, window->width, window->height);
    
    if (window->image == NULL) {
	spCreateImage(window->canvas, "waveImage",
		      window->width, window->height,
		      SppCallbackFunc, swDrawWaveImageCB,
		      SppCallbackData, window,
		      NULL);
    } else {
	swUpdateMeterWidth(window);
	
	spRedrawImage(window->image, window->width, window->height);

	window->point_d = swSampToDisp(window, window->point);
	swDrawCursor(window);
    }
	
    spDebug(30, "swDrawWave", "done\n");
    
    return;
}

void swDrawOverviewCB(spComponent component, swWindow window)
{
    if (window == NULL) return;

    spDebug(30, "swDrawOverviewCB", "reason = %d\n", spGetCallbackReason(component));
    
    if (window->overview_canvas == NULL) {
	window->overview_canvas = component;
	spAddCallback(window->overview_canvas, SP_BUTTON_MOTION_CALLBACK |
		      SP_BUTTON_PRESS_CALLBACK | SP_BUTTON_RELEASE_CALLBACK,
		      (spCallbackFunc)swOverviewButtonCB, (void *)window);
	spAddCallback(window->overview_canvas, SP_POINTER_MOTION_CALLBACK,
		      (spCallbackFunc)swOverviewPointerMotionCB, (void *)window);
	spAddCallback(window->overview_canvas, SP_KEY_PRESS_CALLBACK,
		      (spCallbackFunc)swKeyPressCB, (void *)window);
    }

    swDrawOverview(window);
    
    return;
}

void swDrawWaveCB(spComponent component, swWindow window)
{
    spComponent menu, menu_item, sub_menu;
    
    if (window == NULL) return;
    
    spDebug(30, "swDrawWaveCB", "in\n");

    if (window->canvas == NULL) {
	spDebug(30, "swDrawWaveCB", "create popup menu\n");
	window->canvas = component;
	spAddCallback(window->canvas, SP_BUTTON_MOTION_CALLBACK |
		      SP_BUTTON_PRESS_CALLBACK | SP_BUTTON_RELEASE_CALLBACK,
		      (spCallbackFunc)swButtonMotionCB, (void *)window);
	spAddCallback(window->canvas, SP_POINTER_MOTION_CALLBACK,
		      (spCallbackFunc)swMoveCursorCB, (void *)window);
	spAddCallback(window->canvas, SP_KEY_PRESS_CALLBACK,
		      (spCallbackFunc)swKeyPressCB, (void *)window);

	menu = spCreatePopupMenu(window->canvas, "popupMenu",
				 NULL);
	menu_item = spAddMenuItem(menu, "playWave",
				  SppTitle, SW_MENU_PLAY_LABEL,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayRegionCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "playWindow",
				  SppTitle, SW_MENU_PLAY_WINDOW_LABEL,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_WINDOW_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "playFile",
				  SppTitle, SW_MENU_PLAY_FILE_LABEL,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayFileCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_FILE_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "playStop",
				  SppTitle, SW_MENU_PLAY_STOP_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_PLAY_TIME_DATA,
				  SppCallbackFunc, swPlayStopCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_STOP_SHORTCUT,
				  NULL);
	spAddMenuSeparator(menu, "popupMenuPlaySeparator", NULL);
	menu_item = spAddMenuItem(menu, "zoomIn",
				  SppTitle, SW_MENU_ZOOM_IN_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomInCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_IN_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "zoomOut",
				  SppTitle, SW_MENU_ZOOM_OUT_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomOutCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_OUT_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "zoomFullOut",
				  SppTitle, SW_MENU_ZOOM_FULL_OUT_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomFullOutCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_FULL_OUT_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "zoomRegion",
				  SppTitle, SW_MENU_ZOOM_REGION_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_WAVE,
				  SppCallbackFunc, swZoomRegionCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_REGION_SHORTCUT,
				  NULL);
	spAddMenuSeparator(menu, "popupMenuZoomSeparator", NULL);
	menu_item = spAddMenuItem(menu, "alignWindow",
				  SppTitle, SW_MENU_ALIGN_WINDOW_LABEL,
				  SppGroupId, SW_WINDOW_GROUP_ID,
				  SppSenseLevel, SW_STATE_STOP_SOME_WINDOWS,
				  SppCallbackFunc, swAlignWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ALIGN_WINDOW_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "selectAllRegion",
				  SppTitle, SW_MENU_SELECT_ALL_REGION_LABEL,
				  SppGroupId, SW_WINDOW_GROUP_ID,
				  SppSenseLevel, SW_STATE_SOME_WINDOWS,
				  SppCallbackFunc, swSelectAllRegionCB,
				  SppCallbackData, window,
				  SppShortcut, SW_SELECT_ALL_REGION_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "selectFromHere",
				  SppTitle, SW_MENU_SELECT_FROM_HERE_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swSelectFromHereCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(menu, "selectToHere",
				  SppTitle, SW_MENU_SELECT_TO_HERE_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swSelectToHereCB,
				  SppCallbackData, window,
				  NULL);

#ifdef SW_SUPPORT_CLIPBOARD
	if (swIsClipboardWindow(window) == SP_FALSE) {
	    spAddMenuSeparator(menu, "popupMenuEditSeparator", NULL);
	    sub_menu = spAddSubMenu(menu, "editPopupMenu",
				    SppTitle, SW_MENU_EDIT_LABEL,
				    NULL);
	    menu_item = spAddMenuItem(sub_menu, "cutWave",
				      SppTitle, SW_MENU_CUT_LABEL,
				      SppGroupId, SW_CLIPBOARD_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_SELECT_TIME_CHANNELS,
				      SppCallbackFunc, swCutWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_CUT_SHORTCUT,
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "copyWave",
				      SppTitle, SW_MENU_COPY_LABEL,
				      SppGroupId, SW_CLIPBOARD_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
				      SppCallbackFunc, swCopyWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_COPY_SHORTCUT,
				      NULL);
	    spAddMenuSeparator(sub_menu, "popupSubMenuClipboardSeparator", NULL);
	    menu_item = spAddMenuItem(sub_menu, "paste",
				      SppTitle, SW_MENU_PASTE_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				      SppCallbackFunc, swPasteWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_PASTE_SHORTCUT,
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "mix",
				      SppTitle, SW_MENU_MIX_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				      SppCallbackFunc, swMixWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_MIX_SHORTCUT,
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "insert",
				      SppTitle, SW_MENU_INSERT_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				      SppCallbackFunc, swInsertWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_INSERT_SHORTCUT,
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "replace",
				      SppTitle, SW_MENU_REPLACE_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD_SELECTED,
				      SppCallbackFunc, swReplaceWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_REPLACE_SHORTCUT,
				      NULL);
	    spAddMenuSeparator(sub_menu, "popupSubMenuEditSeparator", NULL);
	    menu_item = spAddMenuItem(sub_menu, "valueChange",
				      SppTitle, SW_MENU_CHANGE_VALUE_LABEL,
				      SppGroupId, SW_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_TIME_DATA,
				      SppCallbackFunc, swPopupValueChangeDialogCB,
				      SppCallbackData, window,
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "insertPause",
				      SppTitle, SW_MENU_INSERT_PAUSE_LABEL,
				      SppGroupId, SW_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_TIME_DATA,
				      SppCallbackFunc, swPopupInsertPauseDialogCB,
				      SppCallbackData, window,
				      NULL);
	}
#endif
	
	spAddMenuSeparator(menu, "popupMenuLabelSeparator", NULL);
	sub_menu = spAddSubMenu(menu, "labelPopupMenu",
				SppTitle, SW_MENU_LABEL_LABEL,
				NULL);
	menu_item = spAddMenuItem(sub_menu, "insertSimpleLabel",
				  SppTitle, SW_MENU_INSERT_SIMPLE_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swInsertSimpleLabelCB,
				  SppCallbackData, window,
				  SppShortcut, SW_INSERT_SIMPLE_LABEL_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "insertValueLabel",
				  SppTitle, SW_MENU_INSERT_VALUE_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swInsertValueLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "insertLabel",
				  SppTitle, SW_MENU_INSERT_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swInsertLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "changeLabel",
				  SppTitle, SW_MENU_CHANGE_LABEL_LABEL,
				  SppGroupId, SW_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppCallbackFunc, swChangeLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "eraseLabel",
				  SppTitle, SW_MENU_ERASE_LABEL_LABEL,
				  SppGroupId, SW_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppShortcut, SW_ERASE_LABEL_SHORTCUT,
				  SppCallbackFunc, swEraseLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "eraseLabelRegion",
				  SppTitle, SW_MENU_ERASE_LABEL_REGION_LABEL,
				  SppGroupId, SW_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swEraseLabelRegionCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "selectBetweenLabels",
				  SppTitle, SW_MENU_SELECT_BETWEEN_LABELS_LABEL,
				  /*SppGroupId, SW_LABEL_GROUP_ID,*/
				  SppGroupId, SW_NORMAL_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swSelectBetweenLabelsCB,
				  SppCallbackData, window,
				  NULL);
	spAddMenuSeparator(sub_menu, "popupMenuRegionLabelSeparator", NULL);
	menu_item = spAddMenuItem(sub_menu, "setRegionLabel",
				  SppTitle, SW_MENU_SET_REGION_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_SELECT_WAVE,
				  SppShortcut, SW_SET_REGION_LABEL_SHORTCUT,
				  SppCallbackFunc, swSetRegionLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "setRegionLabelAsRegion",
				  SppTitle, SW_MENU_SET_REGION_LABEL_AS_REGION_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppCallbackFunc, swSetRegionLabelAsRegionCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "catRegionLabelsRegion",
				  SppTitle, SW_MENU_CAT_REGION_LABELS_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppShortcut, SW_CAT_REGION_LABELS_SHORTCUT,
				  SppCallbackFunc, swCatRegionLabelsCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "divideRegionLabelRegion",
				  SppTitle, SW_MENU_DIVIDE_REGION_LABEL_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swDivideRegionLabelCB,
				  SppCallbackData, window,
				  NULL);
	
#ifdef SW_USE_ANALYSIS
	spAddMenuSeparator(menu, "popupMenuAnalysisSeparator", NULL);
	sub_menu = spAddSubMenu(menu, "analysisPopupMenu",
				SppTitle, SW_MENU_ANALYSIS_LABEL,
				NULL);
	menu_item = spAddMenuItem(sub_menu, "analysisRegion",
				  SppTitle, SW_MENU_ANALYSIS_REGION_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
				  SppShortcut, SW_ANALYSIS_REGION_SHORTCUT,
				  SppCallbackFunc, swAnalysisRegionCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "analysisWide",
				  SppTitle, SW_MENU_ANALYSIS_WIDE_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_TIME_DATA,
				  SppCallbackFunc, swAnalysisWideCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "analysisNarrow",
				  SppTitle, SW_MENU_ANALYSIS_NARROW_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_TIME_DATA,
				  SppCallbackFunc, swAnalysisNarrowCB,
				  SppCallbackData, window,
				  NULL);
#endif
	
	swSetSenseLevel(window);
    }

    swDrawWave(window);
    
    spDebug(30, "swDrawWaveCB", "done\n");
    
    return;
}

void swRedrawWave(swWindow window)
{
    if (window == NULL) return;

    spDebug(20, "swRedrawWave", "in\n");

    /* draw wave */
    swDrawWave(window);
    
    /* set parameter value of window  */
    swSetWindowValue(window);

    spDebug(20, "swRedrawWave", "done\n");
    
    return;
}

void swReloadWave(swWindow window, spBool in_thread)
{
    if (window == NULL) return;

    spDebug(20, "swReloadWave", "in: in_thread = %d\n", in_thread);

    window->drawn_pos = 0;
    swSetWindowValue(window);
    
    /* read wave */
    spDebug(20, "swReloadWave", "read: offset = %ld, length = %ld\n",
	    window->offset, window->length);
    swReadWave(window->wave, in_thread, window->offset, window->length);
    
    swDrawWave(window);

    spDebug(20, "swReloadWave", "done\n");

    return;
}

void swDrawAllWave(swWindow window)
{
    spComponent next = NULL;
    swWindow next_window = NULL;
    
    if (window == NULL) return;

    spDebug(10, "swDrawAllWave", "in\n");
    
    swDrawWave(window);
    swDrawOverview(window);
    
    next = window->window;
    while (1) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}

	if ((next_window = (swWindow)spGetUserData(next)) != NULL) {
	    swDrawWave(next_window);
	    swDrawOverview(next_window);
	}
    }

    return;
}

void swAlignWindowCB(spComponent component, swWindow window)
{
    long k;
    long offset, length;
    double offset_f, length_f;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL || window->wave == NULL) return;

    offset_f = swSampToDim(window, window->offset);
    length_f = swSampToDim(window, window->length);
    
    next = window->window;
    for (k = 0;; k++) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && next_window->wave != NULL) {
	    if (next_window->data_type == window->data_type) {
		offset = swDimToSamp(next_window, offset_f);
		length = swDimToSamp(next_window, length_f);
		length = MIN(length, next_window->wave->total_length - offset);
	    
		swZoomRegion(next_window, offset, length, SP_FALSE);
	    }
	}
    }
    
    return;
}

void swZoomRegion(swWindow window, long offset, long length, spBool in_thread)
{
    if (window->process_flag == SP_TRUE) return;
    
    spDebug(50, "swZoomRegion", "in\n");
    
    if (offset >= 0 && length > 1 && 
	offset + length <= window->wave->total_length &&
	!(offset == window->offset && length == window->length)) {
	window->process_flag = SP_TRUE;
	swLockWindowMutex(window);
	
	swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_FALSE);
	
	window->offset = offset;
	window->length = length;
	
	spDebug(10, "swZoomRegion", "zoom: %ld, %ld\n", window->offset, window->length);

	swReloadWave(window, in_thread);
	
	swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_TRUE);
	
	swUnlockWindowMutex(window);
	window->process_flag = SP_FALSE;
    }
    
    spDebug(50, "swZoomRegion", "done\n");
    
    return;
}

void swZoomRegionCB(spComponent component, swWindow window)
{
    long offset, length;

    if (window != NULL && window->wave != NULL) { 
	offset = window->sel_st;
	length = window->sel_ed - window->sel_st + 1;

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

void swZoomInCB(spComponent component, swWindow window)
{
    long offset, length;

    if (window != NULL && window->wave != NULL) {
	length = window->length / 2;
	length = MAX(length, 2L);

	if (swIsProcessing(window) == SP_TRUE) {
	    offset = window->offset + window->length / 2 - length / 2;
	} else {
	    offset = window->point - length / 2;
	}
	offset = MAX(offset, 0L);
	offset = MIN(offset, window->wave->total_length - length);

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

void swZoomOutCB(spComponent component, swWindow window)
{
    long offset, length;

    if (window != NULL && window->wave != NULL) {
	length = window->length * 2L;
	length = MIN(length, window->wave->total_length);

	if (swIsProcessing(window) == SP_TRUE) {
	    offset = window->offset + window->length / 2 - length / 2;
	} else {
	    offset = window->point - length / 2;
	}
	offset = MAX(offset, 0L);
	offset = MIN(offset, window->wave->total_length - length);

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

void swZoomFullOutCB(spComponent component, swWindow window)
{
    long offset, length;

    if (window != NULL && window->wave != NULL) {
	length = window->wave->total_length;
	offset = 0L;

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

spBool swScrollWindow(swWindow window, long offset, spBool move_cursor, spBool in_thread)
{
    spBool flag = SP_FALSE;

    if (window->process_flag == SP_TRUE) return SP_FALSE;
    
    spDebug(50, "swScrollWindow", "in_thread = %d\n", in_thread);
    
    window->process_flag = SP_TRUE;
    swLockWindowMutex(window);

    offset = MIN(offset, window->wave->total_length - window->length);
    offset = MAX(offset, 0);
    
    if (offset != window->offset) {
	spDebug(50, "swScrollWindow", "window->offset = %ld, offset = %ld\n",
		window->offset, offset);
	
	swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_FALSE);
	
	window->offset = offset;

	spDebug(50, "swScrollWindow", "reloading: in_thread = %d\n", in_thread);
	swReloadWave(window, in_thread);
	spDebug(50, "swScrollWindow", "reloading done: in_thread = %d\n", in_thread);

	if (move_cursor == SP_TRUE && swIsProcessing(window) == SP_FALSE) {
	    swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_FALSE);
	    swUpdatePoint(window, window->offset + window->length / 2, SP_TRUE);
	} else {
	    swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_TRUE);
	}
	
	flag = SP_TRUE;
    }
    
    swUnlockWindowMutex(window);
    window->process_flag = SP_FALSE;

    if (flag == SP_TRUE && move_cursor == SP_TRUE) {
	swMoveAllCursor(window);
	/*swMoveCursor(window, window->offset + window->length / 2, SP_TRUE);*/
    }
	
    spDebug(50, "swScrollWindow", "done\n");
	
    return flag;
}

void swScrollCB(spComponent component, swWindow window)
{
    int value;
    
    spDebug(80, "swScrollCB", "in\n");
	
    if (window != NULL && window->wave != NULL) {
	if (spGetSliderValue(component, &value) == SP_FALSE) return;
	spDebug(80, "swScrollCB", "value = %d\n", value);

	swScrollWindow(window, (long)value * window->scroll_coef, SP_TRUE, SP_FALSE);
    }

    spDebug(80, "swScrollCB", "done\n");
    
    return;
}

void swForwardCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindow(window, window->offset + MAX(window->length / 4, 1), SP_TRUE, SP_FALSE);
    }
    
    return;
}

void swBackwardCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindow(window, window->offset - MAX(window->length / 4, 1), SP_TRUE, SP_FALSE);
    }

    return;
}

void swGoHeadCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindow(window, 0, SP_TRUE, SP_FALSE);
    }

    return;
}

void swGoTailCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindow(window, window->wave->total_length - window->length, SP_TRUE, SP_FALSE);
    }

    return;
}

void swNextWindowCB(spComponent component, swWindow window)
{
    swWindow next_window;
    
    if ((next_window = swGetNextWindow(window)) != NULL) {
	spPopupWindow(next_window->window);
    }
    
    return;
}

void swPrevWindowCB(spComponent component, swWindow window)
{
    swWindow prev_window;

    if ((prev_window = swGetPrevWindow(window)) != NULL) {
	spPopupWindow(prev_window->window);
    }
    
    return;
}
