/*
 *	swEdit.c
 *
 *	Last modified: <2002-02-17 17:17:28 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>

#include "swWindow.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swLabelDialog.h"
#include "swWave.h"

spBool swGetEdge(swWindow window, long st, long ed, long *offset, long *length)
{
    if (swIsNoWave(window) == SP_TRUE ||
	(offset == NULL && length == NULL)) return SP_FALSE;
    
    spDebug(30, "swGetEdge", "in\n");
    
    /* get edge */
    if ((st < 0 && ed < 0) || st == ed) {
	return SP_FALSE;
    } else if (st >= window->wave->total_length && ed >= window->wave->total_length) {
	return SP_FALSE;
    } else if (st < ed) {
	st = MAX(st, 0);
	st = MIN(st, window->wave->total_length - 1);
	ed = MAX(ed, st + 1);
	ed = MIN(ed, window->wave->total_length - 1);

	if (offset != NULL) *offset = st;
	if (length != NULL) *length = ed - st + 1;
    } else {
	ed = MAX(ed, 0);
	ed = MIN(ed, window->wave->total_length - 1);
	st = MAX(st, ed + 1);
	st = MIN(st, window->wave->total_length - 1);

	if (offset != NULL) *offset = ed;
	if (length != NULL) *length = st - ed + 1;
    }

    spDebug(10, "swGetEdge", "st = %ld, ed = %ld\n", st, ed);
    
    return SP_TRUE;
}

#ifdef SW_SUPPORT_EDIT
swWindow swExtractWindowAt(swWindow window, long st, long ed, spBool autosave_flag,
			   int x, int y)
{
    long offset;
    long length;
    swWave wave;
    swWindow new_window;

    if (swIsNoWave(window) == SP_TRUE) return NULL;

    /* get edge */
    if (swGetEdge(window, st, ed, &offset, &length) != SP_TRUE) {
	spDisplayError(window->window, NULL, SW_REGION_ERROR_MESSAGE);
	return NULL;
    }

#ifdef SW_SUPPORT_AUTOSAVE
    if (autosave_flag == SP_TRUE) {
	if (swGetNumRegionLabel(window->wave) > 0
	    && window->autosave_started == SP_FALSE) {
	    /* popup yes-no dialog */
	    if (spCreateMessageBox(window->window, NULL,
				   SW_AUTOSAVE_CLEAR_LABEL_QUESTION_MESSAGE,
				   SppDialogType, SP_QUESTION_DIALOG,
				   SppMessageBoxButtonType, SP_MB_YES_NO,
				   NULL) == SP_DR_NO) {
		return NULL;
	    }
	}
    }
#endif

    /* get new wave */
    if ((wave = swExtractWave(window->wave, offset, length)) == NULL) {
	spDisplayError(window->window, NULL, SW_EDIT_ERROR_MESSAGE);
	return NULL;
    }

    /* create new window */
    new_window = swCreateWaveWindowAt(wave, window->config, window->data_type, x, y);
    new_window->wave->edit_flag = SP_TRUE;
    
    if (autosave_flag == SP_FALSE) {
	/* extract label */
	swExtractLabel(new_window, window, offset, length);
    }

    /* set window title */
    swSetWindowTitle(new_window);

    swDrawOverview(new_window);
    
    return new_window;
}

swWindow swExtractWindow(swWindow window, long st, long ed, spBool autosave_flag)
{
    return swExtractWindowAt(window, st, ed, autosave_flag, -1, -1);
}

void swExtractWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swExtractWindow(window, window->sel_st, window->sel_ed, SP_FALSE);
    }
    return;
}

#ifdef SW_SUPPORT_AUTOSAVE
static swAutosaveIdType getAutosaveId(swWindow window, int serial_number, char *autosave_id)
{
    swAutosaveIdType autosave_id_type;
    
    autosave_id_type = window->config->autosave_id_type;
    
    if (autosave_id_type == SW_AUTOSAVE_ID_EDGE_POINTS) {
	sprintf(autosave_id, "%ld-%ld", window->sel_st, window->sel_ed);
    } else {
	switch (autosave_id_type) {
	  case SW_AUTOSAVE_ID_2DIGIT_SERIAL:
	    sprintf(autosave_id, "%.2d", serial_number);
	    break;
	  case SW_AUTOSAVE_ID_3DIGIT_SERIAL:
	    sprintf(autosave_id, "%.3d", serial_number);
	    break;
	  case SW_AUTOSAVE_ID_4DIGIT_SERIAL:
	    sprintf(autosave_id, "%.4d", serial_number);
	    break;
	  default:		/* SW_AUTOSAVE_ID_SERIAL */
	    sprintf(autosave_id, "%d", serial_number);
	    break;
	}
    }
    
    return autosave_id_type;
}

static void getCoreName(swWindow window, char *corename, char *suffix)
{
    char buf[SP_MAX_PATHNAME];
    char dirname[SP_MAX_PATHNAME];
    
    /* remove original suffix */
    spStrCopy(buf, SP_MAX_PATHNAME, window->wave->orig_filename);
    spRemoveSuffix(buf, suffix);
    
    /* get new filename */
    if (spIsDir(window->config->autosave_dir) == SP_TRUE) {
	spStrCopy(dirname, SP_MAX_PATHNAME, window->config->autosave_dir);
	spAddDirSeparator(dirname);
		
	sprintf(corename, "%s%s", dirname, spGetBaseName(buf));
    } else {
	spStrCopy(corename, SP_MAX_PATHNAME, buf);
    }

    return;
}

static void getAutosaveFileName(swWindow window, char *filename, char *corename)
{
    int serial_number;
    swAutosaveIdType autosave_id_type;
    char autosave_id[SP_MAX_PATHNAME];
    char suffix[SP_MAX_PATHNAME];
    
    for (serial_number = 1;; serial_number++) {
	/* get autosave ID */
	autosave_id_type = getAutosaveId(window, serial_number, autosave_id);
	
	/* get core name */
	getCoreName(window, corename, suffix);
	
	/* get new filename */
	sprintf(filename, "%s%s%s%s", corename, window->config->autosave_id_prefix,
		autosave_id, suffix);

	if (autosave_id_type == SW_AUTOSAVE_ID_EDGE_POINTS
	    || spIsFile(filename) == SP_FALSE) {
	    break;
	}
    }

    return;
}

void swExtractAutosaveWindowAt(swWindow window, int x, int y)
{
    double point_f, end_point_f;
    swWindow new_window;
    char *string;
    char corename[SP_MAX_PATHNAME];
    char filename[SP_MAX_PATHNAME];
    char labelfile[SP_MAX_PATHNAME];
    
    if (window != NULL) {
	if (strnone(window->wave->orig_filename)) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_SAVE_FIRST_ERROR_MESSAGE);
	    return;
	}
	
	if ((new_window = swExtractWindowAt(window, window->sel_st, window->sel_ed,
					    SP_TRUE, x, y)) != NULL) {
	    /* get autosave filename */
	    getAutosaveFileName(window, filename, corename);
	    spDebug(10, "swExtractSaveWindowCB", "filename = %s\n", filename);

	    swSetMouseCursor(new_window, SP_CURSOR_WAIT);
	    
	    /* save wave */
	    if (swWriteWave(filename, NULL, NULL, new_window->wave,
			    SP_FALSE, SP_FALSE) == SP_FALSE) {
		spDisplayError(new_window->window, SW_ERROR_TITLE, SW_SAVE_ERROR_MESSAGE, filename);
	    } else {
		/* set title of new window */
		swSetWindowTitle(new_window);
		
		/* set filename */
		sprintf(labelfile, "%s%s", corename, window->config->autosave_label_suffix);

		if (spIsFile(labelfile) == SP_TRUE) {
		    if (window->autosave_started == SP_TRUE
			&& streq(swGetRegionLabelFileName(window->wave->labels), labelfile)) {
		    } else {
			/* read labels */
			swReadLabel(window->wave, labelfile, SW_TIME_FORMAT_SEC, SP_TRUE);
		    }
		} else {
		    /* clear labels */
		    swClearLabels(window->wave->labels, SW_LABEL_TYPE_REGION);
		    swSetLabelFileName(window->wave->labels, labelfile);
		}

		window->autosave_started = SP_TRUE;

		if (window->config->autosave_label_fullpath == SP_TRUE) {
		    string = filename;
		} else {
		    string = spGetBaseName(filename);
		}

		/* get end edge */
		end_point_f = swSampToDim(window, window->sel_ed);

		/* add label */
		point_f = swSampToDim(window, window->sel_st);
		swAddLabel(window->wave, point_f, end_point_f, string, NULL);
		
		/* write label */
		swWriteLabel(window->wave, window->name, labelfile,
			     SW_TIME_FORMAT_SEC, SP_TRUE);

		/* redraw wave */
		swSetSenseLevel(window);
		swRedrawWave(window);
	    }
	    swSetMouseCursor(new_window, SP_CURSOR_UNKNOWN);
	}
    }
    
    return;
}

void swExtractAutosaveWindow(swWindow window)
{
    swExtractAutosaveWindowAt(window, -1, -1);
    return;
}

void swExtractAutosaveWindowCB(spComponent component, swWindow window)
{
    swExtractAutosaveWindow(window);
    return;
}
#endif

#undef SW_UNDO_RELOAD

static spBool undoWindow(swWindow window, spBool redo_flag)
{
    long edit_offset, edit_length;
    spBool flag = SP_FALSE;
    
    if (window != NULL) {
	if (redo_flag == SP_TRUE) {
	    edit_offset = window->wave->edit_offset;
	    edit_length = window->wave->edit_length;
	    flag = swRedoWave(&window->wave);
	} else {
	    flag = swUndoWave(&window->wave);
	    edit_offset = window->wave->edit_offset;
	    edit_length = window->wave->edit_length;
	}
	if (flag == SP_TRUE) {
	    window->offset = window->wave->offset;
	    window->length = window->wave->data_length;
#ifdef SW_UNDO_RELOAD
	    swResetWindow(window, SP_FALSE);
#else
	    swRedrawWindow(window);
	    swDrawOverview(window);
#endif
	    swUpdateLabels(window);
	    if (edit_length > 0) {
		swSelectRegion(window, window->wave->selected_channel, edit_offset,
			       edit_offset + edit_length - 1);
	    }
	}
    }

    return flag;
}

void swUndoWindowCB(spComponent component, swWindow window)
{
    undoWindow(window, SP_FALSE);
    return;
}

void swRedoWindowCB(spComponent component, swWindow window)
{
    undoWindow(window, SP_TRUE);
    return;
}

static spBool checkAmplitude(swWindow window, double value)
{
    if ((value > 1.0 || value <= -1.0)) {
	/* overflow check */
	if (swIsOverflow(window->wave->samp_bit,
			 value * swGetWaveMax(window->wave)) == SP_TRUE ||
	    swIsOverflow(window->wave->samp_bit,
			 value * swGetWaveMin(window->wave)) == SP_TRUE) {
	    if (spCreateMessageBox(window->window, NULL,
				   SW_OVERFLOW_WARNING_MESSAGE,
				   SppDialogType, SP_WARNING_DIALOG,
				   SppMessageBoxButtonType, SP_MB_YES_NO,
				   NULL) != SP_DR_YES) {
		return SP_FALSE;
	    }
	}
    }

    return SP_TRUE;
}

spBool swEditFinish(swWindow window, swWave wave, spBool in_thread, swEditType edit_type)
{
    spBool flag = SP_TRUE;
    int selected_channel;
    long offset, length;
    long sel_st = -1, sel_ed = -1;
    
    if (swIsNoWave(window) == SP_TRUE || edit_type == SW_EDIT_EXTRACT) return SP_FALSE;

    spDebug(10, "swEditFinish", "edit_type = %d\n", edit_type);

    if (wave == NULL) {
	spDisplayError(window->window, NULL, SW_EDIT_ERROR_MESSAGE);
	flag = SP_FALSE;
    } else if (edit_type == SW_EDIT_WRITE) {
	window->offset = 0;
	window->length = 0;

	swInitWaveRange(window->wave);
	spDebug(10, "swEditFinish", "swInitWaveRange done\n");
		
	/* reset window */
	swResetWindow(window, in_thread);
    } else {
	selected_channel = window->wave->selected_channel;
	
	if (edit_type == SW_EDIT_CROP) {
	    offset = window->wave->edit_offset;
	    length = window->wave->edit_length;
	
	    window->wave = wave;
	    window->offset = 0;
	    window->length = 0;
		
	    /* crop label */
	    swCropLabel(window, offset, length);
	} else if (edit_type == SW_EDIT_DELETE) {
	    offset = window->wave->edit_offset;
	    length = window->wave->edit_length;
	    
	    window->wave = wave;
	    
	    /* delete label */
	    swDeleteLabel(window, offset, length);
	} else if (edit_type == SW_EDIT_SAMP_RATE_CONV) {
	    double st_f, ed_f;
	    double sel_st_f, sel_ed_f;
	    
	    st_f = swSampToDim(window, window->offset);
	    ed_f = swSampToDim(window, window->offset + window->length - 1);
	    sel_st_f = swSampToDim(window, window->sel_st);
	    sel_ed_f = swSampToDim(window, window->sel_ed);

	    window->wave = wave;
	    
	    window->offset = swDimToSamp(window, st_f);
	    window->length = swDimToSamp(window, ed_f) - window->offset + 1;
	    sel_st = swDimToSamp(window, sel_st_f);
	    sel_ed = swDimToSamp(window, sel_ed_f);
	} else {
	    sel_st = window->sel_st;
	    sel_ed = window->sel_ed;

	    window->wave = wave;
	}
	
	/* reset window */
	swResetWindow(window, in_thread);
	
	if (sel_st >= 0 && sel_ed >= 0) {
	    swSelectRegion(window, selected_channel, sel_st, sel_ed);
	}
	
	if (window->wave->overflow == SP_TRUE) {
	    spDisplayWarning(window->window, NULL, SW_OVERFLOW_HAPPENED_WARNING_MESSAGE);
	    window->wave->overflow = SP_FALSE;
	}
	if (edit_type == SW_EDIT_SAMP_RATE_CONV
	    && window->wave->exact_samp_rate != window->wave->samp_rate) {
	    spDisplayWarning(window->window, NULL, SW_SAMP_RATE_CHANGED_WARNING_MESSAGE,
			     window->wave->exact_samp_rate);
	}
    }

    spDebug(40, "swEditFinish", "done\n");

    return flag;
}

static spBool editWindow(swWindow window, swEditType edit_type,
			 long st, long ed, double value, int samp_bit, double samp_rate)
{
    spBool flag = SP_TRUE;
    long offset = 0;
    long length = 0;
    swWave wave = NULL;

    if (swIsNoWave(window) == SP_TRUE) return SP_FALSE;

    if (edit_type != SW_EDIT_BIT_CONV && edit_type != SW_EDIT_SAMP_RATE_CONV
	&& edit_type != SW_EDIT_MONAURALIZE && edit_type != SW_EDIT_CHANGE_VALUE
	&& edit_type != SW_EDIT_PASTE && edit_type != SW_EDIT_MIX
	&& edit_type != SW_EDIT_INSERT && edit_type != SW_EDIT_INSERT_PAUSE) {
	/* get edge */
	if (swGetEdge(window, st, ed, &offset, &length) != SP_TRUE) {
	    spDisplayError(window->window, NULL, SW_REGION_ERROR_MESSAGE);
	    return SP_FALSE;
	}
    }
    
    if (edit_type == SW_EDIT_PASTE || edit_type == SW_EDIT_MIX
	|| edit_type == SW_EDIT_INSERT || edit_type == SW_EDIT_REPLACE) {
	int clipboard_num_channel;
	double clipboard_samp_rate;
	
	if (window->config->toplevel->clipboard_window == NULL
	    || swIsWaveNone(window->config->toplevel->clipboard_window->wave) == SP_TRUE) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_CB_NOT_FOUND_ERROR_MESSAGE);
	    return SP_FALSE;
	}
	if (swIsWaveProcessing(window->config->toplevel->clipboard_window->wave) == SP_TRUE
	    && swIsWavePlaying(window->config->toplevel->clipboard_window->wave) == SP_FALSE) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_CB_PROCESSING_ERROR_MESSAGE);
	    return SP_FALSE;
	}
	
	clipboard_num_channel = swGetWaveNumChannel(window->config->toplevel->clipboard_window->wave);
	
	if (clipboard_num_channel != 1
	    && clipboard_num_channel != swGetWaveNumChannel(window->wave)) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_CB_DIFFERENT_CHANNEL_ERROR_MESSAGE);
	    return SP_FALSE;
	}
	
	clipboard_samp_rate = swGetWaveSampleRate(window->config->toplevel->clipboard_window->wave);
	if (FABS(clipboard_samp_rate - swGetWaveSampleRate(window->wave)) > 100.0) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_CB_DIFFERENT_SAMP_RATE_ERROR_MESSAGE);
	    return SP_FALSE;
	}
	spDebug(10, "editWindow", "target_channel = %d\n", window->target_channel);
    }
    
    if (edit_type == SW_EDIT_AMPLIFY
	&& checkAmplitude(window, value) == SP_FALSE) {
	return SP_FALSE;
    }
    
    if (edit_type == SW_EDIT_CROP) {
	/* crop wave */
	wave = swCropWave(window->wave, offset, length);
    } else if (edit_type == SW_EDIT_DELETE) {
	if (length >= window->wave->total_length) {
	    spDisplayError(window->window, NULL, SW_DELETE_WHOLE_ERROR_MESSAGE);
	    flag = SP_FALSE;
	} else {
	    /* delete wave */
	    wave = swDeleteWave(window->wave, offset, length);
	}
    } else {
	if (edit_type == SW_EDIT_ERASE) {
	    /* erase wave */
	    wave = swEraseWave(window->wave, offset, length);
	} else if (edit_type == SW_EDIT_AMPLIFY) {
	    /* amplify wave */
	    wave = swAmplifyWave(window->wave, offset, length, value);
	} else if (edit_type == SW_EDIT_FADE_IN) {
	    /* fade in */
	    wave = swFadeInWave(window->wave, offset, length);
	} else if (edit_type == SW_EDIT_FADE_OUT) {
	    /* fade out */
	    wave = swFadeOutWave(window->wave, offset, length);
	} else if (edit_type == SW_EDIT_CHANNEL_SWAP) {
	    /* amplify wave */
	    wave = swSwapWaveChannel(window->wave, offset, length);
	} else if (edit_type == SW_EDIT_BIT_CONV) {
	    wave = swConvertWaveBit(window->wave, samp_bit);
	} else if (edit_type == SW_EDIT_SAMP_RATE_CONV) {
	    wave = swConvertWaveSampFreq(window->wave, samp_rate);
	} else if (edit_type == SW_EDIT_MONAURALIZE) {
	    wave = swMonauralizeWave(window->wave);
	} else if (edit_type == SW_EDIT_CHANGE_VALUE) {
	    wave = swChangeWaveValue(window->wave, st, ed, value);
	} else if (edit_type == SW_EDIT_PASTE) {
	    wave = swPasteWave(window->wave, window->config->toplevel->clipboard_window->wave,
			       st, window->target_channel);
	} else if (edit_type == SW_EDIT_MIX) {
	    wave = swMixWave(window->wave, window->config->toplevel->clipboard_window->wave,
			     st, window->target_channel);
	} else if (edit_type == SW_EDIT_INSERT) {
	    wave = swInsertWave(window->wave, window->config->toplevel->clipboard_window->wave,
				st, window->target_channel);
	} else if (edit_type == SW_EDIT_INSERT_PAUSE) {
	    wave = swInsertPauseWave(window->wave, st, ed);
	} else if (edit_type == SW_EDIT_REPLACE) {
	    wave = swReplaceWave(window->wave, window->config->toplevel->clipboard_window->wave, offset, length);
	}
    }

    return flag;
}

spBool swEditWindow(swWindow window, swEditType edit_type,
		    long st, long ed, double value)
{
    return editWindow(window, edit_type, st, ed, value, 0, 0.0);
}

void swCropWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_CROP, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}

void swDeleteWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_DELETE, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}

void swEraseWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_ERASE, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}

void swInvertWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_AMPLIFY, window->sel_st, window->sel_ed, -1.0);
    }
    return;
}

void swFadeInWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_FADE_IN, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}

void swFadeOutWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_FADE_OUT, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}

void swSwapWaveChannelCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	if (window->wave != NULL && window->wave->num_channel >= 2) {
	    swEditWindow(window, SW_EDIT_CHANNEL_SWAP, window->sel_st, window->sel_ed, 0.0);
	}
    }
    return;
}

void swChangeWindowValue(swWindow window, int channel, long point, double value)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_CHANGE_VALUE, channel, point, value);
    }
    return;
}

spBool swConvertWindow(swWindow window, int samp_bit, double samp_rate)
{
    if (samp_bit > 0) {
	return editWindow(window, SW_EDIT_BIT_CONV, 0, 0, 0.0, samp_bit, samp_rate);
    } else {
	return editWindow(window, SW_EDIT_SAMP_RATE_CONV, 0, 0, 0.0, samp_bit, samp_rate);
    }
}

spBool swConvertBit(swWindow window, int samp_bit)
{
    return swConvertWindow(window, samp_bit, 0.0);
}

spBool swConvertSampFreq(swWindow window, double samp_rate)
{
    return swConvertWindow(window, 0, samp_rate);
}

void swMonauralizeCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	if (window->wave != NULL && window->wave->num_channel >= 2) {
	    editWindow(window, SW_EDIT_MONAURALIZE, 0, 0, 0.0, 0, 0.0);
	}
    }
    return;
}

#ifdef SW_SUPPORT_CLIPBOARD
spBool swCopyWindow(swWindow window, long st, long ed)
{
    long offset;
    long length;
    swWave wave;
    swWindow clipboard;
    
    if (swIsNoWave(window) == SP_TRUE) return SP_FALSE;

    /* get edge */
    if (swGetEdge(window, st, ed, &offset, &length) != SP_TRUE) {
	spDisplayError(window->window, NULL, SW_REGION_ERROR_MESSAGE);
	return SP_FALSE;
    }

    /* get new wave */
    if ((wave = swExtractWave(window->wave, offset, length)) == NULL) {
	spDisplayError(window->window, NULL, SW_EDIT_ERROR_MESSAGE);
	return SP_FALSE;
    }

    clipboard = window->config->toplevel->clipboard_window;
    swSetWave(clipboard, wave);
    
    /* extract label */
    swExtractLabel(clipboard, window, offset, length);
    
    /* reload wave */
    swReloadWave(clipboard, SP_FALSE);
    swDrawOverview(clipboard);

    /* set window title */
    swSetWindowTitle(clipboard);
    
    swSetSenseLevel(clipboard);
    
    return SP_TRUE;
}

void swCopyWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swCopyWindow(window, window->sel_st, window->sel_ed);
    }
    return;
}

void swCutWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swCopyWindow(window, window->sel_st, window->sel_ed);
	swEditWindow(window, SW_EDIT_DELETE, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}

void swPasteWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_PASTE, window->point, 0, 0.0);
    }
    return;
}

void swMixWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_MIX, window->point, 0, 0.0);
    }
    return;
}

void swInsertWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_INSERT, window->point, 0, 0.0);
    }
    return;
}

void swCatWindowCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swEditWindow(window, SW_EDIT_INSERT, window->wave->total_length, 0, 0.0);
    }
    return;
}

void swCatTopWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_INSERT, 0, 0, 0.0);
    }
    return;
}

void swReplaceWindowCB(spComponent component, swWindow window)
{
    if (window != NULL) {
	swEditWindow(window, SW_EDIT_REPLACE, window->sel_st, window->sel_ed, 0.0);
    }
    return;
}
#endif
#endif
