/*
 *	swInfoDialog.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spAudioLib.h>
#include <sp/spComponentLib.h>

#include "swWindow.h"
#include "swDraw.h"
#include "swEdit.h"
#include "swDialog.h"
#include "swInfoDialog.h"

#ifdef SW_SUPPORT_INFO_DIALOG

static char *sw_genres[] =
{
    "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge", "Hip-Hop",
    "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B", "Rap", "Reggae", "Rock",
    "Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks", "Soundtrack",
    "Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance",
    "Classical", "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise",
    "Alt", "Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
    "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial",
    "Electronic", "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult",
    "Gangsta Rap", "Top 40", "Christian Rap", "Pop/Funk", "Jungle", "Native American",
    "Cabaret", "New Wave", "Psychedelic", "Rave", "Showtunes", "Trailer", "Lo-Fi",
    "Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll",
    "Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing", "Fast-Fusion", "Bebob",
    "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock",
    "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band",
    "Chorus", "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson", "Opera",
    "Chamber Music", "Sonata", "Symphony", "Booty Bass", "Primus", "Porn Groove",
    "Satire", "Slow Jam", "Club", "Tango", "Samba", "Folklore", "Ballad", "Power Ballad",
    "Rhythmic Soul", "Freestyle", "Duet", "Punk Rock", "Drum Solo", "A Cappella",
    "Euro-House", "Dance Hall", "Goa", "Drum & Bass", "Club-House", "Hardcore", "Terror",
    "Indie", "BritPop", "Negerpunk", "Polsk Punk", "Beat", "Christian Gangsta Rap",
    "Heavy Metal", "Black Metal", "Crossover", "Contemporary Christian",
    "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop", "Synthpop",
    "Unknown Genre", NULL,
};

static char *sw_tracks[] =
{
    "", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10",
    "11", "12", "13", "14", "15", "16", "17", "18", "19", "20",
    "21", "22", "23", "24", "25", "26", "27", "28", "29", "30",
    "31", "32", "33", "34", "35", "36", "37", "38", "39", "40",
    "41", "42", "43", "44", "45", "46", "47", "48", "49", "50",
    NULL,
};

static swInfoDialog createInfoDialog(swConfig config)
{
    swInfoDialog dialog;
    int field_height = 0;
    int field_offset = 100;
    int field_size = 260;
    int field_spacing = 20;
    int release_field_size = 110;
    int track_field_offset = 60;
    int track_field_size = 40;
    
    dialog = xalloc(1, struct _swInfoDialog);
    memset(dialog, 0, sizeof(struct _swInfoDialog));
    dialog->config = config;
    
    /* create dialog */
    dialog->window = spCreateDialogBox("infoDialog",
				       SppTitle, SW_INFO_DIALOG_TITLE,
				       SppCallbackFunc, swPopdownInfoDialogCB,
				       SppCallbackData, dialog,
				       SppDialogBoxButtonType, SP_DB_OK_CANCEL,
				       SppCloseStyle, SP_UNMAP_CLOSE,
				       NULL);
    
    /* create text field */
    dialog->title_text = spCreateParamField(dialog->window, "infoTitleText", field_height,
					    SppTitle, SW_INFO_DIALOG_TITLE_LABEL,
					    SppFieldType, SP_FIELD_TYPE_TEXT,
					    SppEditable, SP_TRUE,
					    SppFieldOffset, field_offset,
					    SppFieldSize, field_size,
					    NULL);
    dialog->artist_text = spCreateParamField(dialog->window, "infoArtistText", field_height,
					     SppTitle, SW_INFO_DIALOG_ARTIST_LABEL,
					     SppFieldType, SP_FIELD_TYPE_TEXT,
					     SppEditable, SP_TRUE,
					     SppFieldOffset, field_offset,
					     SppFieldSize, field_size,
					     NULL);
    dialog->album_text = spCreateParamField(dialog->window, "infoAlbumText", field_height,
					    SppTitle, SW_INFO_DIALOG_ALBUM_LABEL,
					    SppFieldType, SP_FIELD_TYPE_TEXT,
					    SppEditable, SP_TRUE,
					    SppFieldOffset, field_offset,
					    SppFieldSize, field_size,
					    NULL);
    dialog->genre_text = spCreateParamField(dialog->window, "infoGenreText", field_height,
					    SppTitle, SW_INFO_DIALOG_GENRE_LABEL,
					    SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					    SppEditable, SP_TRUE,
					    SppFieldStrings, sw_genres,
					    SppFieldOffset, field_offset,
					    SppFieldSize, field_size,
					    NULL);

    dialog->container = spCreateBox(dialog->window, "infoContainer", 0,
				    SppOrientation, SP_HORIZONTAL,
				    SppUseTextHeight, SP_TRUE,
				    NULL);
    dialog->release_text = spCreateParamField(dialog->container, "infoReleaseText",
					      field_offset + release_field_size + field_spacing,
					      SppTitle, SW_INFO_DIALOG_RELEASE_LABEL,
					      SppFieldType, SP_FIELD_TYPE_TEXT,
					      SppEditable, SP_TRUE,
					      SppFieldOffset, field_offset,
					      SppFieldSize, release_field_size,
					      NULL);
    track_field_size = MAX(track_field_size, field_size - (release_field_size + field_spacing + track_field_offset));
    dialog->track_text = spCreateParamField(dialog->container, "infoTrackText",
					    track_field_offset + track_field_size + field_spacing,
					    SppTitle, SW_INFO_DIALOG_TRACK_LABEL,
					    SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
					    SppEditable, SP_TRUE,
					    SppFieldStrings, sw_tracks,
					    SppFieldOffset, track_field_offset,
					    SppFieldSize, track_field_size,
					    NULL);
    dialog->copyright_text = spCreateParamField(dialog->window, "infoCopyrightText", field_height,
						SppTitle, SW_INFO_DIALOG_COPYRIGHT_LABEL,
						SppFieldType, SP_FIELD_TYPE_TEXT,
						SppEditable, SP_TRUE,
						SppFieldOffset, field_offset,
						SppFieldSize, field_size,
						NULL);
    dialog->engineer_text = spCreateParamField(dialog->window, "infoEngineerText", field_height,
					       SppTitle, SW_INFO_DIALOG_ENGINEER_LABEL,
					       SppFieldType, SP_FIELD_TYPE_TEXT,
					       SppEditable, SP_TRUE,
					       SppFieldOffset, field_offset,
					       SppFieldSize, field_size,
					       NULL);
    dialog->source_text = spCreateParamField(dialog->window, "infoSourceText", field_height,
					     SppTitle, SW_INFO_DIALOG_SOURCE_LABEL,
					     SppFieldType, SP_FIELD_TYPE_TEXT,
					     SppEditable, SP_TRUE,
					     SppFieldOffset, field_offset,
					     SppFieldSize, field_size,
					     NULL);
    dialog->software_text = spCreateParamField(dialog->window, "infoSoftwareText", field_height,
					       SppTitle, SW_INFO_DIALOG_SOFTWARE_LABEL,
					       SppFieldType, SP_FIELD_TYPE_TEXT,
					       SppEditable, SP_TRUE,
					       SppFieldOffset, field_offset,
					       SppFieldSize, field_size,
					       NULL);
    dialog->subject_text = spCreateParamField(dialog->window, "infoSubjectText", field_height,
					      SppTitle, SW_INFO_DIALOG_SUBJECT_LABEL,
					      SppFieldType, SP_FIELD_TYPE_TEXT,
					      SppEditable, SP_TRUE,
					      SppFieldOffset, field_offset,
					      SppFieldSize, field_size,
					      NULL);
    dialog->comment_text = spCreateParamField(dialog->window, "infoCommentText", field_height,
					      SppTitle, SW_INFO_DIALOG_COMMENT_LABEL,
					      SppFieldType, SP_FIELD_TYPE_TEXT,
					      SppEditable, SP_TRUE,
					      SppFieldOffset, field_offset,
					      SppFieldSize, field_size,
					      NULL);
    
    return dialog;
}

static void setInfoString(unsigned long mask1, unsigned long mask2,
			  spComponent text, char *string)
{
    spDebug(10, "setInfoString", "string = %s\n", string);
    
    if (mask1 & mask2) {
	spSetSensitive(text, SP_TRUE);
	spSetTextString(text, string);
    } else {
	spSetTextString(text, "");
	spSetSensitive(text, SP_FALSE);
    }

    return;
}

void swUpdateInfoDialog(swInfoDialog info_dialog, swWindow window)
{
    unsigned long info_mask;
    char buf[SP_MAX_LINE];
    
    if (info_dialog == NULL || window->wave == NULL) return;

    info_mask = swGetSongInfoMask(window->wave);

    setInfoString(info_mask, SP_SONG_TITLE_MASK,
		  info_dialog->title_text, window->wave->song_info.title);
    setInfoString(info_mask, SP_SONG_ARTIST_MASK,
		  info_dialog->artist_text, window->wave->song_info.artist);
    setInfoString(info_mask, SP_SONG_ALBUM_MASK,
		  info_dialog->album_text, window->wave->song_info.album);
    setInfoString(info_mask, SP_SONG_GENRE_MASK,
		  info_dialog->genre_text, window->wave->song_info.genre);
    setInfoString(info_mask, SP_SONG_RELEASE_MASK,
		  info_dialog->release_text, window->wave->song_info.release);
    setInfoString(info_mask, SP_SONG_COPYRIGHT_MASK,
		  info_dialog->copyright_text, window->wave->song_info.copyright);
    setInfoString(info_mask, SP_SONG_ENGINEER_MASK,
		  info_dialog->engineer_text, window->wave->song_info.engineer);
    setInfoString(info_mask, SP_SONG_SOURCE_MASK,
		  info_dialog->source_text, window->wave->song_info.source);
    setInfoString(info_mask, SP_SONG_SOFTWARE_MASK,
		  info_dialog->software_text, window->wave->song_info.software);
    setInfoString(info_mask, SP_SONG_SUBJECT_MASK,
		  info_dialog->subject_text, window->wave->song_info.subject);
    setInfoString(info_mask, SP_SONG_COMMENT_MASK,
		  info_dialog->comment_text, window->wave->song_info.comment);
    if (window->wave->song_info.track > 0) {
	sprintf(buf, "%d", window->wave->song_info.track);
    } else {
	strcpy(buf, "");
    }
    setInfoString(info_mask, SP_SONG_TRACK_MASK,
		  info_dialog->track_text, buf);

    return;
}

swInfoDialog swCreateInfoDialog(swConfig config)
{
    static swInfoDialog info_dialog = NULL;
    
    if (info_dialog == NULL) {
	info_dialog = createInfoDialog(config);
    }
    return info_dialog;
}

void swPopupInfoDialogCB(spComponent component, swWindow window)
{
    swInfoDialog info_dialog;
    
    window->config->toplevel->current_window = window;
    
    info_dialog = swCreateInfoDialog(window->config);

    swUpdateInfoDialog(info_dialog, window);

    /* popup dialog */
    spPopupWindow(info_dialog->window);

    return;
}

static spBool getInfoString(unsigned long *mask1, unsigned long mask2,
			    spComponent text, char *string)
{
    char *p;
    
    if (spIsSensitive(text) == SP_TRUE) {
	if ((p = spGetTextString(text)) != NULL && !strnone(p)) {
	    *mask1 |= mask2;
	    spStrCopy(string, SP_SONG_INFO_SIZE, p);
	    spDebug(10, "getInfoString", "string = %s\n", string);
	    return SP_TRUE;
	}
    }

    return SP_FALSE;
}

void swPopdownInfoDialogCB(spComponent component, swInfoDialog dialog)
{
    spCallbackReason reason = SP_CR_NONE;
    swWindow window;
    spSongInfo song_info;
    char buf[SP_SONG_INFO_SIZE];

    window = dialog->config->toplevel->current_window;

    if (dialog == NULL || spIsCreated(dialog->window) == SP_FALSE
	|| window == NULL) return;

    reason = spGetCallbackReason(component);

    if (reason == SP_CR_OK || reason == SP_CR_CANCEL) {
	/* popdown dialog */
	spPopdownWindow(dialog->window);
    }
    
    if (reason == SP_CR_OK || reason == SP_CR_APPLY) {
	spInitSongInfo(&song_info);
	
	getInfoString(&song_info.info_mask, SP_SONG_TITLE_MASK, 
		      dialog->title_text, song_info.title);
	getInfoString(&song_info.info_mask, SP_SONG_ARTIST_MASK,
		      dialog->artist_text, song_info.artist);
	getInfoString(&song_info.info_mask, SP_SONG_ALBUM_MASK,
		      dialog->album_text, song_info.album);
	getInfoString(&song_info.info_mask, SP_SONG_GENRE_MASK,
		      dialog->genre_text, song_info.genre);
	getInfoString(&song_info.info_mask, SP_SONG_RELEASE_MASK,
		      dialog->release_text, song_info.release);
	getInfoString(&song_info.info_mask, SP_SONG_COPYRIGHT_MASK,
		      dialog->copyright_text, song_info.copyright);
	getInfoString(&song_info.info_mask, SP_SONG_ENGINEER_MASK,
		      dialog->engineer_text, song_info.engineer);
	getInfoString(&song_info.info_mask, SP_SONG_SOURCE_MASK,
		      dialog->source_text, song_info.source);
	getInfoString(&song_info.info_mask, SP_SONG_SOFTWARE_MASK,
		      dialog->software_text, song_info.software);
	getInfoString(&song_info.info_mask, SP_SONG_SUBJECT_MASK,
		      dialog->subject_text, song_info.subject);
	getInfoString(&song_info.info_mask, SP_SONG_COMMENT_MASK,
		      dialog->comment_text, song_info.comment);

	if (getInfoString(&song_info.info_mask, SP_SONG_TRACK_MASK,
			  dialog->track_text, buf) == SP_TRUE) {
	    if (sscanf(buf, "%d", &song_info.track) <= 0) {
		song_info.info_mask &= ~SP_SONG_TRACK_MASK;
	    }
	}

	if (spEqSongInfo(&window->wave->song_info, &song_info) == SP_FALSE) {
	    swEditWindow(window, SW_EDIT_CROP, 0, window->wave->total_length, 0.0);
	    spCopySongInfo(&window->wave->song_info, &song_info);
	}
    }
    
    return;
}
#endif
