/*
 *	swLabel.h
 */

#ifndef __SWLABEL_H
#define __SWLABEL_H

#include "swWave.h"

#ifdef __cplusplus
extern "C" {
#endif

#define SW_MAX_LABEL_STRING 512
#define SW_NUM_LABEL_BUFFER 16

typedef enum {
    SW_LABEL_TYPE_UNKNOWN = -1,
    SW_LABEL_TYPE_ALL = 0,
    SW_LABEL_TYPE_NORMAL = 1,
    SW_LABEL_TYPE_REGION = 2,
} swLabelType;

struct _swLabel {
    long index;
    double time;
    double end_time;
    char label[SW_MAX_LABEL_STRING];
    char data[SW_MAX_LABEL_STRING];
    char string[SW_MAX_LABEL_STRING];
};

struct _swLabels {
    char *filename;
    char *region_filename;
    swTimeFormat format;
    long num_buffer;
    long num_label;
    long num_region;
    swLabel *label;
};

#if defined(MACOS)
#pragma import on
#endif

extern spBool swIsWaveLabelNone(swWave wave);
extern long swAddLabel(swWave wave, double time, double end_time,
		       char *data, char *string);
extern long swChangeLabel(swWave wave, long index, double time, double end_time,
			  char *data, char *string);
extern spBool swReadLabel(swWave wave, char *filename, swTimeFormat format, spBool region_flag);
extern spBool swWriteLabel(swWave wave, char *name, char *filename, swTimeFormat format, spBool region_flag);
extern spBool swCopyLabels(swWave dest_wave, swWave src_wave);
extern long swGetNumRegionLabel(swWave wave);
extern long swGetNumNormalLabel(swWave wave);
extern long swGetNumLabel(swWave wave);

extern long swFindNeighborLabelIndex(swWave wave, long current_index, spBool prev_flag);
    
extern swLabels swAllocLabels(swWaveConfig config, spBool region_flag);
extern void swFreeLabels(swLabels labels);
extern swLabels swReallocLabels(swLabels labels, swWaveConfig config, spBool region_flag);
extern void swClearLabel(swLabels labels, long index);
extern void swClearLabels(swLabels labels, swLabelType label_type);
extern void swSortLabels(swLabels labels, long *converted_index);
extern void swSetLabelFormat(swLabels labels, swTimeFormat format);
extern void swSetRegionLabelFileName(swLabels labels, char *filename);
extern void swSetLabelFileName(swLabels labels, char *filename);
extern char *swGetRegionLabelFileName(swLabels labels);
extern char *swGetLabelFileName(swLabels labels);

extern spBool swIsTimeFormatWithData(swTimeFormat swformat);
extern spTimeFormat swConvertTimeFormatFrom(swTimeFormat swformat);
extern swTimeFormat swConvertTimeFormatTo(spTimeFormat format, spBool with_data);
extern spBool swGetTimeString(double sec, swTimeFormat swformat, char *buf);
extern spBool swConvertTimeString(char *buf, swTimeFormat swformat, double *sec);

#if defined(MACOS)
#pragma import off
#endif
    
#ifdef __cplusplus
}  /* Close scope of 'extern "C"' declaration */
#endif

#endif /* __SWLABEL_H */
