/*
 *	swLabelDialog.c
 *
 *	Last modified: <2002-03-05 14:58:24 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>

#include "swWindow.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swEdit.h"
#include "swLabel.h"
#include "swLabelList.h"
#include "swLabelDialog.h"

#define SW_LABEL_DIALOG_FIELD_HEIGHT 60
#define SW_LABEL_DIALOG_FIELD_OFFSET 60
#define SW_LABEL_DIALOG_FIELD_SIZE 250

static char *label_file_filter[] =
{
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    NULL,
};

static char *label_file_type[] =
{
    "Seconds",
    "Milli Seconds",
    "Points",
    "Separated Seconds (hh:mm:ss.ssss)",
    "Floored Separated Seconds (hh:mm:ss)",
    "Indexed Separated Seconds",
    "Seconds with Data",
    "Milli Seconds with Data",
    "Points with Data",
    "Separated Seconds (hh:mm:ss.ssss) with Data",
    "Floored Separated Seconds (hh:mm:ss) with Data",
    "Indexed Separated Seconds with Data",
    "ESPS",
    NULL,
};

static char *region_label_file_filter[] =
{
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    NULL,
};

static char *region_label_file_type[] =
{
    "Seconds",
    "Milli Seconds",
    "Points",
    "Separated Seconds (hh:mm:ss.ssss)",
    "Floored Separated Seconds (hh:mm:ss)",
    "Indexed Separated Seconds",
    "Seconds with Data",
    "Milli Seconds with Data",
    "Points with Data",
    "Separated Seconds (hh:mm:ss.ssss) with Data",
    "Floored Separated Seconds (hh:mm:ss) with Data",
    "Indexed Separated Seconds with Data",
    NULL,
};

static swLabelDialog createLabelInsertDialog(swConfig config)
{
    static swLabelDialog label_dialog = NULL;
    
    if (label_dialog != NULL) return label_dialog;

    /* initialize dialog */
    label_dialog = xalloc(1, struct _swLabelDialog);
    memset(label_dialog, 0, sizeof(struct _swLabelDialog));
    
    label_dialog->region_flag = SP_FALSE;
    label_dialog->current_index = -1;
    label_dialog->current_window = NULL;
    
    /* create dialog */
    label_dialog->window =
	spCreateDialogBox("labelDialog",
			  SppCallbackFunc, swPopdownLabelInsertDialog,
			  SppCallbackData, label_dialog,
			  SppPopupStyle, SP_MODAL_POPUP,
			  SppCloseStyle, SP_UNMAP_CLOSE,
			  NULL);
    
    /* create text field */
    label_dialog->time_field =
	spCreateParamField(label_dialog->window, "labelDialogTimeField", SW_LABEL_DIALOG_FIELD_HEIGHT,
			   SppTitle, SW_LABEL_DIALOG_TIME_LABEL,
			   SppFieldOffset, SW_LABEL_DIALOG_FIELD_OFFSET,
			   SppFieldSize, SW_LABEL_DIALOG_FIELD_SIZE,
			   NULL);
    label_dialog->end_time_field =
	spCreateParamField(label_dialog->window, "labelDialogEndTimeField", SW_LABEL_DIALOG_FIELD_HEIGHT,
			   SppTitle, SW_LABEL_DIALOG_END_TIME_LABEL,
			   SppFieldOffset, SW_LABEL_DIALOG_FIELD_OFFSET,
			   SppFieldSize, SW_LABEL_DIALOG_FIELD_SIZE,
			   NULL);
    label_dialog->symbol_field =
	spCreateParamField(label_dialog->window, "labelDialogSymbolField", SW_LABEL_DIALOG_FIELD_HEIGHT,
			   SppTitle, SW_LABEL_DIALOG_SYMBOL_LABEL,
			   SppFieldOffset, SW_LABEL_DIALOG_FIELD_OFFSET,
			   SppFieldSize, SW_LABEL_DIALOG_FIELD_SIZE,
			   NULL);
    label_dialog->data_field =
	spCreateParamField(label_dialog->window, "labelDialogDataField", SW_LABEL_DIALOG_FIELD_HEIGHT,
			   SppTitle, SW_LABEL_DIALOG_DATA_LABEL,
			   SppFieldOffset, SW_LABEL_DIALOG_FIELD_OFFSET,
			   SppFieldSize, SW_LABEL_DIALOG_FIELD_SIZE,
			   NULL);
    
    return label_dialog;
}

static void setLabelInsertText(swWindow window, swLabelDialog label_dialog, long index)
{
    double point_f;
    double weight;
    char time_string[SW_MAX_LABEL_STRING];
    char end_time_string[SW_MAX_LABEL_STRING];
    char data_string[SW_MAX_LABEL_STRING];
    char symbol_string[SW_MAX_LABEL_STRING];

    if (window == NULL || window->wave == NULL)
	return;

    if (window->config->time_format == SW_TIME_FORMAT_POINT) {
	weight = window->wave->samp_rate;
    } else {
	weight = 1.0;
    }

    end_time_string[0] = NUL;
    data_string[0] = NUL;
    symbol_string[0] = NUL;
    
    if (window->wave->labels == NULL 
	|| index < 0 || index >= window->wave->labels->num_buffer 
	|| window->wave->labels->label[index].time < 0.0) {
	point_f = swSampToDim(window, window->point);
    } else {
	if (window->wave->labels->label[index].end_time >= 0.0) {
	    point_f = window->wave->labels->label[index].end_time;
	    swGetTimeString(weight * point_f, window->config->time_format, end_time_string);
	}
	
	point_f = window->wave->labels->label[index].time;
	sprintf(data_string, "%s", window->wave->labels->label[index].data);
	sprintf(symbol_string, "%s", window->wave->labels->label[index].string);
    }

    swGetTimeString(weight * point_f, window->config->time_format, time_string);

    spSetTextString(label_dialog->time_field, time_string);
    spSetTextString(label_dialog->end_time_field, end_time_string);
    spSetTextString(label_dialog->symbol_field, symbol_string);
    spSetTextString(label_dialog->data_field, data_string);

    spDebug(10, "setLabelInsertText", "time: %s\n", time_string);

    return;
}

void swPopupLabelInsertDialog(swWindow window, long index)
{
    swLabelDialog label_dialog;

    if (window == NULL || window->wave == NULL)
	return;

    /* create dialog */
    label_dialog = createLabelInsertDialog(window->config);
    label_dialog->region_flag = SP_FALSE;

    if (index >= 0) {
	if (window->wave->labels->label[index].end_time >= 0.0) {
	    label_dialog->region_flag = SP_TRUE;
	    spSetSensitive(label_dialog->end_time_field, SP_TRUE);
	    /*spSetSensitive(label_dialog->symbol_field, SP_FALSE);*/
	} else {
	    spSetSensitive(label_dialog->end_time_field, SP_FALSE);
	    /*spSetSensitive(label_dialog->symbol_field, SP_TRUE);*/
	}
	spSetParams(label_dialog->window,
		    SppTitle, "Change Label",
		    NULL);
    } else {
	spSetSensitive(label_dialog->end_time_field, SP_FALSE);
	/*spSetSensitive(label_dialog->symbol_field, SP_TRUE);*/
	spSetParams(label_dialog->window,
		    SppTitle, "Insert Label",
		    NULL);
    }

    /* set dialog text */
    setLabelInsertText(window, label_dialog, index);

    /* set current window */
    label_dialog->current_window = window;
    label_dialog->current_index = index;

    /* popup dialog */
    spPopupWindow(label_dialog->window);

    return;
}

void swPopdownLabelInsertDialog(spComponent component, swLabelDialog label_dialog)
{
    long index;
    double weight;
    double time, end_time;
    char *time_string;
    char *end_time_string;
    char *data_string;
    char *symbol_string;
    swTimeFormat format;
    spCallbackReason reason;
    
    if (label_dialog == NULL || spIsCreated(label_dialog->window) == SP_FALSE) return;
    
    reason = spGetCallbackReason(component);
    
    if (reason == SP_CR_OK) {
	if (label_dialog->current_window != NULL
	    && label_dialog->current_window->wave != NULL) {
	    format = label_dialog->current_window->config->time_format;
	    if (format == SW_TIME_FORMAT_POINT) {
		weight = label_dialog->current_window->wave->samp_rate;
	    } else {
		weight = 1.0;
	    }

	    /* get string */
	    time_string = xspGetTextString(label_dialog->time_field);
	    end_time_string = xspGetTextString(label_dialog->end_time_field);
	    symbol_string = xspGetTextString(label_dialog->symbol_field);
	    data_string = xspGetTextString(label_dialog->data_field);

	    if (swConvertTimeString(time_string, format, &time) == SP_TRUE
		&& (label_dialog->region_flag == SP_FALSE
		    || swConvertTimeString(end_time_string, format, &end_time) == SP_TRUE)) {
		time /= weight;

		if (label_dialog->region_flag == SP_TRUE) {
		    end_time /= weight;
		} else {
		    end_time = -1.0;
		}
		
		if (label_dialog->current_index >= 0) {
		    if ((index = swChangeLabel(label_dialog->current_window->wave,
					       label_dialog->current_index, 
					       time, end_time, data_string, symbol_string)) >= 0) {
			label_dialog->current_window->active_label_index = index;
			swUpdateLabels(label_dialog->current_window);
		    }
		} else {
		    if ((index = swAddLabel(label_dialog->current_window->wave,
					    time, end_time, data_string, symbol_string)) >= 0) {
			label_dialog->current_window->active_label_index = index;
			swSetSenseLevel(label_dialog->current_window);
			swUpdateLabels(label_dialog->current_window);
		    }
		}
	    } else {
		spDisplayError(component, SW_ERROR_TITLE,
			       SW_TIME_FORMAT_ERROR_MESSAGE);
	    }

	    /* memory free */
	    if (time_string != NULL) xfree(time_string);
	    if (end_time_string != NULL) xfree(end_time_string);
	    if (data_string != NULL) xfree(data_string);
	    if (symbol_string != NULL) xfree(symbol_string);
	}
    }
    
    /* popdown dialog */
    spPopdownWindow(label_dialog->window);
    
    return;
}


static void openLabel(spComponent component, swWindow window, spBool region_flag)
{
    char *filename = NULL;
    swTimeFormat format_index;
    
    if (window == NULL || window->wave == NULL) {
	spDisplayError(component, SW_ERROR_TITLE, SW_FIND_DATA_ERROR_MESSAGE);
	return;
    }

    /* get open file name */
    if (region_flag == SP_TRUE) {
	spDebug(50, "openLabel", "region_label_format = %d\n", window->wave->config->region_label_format);
	filename = xspGetOpenFileName(component, NULL,
				      SppFileMustExist, SP_TRUE,
				      SppFileFilters, region_label_file_filter,
				      SppFileTypes, region_label_file_type,
				      SppFileFilterIndex, &window->wave->config->region_label_format,
				      NULL);
	format_index = window->wave->config->region_label_format;
    } else {
	filename = xspGetOpenFileName(component, NULL,
				      SppFileMustExist, SP_TRUE,
				      SppFileFilters, label_file_filter,
				      SppFileTypes, label_file_type,
				      SppFileFilterIndex, &window->wave->config->label_format,
				      NULL);
	format_index = window->wave->config->label_format;
    }

    if (filename != NULL) {
	/* read label */
	if (swReadLabel(window->wave, filename, format_index, region_flag) == SP_TRUE) {
	    swSetSenseLevel(window);
	    swUpdateLabels(window);
	} else {
	    spDisplayError(component, SW_ERROR_TITLE, SW_OPEN_ERROR_MESSAGE,
			   filename);
	}
	
	xfree(filename);
    }

    return;
}

void swOpenLabelCB(spComponent component, swWindow window)
{
    openLabel(component, window, SP_FALSE);
}

void swOpenRegionLabelCB(spComponent component, swWindow window)
{
    openLabel(component, window, SP_TRUE);
}

static void saveLabel(swWindow window, spBool save_as, spBool region_flag)
{
    char *filename;
    char *label_filename;
    swTimeFormat format_index = SW_TIME_FORMAT_SEC;
    
    if (window == NULL || window->wave == NULL || window->wave->labels == NULL) {
	spDisplayError(window->window, SW_ERROR_TITLE, SW_FIND_DATA_ERROR_MESSAGE);
	return;
    }

    format_index = window->wave->labels->format;

    if (region_flag == SP_TRUE) {
	label_filename = swGetRegionLabelFileName(window->wave->labels);
    } else {
	label_filename = swGetLabelFileName(window->wave->labels);
    }
    
    if (save_as == SP_TRUE || label_filename == NULL) {
	/* get save file name */
	if (region_flag == SP_TRUE) {
	    filename = xspGetSaveFileName(window->window, NULL,
					  SppInitialFileName, label_filename,
					  SppPathMustExist, SP_TRUE,
					  SppFileMustExist, SP_FALSE,
					  SppOverwritePrompt, SP_TRUE,
					  SppFileFilters, region_label_file_filter,
					  SppFileTypes, region_label_file_type,
					  SppFileFilterIndex, &format_index,
					  NULL);
	} else {
	    filename = xspGetSaveFileName(window->window, NULL,
					  SppInitialFileName, label_filename,
					  SppPathMustExist, SP_TRUE,
					  SppFileMustExist, SP_FALSE,
					  SppOverwritePrompt, SP_TRUE,
					  SppFileFilters, label_file_filter,
					  SppFileTypes, label_file_type,
					  SppFileFilterIndex, &format_index,
					  NULL);
	}
	
	if (filename != NULL) {
	    /* write label */
	    if (swWriteLabel(window->wave, window->name,
			     filename, format_index, region_flag) != SP_TRUE) {
		spDisplayError(window->window, SW_ERROR_TITLE, SW_SAVE_ERROR_MESSAGE,
			       filename);
	    }
	
	    xfree(filename);
	}
    } else {
	/* write label */
	if (swWriteLabel(window->wave, window->name,
			 NULL, format_index, region_flag) != SP_TRUE) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_SAVE_ERROR_MESSAGE,
			   label_filename);
	}
    }
    

    return;
}

void swSaveLabelCB(spComponent component, swWindow window)
{
    saveLabel(window, SP_FALSE, SP_FALSE);
    return;
}

void swSaveRegionLabelCB(spComponent component, swWindow window)
{
    saveLabel(window, SP_FALSE, SP_TRUE);
    return;
}

void swSaveAsLabelCB(spComponent component, swWindow window)
{
    saveLabel(window, SP_TRUE, SP_FALSE);
    return;
}

void swSaveAsRegionLabelCB(spComponent component, swWindow window)
{
    saveLabel(window, SP_TRUE, SP_TRUE);
    return;
}

static void clearLabel(spComponent component, swWindow window, swLabelType label_type)
{
    if (window == NULL || window->wave == NULL || window->wave->labels == NULL) return;
    
    if (spCreateMessageBox(window->window, NULL,
			   SW_CLEAR_LABEL_QUESTION_MESSAGE,
			   SppDialogType, SP_QUESTION_DIALOG,
			   SppMessageBoxButtonType, SP_MB_YES_NO,
			   NULL) == SP_DR_YES) {
	swClearLabels(window->wave->labels, label_type);
	
	swSetSenseLevel(window);
	swUpdateLabels(window);
    }
    
    return;
}

void swClearLabelCB(spComponent component, swWindow window)
{
    clearLabel(component, window, SW_LABEL_TYPE_NORMAL);
    return;
}

void swClearRegionLabelCB(spComponent component, swWindow window)
{
    clearLabel(component, window, SW_LABEL_TYPE_REGION);
    return;
}

void swInsertLabelCB(spComponent component, void *data)
{
    swWindow window = (swWindow)data;
    
    swPopupLabelInsertDialog(window, -1);

    return;
}

void swChangeLabelCB(spComponent component, void *data)
{
    long index;
    double point_f;
    swWindow window = (swWindow)data;

    if (window == NULL || window->wave == NULL)
	return;

    point_f = swSampToDim(window, window->point);
    
    if ((index = swFindNearLabelIndex(window, point_f, SP_FALSE, NULL)) >= 0) {
	swPopupLabelInsertDialog(window, index);
    } else {
	spDisplayError(component, SW_ERROR_TITLE, SW_FIND_LABEL_NEAR_CURSOR_ERROR_MESSAGE);
    }

    return;
}

void swInsertSimpleLabelCB(spComponent component, swWindow window)
{
    long index;
    double point_f;

    if (window == NULL || window->wave == NULL)
	return;

    point_f = swSampToDim(window, window->point);

    if ((index = swAddLabel(window->wave, point_f, -1.0, NULL, NULL)) >= 0) {
	window->active_label_index = index;
	swSetSenseLevel(window);
	swUpdateLabels(window);
    }

    return;
}

void swInsertValueLabelCB(spComponent component, swWindow window)
{
    long index;
    double point_f;
    double value;
    long data_offset;
    char *string;
    char buf[SP_MAX_MESSAGE];

    if (window == NULL || window->wave == NULL)
	return;

    point_f = swSampToDim(window, window->point);
    data_offset = (window->point - window->offset) / window->wave->thin_length;

    if (swGetWaveData(window->wave, window->target_channel, data_offset, &value) == SP_FALSE) {
	string = NULL;
    } else {
	if (swIsWaveFloat(window->wave) == SP_TRUE) {
	    sprintf(buf, "%.6f", value);
	} else {
	    sprintf(buf, "%.0f", value);
	}
	string = buf;
    }
    
    if ((index = swAddLabel(window->wave, point_f, -1.0, NULL, string)) >= 0) {
	window->active_label_index = index;
	swSetSenseLevel(window);
	swUpdateLabels(window);
    }

    return;
}

spBool swEraseLabelPrompt(spComponent component, swWindow window)
{
    if (window->config->erase_label_prompt == SP_FALSE)
	return SP_TRUE;
    
    if (spCreateMessageBox(component, NULL,
			   SW_ERASE_LABEL_QUESTION_MESSAGE,
			   SppTitle, SW_ERASE_LABEL_QUESTION_TITLE,
			   SppDialogType, SP_QUESTION_DIALOG,
			   SppMessageBoxButtonType, SP_MB_OK_CANCEL,
			   NULL) == SP_DR_OK) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

void swEraseLabelCB(spComponent component, swWindow window)
{
    double point_f;

    if (window == NULL || window->wave == NULL
	|| swEraseLabelPrompt(window->window, window) == SP_FALSE)
	return;

    point_f = swSampToDim(window, window->point);

    if (swEraseLabel(window, point_f) == SP_TRUE) {
	swSetSenseLevel(window);
	swUpdateLabels(window);
    } else {
	spDisplayError(component, SW_ERROR_TITLE, SW_FIND_LABEL_NEAR_CURSOR_ERROR_MESSAGE);
    }

    return;
}

void swEraseLabelRegionCB(spComponent component, swWindow window)
{
    long offset, length;
    
    if (window == NULL || window->wave == NULL)
	return;
    
    if (swGetEdge(window, window->sel_st, 
		  window->sel_ed, &offset, &length) == SP_TRUE) {
	if (swEraseLabelRegion(window, offset, length) == SP_TRUE) {
	    swSetSenseLevel(window);
	    swUpdateLabels(window);
	}
    } else {
	spDisplayError(component, SW_ERROR_TITLE, SW_REGION_ERROR_MESSAGE);
    }
    
    return;
}

spBool swGetLabelEdge(swWindow window, long index, long *st, long *ed)
{
    if (window == NULL || window->wave == NULL
	|| window->wave->labels == NULL || index < 0
	|| window->wave->labels->label[index].time < 0.0)
	return SP_FALSE;

    if (window->wave->labels->label[index].end_time >= 0.0) {
	*st = swDimToSamp(window, window->wave->labels->label[index].time);
	*ed = swDimToSamp(window, window->wave->labels->label[index].end_time);
    } else {
	long next_index;
    
	*st = swDimToSamp(window, window->wave->labels->label[index].time);
	if ((next_index = swFindNeighborLabelIndex(window->wave, index, SP_FALSE)) >= 0) {
	    *ed = swDimToSamp(window, window->wave->labels->label[next_index].time);
	} else {
	    *ed = window->wave->total_length - 1;
	}
    }

    return SP_TRUE;
}

void swSelectBetweenLabelsCB(spComponent component, swWindow window)
{
    long index;
    long st, ed;
    double point_f;
    spBool flag = SP_FALSE;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL)
	return;

    point_f = swSampToDim(window, window->point);

    if ((index = swFindLabelIndex(window, point_f, -1, SP_TRUE, SP_FALSE, NULL, NULL)) >= 0) {
	if (swGetLabelEdge(window, index, &st, &ed) == SP_TRUE) {
	    swSelectRegion(window, -1, st, ed);
	    flag = SP_TRUE;
	}
    } else if ((index = swFindLabelIndex(window, point_f, 1, SP_TRUE, SP_FALSE, NULL, NULL)) >= 0) {
	ed = swDimToSamp(window, window->wave->labels->label[index].time);
	swSelectRegion(window, -1, 0, ed);
	flag = SP_TRUE;
    }

    if (flag == SP_FALSE) {
	spDisplayError(component, SW_ERROR_TITLE, SW_FIND_LABEL_ERROR_MESSAGE);
    }

    return;
}

void swSetRegionLabelCB(spComponent component, swWindow window)
{
    long index;
    long offset, length;
    
    if (window == NULL || window->wave == NULL)
	return;
    
    if (swGetEdge(window, window->sel_st, 
		  window->sel_ed, &offset, &length) == SP_TRUE) {
	/* add label */
	if ((index = swAddLabel(window->wave, swSampToDim(window, offset),
				swSampToDim(window, offset + length - 1), NULL, NULL)) >= 0) {
	    window->active_label_index = index;
	    swSetSenseLevel(window);
	    swUpdateLabels(window);

	    swMoveCursor(window, offset, SP_TRUE);
	}
    } else {
	spDisplayError(component, SW_ERROR_TITLE, SW_REGION_ERROR_MESSAGE);
    }
    
    return;
}

spBool swSetLabelAsRegion(swWindow window, long index)
{
    long st, ed;
    
    if (window == NULL || window->wave == NULL)
	return SP_FALSE;

    if (swGetLabelEdge(window, index, &st, &ed) == SP_TRUE) {
	swSelectRegion(window, -1, st, ed);
	return SP_TRUE;
    }

    return SP_FALSE;
}

void swSetRegionLabelAsRegionCB(spComponent component, swWindow window)
{
    long index;
    double point_f;

    if (window == NULL || window->wave == NULL)
	return;

    point_f = swSampToDim(window, window->point);
    index = swFindNearLabelIndex(window, point_f, SP_TRUE, NULL);
    
    if (swSetLabelAsRegion(window, index) == SP_FALSE) {
	spDisplayError(component, SW_ERROR_TITLE, SW_FIND_LABEL_NEAR_CURSOR_ERROR_MESSAGE);
    }

    return;
}

#if 1
void swDivideRegionLabelCB(spComponent component, swWindow window)
{
    long index;
    
    if (swIsNoWave(window) == SP_TRUE || window->wave->labels == NULL) {
	return;
    }

    if ((index = swFindLabelIndex(window, window->point_f,
				  -1, SP_FALSE, SP_TRUE, NULL, NULL)) >= 0
	&& window->point_f > window->wave->labels->label[index].time
	&& window->point_f < window->wave->labels->label[index].end_time) {
	swAddLabel(window->wave, window->point_f,
		   window->wave->labels->label[index].end_time,
		   NULL, NULL);
	
	window->wave->labels->label[index].end_time = window->point_f;
	
	swSetSenseLevel(window);
	swUpdateLabels(window);
    } else {
	spDisplayError(component, SW_ERROR_TITLE, SW_FIND_LABEL_ERROR_MESSAGE);
    }

    return;
}
#endif

void swCatRegionLabels(swWindow window, long index, spBool end_is_nearer)
{
    long target_index;
    long start_index, end_index;
    double start_time, end_time;
    char data[SW_MAX_LABEL_STRING];
    char string[SW_MAX_LABEL_STRING];

    if (window == NULL || window->wave == NULL
	|| window->wave->labels == NULL	|| index < 0)
	return;
    
    target_index = swFindNeighborLabelIndex(window->wave, index, spIsFalse(end_is_nearer));

    if (target_index >= 0) {
	if (end_is_nearer == SP_TRUE) {
	    start_index = index;
	    end_index = target_index;
	} else {
	    start_index = target_index;
	    end_index = index;
	}
	start_time = window->wave->labels->label[start_index].time;
	end_time = window->wave->labels->label[end_index].end_time;

	spStrCopy(data, SW_MAX_LABEL_STRING, window->wave->labels->label[start_index].data);
#if 0 /* don't concatenate data */
	if (!strnone(window->wave->labels->label[start_index].data)) {
	    spStrCat(data, SW_MAX_LABEL_STRING, " ");
	}
	spStrCat(data, SW_MAX_LABEL_STRING, window->wave->labels->label[end_index].data);
#endif
	
	spStrCopy(string, SW_MAX_LABEL_STRING, window->wave->labels->label[start_index].string);
	if (!strnone(window->wave->labels->label[start_index].string)) {
	    spStrCat(string, SW_MAX_LABEL_STRING, " ");
	}
	spStrCat(string, SW_MAX_LABEL_STRING, window->wave->labels->label[end_index].string);
	
	swChangeLabel(window->wave, start_index, start_time, end_time, data, string);

	if (end_index == window->active_label_index) {
	    window->active_label_index = start_index;
	}
	swClearLabel(window->wave->labels, end_index);
	
	swSetSenseLevel(window);
	swUpdateLabels(window);
    }
    
    return;
}

void swCatRegionLabelsCB(spComponent component, swWindow window)
{
    long index;
    spBool end_is_nearer = SP_FALSE;

    if (swIsNoWave(window) == SP_TRUE) {
	return;
    }

    if ((index = swFindNearLabelIndex(window, swSampToDim(window, window->point),
				      SP_TRUE, &end_is_nearer)) >= 0) {
	swCatRegionLabels(window, index, end_is_nearer);
    } else if ((index = window->active_label_index) >= 0) {	/* maybe called by shortcut key */
	swCatRegionLabels(window, index, SP_FALSE);
    }
    
    return;
}

spBool swFindIdenticalLabelIndex(swWindow window, double time, 
				spBool region_flag, long *pindex, 
				spBool *start_flag, spBool *end_flag)
{
    long k;
    spBool flag;
    double fragment;

    if (window == NULL || window->wave == NULL
	|| window->wave->labels == NULL || pindex == NULL) return SP_FALSE;

    flag = SP_FALSE;
    if (start_flag != NULL) *start_flag = SP_FALSE;
    if (end_flag != NULL) *end_flag = SP_FALSE;
    
    fragment = 1.0 / window->wave->samp_rate;
    *pindex = MAX(*pindex, -1);

    for (k = *pindex + 1; k < window->wave->labels->num_buffer; k++) {
	if (k == window->active_label_index
	    || (region_flag == SP_TRUE && window->wave->labels->label[k].end_time < 0.0)) {
	    continue;
	}
	
	if (window->wave->labels->label[k].time >= 0.0
	    && FABS(time - window->wave->labels->label[k].time) < fragment) {
	    *pindex = k;
	    flag = SP_TRUE;
	    if (start_flag != NULL) *start_flag = SP_TRUE;
	}
	if (window->wave->labels->label[k].end_time >= 0.0
	    && FABS(time - window->wave->labels->label[k].end_time) < fragment) {
	    *pindex = k;
	    flag = SP_TRUE;
	    if (end_flag != NULL) *end_flag = SP_TRUE;
	}

	if (flag == SP_TRUE) {
	    break;
	}
    }
    
    return flag;
}

spBool swReplaceIdenticalLabelIndex(swWindow window, double orig_time, double target_time,
				    spBool region_flag)
{
    long index;
    spBool start_flag, end_flag;
			    
    index = -1;
    while (1) {
	if (swFindIdenticalLabelIndex(window, orig_time,
				     region_flag, &index, 
				     &start_flag, &end_flag) == SP_FALSE) {
	    break;
	}

	if (start_flag == SP_TRUE) {
	    if (window->wave->labels->label[index].end_time >= 0.0
		&& window->wave->labels->label[index].end_time < target_time) {
		window->wave->labels->label[index].time
		    = window->wave->labels->label[index].end_time;
		window->wave->labels->label[index].end_time = target_time;
	    } else {
		window->wave->labels->label[index].time = target_time;
	    }
	}
	if (end_flag == SP_TRUE) {
	    if (target_time < window->wave->labels->label[index].time) {
		window->wave->labels->label[index].end_time
		    = window->wave->labels->label[index].time;
		window->wave->labels->label[index].time = target_time;
	    } else {
		window->wave->labels->label[index].end_time = target_time;
	    }
	}
    }

    return SP_TRUE;
}
     
long swFindLabelIndex(swWindow window, double time, int direction,
		      spBool normal_only, spBool region_only,
		      spBool *end_is_nearer, long *pmin_dist_l)
{
    long k;
    long index;
    double dist;
    double min_dist;
    long min_dist_l;
    spBool end_flag = SP_FALSE;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL)
	return -1;

    index = -1;
    /* min_dist = FABS(window->wave->labels->label[0].time - time); */
    min_dist = 1000000000.0;
    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if ((region_only == SP_TRUE && window->wave->labels->label[k].end_time < 0.0)
	    || (normal_only == SP_TRUE && window->wave->labels->label[k].end_time >= 0.0)) {
	    continue;
	}
	if ((direction > 0 && window->wave->labels->label[k].time < time)
	    || (direction < 0 && window->wave->labels->label[k].time > time)) {
	    continue;
	}
	
	if (window->wave->labels->label[k].time >= 0.0) {
	    dist = FABS(window->wave->labels->label[k].time - time);
	    
	    if (dist < min_dist
		|| (dist == min_dist && index != window->active_label_index)) {
		end_flag = SP_FALSE;
		min_dist = dist;
		index = k;
	    }
	    
	    if (window->wave->labels->label[k].end_time >= 0.0) {
		dist = FABS(window->wave->labels->label[k].end_time - time);
	    
		if (dist < min_dist) {
		    end_flag = SP_TRUE;
		    min_dist = dist;
		    index = k;
		}
	    }
	}
    }
    spDebug(100, "swFindLabelIndex", "index = %ld, min_dist = %f, end_flag = %d\n",
	    index, min_dist, end_flag);

    min_dist_l = swDimToSamp(window, min_dist);
    if (min_dist_l < window->wave->total_length) {
	if (pmin_dist_l != NULL) {
	    *pmin_dist_l = min_dist_l;
	}
	if (end_is_nearer != NULL) {
	    *end_is_nearer = end_flag;
	}
	return index;
    }

    return -1;
}

long swFindNearLabelIndex(swWindow window, double time, spBool region_only, spBool *end_is_nearer)
{
    long index;
    long min_dist_l;
    int min_dist_d;
    
    if ((index = swFindLabelIndex(window, time, 0, SP_FALSE, region_only, end_is_nearer,
				  &min_dist_l)) >= 0) {
	min_dist_d = swLengthToDisp(window, min_dist_l);
	
	if (min_dist_d <= /*SW_LABEL_DELETE_RANGE / 2 */SW_EDGE_MOTION_RANGE / 2) {
	    return index;
	}
    }

    return -1;
}

spBool swEraseLabel(swWindow window, double time)
{
    long index;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL)
	return SP_FALSE;

    if ((index = swFindNearLabelIndex(window, time, SP_FALSE, NULL)) >= 0) {
	swClearLabel(window->wave->labels, index);
	
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool swCropLabel(swWindow window, long offset, long length)
{
    long k;
    double stp, len;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL) {
	return SP_FALSE;
    }

    stp = swSampToDim(window, offset);
    len = swSampToDim(window, length);

    spDebug(10, "swCropLabel", "stp = %ld, len = %ld\n", stp, len);
    
    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0) {
	    /* shift time */
	    window->wave->labels->label[k].time -= stp;

	    if (window->wave->labels->label[k].end_time >= 0.0) {
		window->wave->labels->label[k].end_time -= stp;
		if (window->wave->labels->label[k].end_time > len) {
		    window->wave->labels->label[k].end_time = len;
		}
		if (window->wave->labels->label[k].end_time > 0.0
		    && window->wave->labels->label[k].time < 0.0) {
		    window->wave->labels->label[k].time = 0.0;
		}
	    }
	    
	    /* if over region, erase label */
	    if (window->wave->labels->label[k].time < 0.0 || window->wave->labels->label[k].time >= len) {
		swClearLabel(window->wave->labels, k);
	    }
	}
    }

    swSetSenseLevel(window);
    swUpdateLabels(window);
    
    return SP_TRUE;
}

spBool swDeleteLabel(swWindow window, long offset, long length)
{
    long k;
    double stp, len;
    spBool clear_flag;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL) {
	return SP_FALSE;
    }

    stp = swSampToDim(window, offset);
    len = swSampToDim(window, length);

    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0) {
	    clear_flag = SP_FALSE;
	    if (window->wave->labels->label[k].time >= stp + len) {
		/* shift time */
		window->wave->labels->label[k].time -= len;
	    } else if (window->wave->labels->label[k].time >= stp) {
		clear_flag = SP_TRUE;
	    }

	    if (window->wave->labels->label[k].end_time > stp + len) {
		if (clear_flag == SP_TRUE) {
		    window->wave->labels->label[k].time = stp;
		    clear_flag = SP_FALSE;
		}
		/* shift time */
		window->wave->labels->label[k].end_time -= len;
	    } else if (window->wave->labels->label[k].end_time > stp) {
		window->wave->labels->label[k].end_time = stp;
	    }

	    if (clear_flag == SP_TRUE) {
		/* if inside region, erase label */
		swClearLabel(window->wave->labels, k);
	    }
	}
    }

    swSetSenseLevel(window);
    swUpdateLabels(window);
    
    return SP_TRUE;
}

spBool swExtractLabel(swWindow newwindow, swWindow window,
		      long offset, long length)
{
    int flag = 0;
    long k;
    double stp, edp;
    double start_time, end_time;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL ||
	newwindow == NULL) {
	return SP_FALSE;
    }

    stp = swSampToDim(window, offset);
    /*edp = stp + swSampToDim(window, length);*/
    edp = swSampToDim(window, offset + length - 1);
    spDebug(80, "swExtractLabel", "stp = %f, edp = %f\n", stp, edp);

    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0) {
	    start_time = -1.0;
	    end_time = -1.0;
	    
	    if (window->wave->labels->label[k].end_time >= 0.0) {
		if (window->wave->labels->label[k].time >= stp
		    && window->wave->labels->label[k].time < edp) {
		    start_time = window->wave->labels->label[k].time - stp;
		    spDebug(80, "swExtractLabel", "start time = %f\n", start_time);
		}
		if (window->wave->labels->label[k].end_time > stp
		    && window->wave->labels->label[k].end_time < edp) {
		    end_time = window->wave->labels->label[k].end_time - stp;
		    spDebug(80, "swExtractLabel", "end time = %f\n", end_time);
		    if (start_time < 0.0) {
			start_time = 0.0;
		    }
		} else if (start_time >= 0.0) {
		    end_time = edp - stp;
		}
	    } else {
		if (window->wave->labels->label[k].time >= stp
		    && window->wave->labels->label[k].time < edp) {
		    start_time = window->wave->labels->label[k].time - stp;
		}
	    }

	    /* if inside region, add label */
	    if (start_time >= 0.0) {
		swAddLabel(newwindow->wave, start_time, end_time,
			   window->wave->labels->label[k].data,
			   window->wave->labels->label[k].string);
		flag = 1;
	    }
	}
    }

    if (flag) {
	swSetSenseLevel(newwindow);
	swUpdateLabels(newwindow);
    
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool swEraseLabelRegion(swWindow window, long offset, long length)
{
    int i;
    long k;
    double stp, edp;
    spBool flag = SP_FALSE;

    if (window == NULL || window->wave == NULL || window->wave->labels == NULL) {
	return SP_FALSE;
    }

    stp = swSampToDim(window, offset);
    edp = stp + swSampToDim(window, length);

    for (i = 0; i < 2; i++) {
	for (k = 0; k < window->wave->labels->num_buffer; k++) {
	    if (window->wave->labels->label[k].time >= 0.0) {
		if (window->wave->labels->label[k].time >= stp
		    && window->wave->labels->label[k].time < edp) {
		    if (i == 0) {
			flag = SP_TRUE;
		    } else {
			swClearLabel(window->wave->labels, k);
		    }
		}
	    }
	}

	if (flag == SP_FALSE) {
	    spDisplayError(window->window, SW_ERROR_TITLE, SW_FIND_LABEL_REGION_ERROR_MESSAGE);
	    break;
	} else if (i == 0 && swEraseLabelPrompt(window->window, window) == SP_FALSE) {
	    flag = SP_FALSE;
	    break;
	}
    }

    if (flag == SP_TRUE) {
	swSetSenseLevel(window);
	swUpdateLabels(window);
    }

    return flag;
}
