/*
 *	swWave.c
 *
 *	Last modified: <2002-04-06 22:35:20 hideki>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spThread.h>
#include <sp/spOption.h>
#include <sp/spWave.h>
#include <sp/spAudio.h>

#include <sp/spTopLevel.h>

#include <sp/sp.h>
#include <sp/vector.h>
#include <sp/voperate.h>
#include <sp/fft.h>
#include <sp/window.h>
#include <sp/filter.h>
#include <sp/sfconv.h>

#include "swWave.h"
#include "swLabel.h"

#define SW_MIN_THIN_LENGTH_USING_PEAK 16/*8*/

spBool swIsWaveNone(swWave wave)
{
    if (wave == NULL) return SP_TRUE;

    if (strnone(wave->plugin_name)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool swIsWaveFloat(swWave wave)
{
    if (wave == NULL) return SP_FALSE;
    
    return wave->float_flag;
}

spBool swIsWaveLong(swWave wave)
{
    if (wave == NULL) return SP_FALSE;
    
    if (swIsWaveFloat(wave) == SP_FALSE
	&& (wave->samp_bit > 16 && wave->samp_bit <= 32)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool swIsWaveOverflow(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    return wave->overflow;
}

spBool swIsWavePlaying(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    return wave->play_flag;
}

spBool swIsWaveProcessing(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->process_flag == SP_TRUE
	/*|| wave->process_finished == SP_FALSE*/) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool swIsWaveProcessFinished(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    return wave->process_finished;
}

spBool swIsWaveEdited(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    return wave->edit_flag;
}

spBool swIsWaveThreadSafe(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    return wave->thread_safe;
}

spBool swIsWavePeakAvailable(swWave wave)
{
    if (wave == NULL || wave->peak_filename == NULL
	|| wave->peak_length <= 0) return SP_FALSE;

    return SP_TRUE;
}

spBool swNeedProcessBlock(swWave src_wave, swWave wave)
{
    if (src_wave == NULL || wave == NULL) return SP_FALSE;

    spDebug(80, "swNeedProcessBlock", "thread_safe = %d, %s %s\n",
	    src_wave->thread_safe, src_wave->plugin_name, wave->plugin_name);

    if (src_wave->thread_safe == SP_FALSE
	&& streq(src_wave->plugin_name, wave->plugin_name)) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

double swGetWaveSampleRate(swWave wave)
{
    if (wave == NULL) return 0.0;

    return wave->samp_rate;
}

double swGetWaveExactSampleRate(swWave wave)
{
    if (wave == NULL) return 0.0;

    return wave->exact_samp_rate;
}

int swGetWaveSampleBit(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->samp_bit;
}

int swGetWaveSampleByte(swWave wave)
{
    int samp_byte;
    
    if (wave == NULL) return 0;

    if (swIsWaveFloat(wave) == SP_TRUE) {
	samp_byte = sizeof(double);
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	samp_byte = sizeof(long);
    } else {
	samp_byte = sizeof(short);
    }
    
    return samp_byte;
}

int swGetWaveNumChannel(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->num_channel;
}

int swGetWaveSelectedChannel(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->selected_channel;
}

long swGetWaveOffset(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->offset;
}

long swGetWaveLength(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->length;
}

long swGetWaveEditOffset(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->edit_offset;
}

long swGetWaveEditLength(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->edit_length;
}

long swGetWaveTotalLength(swWave wave)
{
    if (wave == NULL) return 0;

    return wave->total_length;
}

long swGetWaveThinLength(swWave wave)
{
    if (wave == NULL) return 1;

    return wave->thin_length;
}

char *swGetWaveFileName(swWave wave)
{
    if (wave == NULL) return NULL;

    return wave->filename;
}

char *swGetWaveOriginalFileName(swWave wave)
{
    if (wave == NULL) return NULL;

    return wave->orig_filename;
}

char *swGetWaveFileType(swWave wave)
{
    if (wave == NULL) return NULL;

    return wave->file_type;
}

char *swGetWavePluginName(swWave wave)
{
    if (wave == NULL) return NULL;

    return wave->plugin_name;
}

spSongInfo *swGetWaveSongInfo(swWave wave)
{
    if (wave == NULL) return NULL;

    return &wave->song_info;
}

swLabels swGetWaveLabels(swWave wave)
{
    if (wave == NULL) return NULL;

    return wave->labels;
}

swWaveConfig swGetWaveConfig(swWave wave)
{
    if (wave == NULL) return NULL;

    return wave->config;
}

char *swGetEditName(swWave wave)
{
    if (wave == NULL) return NULL;
    return wave->edit_name;
}

spBool swSetEditName(swWave wave, char *edit_name)
{
    if (wave == NULL) return SP_FALSE;

    if (edit_name != wave->edit_name) {
	if (wave->edit_name != NULL) xfree(wave->edit_name);
	wave->edit_name = strclone(edit_name);
    }

    return SP_TRUE;
}

long swConvertLengthToByte(swWave wave, long length)
{
    if (wave == NULL) return 0;

    return length * (long)wave->num_channel * (long)swGetWaveSampleByte(wave);
}

long swConvertByteToLength(swWave wave, long size)
{
    long length;

    if (wave == NULL) return 0;

    length = size / MAX(wave->num_channel, 1);

    if (swIsWaveFloat(wave) == SP_TRUE) {
	length /= sizeof(double);
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	length /= sizeof(long);
    } else {
	length /= sizeof(short);
    }

    return length;
}

long swLoadIntoBuffer(spPlugin *plugin, swWave wave, int channel, long length, char *buf)
{
    long k;
    long nread;
    
    if (wave == NULL || plugin == NULL || length <= 0) return -1;

    spDebug(50, "swLoadIntoBuffer", "in\n");
    
    if (swIsWaveFloat(wave) == SP_TRUE) {
	double *dbuf = (double *)buf;
	
	nread = spReadPlugin(plugin, dbuf, length * wave->num_channel) / wave->num_channel;
	if (channel >= 0) {
	    for (k = 0; k < nread; k++) {
		dbuf[k] = dbuf[k * wave->num_channel + channel];
	    }
	}
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	long *lbuf = (long *)buf;
	
	nread = spReadPlugin(plugin, lbuf, length * wave->num_channel) / wave->num_channel;
	if (channel >= 0) {
	    for (k = 0; k < nread; k++) {
		lbuf[k] = lbuf[k * wave->num_channel + channel];
	    }
	}
    } else {
	short *sbuf = (short *)buf;
	
	nread = spReadPlugin(plugin, sbuf, length * wave->num_channel) / wave->num_channel;
	if (channel >= 0) {
	    for (k = 0; k < nread; k++) {
		sbuf[k] = sbuf[k * wave->num_channel + channel];
	    }
	}
    }

    spDebug(50, "swLoadIntoBuffer", "done: nread = %ld\n", nread);
    
    return nread;
}


spBool swSetTempDir(swWaveConfig config, char *dir)
{
    char *path = NULL;
    
    if (!strnone(dir) && (path = xspGetExactName(dir)) != NULL) {
	if (spIsDir(path) == SP_FALSE
	    && spCreateDir(path, 0700) == SP_FALSE) {
	    xfree(path); path = NULL;
	}
    }

    if (path == NULL) {
	path = xspGetApplicationTempDir();
    }

    if (path != NULL) {
	spStrCopy(config->temp_dir, SP_MAX_PATHNAME, path);
	spAddDirSeparator(config->temp_dir);
	xfree(path);
    
	spDebug(10, "swSetTempDir", "temp_dir = %s\n", config->temp_dir);
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

char *swGetTempFile(swWaveConfig config, char *suffix)
{
    static long pid = -1;
    static long id = 0;
    static char filename[SP_MAX_PATHNAME] = "";
    static char suffixbuf[SP_MAX_PATHNAME] = "";
#ifdef SW_USE_THREAD
    static void *temp_file_mutex = NULL;
#endif

    if (pid <= 0) {
	pid = spGetProcessId();
    }
#ifdef SW_USE_THREAD
    if (temp_file_mutex == NULL) {
	temp_file_mutex = spCreateMutex(NULL);
    }
    spLockMutex(temp_file_mutex);
#endif

    if (suffix != NULL) {
	spStrCopy(suffixbuf, SP_MAX_PATHNAME, suffix);
    } else {
	spStrCopy(suffixbuf, SP_MAX_PATHNAME, SW_DEFAULT_TEMP_FILE_SUFFIX);
    }

    if (strnone(config->temp_dir)) {
	sprintf(filename, "spwave%ld_%ld%s", pid, id, suffixbuf);
    } else {
	sprintf(filename, "%sspwave%ld_%ld%s", config->temp_dir, pid, id, suffixbuf);
    }
    id++;

#ifdef SW_USE_THREAD
    spUnlockMutex(temp_file_mutex);
#endif
    
    return filename;
}

char *xswBackupOriginal(char *filename)
{
    char buf[SP_MAX_PATHNAME];
    
    if (spIsFile(filename) == SP_FALSE) return NULL;

    spStrCopy(buf, SP_MAX_PATHNAME, filename);
    while (1) {
	spStrCat(buf, SP_MAX_PATHNAME, "~");
	if (spIsExist(buf) == SP_FALSE) {
	    break;
	}
    }
    
    if (spRenameFile(filename, buf) == SP_TRUE) {
	return strclone(buf);
    } else {
	return NULL;
    }
}

spBool swIsOverflow(int samp_bit, double value)
{
    if (samp_bit > 32) {
	return SP_FALSE;
    } else if (samp_bit >= 32){
	if (value >= -2147483648.0 && value <= 2147483647.0) {
	    return SP_FALSE;
	}
    } else if (samp_bit >= 24){
	if (value >= -8388608.0 && value <= 8388607.0) {
	    return SP_FALSE;
	}
    } else {
	if (value >= -32768.0 && value <= 32767.0) {
	    return SP_FALSE;
	}
    }

    return SP_TRUE;
}

static double swGetClippedValue(int samp_bit, double value, spBool *overflow)
{
    spBool flag = SP_FALSE;
    
    if (samp_bit > 32) {
    } else if (samp_bit >= 32){
	if (value < -2147483647.0) {
	    value = -2147483647.0;
	    flag = SP_TRUE;
	} else if (value > 2147483647.0) {
	    value = 2147483647.0;
	    flag = SP_TRUE;
	} else {
	    value = spRound(value);
	}
    } else if (samp_bit >= 24){
	if (value < -8388607.0) {
	    value = -8388607.0;
	    flag = SP_TRUE;
	} else if (value > 8388607.0) {
	    value = 8388607.0;
	    flag = SP_TRUE;
	} else {
	    value = spRound(value);
	}
    } else {
	if (value < -32767.0) {
	    value = -32767.0;
	    flag = SP_TRUE;
	} else if (value > 32767.0) {
	    value = 32767.0;
	    flag = SP_TRUE;
	} else {
	    value = spRound(value);
	}
    }

    if (flag == SP_TRUE && overflow != NULL) {
	spDebug(80, "swGetClippedValue", "overflow\n");
	*overflow = flag;
    }

    return value;
}

double swGetLimitValue(int samp_bit)
{
    double value = 1.0;
    
    if (samp_bit <= 16) {
	value = 32767.0;
    } else if (samp_bit <= 24) {
	value = 8388607.;
    } else if (samp_bit <= 32) {
	value = 2147483647.0;
    }

    return value;
}

spBool swAllocData(swWave wave, long alloc_length)
{
    long alloc_size;
    
    if (wave == NULL) return SP_FALSE;

    alloc_size = swConvertLengthToByte(wave, alloc_length);

    if (alloc_size <= wave->alloc_size) {
	return SP_TRUE;
    } else {
	wave->data = xrealloc(wave->data, alloc_size, char);
	wave->alloc_size = alloc_size;
    }
    spDebug(10, "swAllocData", "done\n");
    
    return SP_TRUE;
}

spBool swFreeData(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->data != NULL) xfree(wave->data);
    wave->data = NULL;
    wave->alloc_size = 0;

    return SP_TRUE;
}

spBool swAllocBuffer(swWave wave, long alloc_length)
{
    long alloc_size;
    
    if (wave == NULL) return SP_FALSE;

    alloc_size = swConvertLengthToByte(wave, alloc_length);

    if (alloc_size <= wave->buf_alloc_size) {
	return SP_TRUE;
    } else {
	wave->buf = xrealloc(wave->buf, alloc_size, char);
	wave->buf_alloc_size = alloc_size;
    }
    spDebug(80, "swAllocBuffer", "done\n");

    return SP_TRUE;
}

spBool swFreeBuffer(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->buf != NULL) xfree(wave->buf);
    wave->buf = NULL;
    wave->buf_alloc_size = 0;
    wave->buf_load_length = 0;

    return SP_TRUE;
}

spBool swAllocPeakBuffer(swWave wave, long alloc_length)
{
    long alloc_size;
    
    if (wave == NULL) return SP_FALSE;

    alloc_size = swConvertLengthToByte(wave, alloc_length);

    if (alloc_size <= wave->peak_buf_alloc_size) {
	return SP_TRUE;
    } else {
	wave->peak_buf = xrealloc(wave->peak_buf, alloc_size, char);
	wave->peak_buf_alloc_size = alloc_size;
    }
    spDebug(80, "swAllocPeakBuffer", "done\n");

    return SP_TRUE;
}

spBool swFreePeakBuffer(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->peak_buf != NULL) xfree(wave->peak_buf);
    wave->peak_buf = NULL;
    wave->peak_buf_alloc_size = 0;
    wave->peak_buf_thin_length = 0;

    return SP_TRUE;
}

spBool swAllocReadBuffer(swWave wave, long alloc_size)
{
    if (wave == NULL) return SP_FALSE;

    if (alloc_size <= wave->readbuf_size) {
	return SP_TRUE;
    } else {
	wave->readbuf = xrealloc(wave->readbuf, alloc_size, char);
	wave->readbuf_size = alloc_size;
    }
    
    return SP_TRUE;
}

spBool swFreeReadBuffer(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->readbuf != NULL) {
	xfree(wave->readbuf);
    }
    wave->readbuf = NULL;
    wave->readbuf_size = 0;
    
    return SP_TRUE;
}

spBool swAllocEditBuffer(swWave wave, long alloc_length)
{
    if (wave == NULL) return SP_FALSE;

    alloc_length *= (long)wave->num_channel;

    if (alloc_length <= wave->editbuf_length) {
	return SP_TRUE;
    } else {
	wave->editbuf = xrealloc(wave->editbuf,
				 alloc_length * sizeof(double), char);
	wave->editbuf_length = alloc_length;
    }
    
    return SP_TRUE;
}

spBool swFreeEditBuffer(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->editbuf != NULL) {
	xfree(wave->editbuf);
    }
    wave->editbuf = NULL;
    wave->editbuf_length = 0;
    
    return SP_TRUE;
}

spBool swInitWaveRange(swWave wave)
{
    if (wave == NULL) return SP_FALSE;

    wave->min = 0; wave->max = -1;
    wave->lmin = 0; wave->lmax = -1;
    wave->dmin = 0.0; wave->dmax = -1.0;
    wave->range_available = SP_FALSE;

    return SP_TRUE;
}

void swRemovePeakFile(swWave wave)
{
    if (wave->peak_filename != NULL) {
	spRemoveFile(wave->peak_filename);
	xfree(wave->peak_filename);
    }
    wave->peak_filename = NULL;
    wave->peak_length = 0;
    
    return;
}

swWave swInitWave(swWaveConfig config, char *filename, char *plugin_name,  char *file_type,
		  int samp_bit, int num_channel, double samp_rate, spBool orig_flag)
{
    swWave wave = NULL;

    wave = xalloc(1, struct _swWave);

    wave->edit_thread = NULL;
    wave->audio_thread = NULL;
#ifdef SW_USE_THREAD
    wave->mutex = spCreateMutex(NULL);
    wave->io_mutex = spCreateMutex(NULL);
#else
    wave->mutex = NULL;
    wave->io_mutex = NULL;
#endif
    
    if (plugin_name != NULL) {
	wave->plugin_name = strclone(plugin_name);
    } else {
	wave->plugin_name = NULL;
    }
    if (filename != NULL) {
	wave->filename = strclone(filename);
    } else {
	wave->filename = NULL;
    }
    if (file_type != NULL) {
	spDebug(40, "swInitWave", "file_type = %s\n", file_type);
	wave->file_type = strclone(file_type);
	wave->orig_file_type = strclone(wave->file_type);
    } else {
	wave->file_type = NULL;
	wave->orig_file_type = strclone("None");
    }
    if (orig_flag == SP_TRUE) {
	if (wave->plugin_name != NULL) {
	    wave->orig_plugin_name = strclone(wave->plugin_name);
	} else {
	    wave->orig_plugin_name = NULL;
	}
	if (wave->filename != NULL) {
	    wave->orig_filename = strclone(wave->filename);
	} else {
	    wave->orig_filename = NULL;
	}
    } else {
	wave->orig_plugin_name = NULL;
	wave->orig_filename = NULL;
    }
    wave->new_plugin_name = NULL;
    wave->new_filename = NULL;
    wave->new_file_type = NULL;
    
    wave->orig_flag = orig_flag;
    wave->edit_flag = SP_FALSE;
    wave->overflow = SP_FALSE;
    wave->thread_safe = SP_FALSE;
    wave->num_channel = num_channel;
    wave->selected_channel = -1;
    wave->samp_bit = samp_bit;
    wave->samp_rate = samp_rate;
    wave->exact_samp_rate = samp_rate;
    wave->song_info_mask = 0;
    spInitSongInfo(&wave->song_info);

    wave->offset = 0;
    wave->length = 0;
    wave->data_length = 0;
    wave->total_length = 0;
    wave->thin_length = 1;
    wave->alloc_size = 0;
    if (wave->samp_bit > 32) {
	wave->float_flag = SP_TRUE;
    } else {
	wave->float_flag = SP_FALSE;
    }

    wave->data = NULL;
    wave->min = 0;
    wave->max = -1;
    wave->lmin = 0;
    wave->lmax = -1;
    wave->dmin = 0.0;
    wave->dmax = -1.0;
    wave->range_available = SP_FALSE;

    wave->peak_filename = NULL;
    wave->peak_length = 0;
    wave->peak_thin_length = SW_DEFAULT_PEAK_THIN_LENGTH;

    wave->peak_buf_alloc_size = 0;
    wave->peak_buf_length = 0;
    wave->peak_buf_thin_length = 0;
    wave->peak_buf = NULL;

    wave->buf_alloc_size = 0;
    wave->buf_length = 0;
    wave->buf_load_length = 0;
    wave->buf = NULL;
    swAllocBuffer(wave, config->buffer_length);

    wave->readbuf_size = 0;
    wave->readbuf = NULL;
    swAllocReadBuffer(wave, swConvertLengthToByte(wave, config->normal_read_length));
    
    wave->editbuf_length = 0;
    wave->editbuf = NULL;
    
    wave->edit_offset = 0;
    wave->edit_length = 0;
    wave->play_offset = 0;
    wave->play_length = 0;

    wave->process_finished = SP_TRUE;
    wave->process_flag = SP_FALSE;
    wave->play_flag = SP_FALSE;
    wave->call_data = NULL;

    wave->labels = NULL;
    
    wave->config = config;

    wave->edit_name = NULL;
    wave->prev_wave = NULL;
    wave->next_wave = NULL;

    return wave;
}

void _swFreeWave(swWave wave)
{
    if (wave != NULL) {
	if (wave->prev_wave != NULL) {
	    wave->prev_wave->next_wave = wave->next_wave;
	}
	if (wave->next_wave != NULL) {
	    wave->next_wave->prev_wave = wave->prev_wave;
	}

	if (wave->edit_thread != NULL) {
	    spThreadLeave();
	    spWaitThread(wave->edit_thread);
	    spThreadEnter();
	    spDestroyThread(wave->edit_thread);
	}
	if (wave->audio_thread != NULL) {
	    spThreadLeave();
	    spWaitThread(wave->audio_thread);
	    spThreadEnter();
	    spDestroyThread(wave->audio_thread);
	}
	if (wave->mutex != NULL) {
	    spDestroyMutex(wave->mutex);
	}
	if (wave->io_mutex != NULL) {
	    spDestroyMutex(wave->io_mutex);
	}
	
	swRemovePeakFile(wave);

	if (wave->orig_flag == SP_FALSE) {
	    spRemoveFile(wave->filename);
	    spDebug(10, "swFreeWave", "remove file done: %s\n", wave->filename);
	}
	
	if (wave->plugin_name != NULL) xfree(wave->plugin_name);
	if (wave->filename != NULL) xfree(wave->filename);
	if (wave->file_type != NULL) xfree(wave->file_type);
	if (wave->orig_plugin_name != NULL) xfree(wave->orig_plugin_name);
	if (wave->orig_filename != NULL) xfree(wave->orig_filename);
	if (wave->orig_file_type != NULL) xfree(wave->orig_file_type);
	if (wave->new_plugin_name != NULL) xfree(wave->new_plugin_name);
	if (wave->new_filename != NULL) xfree(wave->new_filename);
	if (wave->new_file_type != NULL) xfree(wave->new_file_type);

	/* free buffers */
	swFreeData(wave);
	swFreeBuffer(wave);
	swFreeReadBuffer(wave);
	swFreeEditBuffer(wave);
	swFreePeakBuffer(wave);
	
	if (wave->labels != NULL) swFreeLabels(wave->labels);
	
	if (wave->edit_name != NULL) xfree(wave->edit_name);
	
	xfree(wave);
    }
    return;
}

void _swDestroyWave(swWave wave)
{
    swWave temp, temp2;
    
    if (wave != NULL) {
	temp = wave->prev_wave;
	while (temp != NULL) {
	    temp2 = temp->prev_wave;
	    swFreeWave(temp);
	    temp = temp2;
	}

	temp = wave->next_wave;
	while (temp != NULL) {
	    temp2 = temp->next_wave;
	    swFreeWave(temp);
	    temp = temp2;
	}
	
	swFreeWave(wave);
    }
    return;
}

unsigned long swGetSongInfoMask(swWave wave)
{
    if (wave == NULL) {
	return SP_SONG_NO_INFO;
    }

    return wave->song_info_mask;
}

typedef struct _swCallbackData {
    swWaveConfig config;
    swWave wave;
    spBool in_thread;
    void *call_data;
} swCallbackData;

static spBool pluginOpenCB(spPlugin *o_plugin, spPluginCallbackReason reason,
			   void *host_data, void *data)
{
    spBool flag = SP_TRUE;
    swCallbackData *call_data = (swCallbackData *)data;

    if (reason == SP_PLUGIN_CR_OPTION) {
	spOptions options = (spOptions)host_data;
	
	if (call_data->wave != NULL && call_data->config->option_func != NULL) {
	    if (call_data->config->option_func(call_data->wave, call_data->in_thread,
					       options, call_data->call_data) == SP_FALSE) {
		flag = SP_FALSE;
	    }
	}
    } else if (reason == SP_PLUGIN_CR_ERROR) {
	int error = (int)host_data;
	
	if (call_data->config->error_func != NULL) {
	    if (call_data->config->error_func(call_data->wave, call_data->in_thread,
					      error, call_data->call_data) == SP_FALSE) {
		flag = SP_FALSE;
	    }
	}
    }

    return flag;
}

static swWave getWave(swWaveConfig config, char *filename, char *plugin_name, char *file_type,
		      int samp_bit, int num_channel, double samp_rate,
		      spBool orig_flag, spBool callback_flag, spPluginError *error, spBool in_thread)
{
    swWave wave;
    spPlugin *plugin;
    spWaveInfo wave_info;
    spSongInfo song_info;
    spPluginError err;
    spPluginOpenCallback opencb = NULL;
    swCallbackData call_data = {NULL, NULL, SP_FALSE, NULL};
    
    if (strnone(filename)) return NULL;

    /* initialize option */
    spInitWaveInfo(&wave_info);
    if (orig_flag == SP_TRUE) {
	wave_info.samp_bit = samp_bit;
    } else {
	wave_info.samp_bit = MAX(samp_bit, 16);
    }
    wave_info.num_channel = num_channel;
    wave_info.samp_rate = samp_rate;
    if (!strnone(file_type)) {
	spStrCopy(wave_info.file_type, SP_WAVE_FILE_TYPE_SIZE, file_type);
    }

    if (callback_flag == SP_TRUE) {
	opencb = (spPluginOpenCallback)pluginOpenCB;
	call_data.config = config;
	call_data.in_thread = in_thread;
	call_data.wave = NULL;
	call_data.call_data = NULL;
    }
    
    /* open file */
    if ((plugin = spOpenFilePlugin(NULL, filename, "r",
				   SP_PLUGIN_DEVICE_FILE,
				   &wave_info, &song_info, opencb, (void *)&call_data, &err)) == NULL) {
	if (error != NULL) *error = err;
	
	if (err != SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND && err != SP_PLUGIN_ERROR_FAILURE) {
	    return NULL;
	}
	spDebug(10, "getWave", "can't find suitable plugin: err = %d\n", err);
	if (strnone(plugin_name)) {
	    if (strnone(file_type)) {
		return NULL;
	    }
		
	    if (strncaseeq(file_type, "text", 4)
		|| strcaseeq(file_type, "time")
		|| strcaseeq(file_type, "freq")
		|| strcaseeq(file_type, SP_WAVE_FORMAT_TEXT_LABEL)
		|| strcaseeq(file_type, SP_WAVE_FORMAT_TEXT_TIME_LABEL)
		|| strcaseeq(file_type, SP_WAVE_FORMAT_TEXT_FREQ_LABEL)) {
		plugin_name = "input_text";
	    } else {
		spStrCopy(wave_info.file_type, SP_WAVE_FILE_TYPE_SIZE, file_type);
		plugin_name = "input_raw";
	    }

	    spDebug(10, "getWave", "before open: file_type = %s\n", file_type);
	}
	spDebug(10, "getWave", "before open: plugin_name = %s\n", plugin_name);
	
	if ((plugin = spOpenFilePlugin(plugin_name, filename, "r",
				       SP_PLUGIN_DEVICE_FILE,
				       &wave_info, &song_info, /*opencb*/NULL, (void *)&call_data, error)) == NULL) {
	    spDebug(10, "getWave", "can't open file: %s\n", filename);
	    return NULL;
	}
    }
    if (orig_flag == SP_TRUE) {
	samp_bit = wave_info.samp_bit;
    }

    spDebug(10, "getWave", "after open: file_type = %s\n", wave_info.file_type);
		
    wave = swInitWave(config, filename, spGetPluginName(plugin),
		      spGetPluginFileType(plugin, SP_TRUE),
		      samp_bit, wave_info.num_channel, wave_info.samp_rate,
		      orig_flag);
    spCopySongInfo(&wave->song_info, &song_info);
    if (orig_flag == SP_TRUE) {
	spGetPluginSongInfoMask(plugin, &wave->song_info_mask);
    }

    wave->thread_safe = spIsPluginCapable(plugin, SP_PLUGIN_CAPS_THREAD_SAFE);
    
    if ((wave->total_length = spGetPluginTotalLength(plugin)) <= 0) {
	wave->total_length = config->max_read_length;
    }
    spCloseFilePlugin(plugin);

    spDebug(10, "getWave", "total_length = %ld\n", wave->total_length);
    
    return wave;
}

swWave swGetWave(swWaveConfig config, char *filename, char *plugin_name, char *file_type,
		 int samp_bit, int num_channel, double samp_rate,
		 spBool orig_flag, spBool callback_flag, spPluginError *error)
{
    return getWave(config, filename, plugin_name, file_type,
		   samp_bit, num_channel, samp_rate,
		   orig_flag, callback_flag, error, SP_FALSE);
}

spPlugin *swOpenWave(swWave wave, char *mode)
{
    spWaveInfo wave_info;
    char *plugin_name;
    spPlugin *plugin;
    
    if (wave == NULL || strnone(wave->plugin_name) || mode == NULL) {
	return NULL;
    }

    if (strnone(wave->filename)) return SP_FALSE;
    
    if (mode[0] == 'w') {
	if ((plugin_name = xspFindRelatedPluginFile(wave->plugin_name)) == NULL) {
	    return NULL;
	}
    } else {
	plugin_name = strclone(wave->plugin_name);
    }
    spDebug(20, "swOpenWave", "plugin_name = %s\n", plugin_name);
    
    /* initialize option */
    spInitWaveInfo(&wave_info);
    if (wave->file_type != NULL) {
	spDebug(20, "swOpenWave", "format = %s\n", wave->file_type);
	spStrCopy(wave_info.file_type, SP_WAVE_FILE_TYPE_SIZE, wave->file_type);
    }
    if (wave->orig_flag == SP_TRUE) {
	wave_info.samp_bit = wave->samp_bit;
    } else {
	wave_info.samp_bit = MAX(wave->samp_bit, 16);
    }
    wave_info.num_channel = wave->num_channel;
    wave_info.samp_rate = wave->samp_rate;
    spDebug(20, "swOpenWave", "num_channel = %d, samp_bit = %d\n",
	    wave_info.num_channel, wave_info.samp_bit);

    /* open file */
    if ((plugin = spOpenFilePlugin(plugin_name, wave->filename, mode,
				   SP_PLUGIN_DEVICE_FILE,
				   &wave_info, NULL, NULL, NULL, NULL)) == NULL) {
	spDebug(20, "swOpenWave", "spOpenFilePlugin error\n");
	xfree(plugin_name);
	return NULL;
    }

    spDebug(10, "swOpenWave", "file_type = %s\n", wave_info.file_type);
    spDebug(1, "swOpenWave", "total_length = %ld\n", wave->total_length);
    
    xfree(plugin_name);
    
    return plugin;
}

static void getThinReadLength(swWave wave, long length, long thin_length, long *read_length)
{
    if (read_length != NULL) {
	*read_length = 2 * (long)ceil((double)((length) * (long)wave->num_channel)
				      / (double)(2 * thin_length * (long)wave->num_channel));
    }

    return;
}

static long getThinLength(swWave wave, long length, long thin_length, long *read_length)
{
    if (thin_length != 1 /*&& length >= wave->config->max_read_length*/) {
	if (thin_length <= 0) {
	    thin_length = (long)ceil((double)length / (double)wave->config->max_read_length);
	    if (thin_length >= SW_MIN_THIN_LENGTH_USING_PEAK) {
		thin_length = MAX(wave->peak_thin_length, thin_length);
	    }
	}
	getThinReadLength(wave, length, thin_length, read_length);
    } else {
	thin_length = 1;
	if (read_length != NULL) {
	    *read_length = length;
	}
    }

    return thin_length;
}

static long readPeakFile(swWave wave, long offset, long thin_length, long read_length,
			 char *data, spBool callback_flag, spBool in_thread)
{
    spBool flag = SP_TRUE;
    long k, l;
    long ndata;
    long samp_byte;
    long pos, prev_pos;
    long nread;
    long buf_pos;
    long data_pos;
    long prev_call_pos;
    long read_pos;
    long data_length;
    long buf_length;
    double value;
    long *minindex, *maxindex;
    double *minvalue, *maxvalue;
    short *sbuf, *sdata;
    long *lbuf, *ldata;
    double *dbuf, *ddata;
    spBool play_flag, prev_process_flag;
    FILE *peak_fp;
    
    if (wave == NULL || offset < 0 || read_length <= 0
	|| thin_length < wave->peak_thin_length	|| wave->peak_filename == NULL) return -1;

    spLockMutex(wave->io_mutex);
    
    if ((peak_fp = fopen(wave->peak_filename, "rb")) == NULL) {
	spUnlockMutex(wave->io_mutex);
	return -1;
    }
    spDebug(10, "readPeakFile", "wave->peak_filename = %s\n", wave->peak_filename);
    
    if (callback_flag == SP_TRUE && wave->config->read_func != NULL) {
	flag = wave->config->read_func(wave, in_thread, SW_PROCESS_STARTED, wave->call_data);
    }
    
    data_pos = 0;
    prev_call_pos = 0;
    read_pos = 0;
    
    if (flag == SP_TRUE) {
	buf_length = MIN(swConvertByteToLength(wave, wave->readbuf_size), read_length);
	data_length = read_length * (long)wave->num_channel;

	samp_byte = swGetWaveSampleByte(wave);
	
	spDebug(10, "readPeakFile", "buf_length = %ld, data_length = %ld\n", buf_length, data_length);
    
	fseek(peak_fp, samp_byte * 2 * (offset / (2 * wave->peak_thin_length)) * (long)wave->num_channel, 0);
	
	sbuf = (short *)wave->readbuf;
	lbuf = (long *)wave->readbuf;
	dbuf = (double *)wave->readbuf;
	sdata = (short *)data;
	ldata = (long *)data;
	ddata = (double *)data;

	minindex = xalloc(wave->num_channel, long);
	maxindex = xalloc(wave->num_channel, long);
	minvalue = xalloc(wave->num_channel, double);
	maxvalue = xalloc(wave->num_channel, double);
	for (l = 0; l < wave->num_channel; l++) {
	    minindex[l] = -1; maxindex[l] = -1;
	    minvalue[l] = 0.0; maxvalue[l] = 0.0;
	}
    
	prev_process_flag = wave->process_flag;
	play_flag = wave->play_flag;

	prev_pos = 0;
	while (1) {
	    if ((nread = fread(wave->readbuf, samp_byte, buf_length * wave->num_channel, peak_fp)) <= 0) {
		break;
	    }
	
	    for (buf_pos = 0; buf_pos < nread; buf_pos += wave->num_channel) {
		if (thin_length == wave->peak_thin_length) {
		    for (l = 0; l < wave->num_channel; l++) {
			k = buf_pos + l;
			if (swIsWaveFloat(wave) == SP_TRUE) {
			    ddata[data_pos] = dbuf[k];
			} else if (swIsWaveLong(wave) == SP_TRUE) {
			    ldata[data_pos] = lbuf[k];
			} else {
			    sdata[data_pos] = sbuf[k];
			}
			data_pos++;
		    }
		} else {
		    pos = ((read_pos + buf_pos) / wave->num_channel) * wave->peak_thin_length;
		    for (l = 0; l < wave->num_channel; l++) {
			k = buf_pos + l;
			if (swIsWaveFloat(wave) == SP_TRUE) {
			    value = dbuf[k];
			} else if (swIsWaveLong(wave) == SP_TRUE) {
			    value = (double)lbuf[k];
			} else {
			    value = (double)sbuf[k];
			}
		    
			if (maxindex[l] < 0 || value > maxvalue[l]) {
			    maxvalue[l] = value;
			    maxindex[l] = pos;
			}
			if (minindex[l] < 0 || value < minvalue[l]) {
			    minvalue[l] = value;
			    minindex[l] = pos;
			}
		    }
		    
		    if ((pos / wave->peak_thin_length + 1) % 2 == 0 
			&& pos - prev_pos + wave->peak_thin_length >= 2 * thin_length) {
			for (k = 0; k < 2; k++) {
			    for (l = 0; l < wave->num_channel; l++) {
				if (k == 0) {
				    if (minindex[l] < maxindex[l]) {
					value = minvalue[l];
				    } else {
					value = maxvalue[l];
				    }
				} else {
				    if (minindex[l] < maxindex[l]) {
					value = maxvalue[l];
				    } else {
					value = minvalue[l];
				    }
				}
			
				if (swIsWaveFloat(wave) == SP_TRUE) {
				    ddata[data_pos] = value;
				} else if (swIsWaveLong(wave) == SP_TRUE) {
				    ldata[data_pos] = (long)value;
				} else {
				    sdata[data_pos] = (short)value;
				}
				data_pos++;
			    }
			}
			
			for (l = 0; l < wave->num_channel; l++) {
			    minindex[l] = -1; maxindex[l] = -1;
			    minvalue[l] = 0.0; maxvalue[l] = 0.0;
			}
		
			prev_pos += 2 * thin_length;
		    }
		}
		
		if (data_pos >= data_length) {
		    nread = 0;
		    break;
		}
	    }

	    if (nread <= 0) {
		break;
	    }
	    if (play_flag == SP_FALSE && prev_process_flag != wave->process_flag) {
		data_pos = 0;
		break;
	    }

	    spDebug(80, "readPeakFile", "data_pos = %ld, prev_call_pos = %ld\n", data_pos, prev_call_pos);
	    if (callback_flag == SP_TRUE && wave->config->read_func != NULL
		&& thin_length * (data_pos - prev_call_pos) >= wave->config->read_callback_length) {
		flag = wave->config->read_func(wave, in_thread, data_pos / (long)wave->num_channel, wave->call_data);
		prev_call_pos = data_pos;
		if (flag == SP_FALSE) {
		    break;
		}
	    }

	    read_pos += nread;
	}
    
	xfree(minindex); xfree(maxindex);
	xfree(minvalue); xfree(maxvalue);
    }
    
    if (callback_flag == SP_TRUE && wave->config->read_func != NULL) {
	flag = wave->config->read_func(wave, in_thread, SW_PROCESS_FINISHED, wave->call_data);
    }
    
    fclose(peak_fp);
    
    spUnlockMutex(wave->io_mutex);

    ndata = data_pos / (long)wave->num_channel;
    spDebug(80, "readPeakFile", "thin_length = %ld, ndata = %ld\n",
	    thin_length, ndata);
	
    return ndata;
}

static long readThinWaveUsingPeak(swWave wave, long offset, long thin_length, long read_length,
				  spBool callback_flag, spBool in_thread)
{
    long nread;
    
    spDebug(80, "readThinWaveUsingPeak",
	    "offset = %ld, thin_length = %ld, read_length = %ld\n",
	    offset, thin_length, read_length);
    
    swAllocData(wave, 256 * (((read_length - 1) / 256) + 1));

    spDebug(80, "readThinWaveUsingPeak",
	    "peak_buf_length = %ld, peak_buf_thin_length = %ld\n",
	    wave->peak_buf_length, wave->peak_buf_thin_length);
    
    if (thin_length == wave->peak_buf_thin_length) {
	long offsetbyte;
	long readbyte;
	long peak_offset;
	long avail_length;

	spLockMutex(wave->io_mutex);
	
	peak_offset = 2 * (offset / (2 * thin_length));
	avail_length = wave->peak_buf_length - peak_offset;
	spDebug(80, "readThinWaveUsingPeak", "avail_length = %ld\n", avail_length);
	
	if (read_length > avail_length) {
	    spDebug(80, "readThinWaveUsingPeak",
		    "******** read_length %ld over peak_buf_length %ld\n",
		    read_length, avail_length);
	    read_length = avail_length;
	}

	offsetbyte = swConvertLengthToByte(wave, peak_offset);
	readbyte = swConvertLengthToByte(wave, read_length);
	spDebug(100, "readThinWaveUsingPeak",
		"peak_offset = %ld, offsetbyte = %ld, readbyte = %ld\n",
		peak_offset, offsetbyte, readbyte);
	memmove(wave->data, wave->peak_buf + offsetbyte, readbyte);
	
	nread = read_length;
	
	spUnlockMutex(wave->io_mutex);
    } else {
	nread = readPeakFile(wave, offset, thin_length, read_length,
			     wave->data, callback_flag, in_thread);
    }

    if (nread > 0) {
	swLockMutex(wave);
	wave->length = nread;
	swUnlockMutex(wave);
    }

    return nread;
}

static long readPeakBuffer(swWave wave, spBool in_thread)
{
    long thin_length;
    long read_length;
    long peak_buf_length;
    
    thin_length = getThinLength(wave, wave->total_length, 0, &read_length);
    if (thin_length < wave->peak_thin_length) {
	thin_length = wave->peak_thin_length;
	getThinReadLength(wave, wave->total_length, thin_length, &read_length);
    }

    spDebug(80, "readPeakBuffer", "thin_length = %ld, peak_thin_length = %ld, read_length = %ld\n",
	    thin_length, wave->peak_thin_length, read_length);
    
    swAllocPeakBuffer(wave, 256 * (((read_length - 1) / 256) + 1));

    if ((peak_buf_length = readPeakFile(wave, 0, thin_length, read_length,
					wave->peak_buf, SP_FALSE, in_thread)) > 0) {
	swLockMutex(wave);
	wave->peak_buf_length = peak_buf_length;
	wave->peak_buf_thin_length = thin_length;
	spDebug(80, "readPeakBuffer", "peak_buf_length = %ld, peak_buf_thin_length = %ld\n",
		wave->peak_buf_length, wave->peak_buf_thin_length);
	swUnlockMutex(wave);
    }
    
    spDebug(80, "readPeakBuffer", "peak_buf_length = %ld\n", peak_buf_length);

    return peak_buf_length;
}

static long readData(spPlugin *plugin, swWave wave, long read_length,
		     spBool get_range, spBool callback_flag, spBool in_thread)
{
    spBool flag = SP_TRUE;
    spBool make_peak = SP_FALSE;
    long k, l;
    long offset;
    long nread, nwrite;
    long buf_pos, data_pos;
    long prev_call_pos;
    long read_pos;
    long data_length;
    long buf_length;
    double value;
    long *minindex, *maxindex;
    double *minvalue, *maxvalue;
    long *peak_minindex, *peak_maxindex;
    double *peak_minvalue, *peak_maxvalue;
    short *sbuf, *sdata;
    long *lbuf, *ldata;
    double *dbuf, *ddata;
    spBool play_flag, prev_process_flag;
    char *filename;
    FILE *peak_fp = NULL;
    
    if (wave == NULL || plugin == NULL || read_length <= 0) return -1;

    spLockMutex(wave->io_mutex);

    buf_length = MIN(swConvertByteToLength(wave, wave->readbuf_size), read_length);
    data_length = read_length * (long)wave->num_channel;
    spDebug(10, "readData", "buf_length = %ld, data_length = %ld\n", buf_length, data_length);

    if (get_range == SP_TRUE
	|| swIsWaveRangeAvailable(wave) == SP_FALSE) {
	get_range = SP_TRUE;
	swInitWaveRange(wave);
	
	swLockMutex(wave);
	filename = swGetTempFile(wave->config, SW_PEAK_FILE_SUFFIX);
	spDebug(10, "readData", "create new peak: %s\n", filename);
	    
	if ((peak_fp = fopen(filename, "wb")) != NULL) {
	    swRemovePeakFile(wave);
	    wave->peak_filename = strclone(filename);
	    make_peak = SP_TRUE;

	    wave->peak_thin_length = 1 + wave->total_length / wave->config->max_read_length;
	    wave->peak_thin_length = MIN(wave->peak_thin_length, SW_DEFAULT_PEAK_THIN_LENGTH);
	}
	swUnlockMutex(wave);
    }

    if (callback_flag == SP_TRUE && wave->config->read_func != NULL) {
	flag = wave->config->read_func(wave, in_thread, SW_PROCESS_STARTED, wave->call_data);
    }
    
    data_pos = 0;
    prev_call_pos = 0;
    read_pos = 0;
    
    if (flag == SP_TRUE) {
	sbuf = (short *)wave->readbuf;
	lbuf = (long *)wave->readbuf;
	dbuf = (double *)wave->readbuf;
	sdata = (short *)wave->data;
	ldata = (long *)wave->data;
	ddata = (double *)wave->data;

	minindex = xalloc(wave->num_channel, long);
	maxindex = xalloc(wave->num_channel, long);
	minvalue = xalloc(wave->num_channel, double);
	maxvalue = xalloc(wave->num_channel, double);
	for (l = 0; l < wave->num_channel; l++) {
	    minindex[l] = -1; maxindex[l] = -1;
	    minvalue[l] = 0.0; maxvalue[l] = 0.0;
	}
	if (make_peak == SP_TRUE) {
	    peak_minindex = xalloc(wave->num_channel, long);
	    peak_maxindex = xalloc(wave->num_channel, long);
	    peak_minvalue = xalloc(wave->num_channel, double);
	    peak_maxvalue = xalloc(wave->num_channel, double);
	    for (l = 0; l < wave->num_channel; l++) {
		peak_minindex[l] = -1; peak_maxindex[l] = -1;
		peak_minvalue[l] = 0.0; peak_maxvalue[l] = 0.0;
	    }
	} else {
	    peak_minindex = NULL; peak_maxindex = NULL;
	    peak_minvalue = NULL; peak_maxvalue = NULL;
	}

	prev_process_flag = wave->process_flag;
	play_flag = wave->play_flag;
	
	while (1) {
	    if ((nread = swLoadIntoBuffer(plugin, wave, -1, buf_length, wave->readbuf)) <= 0) {
		break;
	    }
	    nread *= wave->num_channel;

	    for (buf_pos = 0; buf_pos < nread; buf_pos += wave->num_channel) {
		offset = (read_pos + buf_pos) / wave->num_channel;
		for (l = 0; l < wave->num_channel; l++) {
		    k = buf_pos + l;
		    if (swIsWaveFloat(wave) == SP_TRUE) {
			value = dbuf[k];
			if (get_range == SP_TRUE) {
			    if (dbuf[k] < wave->dmin) wave->dmin = dbuf[k];
			    if (dbuf[k] > wave->dmax) wave->dmax = dbuf[k];
			}
		    } else if (swIsWaveLong(wave) == SP_TRUE) {
			value = (double)lbuf[k];
			if (get_range == SP_TRUE) {
			    if (lbuf[k] < wave->lmin) wave->lmin = lbuf[k];
			    if (lbuf[k] > wave->lmax) wave->lmax = lbuf[k];
			}
		    } else {
			value = (double)sbuf[k];
			if (get_range == SP_TRUE) {
			    if (sbuf[k] < wave->min) wave->min = sbuf[k];
			    if (sbuf[k] > wave->max) wave->max = sbuf[k];
			}
		    }
		    
		    if (maxindex[l] < 0 || value > maxvalue[l]) {
			maxvalue[l] = value;
			maxindex[l] = offset;
		    }
		    if (minindex[l] < 0 || value < minvalue[l]) {
			minvalue[l] = value;
			minindex[l] = offset;
		    }

		    if (make_peak == SP_TRUE && wave->peak_length >= 0) {
			if (peak_maxindex[l] < 0 || value > peak_maxvalue[l]) {
			    peak_maxvalue[l] = value;
			    peak_maxindex[l] = offset;
			}
			if (peak_minindex[l] < 0 || value < peak_minvalue[l]) {
			    peak_minvalue[l] = value;
			    peak_minindex[l] = offset;
			}
		    }
		}

		if (make_peak == SP_TRUE && wave->peak_length >= 0) {
		    if ((offset + 1) % (2 * wave->peak_thin_length) == 0) {
			for (k = 0; k < 2; k++) {
			    for (l = 0; l < wave->num_channel; l++) {
				if (k == 0) {
				    if (peak_minindex[l] < peak_maxindex[l]) {
					value = peak_minvalue[l];
				    } else {
					value = peak_maxvalue[l];
				    }
				} else {
				    if (peak_minindex[l] < peak_maxindex[l]) {
					value = peak_maxvalue[l];
				    } else {
					value = peak_minvalue[l];
				    }
				}

				if (swIsWaveFloat(wave) == SP_TRUE) {
				    nwrite = fwritedouble(&value, 1, 0, peak_fp);
				} else if (swIsWaveLong(wave) == SP_TRUE) {
				    nwrite = fwritedoubletol(&value, 1, 0, peak_fp);
				} else {
				    nwrite = fwritedoubletos(&value, 1, 0, peak_fp);
				}

				if (nwrite > 0) {
				    wave->peak_length += nwrite;
				} else {
				    wave->peak_length = -1;
				    break;
				}
			    }
			    if (wave->peak_length < 0) {
				break;
			    }
			}
			
			for (l = 0; l < wave->num_channel; l++) {
			    peak_minindex[l] = -1; peak_maxindex[l] = -1;
			    peak_minvalue[l] = 0.0; peak_maxvalue[l] = 0.0;
			}
		    }
		}
		
		if (wave->thin_length > 1) {
		    if ((offset + 1) % (2 * wave->thin_length) == 0) {
			for (k = 0; k < 2; k++) {
			    for (l = 0; l < wave->num_channel; l++) {
				if (k == 0) {
				    if (minindex[l] < maxindex[l]) {
					value = minvalue[l];
				    } else {
					value = maxvalue[l];
				    }
				} else {
				    if (minindex[l] < maxindex[l]) {
					value = maxvalue[l];
				    } else {
					value = minvalue[l];
				    }
				}
			
				if (swIsWaveFloat(wave) == SP_TRUE) {
				    ddata[data_pos] = value;
				} else if (swIsWaveLong(wave) == SP_TRUE) {
				    ldata[data_pos] = (long)value;
				} else {
				    sdata[data_pos] = (short)value;
				}
				data_pos++;
			    }
			}
			
			for (l = 0; l < wave->num_channel; l++) {
			    minindex[l] = -1; maxindex[l] = -1;
			    minvalue[l] = 0.0; maxvalue[l] = 0.0;
			}
		    }
		} else {
		    for (l = 0; l < wave->num_channel; l++) {
			k = buf_pos + l;
			if (swIsWaveFloat(wave) == SP_TRUE) {
			    ddata[data_pos] = dbuf[k];
			} else if (swIsWaveLong(wave) == SP_TRUE) {
			    ldata[data_pos] = lbuf[k];
			} else {
			    sdata[data_pos] = sbuf[k];
			}
			data_pos++;
		    }
		}
		
		if (data_pos >= data_length) {
		    nread = 0;
		    break;
		}
	    }

	    if (nread <= 0) {
		break;
	    }
	    if (play_flag == SP_FALSE && prev_process_flag != wave->process_flag) {
		data_pos = 0;
		break;
	    }

	    spDebug(80, "readData", "data_pos = %ld, prev_call_pos = %ld\n", data_pos, prev_call_pos);
	    if (callback_flag == SP_TRUE && wave->config->read_func != NULL
		&& wave->thin_length * (data_pos - prev_call_pos) >= wave->config->read_callback_length) {
		flag = wave->config->read_func(wave, in_thread, data_pos / (long)wave->num_channel, wave->call_data);
		prev_call_pos = data_pos;
		if (flag == SP_FALSE) {
		    break;
		}
	    }

	    read_pos += nread;
	}
	
	if (make_peak == SP_TRUE) {
	    xfree(peak_minindex); xfree(peak_maxindex);
	    xfree(peak_minvalue); xfree(peak_maxvalue);
	}
	xfree(minindex); xfree(maxindex);
	xfree(minvalue); xfree(maxvalue);
    }

    if (callback_flag == SP_TRUE && wave->config->read_func != NULL) {
	flag = wave->config->read_func(wave, in_thread, SW_PROCESS_FINISHED, wave->call_data);
    }
    spDebug(10, "readData", "data_pos = %ld\n", data_pos);

    if (make_peak == SP_TRUE) {
	swLockMutex(wave);
	fclose(peak_fp);

	spDebug(10, "readData", "wave->peak_length = %ld\n", wave->peak_length);
	if (wave->peak_length <= 0) {
	    swRemovePeakFile(wave);
	} else {
	    wave->peak_length /= wave->num_channel;
	}
	swUnlockMutex(wave);
    }
    
    if (data_pos > 0 && get_range == SP_TRUE) {
	swLockMutex(wave);
	wave->range_available = SP_TRUE;
	swUnlockMutex(wave);
    }
    
    spUnlockMutex(wave->io_mutex);
    
    if (make_peak == SP_TRUE && wave->peak_length > 0) {
	readPeakBuffer(wave, in_thread);
    }
    
    return data_pos / (long)wave->num_channel;
}

static long readThinWaveDirectly(swWave wave, long offset, long thin_length, long read_length,
				 spBool get_range, spBool callback_flag, spBool in_thread)
{
    long ndata;
    spPlugin *plugin;
    
    if (wave == NULL || read_length <= 0 || offset < 0) return -1;

    spDebug(20, "readThinWaveDirectly", "offset = %ld, read_length = %ld\n", offset, read_length);
    
    if ((plugin = swOpenWave(wave, "r")) == NULL) {
	return -1;
    }
    spDebug(20, "readThinWaveDirectly", "open done\n");

    swAllocData(wave, 256 * (((read_length - 1) / 256) + 1));

    spSeekPlugin(plugin, offset);
    spDebug(20, "readThinWaveDirectly", "seek done\n");
    
    if ((ndata = readData(plugin, wave, read_length, get_range, callback_flag, in_thread)) > 0) {
	swLockMutex(wave);
	wave->length = ndata;
	swUnlockMutex(wave);
    }

    spDebug(10, "readThinWaveDirectly", "thin_length = %ld, ndata = %ld\n",
	    thin_length, ndata);
    
    spCloseFilePlugin(plugin);

    return ndata;
}

static long readThinWaveMain(swWave wave, long offset, long length, long thin_length, 
			     spBool get_range, spBool callback_flag, spBool in_thread)
{
    long read_length;

    spDebug(80, "readThinWaveMain", "offset = %ld, length = %ld, thin_length = %ld\n",
	    offset, length, thin_length);
    
    length = MIN(length, wave->total_length - offset);
    thin_length = getThinLength(wave, length, thin_length, &read_length);
    spDebug(80, "readThinWaveMain", "thin_length = %ld, read_length = %ld\n",
	    thin_length, read_length);
    
    if (swIsWaveThreadSafe(wave) == SP_FALSE
	&& swIsWaveProcessing(wave) == SP_TRUE
	&& (swIsWavePeakAvailable(wave) == SP_FALSE
	    || thin_length < MIN(wave->peak_thin_length, wave->peak_buf_thin_length))) {
	return -1;
    }
    
    swLockMutex(wave);
    wave->thin_length = thin_length;
    wave->offset = offset;
    wave->length = read_length;
    wave->data_length = length;
    swUnlockMutex(wave);

    spDebug(80, "readThinWaveMain", "wave->thin_length = %ld, wave->peak_thin_length = %ld\n",
	    wave->thin_length, wave->peak_thin_length);
    
    if (wave->thin_length >= 1 && wave->thin_length >= wave->peak_thin_length
	&& swIsWaveRangeAvailable(wave) == SP_TRUE) {
	return readThinWaveUsingPeak(wave, offset, wave->thin_length, read_length,
				     callback_flag, in_thread);
    } else {
	if (length >= wave->config->read_callback_length) {
	    callback_flag = SP_TRUE;
	}
	return readThinWaveDirectly(wave, offset, wave->thin_length, read_length,
				    get_range, callback_flag, in_thread);
    }
}

spBool swIsWaveInstantiatable(swWave wave)
{
    spBool flag = SP_TRUE;
    
    if (wave == NULL || swIsWaveProcessing(wave) == SP_TRUE) {
	return SP_FALSE;
    }
    
    if (swIsWaveThreadSafe(wave) == SP_FALSE) {
	spPlugin *plugin;

	plugin = spLoadPlugin(wave->plugin_name);
	
	flag = spIsPluginInstantiatable(plugin);
	spDebug(80, "swIsWaveInstantiatable", "flag = %d\n", flag);
	
	spFreePlugin(plugin);
    }

    return flag;
}

static long readThinWave(swWave wave, long offset, long length, long thin_length, 
			 spBool get_range, spBool callback_flag, spBool in_thread)
{
    if (wave == NULL) return -1;

    spDebug(80, "readThinWave", "offset = %ld, length = %ld, total_length = %ld\n",
	    offset, length, wave->total_length);
    
    if (swIsWaveRangeAvailable(wave) == SP_FALSE
	&& (offset != 0 || length != wave->total_length)) {
	/* create peak */
	spDebug(80, "readThinWave", "read just for creating peak file\n");
	if (readThinWaveMain(wave, 0, wave->total_length, 0, 
			     SP_TRUE, SP_FALSE, in_thread) > 0
	    && wave->peak_length > 0) {
	    get_range = SP_FALSE;
	}
    }

    if (thin_length <= 0 && swIsWaveInstantiatable(wave) == SP_FALSE) {
	callback_flag = SP_FALSE;
	thin_length = wave->peak_buf_thin_length;
    }
    spDebug(80, "readThinWave", "thin_length = %ld\n", thin_length);
    
    return readThinWaveMain(wave, offset, length, thin_length, 
			    get_range, callback_flag, in_thread);
}

long swReadWave(swWave wave, spBool in_thread, long offset, long length)
{
    return readThinWave(wave, offset, length, 0, SP_FALSE, SP_FALSE, in_thread);
}

long swReadTotalWave(swWave wave, spBool in_thread)
{
    if (wave == NULL) return -1;

    return readThinWave(wave, 0, wave->total_length, 0, SP_TRUE, SP_TRUE, in_thread);
}

long swReadBuffer(spPlugin *plugin, swWave wave, int channel, long length)
{
    long nread;
    
    if (wave == NULL || plugin == NULL || length <= 0) return -1;

    spDebug(50, "swReadBuffer", "in\n");
    
    swAllocBuffer(wave, length);

    nread = swLoadIntoBuffer(plugin, wave, channel, length, wave->buf);

    if (nread > 0) {
	wave->buf_length = nread;
    } else {
	wave->buf_length = 0;
    }

    spDebug(50, "swReadBuffer", "done\n");
    
    return nread;
}

long swWriteData(spPlugin *plugin, swWave wave)
{
    long nwrite;
    
    if (wave == NULL || plugin == NULL) return -1;

    nwrite = spWritePlugin(plugin, wave->data, wave->length * wave->num_channel);

    spDebug(10, "swWriteData", "wave->length = %ld, nwrite = %ld\n", wave->length, nwrite);

    if (nwrite > 0) {
	nwrite /= wave->num_channel;
    }

    return nwrite;
}

long swWriteBuffer(spPlugin *plugin, swWave wave, long offset, long length)
{
    long nwrite;
    
    if (wave == NULL || plugin == NULL) return -1;

    nwrite = spWritePlugin(plugin,
			   wave->buf + swConvertLengthToByte(wave, offset),
			   length * wave->num_channel);

    spDebug(10, "swWriteBuffer", "wave->length = %ld, nwrite = %ld\n", wave->length, nwrite);

    if (nwrite > 0) {
	nwrite /= wave->num_channel;
    }

    return nwrite;
}

static spBool swSetEditBuffer(spPlugin *plugin, swWave wave, long channel, long pos,
			      double value, spBool *overflow)
{
    int o_samp_bit;
    long offset;
    
    if (wave == NULL) {
	return SP_FALSE;
    }

    if (channel >= 0) {
	offset = pos * wave->num_channel + channel;
    } else {
	offset = pos;
    }

    if (offset >= wave->editbuf_length) {
	return SP_FALSE;
    }

    if (spGetPluginSampleBit(plugin, &o_samp_bit) == SP_FALSE) {
	o_samp_bit = wave->samp_bit;
    }

    if (o_samp_bit > 32) {
	double *buf = (double *)wave->editbuf;
	buf[offset] = value;
    } else if (o_samp_bit > 16) {
	long *buf = (long *)wave->editbuf;
	buf[offset] = (long)swGetClippedValue(o_samp_bit, value, overflow);
    } else {
	short *buf = (short *)wave->editbuf;
	buf[offset] = (short)swGetClippedValue(o_samp_bit, value, overflow);
    }
    
    return SP_TRUE;
}

long swWriteEditBuffer(spPlugin *plugin, swWave wave, long length, spBool single_channel)
{
    int num_channel;
    long nwrite;
    
    if (wave == NULL || plugin == NULL) return -1;

    if (single_channel == SP_TRUE) {
	num_channel = 1;
    } else {
	num_channel = wave->num_channel;
    }
    
    nwrite = spWritePlugin(plugin, wave->editbuf, length * num_channel);
    
    if (nwrite > 0) {
	nwrite /= num_channel;
    }
    
    return nwrite;
}

static spPlugin *openWaveToWrite(swWave wave, spWaveInfo *wave_info, char *filename,
				 char *o_plugin_name, char *file_type,
				 int samp_bit, int num_channel, double samp_rate,
				 long length, spBool orig_flag, spBool callback_flag,
				 spBool in_thread)
{
    spPlugin *o_plugin;
    spSongInfo song_info;
    
    spDebug(50, "openWaveToWrite", "orig_flag = %d, callback_flag = %d\n",
	    orig_flag, callback_flag);
    
    /* initialize option */
    spInitWaveInfo(wave_info);
    if (file_type != NULL) {
	spStrCopy(wave_info->file_type, SP_WAVE_FILE_TYPE_SIZE, file_type);
    } else {
	if (orig_flag == SP_TRUE && wave->orig_file_type != NULL) {
	    spStrCopy(wave_info->file_type, SP_WAVE_FILE_TYPE_SIZE, wave->orig_file_type);
	} else if (wave->file_type != NULL) {
	    spStrCopy(wave_info->file_type, SP_WAVE_FILE_TYPE_SIZE, wave->file_type);
	}
    }
    if (samp_bit <= 0) {
	samp_bit = wave->samp_bit;
    }
    if (orig_flag == SP_TRUE) {
	wave_info->samp_bit = samp_bit;
    } else {
	wave_info->samp_bit = MAX(samp_bit, 16);
    }
    if (num_channel >= 1) {
	wave_info->num_channel = num_channel;
    } else {
	wave_info->num_channel = wave->num_channel;
    }
    if (samp_rate > 0.0) {
	wave_info->samp_rate = samp_rate;
    } else {
	wave_info->samp_rate = wave->samp_rate;
    }
    if (length > 0) {
	wave_info->length = length;
    } else {
	wave_info->length = wave->total_length;
    }
    spCopySongInfo(&song_info, &wave->song_info);

    spDebug(10, "openWaveToWrite", "wave->samp_bit = %d, wave_info->samp_bit = %d\n",
	    wave->samp_bit, wave_info->samp_bit);
    
    /* open output file */
    if (orig_flag == SP_TRUE && callback_flag == SP_TRUE) {
	swCallbackData call_data = {NULL, NULL, SP_FALSE, NULL};

	call_data.config = wave->config;
	call_data.wave = wave;
	call_data.in_thread = in_thread;
	call_data.call_data = wave->call_data;
	
	o_plugin = spOpenFilePlugin(o_plugin_name, filename, "w",
				    SP_PLUGIN_DEVICE_FILE,
				    wave_info, &song_info,
				    (spPluginOpenCallback)pluginOpenCB, (void *)&call_data, NULL);
    } else {
	o_plugin = spOpenFilePlugin(o_plugin_name, filename, "w",
				    SP_PLUGIN_DEVICE_FILE,
				    wave_info, &song_info, NULL, NULL, NULL);
    }

    return o_plugin;
}

static double getBitConvWeight(swWave wave, int o_samp_bit)
{
    double weight = 1.0;
    
    if (wave->samp_bit > 32 || o_samp_bit > 32) {
	if (wave->config->float_normalized == SP_TRUE
	    && (wave->samp_bit <= 32 || o_samp_bit <= 32)) {
	    if (o_samp_bit <= 32) {
		weight = swGetLimitValue(o_samp_bit);
	    } else if (wave->samp_bit <= 32) {
		weight = 1.0 / swGetLimitValue(wave->samp_bit);
	    }
	}
    } else if (wave->samp_bit <= 16) {
	if (o_samp_bit <= 16) {
	    weight = 1.0;
	} else if (o_samp_bit <= 24) {
	    weight = 256.0;
	} else if (o_samp_bit <= 32) {
	    weight = 65536.0;
	}
    } else if (wave->samp_bit <= 24) {
	if (o_samp_bit <= 16) {
	    weight = 1.0 / 256.0;
	} else if (o_samp_bit <= 24) {
	    weight = 1.0;
	} else if (o_samp_bit <= 32) {
	    weight = 256.0;
	}
    } else if (wave->samp_bit <= 32) {
	if (o_samp_bit <= 16) {
	    weight = 1.0 / 65536.0;
	} else if (o_samp_bit <= 24) {
	    weight = 1.0 / 256.0;
	} else if (o_samp_bit <= 32) {
	    weight = 1.0;
	}
    }

    return weight;
}

static long cropWaveRegion(spPlugin *i_plugin, spPlugin *o_plugin, swWave wave, spWaveInfo *wave_info,
			   swEditType edit_type, long offset, long length,
			   spBool *overflow, spBool callback_flag, spBool in_thread)
{
    long k, l;
    long len;
    long total_len;
    long nread, nwrite;
    double value;
    double weight = 1.0;
    
    spDebug(50, "cropWaveRegion", "callback_flag = %d\n", callback_flag);
    
    swLockMutex(wave);
    wave->process_finished = SP_FALSE;
    swUnlockMutex(wave);
    
    if (callback_flag == SP_TRUE && wave->config->edit_func != NULL) {
	 wave->config->edit_func(wave, in_thread, SW_PROCESS_STARTED, edit_type, wave->call_data);
	 swWaitMutex(wave);
	 wave->config->edit_func(wave, in_thread, offset, edit_type, wave->call_data);
	 swWaitMutex(wave);
    }
    
    if (edit_type == SW_EDIT_BIT_CONV) {
	int o_samp_bit;
	
	o_samp_bit = wave->samp_bit;
	spGetPluginSampleBit(o_plugin, &o_samp_bit);
	
	weight = getBitConvWeight(wave, o_samp_bit);
    }
    
    spSeekPlugin(i_plugin, offset);
    
    if (length <= 0) length = wave->total_length;
	
    total_len = 0;
    len = MIN(wave->config->normal_read_length, wave->total_length - offset);
    swAllocEditBuffer(wave, len);
    
    while ((nread = swReadBuffer(i_plugin, wave, -1, len)) > 0) {
	if (total_len + nread > length) {
	    nread = length - total_len;
	}
	if (edit_type == SW_EDIT_BIT_CONV) {
	    for (k = 0; k < nread; k++) {
		for (l = 0; l < wave->num_channel; l++) {
		    value = swGetBufferData(wave, l, k);
		    swSetEditBuffer(o_plugin, wave, l, k, weight * value, overflow);
		}
	    }
	    nwrite = swWriteEditBuffer(o_plugin, wave, nread, SP_FALSE);
	} else if (edit_type == SW_EDIT_MONAURALIZE) {
	    for (k = 0; k < nread; k++) {
		value = 0.0;
		for (l = 0; l < wave->num_channel; l++) {
		    value += swGetBufferData(wave, l, k);
		}
		value /= (double)wave->num_channel;
		swSetEditBuffer(o_plugin, wave, -1, k, value, overflow);
	    }
	    nwrite = swWriteEditBuffer(o_plugin, wave, nread, SP_TRUE);
	} else if (edit_type != SW_EDIT_COPY && wave->selected_channel >= 0) {
	    for (k = 0; k < nread; k++) {
		value = swGetBufferData(wave, wave->selected_channel, k);
		swSetEditBuffer(o_plugin, wave, -1, k, value, overflow);
	    }
	    nwrite = swWriteEditBuffer(o_plugin, wave, nread, SP_TRUE);
	} else {
	    nwrite = swWriteBuffer(o_plugin, wave, 0, nread);
	}
	if (nwrite <= 0) {
	    total_len = -1;
	    break;
	}
	
	total_len += nread;

	if (total_len >= length) {
	    break;
	}
	
	if (callback_flag == SP_TRUE && wave->config->edit_func != NULL) {
	    wave->config->edit_func(wave, in_thread, offset + total_len, edit_type, wave->call_data);
	    swWaitMutex(wave);
	}
	if (callback_flag == SP_TRUE && wave->process_flag == SP_FALSE) {
	    break;
	}
    }

    swLockMutex(wave);
    if (callback_flag == SP_TRUE && wave->process_flag == SP_FALSE) {
	total_len = 0;
    }
    swUnlockMutex(wave);

    spDebug(10, "cropWaveRegion", "total_len = %ld, process_flag = %d\n",
	    total_len, wave->process_flag);
    
    return total_len;
}

static long editWaveRegion(spPlugin *i_plugin, spPlugin *o_plugin, swWave wave, spWaveInfo *wave_info,
			   swEditType edit_type, long offset, long length,
			   double value, spBool *overflow, spBool in_thread)
{
    int target_channel = 0;
    long i, j;
    long k, n;
    long len;
    long total_len;
    long nread;
    long edit_count;
    double edit_value;
    int flag1 = 0, flag2 = 0;

    swLockMutex(wave);
    wave->process_finished = SP_FALSE;
    swUnlockMutex(wave);
    
    if (wave->config->edit_func != NULL) {
	wave->config->edit_func(wave, in_thread, SW_PROCESS_STARTED, edit_type, wave->call_data);
	swWaitMutex(wave);
	wave->config->edit_func(wave, in_thread, 0, edit_type, wave->call_data);
	swWaitMutex(wave);
    }

    if (edit_type == SW_EDIT_CHANGE_VALUE) {
	target_channel = MAX((int)length, 0);
	target_channel = MIN(target_channel, wave->num_channel - 1);
	length = 1;
    } else {
	if (length <= 0) length = wave->total_length;
    }
    
    edit_count = 0;
    total_len = 0;
    len = MIN(wave->config->normal_read_length, wave->total_length);
    swAllocEditBuffer(wave, len);
    
    while ((nread = swReadBuffer(i_plugin, wave, -1, len)) > 0) {
	if (total_len + nread >= offset && total_len < offset + length) {
	    if (!flag1) {
		/* before the region */
		n = offset - total_len;
		if (n > 0) {
		    if (swWriteBuffer(o_plugin, wave, 0, n) <= 0) {
			total_len = -1;
			break;
		    }
		}
	    } else {
		n = 0;
	    }
	    flag1 = 1;

	    if (edit_type != SW_EDIT_DELETE) {
		k = 0;
		for (i = n; i < nread; i++) {
		    if (total_len + i >= offset + length) {
			break;
		    }
			    
		    if (edit_type == SW_EDIT_CHANNEL_SWAP) {
			for (j = 0; j < wave_info->num_channel; j++) {
			    swSetEditBuffer(o_plugin, wave, j, k,
					    swGetBufferData(wave, wave_info->num_channel - 1 - j, i),
					    overflow);
			}
		    } else if (edit_type == SW_EDIT_CHANGE_VALUE) {
			for (j = 0; j < wave_info->num_channel; j++) {
			    if (j == target_channel) {
				edit_value = value;
			    } else {
				edit_value = swGetBufferData(wave, j, i);
			    }
			    swSetEditBuffer(o_plugin, wave, j, k, edit_value, overflow);
			}
		    } else {
			for (j = 0; j < wave_info->num_channel; j++) {
			    if (wave->selected_channel < 0 || j == wave->selected_channel) {
				if (edit_type == SW_EDIT_AMPLIFY) {
				    edit_value = value * swGetBufferData(wave, j, i);
				} else if (edit_type == SW_EDIT_FADE_IN) {
				    edit_value = ((double)edit_count / (double)length)
					* swGetBufferData(wave, j, i);
				} else if (edit_type == SW_EDIT_FADE_OUT) {
				    edit_value = ((double)(length - 1 - edit_count) / (double)length)
					* swGetBufferData(wave, j, i);
				} else {
				    edit_value = 0.0;
				}
			    } else {
				edit_value = swGetBufferData(wave, j, i);
			    }
			    swSetEditBuffer(o_plugin, wave, j, k, edit_value, overflow);
			}
		    }
		    k++;
		    edit_count++;
		}
		
		if (k > 0) {
		    if (swWriteEditBuffer(o_plugin, wave, k, SP_FALSE) <= 0) {
			total_len = -1;
			break;
		    }
		}
	    }
		
	    if (!flag2) {
		/* after the region */
		k = offset + length - total_len;
		n = total_len + nread - (offset + length);
		if (k >= 0 && n > 0) {
		    if (swWriteBuffer(o_plugin, wave, k, n) <= 0) {
			total_len = -1;
			break;
		    }
		    flag2 = 1;
		}
	    }
	} else {
	    /* before & after the region */
	    if (swWriteBuffer(o_plugin, wave, 0, nread) <= 0) {
		total_len = -1;
		break;
	    }
	}
	total_len += nread;

	if (wave->config->edit_func != NULL) {
	    wave->config->edit_func(wave, in_thread, total_len, edit_type, wave->call_data);
	    swWaitMutex(wave);
	}
	if (wave->process_flag == SP_FALSE) {
	    break;
	}
    }

    swLockMutex(wave);
    if (wave->process_flag == SP_FALSE) {
	total_len = 0;
    }
    swUnlockMutex(wave);
    
    return total_len;
}

static long pasteWave(spPlugin *i_plugin, spPlugin *o_plugin, swWave src_wave, swWave wave, 
		      spWaveInfo *wave_info, swEditType edit_type, long offset, long length,
		      spBool *overflow, spBool in_thread)
{
    int target_channel;
    long i, j;
    long k, n;
    long len;
    long src_len = 0;
    long total_len;
    long nread, src_nread;
    long edit_len;
    double edit_value;
    spPlugin *src_plugin = NULL;
    int flag1 = 0, flag2 = 0;

    if (edit_type == SW_EDIT_INSERT_PAUSE) {
	src_len = length;
    } else {
	if (swIsWaveNone(src_wave) == SP_TRUE) {
	    return 0;
	}
	
	swLockMutex(src_wave);
	if (swIsWaveProcessing(src_wave) == SP_TRUE
	    && swIsWavePlaying(src_wave) == SP_FALSE) {
	    swUnlockMutex(src_wave);
	    return 0;
	}
	swUnlockMutex(src_wave);
	
	if (src_wave->num_channel != 1 && src_wave->num_channel != wave->num_channel) {
	    return 0;
	}
	
	if ((src_plugin = swOpenWave(src_wave, "r")) == NULL) {
	    return 0;
	}
    }

    swLockMutex(wave);
    wave->process_finished = SP_FALSE;
    swUnlockMutex(wave);

    if (wave->config->edit_func != NULL) {
	wave->config->edit_func(wave, in_thread, SW_PROCESS_STARTED, edit_type, wave->call_data);
	swWaitMutex(wave);
	wave->config->edit_func(wave, in_thread, 0, edit_type, wave->call_data);
	swWaitMutex(wave);
    }

    target_channel = MAX((int)length, 0);
    target_channel = MIN(target_channel, wave->num_channel - 1);
    
    if (edit_type == SW_EDIT_INSERT || edit_type == SW_EDIT_INSERT_PAUSE) {
	length = 0;
    } else if (edit_type == SW_EDIT_REPLACE) {
	target_channel = wave->selected_channel;
    } else {
	length = MIN(src_wave->total_length, wave->total_length - offset);
    }
    spDebug(50, "pasteWave", "target_channel = %d, length = %ld\n",
	    target_channel, length);
    
    total_len = 0;
    edit_len = 0;
    
    len = MIN(wave->config->normal_read_length, wave->total_length);
    if (edit_type != SW_EDIT_INSERT_PAUSE) {
	src_len = MIN(wave->config->normal_read_length, src_wave->total_length);
    }
    swAllocEditBuffer(wave, len);

    while ((nread = swReadBuffer(i_plugin, wave, -1, len)) > 0) {
	spDebug(80, "pasteWave", "nread = %ld, total_len = %ld\n",
		nread, total_len);
	/* if (total_len + nread >= offset && total_len < offset + length) { */
	if (offset <= total_len + nread && total_len <= offset + length) {
	    if (!flag1) {
		/* before the region */
		n = offset - total_len;
		if (n > 0) {
		    if (swWriteBuffer(o_plugin, wave, 0, n) <= 0) {
			total_len = -1;
			break;
		    }
		}
	    } else {
		n = 0;
	    }
	    flag1 = 1;

	    if (edit_type == SW_EDIT_INSERT || edit_type == SW_EDIT_REPLACE
		|| edit_type == SW_EDIT_INSERT_PAUSE) {
		if (edit_len <= 0) {
		    while (1) {
			if (edit_type == SW_EDIT_INSERT_PAUSE) {
			    src_nread = MIN(wave->config->normal_read_length, src_len);
			    if (src_nread > 0) {
				for (i = 0; i < src_nread; i++) {
				    for (j = 0; j < wave_info->num_channel; j++) {
					swSetEditBuffer(o_plugin, wave, j, i, 0.0, overflow);
				    }
				}
				if (swWriteEditBuffer(o_plugin, wave, src_nread, SP_FALSE) <= 0) {
				    total_len = -1;
				}
				edit_len += src_nread;
				src_len -= src_nread;
			    }
			} else {
			    swLockMutex(src_wave);
			    if ((src_nread = swReadBuffer(src_plugin, src_wave, -1, src_len)) > 0) {
				if (src_wave->num_channel == wave->num_channel) {
				    if (swWriteBuffer(o_plugin, src_wave, 0, src_nread) <= 0) {
					total_len = -1;
				    }
				} else {
				    for (i = 0; i < src_nread; i++) {
					for (j = 0; j < wave_info->num_channel; j++) {
					    if (j == target_channel) {
						edit_value = swGetBufferData(src_wave, 0, i);
					    } else {
						edit_value = 0.0;
					    }
					    swSetEditBuffer(o_plugin, wave, j, i, edit_value, overflow);
					}
				    }
				    if (swWriteEditBuffer(o_plugin, wave, src_nread, SP_FALSE) <= 0) {
					total_len = -1;
				    }
				}
				edit_len += src_nread;
			    }
			    swUnlockMutex(src_wave);
			}
			
			if (total_len < 0 || src_nread <= 0 || wave->process_flag == SP_FALSE) {
			    break;
			}
		    
			if (wave->config->edit_func != NULL) {
			    wave->config->edit_func(wave, in_thread, total_len, edit_type, wave->call_data);
			    swWaitMutex(wave);
			}
		    }
		}
	    } else {
		swLockMutex(src_wave);
		if ((src_nread = swReadBuffer(src_plugin, src_wave, -1, nread - n)) > 0) {
		    for (i = 0; i < src_nread; i++) {
			for (j = 0; j < wave_info->num_channel; j++) {
			    if (src_wave->num_channel == wave->num_channel) {
				edit_value = swGetBufferData(src_wave, j, i);
			    } else if (j == target_channel) {
				edit_value = swGetBufferData(src_wave, 0, i);
			    } else if (edit_type == SW_EDIT_PASTE) {
				edit_value = swGetBufferData(wave, j, n + i);
			    } else {
				edit_value = 0.0;
			    }
			    if (edit_type == SW_EDIT_MIX) {
				edit_value += swGetBufferData(wave, j, n + i);
			    }
			    swSetEditBuffer(o_plugin, wave, j, i, edit_value, overflow);
			}
		    }
		    if (swWriteEditBuffer(o_plugin, wave, src_nread, SP_FALSE) <= 0) {
			total_len = -1;
		    }
		}
		swUnlockMutex(src_wave);
	    }

	    if (total_len < 0) break;
		
	    if (!flag2) {
		/* after the region */
		k = offset + length - total_len;
		n = total_len + nread - (offset + length);
		if (k >= 0 && n > 0) {
		    if (swWriteBuffer(o_plugin, wave, k, n) <= 0) {
			total_len = -1;
			break;
		    }
		    flag2 = 1;
		}
	    }
	} else {
	    /* before & after the region */
	    if (swWriteBuffer(o_plugin, wave, 0, nread) <= 0) {
		total_len = -1;
		break;
	    }
	}
	total_len += nread;

	if (wave->config->edit_func != NULL) {
	    wave->config->edit_func(wave, in_thread, total_len, edit_type, wave->call_data);
	    swWaitMutex(wave);
	}
	if (wave->process_flag == SP_FALSE) {
	    break;
	}
    }
    
    if (total_len >= 0 && (edit_type == SW_EDIT_INSERT || edit_type == SW_EDIT_INSERT_PAUSE)) {
	total_len += edit_len;
    }

    if (edit_type != SW_EDIT_INSERT_PAUSE) {
	/* close source file */
	spCloseFilePlugin(src_plugin);
    }
    
    swLockMutex(wave);
    if (wave->process_flag == SP_FALSE) {
	total_len = 0;
    }
    swUnlockMutex(wave);
    
    spDebug(50, "pasteWave", "done\n");
    
    return total_len;
}

static long convSampFreq(spPlugin *i_plugin, spPlugin *o_plugin, swWave wave, spWaveInfo *wave_info,
			 long upratio, long downratio, DVector filter, spBool *overflow,
			 spBool in_thread)
{
    long k, l;
    long fftl;
    long len;
    long total_len;
    long obuf_len;
    long nread;
    long nwrite;
    long pos, prev_pos;
    long shift;
    long syn_pos;
    long syn_offset;
    double value;
    DVector isig;
    DVector *osigs;
    DVector upsig;
    DVector filsig;

    swLockMutex(wave);
    wave->process_finished = SP_FALSE;
    swUnlockMutex(wave);

    if (wave->config->edit_func != NULL) {
	wave->config->edit_func(wave, in_thread, SW_PROCESS_STARTED, SW_EDIT_SAMP_RATE_CONV, wave->call_data);
	swWaitMutex(wave);
	wave->config->edit_func(wave, in_thread, 0, SW_EDIT_SAMP_RATE_CONV, wave->call_data);
	swWaitMutex(wave);
    }
    
    spDebug(50, "convSampFreq", "up = %ld, down = %ld\n", upratio, downratio);

    fftl = 1024;
    len = (wave->config->sfc_buffer_length - filter->length + 1) / upratio;
    spDebug(30, "convSampFreq", "len = %ld\n", len);
    len = MAX(len, 128);
    len = MIN(len, wave->total_length);
    len = MIN(len, fftl);
    
    /* allocate memory */
    isig = xdvalloc(len);
    osigs = xalloc(wave->num_channel, DVector);
    obuf_len = (len * upratio + filter->length - 1) / downratio + 1;
    for (l = 0; l < wave->num_channel; l++) {
	osigs[l] = xdvzeros(obuf_len);
    }
    swAllocEditBuffer(wave, (len * upratio) / downratio + 1);
    
    pos = (-filter->length / 2);
    syn_offset = 0;
    total_len = 0;
    while ((nread = swReadBuffer(i_plugin, wave, -1, len)) > 0) {
	shift = (long)nread * upratio;
	prev_pos = pos;
	pos += shift;

	spDebug(80, "convSampFreq", "nread = %ld, pos = %ld\n", nread, pos);

	nwrite = 0;
	syn_pos = 0;
	for (l = 0; l < wave->num_channel; l++) {
	    if (in_thread == SP_TRUE) spYieldThread();
	
	    for (k = 0; k < nread; k++) {
		isig->data[k] = swGetBufferData(wave, l, k);
	    }
	    isig->length = nread;
	
	    /* up sampling */
	    upsig = xdvupsample(isig, upratio);

	    if (in_thread == SP_TRUE) spYieldThread();

	    /* lowpass filtering */
	    filsig = xdvfftfilt(filter, upsig, fftl);
	    
	    if (in_thread == SP_TRUE) spYieldThread();
	    
	    spDebug(80, "convSampFreq", "osigs[%ld]->length = %ld, filsig->length = %ld\n",
		    l, osigs[l]->length, filsig->length);

	    if (prev_pos < 0) {
		syn_pos = -prev_pos;
	    } else {
		syn_pos = syn_offset;
	    }
	    for (k = 0; k < osigs[l]->length; k++) {
		if (syn_pos >= filsig->length) break;
		    
		osigs[l]->data[k] += filsig->data[syn_pos];
		syn_pos += downratio;
	    }

	    if (pos > 0) {
		nwrite = 0;
		syn_pos = syn_offset;
		for (k = 0; k < osigs[l]->length; k++) {
		    if (syn_pos >= pos) break;

		    value = wave->config->sfc_gain * osigs[l]->data[k];
		    
		    if (swSetEditBuffer(o_plugin, wave, l, nwrite, value, overflow) == SP_FALSE) {
			spDebug(80, "convSampFreq", "swSetEditBuffer error\n");
		    }
		    nwrite++;
		    syn_pos += downratio;
		}
	    
		dvdatashift(osigs[l], -k);
	    }
	    
	    /* free memory */
	    xdvfree(upsig);
	    xdvfree(filsig);
	}
	if (in_thread == SP_TRUE) spYieldThread();
	
	if (pos > 0) {
	    syn_offset = syn_pos - pos;
	    pos = 0;
	
	    if (swWriteEditBuffer(o_plugin, wave, nwrite, SP_FALSE) <= 0) {
		total_len = -1;
		break;
	    }
	}
	
	total_len += nread;

	if (wave->config->edit_func != NULL) {
	    wave->config->edit_func(wave, in_thread, total_len, SW_EDIT_SAMP_RATE_CONV, wave->call_data);
	    swWaitMutex(wave);
	}
	if (wave->process_flag == SP_FALSE) {
	    break;
	}
    }

    swLockMutex(wave);
    if (wave->process_flag == SP_FALSE) {
	total_len = 0;
    }
    swUnlockMutex(wave);

    xdvfree(isig);
    xdvfree(filter);
    for (l = 0; l < wave->num_channel; l++) {
	xdvfree(osigs[l]);
    }
    xfree(osigs);

    return total_len;
}

static long writeWaveRegion(spPlugin *i_plugin, spPlugin *o_plugin, swWave wave,
			    spWaveInfo *wave_info, swEditType edit_type,
			    long offset, long length, double value, double *ovalue,
			    void *data, spBool *overflow, spBool callback_flag, spBool in_thread)
{
    if (edit_type == SW_EDIT_CROP || edit_type == SW_EDIT_COPY || edit_type == SW_EDIT_WRITE
	|| edit_type == SW_EDIT_BIT_CONV || edit_type == SW_EDIT_MONAURALIZE) {
	return cropWaveRegion(i_plugin, o_plugin, wave, wave_info, edit_type,
			      offset, length, overflow, callback_flag, in_thread);
    } else if (edit_type == SW_EDIT_SAMP_RATE_CONV) {
	long upratio, downratio;
	double new_samp_rate;
	double ratio;
	double cutoff;
	double transition;
	DVector filter;

	new_samp_rate = getsfcratio(wave->samp_rate, value, 
				    wave->config->sfc_tolerance / 100.0,
				    &upratio, &downratio);
	spDebug(10, "writeWaveRegion", "New Sampling Frequency: %f\n", new_samp_rate);
	spDebug(10, "writeWaveRegion", "upratio = %ld, downratio = %ld\n", upratio, downratio);
	if (ovalue != NULL) {
	    *ovalue = new_samp_rate;
	}
	
	/* get lowpass filter */
	ratio = (double)MAX(upratio, downratio);
	cutoff = wave->config->sfc_cutoff / ratio;
	transition = wave->config->sfc_transition / ratio;
	filter = xdvlowpass(cutoff, wave->config->sfc_sidelobe, transition, (double)upratio);

	return convSampFreq(i_plugin, o_plugin, wave, wave_info, upratio, downratio, filter,
			    overflow, in_thread);
    } else if (edit_type == SW_EDIT_PASTE || edit_type == SW_EDIT_INSERT || edit_type == SW_EDIT_INSERT_PAUSE
	       || edit_type == SW_EDIT_MIX || edit_type == SW_EDIT_REPLACE) {
	return pasteWave(i_plugin, o_plugin, (swWave)data, wave, wave_info, edit_type,
			 offset, length, overflow, in_thread);
    } else {
	return editWaveRegion(i_plugin, o_plugin, wave, wave_info, edit_type,
			      offset, length, value, overflow, in_thread);
    }
}

static swWave findBackupWave(swWave wave, char *o_filename)
{
    swWave owave;
    swWave prev_wave, next_wave;

    owave = NULL;
    prev_wave = wave->prev_wave;
    while (prev_wave != NULL) {
	if (streq(prev_wave->filename, o_filename)) {
	    owave = prev_wave;
	    break;
	}

	prev_wave = prev_wave->prev_wave;
    }

    if (owave == NULL) {
	next_wave = wave->next_wave;
	while (next_wave != NULL) {
	    if (streq(next_wave->filename, o_filename)) {
		owave = next_wave;
		break;
	    }
	    next_wave = next_wave->next_wave;
	}
    }

    return owave;
}

static spBool createBackupFile(swWave wave, char *o_filename, spBool in_thread)
{
    spBool flag = SP_FALSE;
    char *filename;
    spPlugin *i_plugin;
    spPlugin *o_plugin;
    spWaveInfo wave_info;
    swWave backup_wave;
    swWave prev_wave, next_wave;

    if (wave == NULL || strnone(o_filename)) return SP_FALSE;

    if ((backup_wave = findBackupWave(wave, o_filename)) == NULL) {
	spDebug(10, "createBackupFile", "There is no file that needs backup\n");
	return SP_TRUE;
    }

    filename = swGetTempFile(wave->config, NULL);

    /* open input file */
    if ((i_plugin = swOpenWave(backup_wave, "r")) != NULL) {
	o_plugin = openWaveToWrite(backup_wave, &wave_info, filename,
				   "output_raw", "raw", 0, 0, 0.0, 0, SP_FALSE,
				   SP_TRUE, in_thread);

	if (o_plugin != NULL) {
	    if (writeWaveRegion(i_plugin, o_plugin, backup_wave, &wave_info, SW_EDIT_COPY,
				0, 0, 0.0, NULL, NULL, NULL, SP_FALSE, in_thread) > 0) {
		flag = SP_TRUE;
	    }
	    
	    /* close output file */
	    spCloseFilePlugin(o_plugin);

	    if (flag == SP_FALSE) {
		spDebug(10, "createBackupFile", "remove file: %s\n", filename);
		spRemoveFile(filename);
	    }
	}
	
	/* close input file */
	spCloseFilePlugin(i_plugin);
    }
    
    if (flag == SP_TRUE) {
	prev_wave = wave->prev_wave;
	while (prev_wave != NULL) {
	    if (streq(prev_wave->filename, o_filename)) {
		spDebug(10, "createBackupFile", "o_filename = %s, prev_wave->filename = %s\n",
			o_filename, prev_wave->filename);
	
		swLockMutex(prev_wave);
		if (prev_wave->plugin_name != NULL) xfree(prev_wave->plugin_name);
		prev_wave->plugin_name = strclone("input_raw");

		if (prev_wave->file_type != NULL) xfree(prev_wave->file_type);
		prev_wave->file_type = strclone("raw");

		if (prev_wave->filename != NULL) xfree(prev_wave->filename);
		prev_wave->filename = strclone(filename);
	    
		prev_wave->orig_flag = SP_FALSE;
		prev_wave->edit_flag = SP_TRUE;
		swUnlockMutex(prev_wave);
	    }

	    prev_wave = prev_wave->prev_wave;
	}
	
	next_wave = wave->next_wave;
	while (next_wave != NULL) {
	    if (streq(next_wave->filename, o_filename)) {
		spDebug(10, "createBackupFile", "o_filename = %s, next_wave->filename = %s\n",
			o_filename, next_wave->filename);
	
		swLockMutex(next_wave);
		if (next_wave->plugin_name != NULL) xfree(next_wave->plugin_name);
		next_wave->plugin_name = strclone("input_raw");

		if (next_wave->file_type != NULL) xfree(next_wave->file_type);
		next_wave->file_type = strclone("raw");

		if (next_wave->filename != NULL) xfree(next_wave->filename);
		next_wave->filename = strclone(filename);
	    
		next_wave->orig_flag = SP_FALSE;
		next_wave->edit_flag = SP_TRUE;
		swUnlockMutex(next_wave);
	    }
	
	    next_wave = next_wave->next_wave;
	}

	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static void updateWaveInfo(spPlugin *o_plugin, swWave wave, spWaveInfo *wave_info,
			   char *filename, char *i_plugin_name, spBool orig_flag)
{
    char *string;
    swWave prev_wave, next_wave;

    if (i_plugin_name != NULL) {
	if (orig_flag == SP_TRUE) {
	    swLockMutex(wave);
	    if (wave->file_type != NULL) xfree(wave->file_type);
	    
	    if ((string = spGetPluginFileType(o_plugin, SP_TRUE)) != NULL) {
		wave->file_type = strclone(string);
	    } else {
		wave->file_type = NULL;
	    }
	    
	    if (wave->plugin_name != NULL) xfree(wave->plugin_name);
	    wave->plugin_name = strclone(i_plugin_name);
	    
	    if (wave->filename != filename
		&& !streq(wave->filename, filename)) {
		if (wave->orig_flag == SP_FALSE) {
		    spDebug(10, "updateWaveInfo", "remove file: %s\n", wave->filename);
		    spRemoveFile(wave->filename);
		}
		if (wave->filename != NULL) xfree(wave->filename);
		wave->filename = strclone(filename);
	    }
	    
	    if (wave->orig_plugin_name != NULL) xfree(wave->orig_plugin_name);
	    wave->orig_plugin_name = strclone(i_plugin_name);
	    
	    wave->orig_flag = SP_TRUE;
	    swUnlockMutex(wave);
	}
    }
    
    if (orig_flag == SP_TRUE) {
	swLockMutex(wave);
	if (wave->orig_filename != filename
	    && !streq(wave->orig_filename, filename)) {
	    if (wave->orig_filename != NULL) xfree(wave->orig_filename);
	    wave->orig_filename = strclone(filename);
	}
	
	if (wave->orig_file_type != NULL) xfree(wave->orig_file_type);
	wave->orig_file_type = strclone(spGetPluginFileType(o_plugin, SP_TRUE));
	spDebug(10, "updateWaveInfo", "wave->orig_file_type = %s\n", wave->orig_file_type);

	wave->edit_flag = SP_FALSE;

	wave->samp_bit = wave_info->samp_bit;
	wave->num_channel = wave_info->num_channel;
	wave->samp_rate = wave_info->samp_rate;
	spDebug(10, "updateWaveInfo", "wave->samp_bit = %d\n", wave->samp_bit);
	swUnlockMutex(wave);
	
	prev_wave = wave->prev_wave;
	while (prev_wave != NULL) {
	    swLockMutex(prev_wave);
	    if (prev_wave->orig_plugin_name != NULL) xfree(prev_wave->orig_plugin_name);
	    if (prev_wave->orig_filename != NULL) xfree(prev_wave->orig_filename);
	    if (prev_wave->orig_file_type != NULL) xfree(prev_wave->orig_file_type);
	    
	    prev_wave->orig_plugin_name = strclone(wave->orig_plugin_name);
	    prev_wave->orig_filename = strclone(wave->orig_filename);
	    prev_wave->orig_file_type = strclone(wave->orig_file_type);
	    prev_wave->edit_flag = SP_TRUE;
	    swUnlockMutex(prev_wave);
	    
	    prev_wave = prev_wave->prev_wave;
	}
	
	next_wave = wave->next_wave;
	while (next_wave != NULL) {
	    swLockMutex(next_wave);
	    if (next_wave->orig_plugin_name != NULL) xfree(next_wave->orig_plugin_name);
	    if (next_wave->orig_filename != NULL) xfree(next_wave->orig_filename);
	    if (next_wave->orig_file_type != NULL) xfree(next_wave->orig_file_type);
	    
	    next_wave->orig_plugin_name = strclone(wave->orig_plugin_name);
	    next_wave->orig_filename = strclone(wave->orig_filename);
	    next_wave->orig_file_type = strclone(wave->orig_file_type);
	    next_wave->edit_flag = SP_TRUE;
	    swUnlockMutex(next_wave);
	    
	    next_wave = next_wave->next_wave;
	}
    }
    
    return;
}

static spBool swWriteWaveRegion(char *filename, char *plugin_name, char *file_type,
				int samp_bit, swWave wave, swEditType edit_type,
				long offset, long length, double value, double *ovalue,
				void *data, spBool *overflow, spBool orig_flag,
				spBool callback_flag, spBool in_thread)
{
    spBool flag;
    long nwrite = 0;
    double o_samp_rate;
    char *o_plugin_name;
    char *i_plugin_name;
    spPlugin *i_plugin;
    spPlugin *o_plugin;
    spWaveInfo wave_info;
    char *orig_backup = NULL;
    
    if (swIsWaveNone(wave) == SP_TRUE || strnone(filename)) return SP_FALSE;

    /* find output plugin */
    if (strnone(plugin_name)) {
	if ((o_plugin_name = xspFindSuitablePluginFile(SP_PLUGIN_DEVICE_FILE,
						       filename, "w")) == NULL) {
	    if (0 && wave->orig_plugin_name != NULL/* && spGetSuffix(filename) == NULL*/) {
		o_plugin_name = strclone(wave->orig_plugin_name);
	    } else {
		if (callback_flag == SP_TRUE && wave->config->error_func != NULL) {
		    return wave->config->error_func(wave, in_thread, SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND, wave->call_data);
		}
		return SP_FALSE;
	    }
	}
    } else {
	o_plugin_name = strclone(plugin_name);
    }

    if (orig_flag == SP_TRUE) {
	if (createBackupFile(wave, filename, in_thread) == SP_FALSE) {
	    xfree(o_plugin_name);
	    return SP_FALSE;
	}
	orig_backup = xswBackupOriginal(filename);
    }

    i_plugin_name = xspFindRelatedPluginFile(o_plugin_name);

    flag = SP_FALSE;
    /* open input file */
    if ((i_plugin = swOpenWave(wave, "r")) != NULL) {
	if (edit_type == SW_EDIT_SAMP_RATE_CONV) {
	    o_samp_rate = value;
	} else {
	    o_samp_rate = 0.0;
	}
	
	o_plugin = openWaveToWrite(wave, &wave_info, filename,
				   o_plugin_name, file_type,
				   samp_bit, 0, o_samp_rate,
				   length, orig_flag, callback_flag, in_thread);

	if (o_plugin != NULL) {
	    spDebug(10, "swWriteWaveRegion", "openWaveToWrite done\n");
	    if ((nwrite = writeWaveRegion(i_plugin, o_plugin, wave, &wave_info, edit_type,
					  offset, length, value, ovalue,
					  data, overflow, /*!orig_flag*/callback_flag,
					  in_thread)) > 0) {
		flag = SP_TRUE;
	    }
	}
	    
	/* close input file */
	spCloseFilePlugin(i_plugin);

	if (o_plugin != NULL) {
	    spWaitMutex(wave->io_mutex);
	    
	    if (flag == SP_TRUE) {
		updateWaveInfo(o_plugin, wave, &wave_info, filename, i_plugin_name, orig_flag);
	    }
	    
	    /* close output file */
	    spCloseFilePlugin(o_plugin);

	    if (flag == SP_FALSE) {
		spDebug(10, "swWriteWaveRegion", "failed: remove current file: %s\n", filename);
		spRemoveFile(filename);
	    }
	}
    }

    xfree(o_plugin_name);
    if (i_plugin_name != NULL) xfree(i_plugin_name);

    if (orig_backup != NULL) {
	if (flag == SP_TRUE) {
	    spDebug(10, "swWriteWaveRegion", "remove backup file: %s\n", orig_backup);
	    spRemoveFile(orig_backup);
	} else {
	    spDebug(10, "swWriteWaveRegion", "restore backup file: %s --> %s\n",
		    orig_backup, filename);
	    spRenameFile(orig_backup, filename);
	}
	xfree(orig_backup);
    }

    return flag;
}

static spBool writeWave(char *filename, char *plugin_name, char *file_type, swWave wave,
			spBool callback_flag, spBool in_thread)
{
    spPlugin *plugin;
    long total_length;
    spBool flag = SP_FALSE;
    
    if (swIsWaveNone(wave) == SP_TRUE) return SP_FALSE;

    if (strnone(filename)) {
	if (strnone(wave->orig_filename)) {
	    return SP_FALSE;
	}
	
	filename = wave->orig_filename;
    }
    spDebug(50, "writeWave", "callback_flag = %d\n", callback_flag);

    /* write wave */
    if (!streq(filename, wave->filename)) {
	if ((flag = swWriteWaveRegion(filename, plugin_name, file_type, 0, wave,
				      SW_EDIT_COPY, 0, 0, 0.0, NULL, NULL, NULL,
				      SP_TRUE, callback_flag, in_thread)) == SP_TRUE) {
	    if ((plugin = swOpenWave(wave, "r")) != NULL) {
		swLockMutex(wave);
		wave->thread_safe = spIsPluginCapable(plugin, SP_PLUGIN_CAPS_THREAD_SAFE);
		if ((total_length = spGetPluginTotalLength(plugin)) > 0) {
		    wave->total_length = total_length;
		}
		swUnlockMutex(wave);
		spCloseFilePlugin(plugin);
	    }
	}
    } else {
	flag = SP_TRUE;
    }
    
    return flag;
}

swWave swGetCroppedWave(char *filename, swWave wave, long offset, long length, spBool in_thread)
{
    int num_channel;
    spBool flag;
    
    if (swIsWaveNone(wave) == SP_TRUE || strnone(filename)) return NULL;

    /* write wave */
    flag = swWriteWaveRegion(filename, "output_raw", "raw", 0, wave,
			     SW_EDIT_CROP, offset, length, 0.0, NULL, NULL, NULL,
			     SP_FALSE, SP_TRUE, in_thread);

    if (flag == SP_TRUE) {
	if (wave->selected_channel >= 0) {
	    num_channel = 1;
	} else {
	    num_channel = wave->num_channel;
	}
	
	/* allocate wave */
	return getWave(wave->config, filename, "input_raw", "raw", wave->samp_bit,
		       num_channel, wave->samp_rate, SP_FALSE, SP_FALSE, NULL, in_thread);
    } else {
	return NULL;
    }
}

spBool swSetOriginalName(swWave wave, char *orig_filename, char *orig_plugin_name)
{
    if (wave == NULL) return SP_FALSE;

    if (wave->orig_filename != orig_filename && !strnone(orig_filename)) {
	if (wave->orig_filename != NULL) xfree(wave->orig_filename);
	wave->orig_filename = strclone(orig_filename);
    }
    if (wave->orig_plugin_name != orig_plugin_name && !strnone(orig_plugin_name)) {
	if (wave->orig_plugin_name != NULL) xfree(wave->orig_plugin_name);
	wave->orig_plugin_name = strclone(orig_plugin_name);
    }

    return SP_TRUE;
}

spBool swSetOriginalParams(swWave wave, swWave orig_wave, spBool copy_label)
{
    if (wave == NULL || orig_wave == NULL) return SP_FALSE;

    if (wave->orig_file_type != NULL) xfree(wave->orig_file_type);
    if (orig_wave->orig_file_type != NULL) {
	wave->orig_file_type = strclone(orig_wave->orig_file_type);
    } else {
	wave->orig_file_type = strclone(orig_wave->file_type);
    }
    if (copy_label == SP_TRUE) {
	swCopyLabels(wave, orig_wave);
    }
    
    spCopySongInfo(&wave->song_info, &orig_wave->song_info);
    wave->song_info_mask = orig_wave->song_info_mask;
    
    return SP_TRUE;
}

spBool swSetOriginalRange(swWave wave, swWave orig_wave, double weight)
{
    double min, max;

    if (wave == NULL || orig_wave == NULL) return SP_FALSE;

    min = swGetWaveMin(orig_wave);
    max = swGetWaveMax(orig_wave);
    if (weight < 0.0) {
	return swSetWaveRange(wave, weight * max, weight * min);
    } else {
	return swSetWaveRange(wave, weight * min, weight * max);
    }
}
    
spBool swSetPrevWave(swWave wave, swWave prev_wave)
{
    int i;
    int flag;
    swWave temp, temp2;
    
    if (wave == NULL || prev_wave == NULL) return SP_FALSE;

    flag = 0;
    temp = prev_wave;
    for (i = 0;; i++) {
	if (temp == NULL)
	    break;
	
	if (i >= wave->config->max_num_undo) {
	    temp2 = temp->prev_wave;
	    swFreeWave(temp);
	    temp = temp2;
	    if (i <= 0) {
		flag = 1;
	    }
	} else {
	    temp = temp->prev_wave;
	}
    }

    if (!flag) {
	temp = prev_wave->next_wave;
	for (;;) {
	    if (temp == NULL)
		break;

	    temp2 = temp->next_wave;
	    swFreeWave(temp);
	    temp = temp2;
	}
	
	prev_wave->next_wave = wave;
	wave->prev_wave = prev_wave;
    }

    wave->edit_flag = SP_TRUE;
    
    return SP_TRUE;
}

spBool swCanUndoWave(swWave wave)
{
    if (wave == NULL || wave->prev_wave == NULL) return SP_FALSE;

    return SP_TRUE;
}

spBool swCanRedoWave(swWave wave)
{
    if (wave == NULL || wave->next_wave == NULL) return SP_FALSE;

    return SP_TRUE;
}

spBool swUndoWave(swWave *wave)
{
    if (swCanUndoWave(*wave) == SP_FALSE) return SP_FALSE;

    swFreeEditBuffer(*wave);
    *wave = (*wave)->prev_wave;

    return SP_TRUE;
}

spBool swRedoWave(swWave *wave)
{
    if (swCanRedoWave(*wave) == SP_FALSE) return SP_FALSE;

    swFreeEditBuffer(*wave);
    *wave = (*wave)->next_wave;

    return SP_TRUE;
}

spBool swIsWaveRangeAvailable(swWave wave)
{
    if (swIsWaveNone(wave) == SP_TRUE
	|| wave->range_available == SP_FALSE) return SP_FALSE;

    if ((swIsWaveFloat(wave) == SP_TRUE && wave->dmin > wave->dmax)
	|| (swIsWaveLong(wave) == SP_TRUE && wave->lmin > wave->lmax)
	|| (swIsWaveFloat(wave) == SP_FALSE && swIsWaveLong(wave) == SP_FALSE
	    && wave->min > wave->max)) {
	return SP_FALSE;
    }

    return SP_TRUE;
}

spBool swSetWaveRange(swWave wave, double min, double max)
{
    if (swIsWaveNone(wave) == SP_TRUE) return SP_FALSE;
    
    if (swIsWaveFloat(wave) == SP_TRUE) {
	wave->dmin = min;
	wave->dmax = max;
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	wave->lmin = (long)spRound(min);
	wave->lmax = (long)spRound(max);
    } else {
	wave->min = (short)spRound(min);
	wave->max = (short)spRound(max);
    }

    return SP_TRUE;
}

double swGetWaveMax(swWave wave)
{
    if (swIsWaveNone(wave) == SP_TRUE) return 0.0;
    
    if (swIsWaveFloat(wave) == SP_TRUE) {
	return wave->dmax;
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	return (double)wave->lmax;
    } else {
	return (double)wave->max;
    }
}

double swGetWaveMin(swWave wave)
{
    if (swIsWaveNone(wave) == SP_TRUE) return 0.0;
    
    if (swIsWaveFloat(wave) == SP_TRUE) {
	return wave->dmin;
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	return (double)wave->lmin;
    } else {
	return (double)wave->min;
    }
}

spBool swGetWaveBufferRange(swWave wave, long channel, char *buf, long buf_length,
			    double *minp, double *maxp, long *min_offsetp, long *max_offsetp)
{
    long k, l;
    long min_offset = 0, max_offset = 0;
    
    if (wave == NULL || buf == NULL || buf_length <= 0
	|| channel < 0 || channel >= wave->num_channel) return SP_FALSE;

    if (swIsWaveFloat(wave) == SP_TRUE) {
	double dmin, dmax;
	double *dbuf = (double *)buf;

	dmin = dmax = dbuf[channel]; 
	for (k = 1; k < buf_length; k++) {
	    l = k * wave->num_channel + channel;
	    if (dbuf[l] < dmin) {
		dmin = dbuf[l];
		min_offset = k;
	    }
	    if (dbuf[l] > dmax) {
		dmax = dbuf[l];
		max_offset = k;
	    }
	}
	if (minp != NULL) *minp = dmin;
	if (maxp != NULL) *maxp = dmax;
    } else if (swIsWaveLong(wave) == SP_TRUE) {
	long lmin, lmax;
	long *lbuf = (long *)buf;
	
	lmin = lmax = lbuf[channel]; 
	for (k = 1; k < buf_length; k++) {
	    l = k * wave->num_channel + channel;
	    if (lbuf[l] < lmin) {
		lmin = lbuf[l];
		min_offset = k;
	    }
	    if (lbuf[l] > lmax) {
		lmax = lbuf[l];
		max_offset = k;
	    }
	}
	if (minp != NULL) *minp = (double)lmin;
	if (maxp != NULL) *maxp = (double)lmax;
    } else {
	short smin, smax;
	short *sbuf = (short *)buf;
	
	smin = smax = sbuf[channel]; 
	for (k = 1; k < buf_length; k++) {
	    l = k * wave->num_channel + channel;
	    if (sbuf[l] < smin) {
		smin = sbuf[l];
		min_offset = k;
	    }
	    if (sbuf[l] > smax) {
		smax = sbuf[l];
		max_offset = k;
	    }
	}
	if (minp != NULL) *minp = (double)smin;
	if (maxp != NULL) *maxp = (double)smax;
    }
    
    if (min_offsetp != NULL) {
	*min_offsetp = min_offset;
    }
    if (max_offsetp != NULL) {
	*max_offsetp = max_offset;
    }

    return SP_TRUE;
}

spBool swGetBufferRange(swWave wave, long channel, double *minp, double *maxp,
			long *min_offsetp, long *max_offsetp)
{
    if (wave == NULL || wave->buf == NULL) return SP_FALSE;
    
    return swGetWaveBufferRange(wave, channel, wave->buf, wave->buf_length, minp, maxp, min_offsetp, max_offsetp);
}

spBool swGetWaveBufferData(swWave wave, long channel, long pos, long buf_length, char *buf, double *value)
{
    long k, l;

    if (buf == NULL || value == NULL || pos < 0) return SP_FALSE;
    
    if (swIsWaveNone(wave) == SP_FALSE) {
	if (channel < 0 || channel >= wave->num_channel) channel = 0;

	k = pos;
	if (k >= 0 && k < buf_length) {
	    l = k * wave->num_channel + channel;
	    if (swIsWaveFloat(wave) == SP_TRUE) {
		double *dbuf = (double *)buf;
		*value = dbuf[l];
	    } else if (swIsWaveLong(wave) == SP_TRUE) {
		long *lbuf = (long *)buf;
		*value = (double)lbuf[l];
	    } else {
		short *sbuf = (short *)buf;
		*value = (double)sbuf[l];
	    }
	    return SP_TRUE;
	}
    }
    *value = 0.0;
    
    return SP_FALSE;
}

spBool swSetWaveBufferData(swWave wave, long channel, long pos, long buf_length, char *buf, double value)
{
    long k, l;
    
    if (buf == NULL || pos < 0) return SP_FALSE;
    
    if (swIsWaveNone(wave) == SP_FALSE) {
	if (channel < 0 || channel >= wave->num_channel) channel = 0;

	k = pos;
	if (k >= 0 && k < buf_length) {
	    l = k * wave->num_channel + channel;
	    if (swIsWaveFloat(wave) == SP_TRUE) {
		double *dbuf = (double *)buf;
		dbuf[l] = value;
	    } else if (swIsWaveLong(wave) == SP_TRUE) {
		long *lbuf = (long *)buf;
		lbuf[l] = (long)swGetClippedValue(wave->samp_bit, value, &wave->overflow);
	    } else {
		short *sbuf = (short *)buf;
		sbuf[l] = (short)swGetClippedValue(wave->samp_bit, value, &wave->overflow);
	    }
	    return SP_TRUE;
	}
    }
    
    return SP_FALSE;
}

spBool swGetWaveData(swWave wave, long channel, long pos, double *value)
{
    if (wave != NULL) {
	return swGetWaveBufferData(wave, channel, pos, wave->length, wave->data, value);
    }
    return SP_FALSE;
}

spBool swSetWaveData(swWave wave, long channel, long pos, double value)
{
    if (wave != NULL) {
	return swSetWaveBufferData(wave, channel, pos, wave->length, wave->data, value);
    }
    return SP_FALSE;
}

double swGetBufferData(swWave wave, long channel, long pos)
{
    double value = 0.0;
    
    if (wave != NULL) {
	swGetWaveBufferData(wave, channel, pos, wave->buf_length, wave->buf, &value);
    }
    return value;
}

spBool swSetBufferData(swWave wave, long channel, long pos, double value)
{
    if (wave != NULL) {
	swSetWaveBufferData(wave, channel, pos, wave->buf_length, wave->buf, value);
    }
    return SP_FALSE;
}

spBool swGetPeakData(swWave wave, long channel, long pos, double *value)
{
    if (wave != NULL && wave->peak_buf != NULL
	&& wave->peak_buf_length > 0) {
	return swGetWaveBufferData(wave, channel,
				   /*2 * (pos / (2 * wave->peak_buf_thin_length))*/
				   pos / wave->peak_buf_thin_length,
				   wave->peak_buf_length, wave->peak_buf, value);
    }
    return SP_FALSE;
}

#ifdef SW_USE_THREAD
void *swCreateProcessThread(int priority, spThreadFunc func, swProcessConfig *config)
{
    void *thread;
    
    if ((thread = spCreateThread(0, priority, func, (void *)config)) != NULL) {
	return thread;
    }

    return NULL;
}
#endif

spBool swLockMutex(swWave wave)
{
#ifdef SW_USE_THREAD
    if (wave == NULL) return SP_FALSE;

    if (wave->mutex == NULL) {
	return SP_FALSE;
    }
    
    return spLockMutex(wave->mutex);
#else
    return SP_FALSE;
#endif
}

spBool swUnlockMutex(swWave wave)
{
#ifdef SW_USE_THREAD
    if (wave == NULL) return SP_FALSE;

    if (wave->mutex == NULL) {
	return SP_FALSE;
    }

    return spUnlockMutex(wave->mutex);
#else
    return SP_FALSE;
#endif
}

spBool swWaitMutex(swWave wave)
{
#ifdef SW_USE_THREAD
    if (wave == NULL) return SP_FALSE;

    if (wave->mutex == NULL) {
	return SP_FALSE;
    }

    return spWaitMutex(wave->mutex);
#else
    return SP_FALSE;
#endif
}

static swWave editWaveMain(swWave wave, swEditType edit_type, int samp_bit,
			   long offset, long length, double value, void *data,
			   spBool in_thread)
{
    spBool copy_label;
    spBool overflow;
    spBool process_flag;
    spBool process_finished;
    char *filename;
    char *orig_filename;
    int num_channel;
    double samp_rate;
    double ovalue;
    swWave owave = NULL;

    swLockMutex(wave);
    if (swIsWaveNone(wave) == SP_TRUE || swIsWaveProcessing(wave) == SP_TRUE) {
	swUnlockMutex(wave);
	return NULL;
    }
    wave->process_flag = SP_TRUE;
    swUnlockMutex(wave);

    if (edit_type == SW_EDIT_WRITE) {
	spBool callback_flag;
	
	callback_flag = (spRound(value) == 0.0 ? SP_FALSE : SP_TRUE);
	if (writeWave(wave->new_filename, wave->new_plugin_name,
		      wave->new_file_type, wave, callback_flag, in_thread) == SP_TRUE) {
	    owave = wave;
	} else {
	    swLockMutex(wave);
	    wave->process_flag = SP_FALSE;
	    swUnlockMutex(wave);
	}
    } else {
	filename = swGetTempFile(wave->config, NULL);
	if (edit_type == SW_EDIT_EXTRACT) {
	    orig_filename = NULL;
	    copy_label = SP_FALSE;
	} else {
	    orig_filename = wave->orig_filename;
	    copy_label = SP_TRUE;
	}
	
	if (edit_type == SW_EDIT_CROP || edit_type == SW_EDIT_EXTRACT) {
	    owave = swGetCroppedWave(filename, wave, offset, length, in_thread);
	} else {
	    if (edit_type != SW_EDIT_BIT_CONV) {
		samp_bit = wave->samp_bit;
	    }
	    overflow = SP_FALSE;
	    
	    /* write wave */
	    if (swWriteWaveRegion(filename, "output_raw", "raw", samp_bit, wave,
				  edit_type, offset, length, value, &ovalue, data, &overflow,
				  SP_FALSE, SP_TRUE, in_thread) == SP_TRUE) {
		if (edit_type == SW_EDIT_SAMP_RATE_CONV) {
		    samp_rate = value;
		} else {
		    samp_rate = wave->samp_rate;
		}
		if (edit_type == SW_EDIT_MONAURALIZE) {
		    num_channel = 1;
		} else {
		    num_channel = wave->num_channel;
		}
		
		/* allocate wave */
		if ((owave  = getWave(wave->config, filename, "input_raw", "raw", samp_bit,
				      num_channel, samp_rate, SP_FALSE, SP_FALSE,
				      NULL, in_thread)) != NULL) {
		    spDebug(50, "editWaveMain", "overflow = %d\n", overflow);
		    owave->overflow = overflow;
		    if (edit_type == SW_EDIT_SAMP_RATE_CONV) {
			owave->exact_samp_rate = ovalue;
		    }
		}
	    }
	}
	
	if (owave != NULL) {
	    swLockMutex(wave);
		
	    swSetOriginalName(owave, orig_filename, wave->orig_plugin_name);
	    swSetOriginalParams(owave, wave, copy_label);
	
	    if (edit_type != SW_EDIT_EXTRACT) {
#if 0
		if (edit_type == SW_EDIT_BIT_CONV) {
		    value = getBitConvWeight(wave, owave->samp_bit);
		} else if (edit_type != SW_EDIT_AMPLIFY) {
		    value = 1.0;
		}

		/* set wave range */
		swSetOriginalRange(owave, wave, value);
#endif
	
		/* set previous wave */
		swSetPrevWave(owave, wave);

		wave->edit_offset = offset;
		if (edit_type == SW_EDIT_PASTE
		    || edit_type == SW_EDIT_INSERT || edit_type == SW_EDIT_MIX) {
		    swWave src_wave = (swWave)data;
		    wave->edit_length = MIN(src_wave->total_length, wave->total_length - offset);
		} else {
		    wave->edit_length = length;
		}
	    }
	    
	    swUnlockMutex(wave);
	}
    }

    swLockMutex(wave);
    process_flag = wave->process_flag;
    process_finished = wave->process_finished;
    
    if (wave->new_filename != NULL) xfree(wave->new_filename);
    if (wave->new_plugin_name != NULL) xfree(wave->new_plugin_name);
    if (wave->new_file_type != NULL) xfree(wave->new_file_type);
    wave->new_filename = NULL;
    wave->new_plugin_name = NULL;
    wave->new_file_type = NULL;
    
    wave->process_flag = SP_FALSE;
    wave->process_finished = SP_TRUE;
    swUnlockMutex(wave);
    
    if (process_finished == SP_FALSE
	&& process_flag == SP_TRUE /* not interrupted */
	&& wave->config->edit_finish_func != NULL) {
	wave->config->edit_finish_func(wave, owave, in_thread, edit_type, wave->call_data);
	swWaitMutex(wave);
    }

    if (process_finished == SP_FALSE && wave->config->edit_func != NULL) {
	wave->config->edit_func(wave, in_thread, SW_PROCESS_FINISHED, edit_type, wave->call_data);
	swWaitMutex(wave);
    }
    
    return owave;
}

#ifdef SW_USE_THREAD
spThreadReturn editWaveThread(void *data)
{
    swWave wave;
    swProcessConfig *config = (swProcessConfig *)data;

    if (config == NULL) return SP_THREAD_RETURN_FAILURE;
    
    spDebug(10, "editWaveThread", "in\n");

    wave = editWaveMain(config->wave, config->edit_type, config->samp_bit,
			config->offset, config->length, config->value, config->data, SP_TRUE);

    if (wave != NULL) {
	return SP_THREAD_RETURN_SUCCESS;
    } else {
	return SP_THREAD_RETURN_FAILURE;
    }
}
#endif

static swWave editWave(swWave wave, swEditType edit_type, int samp_bit,
		       long offset, long length, double value, void *data,
		       char *filename, char *plugin_name, char *file_type,
		       spBool use_thread, spBool *thread_flag)
{
    if (wave->new_filename != NULL) xfree(wave->new_filename);
    if (wave->new_plugin_name != NULL) xfree(wave->new_plugin_name);
    if (wave->new_file_type != NULL) xfree(wave->new_file_type);
    
    if (filename != NULL) wave->new_filename = strclone(filename);
    if (plugin_name != NULL) wave->new_plugin_name = strclone(plugin_name);
    if (file_type != NULL) wave->new_file_type = strclone(file_type);
    
#ifdef SW_USE_THREAD
    if (use_thread == SP_TRUE
	&& wave->config->thread_safe == SP_TRUE && wave->config->process_use_thread == SP_TRUE
	&& wave->config->edit_finish_func != NULL && edit_type != SW_EDIT_EXTRACT) {
	wave->process_config.edit_type = edit_type;
	wave->process_config.wave = wave;
	wave->process_config.samp_bit = samp_bit;
	wave->process_config.offset = offset;
	wave->process_config.length = length;
	wave->process_config.value = value;
	wave->process_config.data = data;

	if (wave->edit_thread != NULL) {
	    spThreadLeave();
	    spWaitThread(wave->edit_thread);
	    spThreadEnter();
	    spDestroyThread(wave->edit_thread);
	}
	    
	if ((wave->edit_thread = swCreateProcessThread(SP_THREAD_PRIORITY_BELOW_NORMAL,
						       editWaveThread, &wave->process_config)) != NULL) {
	    if (thread_flag != NULL) *thread_flag = SP_TRUE;
	    return NULL;
	}
	spDebug(1, "editWave", "thread error\n");
    }
#endif

    if (thread_flag != NULL) *thread_flag = SP_FALSE;
    
    return editWaveMain(wave, edit_type, samp_bit, offset, length, value, data, SP_FALSE);
}

static swWave swEditWave(swWave wave, swEditType edit_type, int samp_bit,
			 long offset, long length, double value, void *data)
{
    return editWave(wave, edit_type, samp_bit, offset, length, value, data,
		    NULL, NULL, NULL, SP_TRUE, NULL);
}

spBool swWriteWave(char *filename, char *plugin_name, char *file_type,
		   swWave wave, spBool callback_flag, spBool use_thread)
{
    spBool thread_flag = SP_FALSE;
    
    if (editWave(wave, SW_EDIT_WRITE, 0, 0, 0, (double)callback_flag, NULL,
		 filename, plugin_name, file_type, use_thread, &thread_flag) == NULL
	&& thread_flag == SP_FALSE) {
	return SP_FALSE;
    }
    
    return SP_TRUE;
}

swWave swExtractWave(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_EXTRACT, 0, offset, length, 0.0, NULL);
}

swWave swCropWave(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_CROP, 0, offset, length, 0.0, NULL);
}

swWave swDeleteWave(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_DELETE, 0, offset, length, 0.0, NULL);
}

swWave swEraseWave(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_ERASE, 0, offset, length, 0.0, NULL);
}

swWave swAmplifyWave(swWave wave, long offset, long length, double value)
{
    return swEditWave(wave, SW_EDIT_AMPLIFY, 0, offset, length, value, NULL);
}

swWave swFadeInWave(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_FADE_IN, 0, offset, length, 0.0, NULL);
}

swWave swFadeOutWave(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_FADE_OUT, 0, offset, length, 0.0, NULL);
}

swWave swSwapWaveChannel(swWave wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_CHANNEL_SWAP, 0, offset, length, 0.0, NULL);
}

swWave swConvertWaveBit(swWave wave, int samp_bit)
{
    return swEditWave(wave, SW_EDIT_BIT_CONV, samp_bit, 0, 0, 0.0, NULL);
}

swWave swConvertWaveSampFreq(swWave wave, double samp_rate)
{
    return swEditWave(wave, SW_EDIT_SAMP_RATE_CONV, 0, 0, 0, samp_rate, NULL);
}

swWave swMonauralizeWave(swWave wave)
{
    return swEditWave(wave, SW_EDIT_MONAURALIZE, 0, 0, 0, 0.0, NULL);
}

swWave swChangeWaveValue(swWave wave, int channel, long offset, double value)
{
    return swEditWave(wave, SW_EDIT_CHANGE_VALUE, 0, offset, channel, value, NULL);
}

swWave swPasteWave(swWave wave, swWave src_wave, long offset, int channel)
{
    return swEditWave(wave, SW_EDIT_PASTE, 0, offset, channel, 0.0, (void *)src_wave);
}

swWave swMixWave(swWave wave, swWave src_wave, long offset, int channel)
{
    return swEditWave(wave, SW_EDIT_MIX, 0, offset, channel, 0.0, (void *)src_wave);
}

swWave swInsertWave(swWave wave, swWave src_wave, long offset, int channel)
{
    return swEditWave(wave, SW_EDIT_INSERT, 0, offset, channel, 0.0, (void *)src_wave);
}

swWave swInsertPauseWave(swWave wave, long offset, long pause_length)
{
    return swEditWave(wave, SW_EDIT_INSERT_PAUSE, 0, offset, pause_length, 0.0, NULL);
}

swWave swReplaceWave(swWave wave, swWave src_wave, long offset, long length)
{
    return swEditWave(wave, SW_EDIT_REPLACE, 0, offset, length, 0.0, (void *)src_wave);
}

spBool swSetWaveErrorCallback(swWaveConfig config, swWaveErrorCallback func)
{
    if (config == NULL) return SP_FALSE;

    config->error_func = func;

    return SP_TRUE;
}

spBool swSetWaveOptionCallback(swWaveConfig config, swWaveOptionCallback func)
{
    if (config == NULL) return SP_FALSE;

    config->option_func = func;

    return SP_TRUE;
}

spBool swSetWavePlayCallback(swWaveConfig config, swWavePlayCallback func)
{
    if (config == NULL) return SP_FALSE;

    config->play_func = func;

    return SP_TRUE;
}

spBool swSetWaveReadCallback(swWaveConfig config, swWaveIoCallback func)
{
    if (config == NULL) return SP_FALSE;

    config->read_func = func;

    return SP_TRUE;
}

spBool swSetWaveEditCallback(swWaveConfig config, swWaveEditCallback func)
{
    if (config == NULL) return SP_FALSE;

    config->edit_func = func;

    return SP_TRUE;
}

spBool swSetWaveEditFinishCallback(swWaveConfig config, swWaveEditFinishCallback func)
{
    if (config == NULL) return SP_FALSE;

    config->edit_finish_func = func;

    return SP_TRUE;
}

spBool swSetWaveCallbackData(swWave wave, void *data)
{
    if (wave == NULL) return SP_FALSE;

    wave->call_data = data;

    return SP_TRUE;
}

spBool swProcessStop(swWave wave)
{
    spBool flag = SP_FALSE;
    
    if (wave == NULL) return SP_FALSE;

    if (wave->process_finished == SP_FALSE) {
	if (wave->process_flag == SP_TRUE) {
	    wave->process_flag = SP_FALSE;
	    flag = SP_TRUE;
	}
    }
    spDebug(1, "swProcessStop", "flag = %d\n", flag);
    
    return flag;
}

swWave swAnalysisWave(swWave iwave, long offset, long length,
		      swAnalysisType analysis_type, swWindowType window_type,
		      long fftl, double lifterm)
{
    int n;
    int channel;
    int num_channel;
    long k;
    long hfftl;
    long lifl;
    long nread, nwrite;
    swWave wave;
    spPlugin *plugin;
    static DVector data = NULL;
    static DVector ana_window = NULL;

    swLockMutex(iwave);
    if (swIsWaveNone(iwave) == SP_TRUE || swIsWaveProcessing(iwave) == SP_TRUE) {
	swUnlockMutex(iwave);
	return NULL;
    }
    iwave->process_flag = SP_TRUE;
    iwave->process_finished = SP_FALSE;
    swUnlockMutex(iwave);
    
    fftl = POW2(spNextPow2(fftl));
    hfftl = fftl / 2 + 1;
    lifl = (long)spRound(iwave->samp_rate * lifterm / 1000.0);

    if (iwave->selected_channel >= 0) {
	num_channel = 1;
    } else {
	num_channel = iwave->num_channel;
    }

    wave = swInitWave(iwave->config,
		      swGetTempFile(iwave->config, NULL), "input_raw", "raw",
		      64, num_channel, iwave->samp_rate, SP_FALSE);
    swAllocData(wave, hfftl);
    wave->length = hfftl;
    wave->total_length = hfftl;
	
    if (data != NULL && data->length != fftl) {
	xdvfree(data);
	data = NULL;
    }
    if (data == NULL) {
	data = xdvrizeros(fftl);
    } else {
	dvzeros(data, data->length);
    }

    /* create analysis window */
    if (ana_window != NULL && ana_window->length != length) {
	xdvfree(ana_window);
	ana_window = NULL;
    }
    if (ana_window == NULL) {
	ana_window = xdvalloc(length);
    }
    switch (window_type) {
      case SW_WINDOW_HAMMING:
	hamming(ana_window->data, ana_window->length);
	break;
      case SW_WINDOW_HANNING:
	hanning(ana_window->data, ana_window->length);
	break;
      case SW_WINDOW_BLACKMAN:
	blackman(ana_window->data, ana_window->length);
	break;
      case SW_WINDOW_GAUSS:
	gausswin(ana_window->data, ana_window->length);
	break;
      default:
	dvones(ana_window, ana_window->length);
	break;
    }

    /* read wave */
    if ((plugin = swOpenWave(iwave, "r")) == NULL) {
	nread = 0;
    } else {
	spSeekPlugin(plugin, offset);
	nread = swReadBuffer(plugin, iwave, -1, length);
	spCloseFilePlugin(plugin);

	spDebug(10, "swAnalysisWave", "length = %ld, nread = %ld\n", length, nread);
    }

    if (nread > 0) {
	for (n = 0; n < iwave->num_channel; n++) {
	    if (iwave->selected_channel >= 0 && n != iwave->selected_channel) {
		continue;
	    }
	
	    for (k = 0; k < data->length; k++) {
		if (k < nread) {
		    data->data[k] = swGetBufferData(iwave, n, k) * ana_window->data[k];
		} else {
		    data->data[k] = 0.0;
		}
		if (data->imag != NULL) {
		    data->imag[k] = 0.0;
		}
	    }

	    switch (analysis_type) {
	      case SW_ANALYSIS_SPECTRUM:
	      case SW_ANALYSIS_SMOOTHED_SPECTRUM:
		dvfft(data);
		if (analysis_type == SW_ANALYSIS_SMOOTHED_SPECTRUM
		    && lifl >= 1) {
		    dvspectocep(data);
		    dvlif(data, fftl, lifl);
		    dvceptospec(data);
		}
		if (iwave->config->linear_spectrum == SP_TRUE) {
		    dvabs(data);
		    if (iwave->config->normalize_spectrum == SP_TRUE) {
			double ampmax;

			ampmax = ABS(swGetWaveMax(iwave));
			ampmax = MAX(ABS(swGetWaveMin(iwave)), ampmax);
			if (ampmax > 0.0) {
			    dvscoper(data, "/", ampmax);
			}
		    }
		} else {
		    dvscmax(data, 1.0e-5);
		    dvdecibel(data);
		    if (iwave->config->normalize_spectrum == SP_TRUE) {
			double dbmax;

			dbmax = ABS(swGetWaveMax(iwave));
			dbmax = MAX(ABS(swGetWaveMin(iwave)), dbmax);
			dbmax = dB(dbmax);
			dvscoper(data, "-", dbmax);
		    }
		}
		break;
	      case SW_ANALYSIS_CEPSTRUM:
		dvfft(data);
		dvspectocep(data);
		break;
	      case SW_ANALYSIS_PHASE:
	      case SW_ANALYSIS_UNWRAPPED_PHASE:
		dvfft(data);
		dvangle(data);
		if (analysis_type == SW_ANALYSIS_UNWRAPPED_PHASE) {
		    dvunwrap(data, 0.0);
		}
		break;
	      case SW_ANALYSIS_GROUP_DELAY:
		{
		    DVector tmp;
		    tmp = xdvfftgrpdly(data, fftl);
		    xdvfree(data);
		    data = tmp;
		    dvscoper(data, "-", (double)(length / 2));
		    dvscoper(data, "*", 1000.0 / iwave->samp_rate);
		    /*data->data[hfftl - 1] = 0.0;*/
		}
		break;
	      default:
		nread = 0;
		break;
	    }
	    if (nread <= 0) {
		break;
	    }
	    
	    if (iwave->selected_channel >= 0) {
		channel = 0;
	    } else {
		channel = n;
	    }
	    for (k = 0; k < hfftl; k++) {
		swSetWaveData(wave, channel, k, data->data[k]);
	    }
	}
    }

    spDebug(10, "swAnalysisWave", "set analysis data done\n");

    if (nread > 0) {
	/* write wave */
	if ((plugin = swOpenWave(wave, "w")) != NULL) {
	    if ((nwrite = swWriteData(plugin, wave)) <= 0) {
		nread = 0;
	    }
	    spCloseFilePlugin(plugin);
	    spDebug(10, "swAnalysisWave", "write data done\n");
	} else {
	    nread = 0;
	}
    }

    swLockMutex(iwave);
    iwave->process_flag = SP_FALSE;
    iwave->process_finished = SP_TRUE;
    swUnlockMutex(iwave);
    
    if (nread <= 0) {
	swFreeWave(wave);
	return NULL;
    } else {
	return wave;
    }
}
