/*
 *	swWaveAudio.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spThread.h>
#include <sp/spOption.h>
#include <sp/spWave.h>
#include <sp/spAudio.h>

#include <sp/spTopLevel.h>

#include "swWaveAudio.h"

struct _swAudio {
    spAudio audio;
    long buf_size;
    char *buf;
    swWave wave;
};

static spBool playWaveAudio(swWave wave, spBool loop_flag, spBool in_thread)
{
    spBool flag = SP_FALSE;
    spBool started = SP_FALSE;
    spBool break_flag;
    spBool repeat_flag;
    int num_channel;
    int selected_channel;
    long k;
    long max_loop;
    long pos;
    long offset, length;
    long prev_offset, prev_length;
    long nread, nwrite;
    long play_length;
    long audio_length, min_audio_length;
    long wlength;
    long total_play_length;
    spPlugin *plugin;
    swAudio audio;

    if (swIsWavePlayable(wave) == SP_FALSE
	|| (audio = swBeginAudio(wave)) == NULL) {
	return SP_FALSE;
    }

    swLockMutex(wave);
    wave->process_flag = SP_TRUE;
    wave->process_finished = SP_FALSE;
    wave->play_flag = SP_TRUE;
    
    selected_channel = wave->selected_channel;
    if (selected_channel >= 0) {
	num_channel = 1;
    } else {
	num_channel = wave->num_channel;
    }
    swUnlockMutex(wave);
    
    spDebug(10, "playWaveAudio", "play_read_length = %ld, samp_rate = %f\n",
	    wave->config->play_read_length, wave->samp_rate);
    
    min_audio_length = swConvertByteToLength(wave, audio->buf_size) / 2;
    wlength = POW2(spNextPow2((long)((double)wave->config->play_read_length
				     * (wave->samp_rate / 48000.0))));
    spDebug(10, "playWaveAudio", "wlength = %ld\n", wlength);
    wlength = MAX(wlength, 256);
    wlength = MIN(wlength, min_audio_length);
    
    /* open audio device */
    if (spOpenAudioDevice(audio->audio, "wo") == SP_FALSE) {
	spDebug(10, "playWaveAudio", "Can't open audio device.\n");
    } else {
	/* open wave */
	if ((plugin = swOpenWave(wave, "r")) != NULL) {
	    flag = SP_TRUE;
	    started = SP_TRUE;

	    swLockMutex(wave);
	    offset = wave->play_offset;
	    length = wave->play_length;
	    swUnlockMutex(wave);
	    
	    spDebug(50, "playWaveAudio", "offset = %ld, length = %ld\n", offset, length);
	    
	    if (wave->config->play_func != NULL) {
		flag = wave->config->play_func(wave, in_thread, SW_PROCESS_STARTED, wave->call_data);
		if (flag == SP_TRUE) {
		    flag = wave->config->play_func(wave, in_thread, offset, wave->call_data);
		}
	    }

	    audio_length = 0;
	    
	    prev_offset = offset;
	    prev_length = length;
	    total_play_length = 0;
	    max_loop = 2147483647L / wlength;
	    spDebug(50, "playWaveAudio", "max_loop = %ld\n", max_loop);

	    for (k = 0; k < max_loop && flag == SP_TRUE; k++) {
		spSeekPlugin(plugin, offset);
		    
		spDebug(80, "playWaveAudio", "wlength = %ld, min_audio_length = %ld, audio_length = %ld, buf_size = %ld\n",
			wlength, min_audio_length, audio_length, audio->buf_size);

		repeat_flag = SP_FALSE;
		nread = 0;
		play_length = 0;
		while (flag == SP_TRUE) {
		    if ((nread = swLoadIntoBuffer(plugin, wave, selected_channel,
						  wlength, audio->buf + swConvertLengthToByte(wave, audio_length))) <= 0) {
			spDebug(10, "playWaveAudio", "nread failed: nread = %ld\n", nread);
			break;
		    }
		    if (play_length + nread >= length) {
			nread = length - play_length;
		    }
		    audio_length += nread;
		    play_length += nread;
		    total_play_length += nread;

		    if (audio_length >= min_audio_length) {
			spDebug(80, "playWaveAudio", "audio_length = %ld, min_audio_length = %ld\n",
				audio_length, min_audio_length);
			
			/* write audio data */
			nwrite = spWriteAudio(audio->audio, audio->buf, audio_length * num_channel);
			if (nwrite <= 0) {
			    spDebug(10, "playWaveAudio", "write failed: nwrite = %ld\n", nwrite);
			    flag = SP_FALSE;
			    break;
			}
			audio_length = 0;
			
			if (spGetAudioOutputPosition(audio->audio, &pos) == SP_TRUE) {
			    spDebug(80, "playWaveAudio",
				    "current pos = %ld, play_length = %ld, total_play_length = %ld\n",
				    pos, play_length, total_play_length);
			    pos -= (total_play_length - play_length);
			    if (pos < 0) {
				pos += prev_length;
				pos = MAX(pos, 0);
			    }
			    pos += prev_offset;
			} else {
			    pos = prev_offset + play_length;
			}
			
			if (wave->config->play_func != NULL && flag == SP_TRUE) {
			    flag = wave->config->play_func(wave, in_thread, pos, wave->call_data);
			}
			
			if (wave->process_flag == SP_FALSE) {
			    spStopAudio(audio->audio);
			    loop_flag = SP_FALSE;
			    break;
			}
			
			break_flag = SP_FALSE;
			
			swLockMutex(wave);
			if (wave->play_offset != offset || wave->play_length != length) {
			    if (wave->play_offset != offset
				|| wave->play_offset > offset + length
				|| pos < wave->play_offset
				|| wave->play_offset + wave->play_length < pos) {
				break_flag = SP_TRUE;
				offset = wave->play_offset;
				length = wave->play_length;
			    } else {
				if (offset + length != wave->play_offset + wave->play_length) {
				    length = wave->play_offset + wave->play_length - offset;
				}
			    }
			    repeat_flag = SP_TRUE;
			}
			swUnlockMutex(wave);
			
			if (break_flag == SP_TRUE) {
			    break;
			}
		    }

		    if (play_length >= length) {
			if (loop_flag == SP_FALSE && audio_length > 0) {
			    spWriteAudio(audio->audio, audio->buf, audio_length * num_channel);
			}
			break;
		    }
		}
		spDebug(80, "playWaveAudio", "play_length = %ld, nread = %ld\n", play_length, nread);
		    
		if (nread < 0) {
		    flag = SP_FALSE;
		}
		if (loop_flag == SP_FALSE && repeat_flag == SP_FALSE) {
		    break;
		}
		swLockMutex(wave);
		offset = wave->play_offset;
		length = wave->play_length;
		swUnlockMutex(wave);
		
		prev_offset = offset;
		prev_length = play_length;
	    }
	    spDebug(10, "playWaveAudio", "play finished: total_play_length = %ld\n", total_play_length);

	    /* close file */
	    spCloseFilePlugin(plugin);
	}
	
	spCloseAudioDevice(audio->audio);
	
	swLockMutex(wave);
	wave->process_flag = SP_FALSE;
	swUnlockMutex(wave);
	    
	if (started == SP_TRUE && wave->config->play_func != NULL) {
	    flag = wave->config->play_func(wave, in_thread, SW_PROCESS_FINISHED, wave->call_data);
	}
    }
    
    swLockMutex(wave);
    wave->process_flag = SP_FALSE;
    wave->process_finished = SP_TRUE;
    wave->play_flag = SP_FALSE;
    swUnlockMutex(wave);
    
    swEndAudio(audio);
    
    spDebug(80, "playWaveAudio", "done: flag = %d\n", flag);
    
    return flag;
}

static spBool playWaveFile(swWave wave, long offset, long length)
{
    spBool flag;
    int swap;
    int num_channel;
    int selected_channel;
    long k;
    long nread;
    long play_length;
    long dlength;
    long woffset, wlength;
    long rlength;
    short *dummy;
    spWaveInfo wave_info;
    char filename[SP_MAX_PATHNAME];
    FILE *fp;
    spPlugin *plugin;

    if (swIsWaveFloat(wave) == SP_TRUE) {
	return SP_FALSE;
    }

    if (length > wave->config->max_read_length) {
	if (wave->config->error_func != NULL) {
	    return wave->config->error_func(wave, SP_FALSE, SW_ERROR_PLAY_LONG, wave->call_data);
	}
	
	return SP_FALSE;
    }
    
    /* make temporary file */
    if (wave->config->play_use_wav == SP_TRUE) {
	swap = SP_WAV_NEED_SWAP;
	sprintf(filename, "%s%csp%ld.wav",
		spGetTempDir(), SP_DIR_SEPARATOR, spGetProcessId());
    } else {
	swap = 0;
	sprintf(filename, "%s%csp%ld.ad",
		spGetTempDir(), SP_DIR_SEPARATOR, spGetProcessId());
    }

    if (NULL == (fp = spOpenFile(filename, "wb"))) {
        return SP_FALSE;
    }

    spDebug(10, "playWaveFile", "file = %s\n", filename);

    selected_channel = wave->selected_channel;
    if (selected_channel >= 0) {
	num_channel = 1;
    } else {
	num_channel = wave->num_channel;
    }

    /* get dummy length */
    dlength =(long)(SP_WAVE_PLAY_DUMMY_TIME * wave->samp_rate);
    dlength = dlength * (long)num_channel;
    woffset = offset * (long)num_channel;
    wlength = length * (long)num_channel;
    
    if (wave->config->play_use_wav == SP_TRUE) {
	spInitWaveInfo(&wave_info);
	wave_info.num_channel = num_channel;
	wave_info.samp_rate = wave->samp_rate;
	wave_info.samp_bit = 16;
	wave_info.length = dlength * 2 + wlength;
	
	/* write header */
	spWriteWavInfo(&wave_info, fp);
    }

    /* write dummy */
    dummy = xalloc(dlength, short);
    for (k = 0; k < dlength; k++) dummy[k] = 0;
    fwriteshort(dummy, dlength, swap, fp);

    /* open wave */
    if ((plugin = swOpenWave(wave, "r")) == NULL) {
	return SP_FALSE;
    }

    spSeekPlugin(plugin, offset);

    rlength = MAX(wave->config->play_read_length,
		  wave->config->audio_buffer_size / num_channel);
    spDebug(10, "playWaveFile", "rlength = %ld\n", rlength);
    
    play_length = 0;
    while ((nread = swReadBuffer(plugin, wave, selected_channel, rlength)) > 0) {
	if (play_length + nread >= length) {
	    nread = length - play_length;
	}
	
	/* write signal */
	if (swIsWaveLong(wave) == SP_TRUE) {
	    if (wave->samp_bit == 24) {
		fwritelong24tos((long *)wave->buf, nread * num_channel, swap, fp);
	    } else {
		fwritelong32tos((long *)wave->buf, nread * num_channel, swap, fp);
	    }
	} else {
	    fwriteshort((short *)wave->buf, nread * num_channel, swap, fp);
	}
	play_length += nread;

	if (play_length >= length) {
	    break;
	}
    }
    
    /* write dummy */
    fwriteshort(dummy, dlength, swap, fp);
    xfree(dummy);

    /* close file */
    spCloseFile(fp);
    
    /* execute play command */
    flag = spPlayFile(filename, num_channel, wave->samp_rate);

    /* delete temporary file */
    spRemoveFile(filename);

    /* close input file */
    spCloseFilePlugin(plugin);
    
    return flag;
}

#ifdef SW_USE_THREAD
static spThreadReturn playWaveThread(void *data)
{
    spBool flag;
    swProcessConfig *config = (swProcessConfig *)data;

    if (config == NULL) return SP_THREAD_RETURN_FAILURE;
    
    spDebug(10, "playWaveThread", "in\n");

    flag = playWaveAudio(config->wave, config->flag, SP_TRUE);

    if (flag == SP_FALSE && config->wave->config->play_func != NULL) {
	config->wave->config->play_func(config->wave, SP_TRUE, SW_PROCESS_ERROR, config->wave->call_data);
	swWaitMutex(config->wave);
    }

    spDebug(10, "playWaveThread", "done\n");
    
    if (flag == SP_TRUE) {
	return SP_THREAD_RETURN_SUCCESS;
    } else {
	return SP_THREAD_RETURN_FAILURE;
    }
}
#endif

static void correctPlayRegion(swWave wave)
{
    wave->play_offset = MAX(wave->play_offset, 0);
    wave->play_offset = MIN(wave->play_offset, wave->total_length - 1);
    if (wave->play_length <= 0) wave->play_length = wave->total_length;
    wave->play_length = MIN(wave->play_length, wave->total_length - wave->play_offset);

    return;
}

spBool swSetPlayRegion(swWave wave, long offset, long length)
{
    if (swIsWaveNone(wave) == SP_TRUE) {
	return SP_FALSE;
    }
    
    swLockMutex(wave);
    wave->play_offset = offset;
    wave->play_length = length;
    correctPlayRegion(wave);
    swUnlockMutex(wave);
    
    return SP_TRUE;
}

spBool swPlayWave(swWave wave, spBool loop_flag)
{
    if (swIsWaveNone(wave) == SP_TRUE || swIsWaveProcessing(wave) == SP_TRUE) {
	spDebug(10, "swPlayWave", "playing failed\n");
	return SP_FALSE;
    }

    swLockMutex(wave);
    if (wave->play_length <= 0) {
	wave->selected_channel = -1;
    }
    correctPlayRegion(wave);
    swUnlockMutex(wave);

#ifdef SP_SUPPORT_AUDIO
    if (wave->config->play_use_audio == SP_TRUE && swIsWaveFloat(wave) == SP_FALSE) {
#ifdef SW_USE_THREAD
	if (wave->config->thread_safe == SP_TRUE && wave->config->process_use_thread == SP_TRUE) {
	    wave->process_config.edit_type = SW_EDIT_NONE;
	    wave->process_config.wave = wave;
	    wave->process_config.flag = loop_flag;

	    if (wave->audio_thread != NULL) {
		spDebug(10, "swPlayWave", "waiting thread...\n");
		spThreadLeave();
		spWaitThread(wave->audio_thread);
		spThreadEnter();
		spDestroyThread(wave->audio_thread);
		spDebug(10, "swPlayWave", "waiting thread done\n");
	    }
	    
	    if ((wave->audio_thread = swCreateProcessThread(SP_THREAD_PRIORITY_NORMAL,
							    playWaveThread, &wave->process_config)) != NULL) {
		return SP_TRUE;
	    }
	    spDebug(1, "swPlayWave", "thread error\n");
	}
#endif
	return playWaveAudio(wave, loop_flag, SP_FALSE);
    }
#endif

    return playWaveFile(wave, wave->play_offset, wave->play_length);
}

static swAudio sw_audio = NULL;

spBool swIsWavePlayable(swWave wave)
{
    spBool flag = SP_TRUE;
    
    if (wave == NULL) return SP_FALSE;
    
    swLockMutex(wave);
    if (sw_audio != NULL && sw_audio->wave != NULL) flag = SP_FALSE;
    swUnlockMutex(wave);
    
    return flag;
}

static void swTerminateAudio(void *data)
{
    spDebug(1, "swTerminateAudio", "in\n");
    
    if (sw_audio != NULL) {
	if (sw_audio->audio != NULL) {
	    spFreeAudio(sw_audio->audio);
	}

	if (sw_audio->buf != NULL) {
	    xfree(sw_audio->buf);
	}
	
	xfree(sw_audio);
	sw_audio = NULL;
    }
    
    spDebug(1, "swTerminateAudio", "done\n");
    
    return;
}

swAudio swBeginAudio(swWave wave)
{
    int num_channel;
    long read_size;
    
    if (wave == NULL) return NULL;
    
    swLockMutex(wave);
    
    if (sw_audio == NULL) {
	sw_audio = xalloc(1, struct _swAudio);
	sw_audio->audio = spInitAudio();
	sw_audio->buf_size = 0;
	sw_audio->buf = NULL;
	sw_audio->wave = NULL;

	spAddExitCallback(swTerminateAudio, NULL);
    }

    if (sw_audio->audio != NULL) {
	sw_audio->wave = wave;

	read_size = swConvertLengthToByte(wave, wave->config->play_read_length);
	if (sw_audio->buf_size < read_size) {
	    sw_audio->buf_size = read_size;
	    if (sw_audio->buf == NULL) {
		sw_audio->buf = xalloc(sw_audio->buf_size, char);
	    } else {
		sw_audio->buf = xrealloc(sw_audio->buf, sw_audio->buf_size, char);
	    }
	}
	
	spSetAudioSampleRate(sw_audio->audio, (long)wave->samp_rate);
	spSetAudioBufferSize(sw_audio->audio, wave->config->audio_buffer_size);
	spSetAudioSampleBit(sw_audio->audio, MAX(wave->samp_bit, 16));
	
	if (wave->selected_channel >= 0) {
	    num_channel = 1;
	} else {
	    num_channel = wave->num_channel;
	}
	spSetAudioChannel(sw_audio->audio, num_channel);
    }
    
    swUnlockMutex(wave);

    return sw_audio;
}

spBool swEndAudio(swAudio audio)
{
    swWave wave;
    
    if (audio == NULL || audio->wave == NULL) return SP_FALSE;

    wave = audio->wave;
    swLockMutex(wave);
    audio->wave = NULL;
    swUnlockMutex(wave);
    
    return SP_TRUE;
}
