/*
 *	swWindow.c
 *
 *	Last modified: <2002-04-07 01:44:38 hideki>
 *  Last modified: <2004-02-22 15:59:05 Chien-Wen Danny Su>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spBaseLib.h>
#include <sp/spAudioLib.h>
#include <sp/spComponentLib.h>
#include <sp/spPlugin.h>

#include "swWave.h"
#include "swWaveAudio.h"

#include "swWindow.h"
#include "swDialog.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swEdit.h"
#include "swLabel.h"
#include "swLabelDialog.h"
#include "swAnalysis.h"
#include "swAnalysisDialog.h"
#include "swInfoDialog.h"
#include "swLabelList.h"

#ifdef SW_CLOSE_PROMPT
#define SW_DEFAULT_CLOSE_STYLE SP_CALLBACK_CLOSE
#else
#define SW_DEFAULT_CLOSE_STYLE SP_DESTROY_CLOSE
#endif


/* record module variables */
// NOTE: what's the "volatile" keyword
#define AUDIO_BUFFER_SIZE 1024
#define BUFFER_LENGTH 160000
char *temp_output = NULL;
spBool started = SP_FALSE;
spBool record_successful = SP_FALSE;

int num_channel;
char *filetype;

int default_channel = 0; /* because Julius can handle only single channel data */
spAudio source;
spPlugin *output_plugin;
static void *input_thread_handle = NULL;
long offset;

swWave wave = NULL;
spPluginError err;
/* end record module variables */


#define SW_MAX_PLUGIN 64

#define SW_MIN_DRAW_INTERVAL_F /*0.2*/0.1

spBool saveas = SP_FALSE;
spBool open = SP_FALSE;
spComponent record_button;

swWindow swGetNextWindow(swWindow window)
{
    spComponent next;
    swWindow next_window = NULL;

    if (window == NULL || window->window == NULL) return NULL;

    next = spGetNextWindow(window->window, SP_FALSE);
    while (next != NULL && next != window->window) {
	next_window = (swWindow)spGetUserData(next);
	if (swIsClipboardWindow(next_window) == SP_TRUE) {
	    if (swIsClipboardVisible(next_window) == SP_TRUE) {
		break;
	    }
	} else if (next_window != NULL) {
	    break;
	}
	next_window = NULL;
	spDebug(80, "swGetNextWindow", "next\n");
	next = spGetNextWindow(next, SP_FALSE);
    }

    return next_window;
}

swWindow swGetPrevWindow(swWindow window)
{
    spComponent prev;
    swWindow prev_window = NULL;

    if (window == NULL || window->window == NULL) return NULL;

    prev = spGetPrevWindow(window->window, SP_FALSE);
    while (prev != NULL && prev != window->window) {
	prev_window = (swWindow)spGetUserData(prev);
	if (swIsClipboardWindow(prev_window) == SP_TRUE) {
	    if (swIsClipboardVisible(prev_window) == SP_TRUE) {
		break;
	    }
	} else if (prev_window != NULL) {
	    break;
	}
	prev_window = NULL;
	spDebug(80, "swGetNextWindow", "prev\n");
	prev = spGetPrevWindow(prev, SP_FALSE);
    }

    return prev_window;
}

static void popdownOptionDialogCB(spComponent component, spCallbackReason *preason)
{
    spCallbackReason reason;
    
    reason = spGetCallbackReason(component);

    if (reason == SP_CR_OK || reason == SP_CR_APPLY) {
	spGetOptionDialogData(component);
    }
    
    if (preason != NULL) {
	*preason = reason;
	spDebug(10, "popdownOptionDialogCB", "reason = %d\n", *preason);
    }

    if (reason == SP_CR_OK || reason == SP_CR_CANCEL) {
	spPopdownWindow(component);
    }

    return;
}

static spBool waveOptionCallback(swWave wave, spBool in_thread, spOptions options, void *data)
{
    spBool flag = SP_TRUE;
    spComponent dialog;
    spCallbackReason reason = SP_CR_CANCEL;
    swWindow window = (swWindow)data;

    if (window != NULL && wave != NULL) {
	if (in_thread == SP_TRUE) spThreadEnter();
	
	swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
	
	spDebug(10, "waveOptionCallback", "in\n");
	dialog = spCreateOptionDialog("Options for Output File",
				      options,
				      SppCallbackFunc, popdownOptionDialogCB,
				      SppCallbackData, &reason,
				      SppDialogBoxButtonType, SP_DB_OK_CANCEL,
				      SppCloseStyle, SP_UNMAP_CLOSE,
				      NULL);
	spDebug(10, "waveOptionCallback", "create dialog done\n");
	spPopupWindow(dialog);

	spDestroyWindow(dialog);
	
	spDebug(10, "waveOptionCallback", "reason = %d\n", reason);
	if (reason == SP_CR_OK) {
	    swSetMouseCursor(window, SP_CURSOR_WAIT);
	} else {
	    flag = SP_FALSE;
	}
	
	if (in_thread == SP_TRUE) spThreadLeave();
    }
    
    spDebug(10, "waveOptionCallback", "done\n");
    
    return flag;
}

static spBool waveErrorCallback(swWave wave, spBool in_thread, int error, void *data)
{
    spComponent w;
    swWindow window = (swWindow)data;

    spDebug(10, "waveErrorCallback", "error = %d, in_thread = %d\n", error, in_thread);
    
    if (in_thread == SP_TRUE) spThreadEnter();
    
    if (window == NULL) {
	w = swGetFormatDialogWindow();
    } else {
	w = window->window;
    }

    switch (error) {
      case SP_PLUGIN_ERROR_MULTIPLE_INSTANCE:
	spDisplayError(w, SW_ERROR_TITLE, SW_MULTIPLE_INSTANCE_ERROR_MESSAGE);
	break;
      case SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND:
	spDisplayError(w, SW_ERROR_TITLE, SW_SUFFIX_ERROR_MESSAGE);
	if (in_thread == SP_FALSE) {
	    if (window != NULL) swSaveAsWindow(w, window);
	}
	break;
      case SP_PLUGIN_ERROR_SAMP_RATE:
	spDisplayError(w, SW_ERROR_TITLE, SW_SAMP_RATE_ERROR_MESSAGE);
	break;
      case SP_PLUGIN_ERROR_SAMP_BIT:
	spDisplayError(w, SW_ERROR_TITLE, SW_SAMP_BIT_ERROR_MESSAGE);
	break;
      case SP_PLUGIN_ERROR_NUM_CHANNEL:
	spDisplayError(w, SW_ERROR_TITLE, SW_NUM_CHANNEL_ERROR_MESSAGE);
	break;
      case SW_ERROR_PLAY_LONG:
	spDisplayError(w, SW_ERROR_TITLE, SW_PLAY_LONG_ERROR_MESSAGE);
	break;
      case SW_ERROR_LOOP_PLAY_SHORT:
	spDisplayError(w, SW_ERROR_TITLE, SW_LOOP_PLAY_SHORT_ERROR_MESSAGE);
	break;
      case SP_PLUGIN_ERROR_INSTANTIATE:
      case SP_PLUGIN_ERROR_LOAD:
      case SP_PLUGIN_ERROR_WRONG_PLUGIN:
	spDisplayError(w, SW_ERROR_TITLE, SW_PLUGIN_ERROR_MESSAGE);
	break;
      case SP_PLUGIN_ERROR_FAILURE:
	spDisplayError(w, SW_ERROR_TITLE, SW_FORMAT_NOT_SUPPORTED_MESSAGE);
	break;
      default:
	spDisplayError(w, SW_ERROR_TITLE, SW_SAVE_ERROR_MESSAGE);
	break;
    }
    
    if (in_thread == SP_TRUE) spThreadLeave();

    return SP_TRUE;
}

static spBool waveProcessCallback(swWave wave, spBool in_thread, long pos, swEditType edit_type, void *data)
{
    long point_d;
    double point_f;
    swWindow window = (swWindow)data;

    if (window != NULL && window->wave != NULL) {
	spDebug(80, "waveProcessCallback", "thread = %d\n", in_thread);
	
	if (in_thread == SP_TRUE) spThreadEnter();
    
	if (pos == SW_PROCESS_FINISHED) {		/* process finished */
	    spDebug(30, "waveProcessCallback", "process finished\n");
	    swLockWindowMutex(window);
	    window->point_f = window->prev_point_f;
	    swUpdatePoint(window, swDimToSamp(window, window->point_f), SP_TRUE);
	    window->point_d = swSampToDisp(window, window->point);
	    swRefreshWindow(window, SP_TRUE);
	    swUnlockWindowMutex(window);

	    swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
	    swSetProcessSenseLevel(window, SP_FALSE);
	    
	    if (edit_type == SW_EDIT_PASTE || edit_type == SW_EDIT_INSERT
		|| edit_type == SW_EDIT_MIX || edit_type == SW_EDIT_REPLACE) {
		swLockMainMutex(window);
		window->config->toplevel->using_clipboard = SP_FALSE;
		swUnlockMainMutex(window);
	    }
	} else if (pos == SW_PROCESS_ERROR) {		/* process error */
	    spDebug(30, "waveProcessCallback", "error: edit_type = %d\n", edit_type);
	    
	    if (edit_type == SW_EDIT_NONE) {
		spDisplayError(window->window, NULL, SW_PLAY_ERROR_MESSAGE);
	    }
	} else if (pos == SW_PROCESS_STARTED) {
	    if (edit_type == SW_EDIT_PASTE || edit_type == SW_EDIT_INSERT 
		|| edit_type == SW_EDIT_MIX || edit_type == SW_EDIT_REPLACE) {
		swLockMainMutex(window);
		window->config->toplevel->using_clipboard = SP_TRUE;
		swUnlockMainMutex(window);
	    }
	    spDebug(30, "waveProcessCallback", "process started\n");
	    if (wave != NULL && wave->play_flag == SP_FALSE) {
		swSetMouseCursor(window, SP_CURSOR_WAIT);
	    }
	    swSetProcessSenseLevel(window, SP_TRUE);
	} else {
	    /*spDebug(80, "waveProcessCallback", "processing... %ld\n", pos);*/
	    if (edit_type != SW_EDIT_NONE || window->config->play_draw == SP_TRUE) {
		if (window->drawn_pos >= window->wave->length) {	/* if drawing has been completed */
		    point_d = swSampToDisp(window, pos);
		    point_f = swSampToDim(window, pos);
		    
		    spDebug(80, "waveProcessCallback", "point_f = %f, window->point_f = %f\n",
			    point_f, window->point_f);
		    
		    if (window->prev_point_f != window->point_f
			&& point_d >= window->point_d
			&& point_f - window->point_f < SW_MIN_DRAW_INTERVAL_F) {
			/* do nothing */
		    } else {
			if (window->sync_play == SP_TRUE
			    /*&& swIsWaveThreadSafe(window->wave) == SP_TRUE */
			    && (point_d < 0 || point_d >= window->width)) {
			    swScrollWindow(window, pos, SP_TRUE, in_thread);
			    
			    swLockWindowMutex(window);
			    window->prev_point_f = window->point_f;
			    swUnlockWindowMutex(window);
			} else {
			    swLockWindowMutex(window);
			    swUpdatePoint(window, pos, SP_TRUE);
			    window->point_d = point_d;
			    window->point_f = point_f;
			    swRefreshWindow(window, SP_TRUE);
			    swUnlockWindowMutex(window);
			}
		    }
		}
	    }
	    
	    if (in_thread == SP_FALSE) {
		spDispatchEvent(window->config->toplevel->toplevel);
	    }
	}

	if (in_thread == SP_TRUE) spThreadLeave();
	
	spDebug(80, "waveProcessCallback", "done\n");
    }
    
    return SP_TRUE;
}

static spBool wavePlayCallback(swWave wave, spBool in_thread, long pos, void *data)
{
    return waveProcessCallback(wave, in_thread, pos, SW_EDIT_NONE, data);
}

static spBool waveEditCallback(swWave wave, spBool in_thread, long pos, swEditType edit_type, void *data)
{
    return waveProcessCallback(wave, in_thread, pos, edit_type, data);
}

static spBool waveEditFinishCallback(swWave wave, swWave owave, spBool in_thread,
				     swEditType edit_type, void *data)
{
    spBool flag;
    swWindow window = (swWindow)data;
    
    spDebug(30, "waveEditFinishCallback", "in_thread = %d\n", in_thread);

    if (in_thread == SP_TRUE) spThreadEnter();
    
    flag = swEditFinish(window, owave, in_thread, edit_type);
    
    if (in_thread == SP_TRUE) spThreadLeave();

    return flag;
}

static spBool waveReadCallback(swWave wave, spBool in_thread, long pos, void *data)
{
    long pos2;
    swWindow window = (swWindow)data;

    if (window == NULL) return SP_FALSE;

    spDebug(80, "waveReadCallback", "pos = %ld\n", pos);

    if (pos == SW_PROCESS_FINISHED) {
	pos2 = wave->length;
    } else {
	pos2 = pos;
    }
    swDrawWaveImage(window->image, window, pos2);

    if (pos == SW_PROCESS_STARTED) {
	swSetMouseCursor(window, SP_CURSOR_WAIT);
	spSetSensitive(window->window, SP_FALSE);
	
	swLockWindowMutex(window);
	window->prev_point_f = window->point_f;
	swUnlockWindowMutex(window);
    } else if (pos == SW_PROCESS_FINISHED) {
	swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
	spSetSensitive(window->window, SP_TRUE);
    }
    
    if (pos2 >= 0) {
	/* refresh canvas */
	swRefreshWindow(window, SP_FALSE);
    }
    
    spDebug(80, "waveReadCallback", "done\n");
    
    return SP_TRUE;
}

void swInitialize(spTopLevel toplevel, swConfig config)
{
    swInitTopLevel(toplevel, config);

    swSetColor(config);

    config->wave_config->thread_safe = spIsThreadSafe();

    config->time_format = MIN(config->time_format, SW_TIME_FORMAT_SEPARATED_SEC);

    if (strnone(config->def_format_string)) {
	strcpy(config->def_format_string, "raw");
    }
    if (strnone(config->format_string)) {
	spStrCopy(config->format_string, SP_MAX_SETUP_VALUE, config->def_format_string);
    } else {
	config->format_specified = SP_TRUE;
    }
    if (config->samp_rate <= 0.0) {
	config->samp_rate = config->def_samp_rate;
    } else {
	config->format_specified = SP_TRUE;
    }
    if (config->head_size < 0) {
	config->head_size = config->def_head_size;
    } else {
	config->format_specified = SP_TRUE;
    }
    if (config->num_channel <= 0) {
	config->num_channel = config->def_num_channel;
    } else {
	config->format_specified = SP_TRUE;
    }
    if (config->samp_bit <= 0) {
	config->samp_bit = config->def_samp_bit;
    } else {
	config->format_specified = SP_TRUE;
    }

#ifdef SW_USE_ANALYSIS
    if (!strnone(config->analysis_type_string)) {
	config->analysis_type = swGetAnalysisTypeFromLabel(config->analysis_type_string);
	if (config->analysis_type == SW_ANALYSIS_NONE) {
	    config->analysis_type = SW_ANALYSIS_SPECTRUM;
	}
    } else {
	config->analysis_type = SW_ANALYSIS_SPECTRUM;
    }
    if (!strnone(config->window_type_string)) {
	config->window_type = swGetWindowTypeFromLabel(config->window_type_string);
	if (config->window_type == SW_WINDOW_NONE) {
	    config->window_type = SW_WINDOW_HAMMING;
	}
    } else {
	config->window_type = SW_WINDOW_HAMMING;
    }
#endif

    if (!strnone(config->play_command)) {
	spSetPlayCommand(config->play_command);
	config->wave_config->play_use_audio  =
	    (config->use_play_command == SP_TRUE ? SP_FALSE : SP_TRUE);
    }
    config->wave_config->audio_buffer_size = config->audio_buffer_size;

    swSetTempDir(config->wave_config, config->temp_dir);
    
    swSetWaveOptionCallback(config->wave_config, waveOptionCallback);
    swSetWaveErrorCallback(config->wave_config, waveErrorCallback);
    swSetWavePlayCallback(config->wave_config, wavePlayCallback);
    swSetWaveReadCallback(config->wave_config, waveReadCallback);
    swSetWaveEditCallback(config->wave_config, waveEditCallback);
    swSetWaveEditFinishCallback(config->wave_config, waveEditFinishCallback);
    
    swCreateFormatDialog(config);
    swCreatePreferenceDialog(config);
    swCreateClipboardWindow(config);

    spSetPluginLang(spGetLanguage());

    return;
}

swDataType swGetDataType(swConfig config)
{
    if (config != NULL
	&& (strcaseeq(config->format_string, "freq")
	    || strcaseeq(config->format_string, SP_WAVE_FORMAT_TEXT_FREQ_LABEL))) {
	return SW_FREQ_DATA;
    }
    
    return SW_TIME_DATA;
}

void swCreateWindow(char *filename, char *label_file, spBool is_region_label, swConfig config)
{
    char *format;
    swDataType data_type = SW_TIME_DATA;
    swWave wave = NULL;
    swWindow window = NULL;
    spPluginError err;

    if (!strnone(filename)) {
	if (spIsFile(filename) == SP_FALSE) {
	    spDisplayError(/*swGetFormatDialogWindow()*/NULL, SW_ERROR_TITLE,
			   SW_FILE_NOT_EXIST_ERROR_MESSAGE, filename);
	    return;
	}
	
	if (config->use_def_format == SP_TRUE || config->format_specified == SP_TRUE) {
	    format = config->format_string;
	} else {
	    format = NULL;
	}
	
	if ((wave = swGetWave(config->wave_config, filename, NULL, format, config->samp_bit,
			      config->num_channel, config->samp_rate,
			      SP_TRUE, SP_TRUE, &err)) == NULL) {
	    spDebug(10, "swCreateWindow", "first swGetWave: err = %d\n", err);
	    
	    if (format != NULL ||
		(err != SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND && err != SP_PLUGIN_ERROR_FAILURE)) {
		return;
	    } else {
		if (swPopupFormatDialog(config, filename) == SP_TRUE) {
		    data_type = swGetDataType(config);
		    spDebug(10, "swCreateWindow", "format_string = %s, type = %d\n",
			    config->format_string, data_type);
		    
		    if ((wave = swGetWave(config->wave_config, filename, NULL, config->format_string, config->samp_bit,
					  config->num_channel, config->samp_rate,
					  SP_TRUE, SP_FALSE, &err)) == NULL) {
			spDebug(10, "swCreateWindow", "second swGetWave: err = %d\n", err);
			return;
		    }
		} else {
		    return;
		}
	    }
	}

	if (wave != NULL && !strnone(label_file)) {
	    spDebug(10, "swCreateWindow", "is_region_labe = %d, label_file = %s\n",
		    is_region_label, label_file);
	    swReadLabel(wave, label_file, SW_TIME_FORMAT_UNKNOWN, is_region_label);
	}
    }

    window = swCreateWaveWindow(wave, config, data_type);
    
    return;
}

void swUpdateMeterWidth(swWindow window)
{
    if (window == NULL) return;

    if (window->data_type == SW_FREQ_DATA) {
	window->meter_width = 0;
    } else {
	if (window->config->display_meter == SP_TRUE) {
	    window->meter_width = SW_METER_WIDTH;
	} else {
	    window->meter_width = 0;
	}
    }

    return;
}

swWindow swInitWindow(swConfig config, swDataType data_type)
{
    swWindow window = NULL;

    window = xalloc(1, struct _swWindow);

    window->mutex = spCreateMutex(NULL);

    window->title = NULL;
    window->name = NULL;
    window->data_label = NULL;
    window->data_dim = NULL;
    window->data_type = data_type;
    window->index = 1;
    window->scroll_coef = 1;
    if (data_type == SW_FREQ_DATA) {
	window->width = (config->freq_width > 0 ? config->freq_width :
			 SW_FREQ_WINDOW_WIDTH);
	window->height = (config->freq_height > 0 ? config->freq_height :
			  SW_FREQ_WINDOW_HEIGHT);
    } else {
	window->width = (config->width > 0 ? config->width : SW_WINDOW_WIDTH);
	window->height = (config->height > 0 ? config->height : SW_WINDOW_HEIGHT);
    }
    window->overview_width = 0;
    window->overview_height = 0;
    window->draw_height = 0.0;
    window->draw_width = 0.0;
    window->draw_height = 0.0;
    window->drawn_pos = 0;
    
    window->selecting = SP_FALSE;
    window->sel_st = -1;
    window->sel_ed = -1;
    window->sel_st_d = -1;
    window->sel_ed_d = -1;
    window->point = -1;
    window->point_d = -1;
    window->point_f = -1.0;
    window->prev_point_f = -1.0;
    window->target_channel = 0;
    window->active_label_index = -1;
    
    window->offset = 0;
    window->length = 0;
    window->wave = NULL;

    window->visible_flag = SP_FALSE;
    window->analysis_flag = SP_FALSE;
    window->loop_play = config->use_loop_play;
    window->sync_play = config->use_sync_play;
    window->drag_region = SP_FALSE;
    window->draw_label = SP_TRUE;
    if (data_type == SW_FREQ_DATA) {
	window->display_info_area = config->display_freq_info_area;
    } else {
	window->display_info_area = config->display_info_area;
    }
    window->autosave_started = SP_FALSE;
    window->process_flag = SP_FALSE;
    window->num_blocked = 0;
    window->drag_label_type = SW_DRAG_NO_LABEL;

    window->pointer = 0;
    window->direction = 1;
    window->config = config;
    window->window = NULL;
    window->wave_box = NULL;
    window->overview_canvas = NULL;
    window->canvas = NULL;
    window->image = NULL;
    window->hscroll = NULL;
    window->tool_bar = NULL;
    window->undo_menu = NULL;
    window->redo_menu = NULL;
    window->info_box = NULL;
    window->label_list = NULL;
    window->related_window = NULL;

    swUpdateMeterWidth(window);

    return window;
}

void swSetWave(swWindow window, swWave wave)
{
    if (window != NULL) {
	window->sel_st = -1;
	window->sel_ed = -1;
	window->sel_st_d = -1;
	window->sel_ed_d = -1;
	window->point = -1;
	window->point_d = -1;
	window->point_f = -1.0;
	window->prev_point_f = -1.0;

	if (window->wave != wave) {
#if 1
	    if (window->wave != NULL) {
		swDestroyWave(window->wave);
	    }
#endif
	    if (window->related_window != NULL
		&& window->related_window->related_window == window) {
		window->related_window->related_window = NULL;
	    }
	    window->related_window = NULL;
	}
	
	if (window->data_label != NULL) xfree(window->data_label);
	if (window->data_dim != NULL) xfree(window->data_dim);
	window->data_label = NULL;
	window->data_dim = NULL;

	if (wave != NULL) {
	    if (window->offset <= 0) {
		window->offset = 0;
	    } else {
		window->offset = MIN(window->offset, wave->total_length - 1);
	    }
	    if (window->length <= 0) {
		window->length = wave->total_length;
	    } else {
		window->length = MIN(window->length, wave->total_length - window->offset);
	    }
	    window->wave = wave;
	    swSetWaveCallbackData(window->wave, (void *)window);
	    
	    window->config->toplevel->null_window = NULL;
	} else {
	    window->offset = 0;
	    window->length = 0;
	    window->wave = NULL;
	}
    }

    spDebug(10, "swSetWave", "done\n");
    
    return;
}

void swGetWindowIndex(swWindow window)
{
    long k;
    int index = 1;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL) return;

    next = window->window;
    for (k = 0;; k++) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && streq(next_window->name, window->name)) {
	    index = MAX(next_window->index + 1, index);
	}
    }

    window->index = index;
    
    return;
}

void swSetWindowTitle(swWindow window)
{
    char *path, *string;
    char samp_bit_string[SP_MAX_LINE];
    char name[SP_MAX_MESSAGE];
    char title[SP_MAX_MESSAGE];

    if (window == NULL)
	return;

    if (window->wave != NULL) {
	if ((path = xspGetReadablePath(window->wave->orig_filename)) != NULL) {
	    string = spGetBaseName(path);
	} else {
	    string = NULL;
	}

	if (window->name == NULL
	    || (string != NULL && !streq(string, window->name))) {
	    if (window->name != NULL) {
		xfree(window->name);
	    }
	    if (swIsClipboardWindow(window) == SP_TRUE) {
		window->name = strclone(SW_CLIPBOARD_TITLE);
	    } else if (string == NULL) {
		window->name = strclone(SW_UNTITLED_TITLE);
	    } else {
		window->name = strclone(string);
	    }	
	    spDebug(10, "swSetWindowTitle", "window name: %s\n", window->name);
	    
	    swGetWindowIndex(window);
	}

	if (path != NULL) xfree(path);
	
	if (window->index >= 2) {
	    sprintf(name, "%s<%d>", window->name, window->index);
	} else {
	    strcpy(name, window->name);
	}
	if (swIsWaveEdited(window->wave) == SP_TRUE) {
	    strcat(name, " *");
	}

	swGetSampleBitString(samp_bit_string, swGetWaveSampleBit(window->wave));

	sprintf(title,
		"%s  (Sampling Frequency: %.1f Hz  Data Format: %s  %s Bits/Sample)", 
		name, swGetWaveSampleRate(window->wave),
		swGetWaveFileType(window->wave), samp_bit_string);

	spDebug(10, NULL, "title = %s, name = %s\n", title, name);

	/* change title */
	spSetParams(window->window, SppTitle, title, SppIconName, name, NULL);
    } else {
	if (window->name == NULL) {
	    if (swIsClipboardWindow(window) == SP_TRUE) {
		window->name = strclone(SW_CLIPBOARD_TITLE);
	    } else {
		window->name = strclone(SW_NO_WAVE_TITLE);
	    }
	}
	
	/* set window title */
	strcpy(title, window->name);

	/* change title */
	spSetParams(window->window, SppTitle, title, SppIconName, title, NULL);
    }

    spDebug(10, "swSetWindowTitle", "done\n");
    
    return;
}

spBool swCloseWindow(swWindow window, spBool quit_prompt)
{
    spComponent component;
    swConfig config;
    char *orig_filename;
    
    spDebug(10, "swCloseWindow", "in\n");
    
    if (window != NULL && window->window != NULL
	&& swIsVisible(window) == SP_TRUE) {
	config = window->config;

#ifdef SW_CLOSE_PROMPT
	if (window->wave != NULL
	    && swIsClipboardWindow(window) == SP_FALSE
	    && swIsWaveEdited(window->wave) == SP_TRUE
	    && !strnone(orig_filename = swGetWaveOriginalFileName(window->wave))) {
	    spDialogResponse response = SP_DR_YES;
	    char buf[SP_MAX_MESSAGE];

	    if (window->name != NULL) {
		orig_filename = window->name;
	    } else {
		orig_filename = spGetBaseName(orig_filename);
	    }
	    spMapWindow(window->window);
	    
	    sprintf(buf, SW_SAVE_CHANGES_QUESTION_MESSAGE, orig_filename);

	    response = spCreateMessageBox(window->window, NULL, buf,
					  SppTitle, SW_SAVE_CHANGES_QUESTION_TITLE,
					  SppDialogType, SP_QUESTION_DIALOG,
					  SppMessageBoxButtonType, SP_MB_YES_NO_CANCEL,
					  NULL);
	    
	    if (response == SP_DR_CANCEL) {
		return SP_FALSE;
	    } else if (response == SP_DR_YES) {
		if (swWriteWave(NULL, NULL, NULL, window->wave, SP_TRUE, SP_FALSE) == SP_FALSE) {
		    if (spCreateMessageBox(window->window, NULL,
					   SW_SAVE_ERROR_CLOSE_QUESTION_MESSAGE,
					   SppDialogType, SP_WARNING_DIALOG,
					   SppMessageBoxButtonType, SP_MB_YES_NO,
					   NULL) != SP_DR_YES) {
			return SP_FALSE;
		    }
		}
	    }
	}
	
	spDebug(10, "swCloseWindow", "num_window = %ld\n",
		window->config->toplevel->num_window);
	if (quit_prompt == SP_TRUE && window->config->toplevel->num_window == 1) {
	    if (spQuitPrompt(window->window) == SP_FALSE) {
		return SP_FALSE;
	    }
	}
#endif
	
	if (window == config->toplevel->current_window) {
	    config->toplevel->current_window = NULL;
	}
	
	if (window->related_window != NULL
	    && window->related_window->related_window == window) {
	    window->related_window->related_window = NULL;
	}

	component = window->window;
	window->visible_flag = SP_FALSE;
	
	swLockMainMutex(window);
	--config->toplevel->num_window;
	swUnlockMainMutex(window);
	
	if (swIsClipboardWindow(window) == SP_TRUE) {
	    spDebug(10, "swCloseWindow", "num_window = %ld\n",
		    window->config->toplevel->num_window);
	    spPopdownWindow(window->window);
	} else {
	    window->window = NULL;
	    spCloseWindow(component);
	}

	if (config->toplevel->num_window == 0) {
	    spQuit(0);
	}
	
	swSetSenseLevel(NULL);
    }

    spDebug(10, "swCloseWindow", "done\n");
    
    return SP_TRUE;
}

void swDestroyWindow(swWindow window)
{
    spDebug(10, "swDestroyWindow", "in\n");
    
    if (window != NULL) {
	if (window == window->config->toplevel->clipboard_window) {
	    window->config->toplevel->clipboard_window = NULL;
	}
	if (window->title != NULL) xfree(window->title);
	if (window->name != NULL) xfree(window->name);
	if (window->data_label != NULL) xfree(window->data_label);
	if (window->data_dim != NULL) xfree(window->data_dim);
	if (window->wave != NULL) swDestroyWave(window->wave);
	spDebug(10, "swDestroyWindow", "destroy wave done\n");

	if (window->mutex != NULL) spDestroyMutex(window->mutex);
    
	xfree(window);
    }

    spDebug(10, "swDestroyWindow", "done\n");
    
    return;
}

static void setSelectDataSenseLevel(swWindow window, long group_id, spBool select)
{
    if (select == SP_TRUE) {
	if (window->data_type == SW_TIME_DATA
	    && window->analysis_flag == SP_FALSE) {
	    if (window->wave->selected_channel < 0) {
		if (window->wave->num_channel > 1) {
		    spSetWindowSenseLevel(window->window, group_id,
					  SW_STATE_SELECT_TIME_MULTI_CHANNEL);
		} else {
		    spSetWindowSenseLevel(window->window, group_id,
					  SW_STATE_SELECT_TIME_CHANNELS);
		}
	    } else {
		spSetWindowSenseLevel(window->window, group_id, SW_STATE_SELECT_TIME_DATA);
	    }
	} else {
	    spSetWindowSenseLevel(window->window, group_id, SW_STATE_FREQ_DATA);
	}
    } else {
	if (window->data_type == SW_TIME_DATA
	    && window->analysis_flag == SP_FALSE) {
	    spSetWindowSenseLevel(window->window, group_id, SW_STATE_TIME_DATA);
	} else {
	    spSetWindowSenseLevel(window->window, group_id, SW_STATE_FREQ_DATA);
	}
    }
    
    return;
}

void swSetSelectSenseLevel(swWindow window, spBool select)
{
    if (window == NULL || window->wave == NULL) return;
    
    if (swIsWaveProcessing(window->wave) == SP_TRUE || window->num_blocked > 0) {
	if (window->sync_play == SP_TRUE && swIsWavePlaying(window->wave) == SP_TRUE) {
	    spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_PLAY_WAVE);
	    spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_PLAY_WAVE);
	} else if (window->num_blocked > 0) {
	    if (select == SP_TRUE) {
		spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_SELECT_WAVE);
	    } else {
		spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_EXIST_WAVE);
	    }
	    spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_PLAY_WAVE);
	} else {
	    if (select == SP_TRUE) {
		spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_SELECT_WAVE);
		spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_SELECT_WAVE);
	    } else {
		spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_EXIST_WAVE);
		spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_EXIST_WAVE);
	    }
	}
	spSetWindowSenseLevel(window->window, SW_DATA_GROUP_ID, SW_STATE_PLAY_TIME_DATA);
	spSetWindowSenseLevel(window->window, SW_CLIPBOARD_DATA_GROUP_ID, SW_STATE_PLAY_TIME_DATA);
	spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_PLAY_WAVE);
	
	spSetWindowSenseLevel(window->window, SW_CLIPBOARD_GROUP_ID, SW_STATE_NO_CLIPBOARD);
    } else {
	if (select == SP_TRUE) {
	    spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_SELECT_WAVE);
	    spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_SELECT_WAVE);
	} else {
	    spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_EXIST_WAVE);
	    spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_EXIST_WAVE);
	}

	if (swIsClipboardNone(window) == SP_TRUE
	    || swIsWaveProcessing(window->wave) == SP_TRUE
	    || window->num_blocked > 0) {
	    spSetWindowSenseLevel(window->window, SW_CLIPBOARD_GROUP_ID, SW_STATE_NO_CLIPBOARD);
	} else {
	    if (select == SP_TRUE) {
		spSetWindowSenseLevel(window->window, SW_CLIPBOARD_GROUP_ID, SW_STATE_EXIST_CLIPBOARD_SELECTED);
	    } else {
		spSetWindowSenseLevel(window->window, SW_CLIPBOARD_GROUP_ID, SW_STATE_EXIST_CLIPBOARD);
	    }
	}

	setSelectDataSenseLevel(window, SW_DATA_GROUP_ID, select);
	if (window->config->toplevel->using_clipboard == SP_TRUE) {
	    setSelectDataSenseLevel(window, SW_CLIPBOARD_DATA_GROUP_ID, SP_FALSE);
	} else {
	    setSelectDataSenseLevel(window, SW_CLIPBOARD_DATA_GROUP_ID, select);
	}
    }
    
    return;
}

void swSetProcessSenseLevel(swWindow window, spBool process)
{
    long num_blocked;
    spBool need_block;
    swWindow next_window;
    spCloseStyle close_style = SW_DEFAULT_CLOSE_STYLE;
    
    if (window == NULL) return;

    spDebug(80, "swSetProcessSenseLevel", "process = %d\n", process);

    if (process == SP_TRUE) {
	swLockMainMutex(window);
	if (window->config->toplevel->playable == SP_TRUE) {
	    if (swIsWavePlaying(window->wave) == SP_TRUE) {
		window->config->toplevel->playable = SP_FALSE;
	    }
	}
	window->config->toplevel->num_process++;
	swSetWindowsSenseLevel();
	swUnlockMainMutex(window);
		
	next_window = window;
	while (next_window != NULL) {
	    swLockMutex(next_window->wave);

	    need_block = swNeedWindowProcessBlock(window, next_window);
	    
	    if (need_block == SP_TRUE) {
		next_window->num_blocked++;
	    }
	    swUnlockMutex(next_window->wave);
		
	    spDebug(80, "swSetProcessSenseLevel",
		    "process: need_block = %d, num_process = %d, num_blocked = %d\n",
		    need_block, window->config->toplevel->num_process,
		    next_window->num_blocked);
	    
	    if (window->config->toplevel->num_process == 1 || need_block == SP_TRUE) {
		swSetWindowSenseLevel(next_window, SP_FALSE);

		if (next_window == window || next_window->num_blocked == 1) {
		    spSetParams(next_window->window,
				SppCloseStyle, SP_NO_CLOSE,
				NULL);
		}
	    }
		
	    next_window = swGetNextWindow(next_window);
	    if (next_window == window) {
		break;
	    }
	}
    } else {
	swLockMainMutex(window);
	if (window->config->toplevel->playable == SP_FALSE
	    && swIsWavePlaying(window->wave) == SP_TRUE) {
	    window->config->toplevel->playable = SP_TRUE;
	}
	window->config->toplevel->num_process =
	    MAX(window->config->toplevel->num_process - 1, 0);
	swSetWindowsSenseLevel();
	swUnlockMainMutex(window);

	/* because need_block changes after editing */
	num_blocked = window->num_blocked;
	
	next_window = window;
	while (next_window != NULL) {
	    if (num_blocked > 0) {
		need_block = SP_TRUE;
	    } else {
		need_block = swNeedWindowProcessBlock(window, next_window);
	    }
	    
	    swLockMutex(next_window->wave);
	    if (need_block == SP_TRUE) {
		next_window->num_blocked = MAX(next_window->num_blocked - 1, 0);
	    }
	    swUnlockMutex(next_window->wave);
	    
	    spDebug(80, "swSetProcessSenseLevel",
		    "no process: need_block = %d, num_process = %d, num_blocked = %d\n",
		    need_block, window->config->toplevel->num_process,
		    next_window->num_blocked);
	    
	    if ((need_block == SP_TRUE && next_window->num_blocked == 0)
		|| window->config->toplevel->num_process == 0) {
		swSetWindowSenseLevel(next_window, SP_FALSE);
		
		if (next_window == window || next_window->num_blocked == 0) {
		    spSetParams(next_window->window,
				SppCloseStyle, close_style,
				NULL);
		}
	    }
	    
	    next_window = swGetNextWindow(next_window);
	    if (next_window == window) {
		break;
	    }
	}
    }

    return;
}

void swSetWindowsSenseLevel(void)
{
    if (swGetNumWindow() >= 2) {
	spDebug(10, "swSetWindowsSenseLevel", "SW_STATE_SOME_WINDOWS\n");
	if (swGetNumProcess() > 0) {
	    spSetCurrentSenseLevel(SW_WINDOW_GROUP_ID, SW_STATE_SOME_WINDOWS);
	} else {
	    spSetCurrentSenseLevel(SW_WINDOW_GROUP_ID, SW_STATE_STOP_SOME_WINDOWS);
	}
    } else {
	spSetCurrentSenseLevel(SW_WINDOW_GROUP_ID, SW_STATE_ONE_WINDOW);
    }
    
    if (swGetNumProcess() > 0) {
	spSetCurrentSenseLevel(SW_QUIT_GROUP_ID, SW_STATE_PLAY_WAVE);
    } else {
	spSetCurrentSenseLevel(SW_QUIT_GROUP_ID, SW_STATE_ONE_WINDOW);
    }
    
    return;
}

void swSetUndoSenseLevel(swWindow window)
{
    if (window == NULL) return;
    
    if (window->wave == NULL || swIsWaveProcessing(window->wave) == SP_TRUE) {
	spSetSensitive(window->undo_menu, SP_FALSE);
	spSetSensitive(window->redo_menu, SP_FALSE);
    } else {
	if (swCanRedoWave(window->wave) == SP_TRUE) {
	    spSetSensitive(window->redo_menu, SP_TRUE);
	} else {
	    spSetSensitive(window->redo_menu, SP_FALSE);
	}
	if (swCanUndoWave(window->wave) == SP_TRUE) {
	    spSetSensitive(window->undo_menu, SP_TRUE);
	} else {
	    spSetSensitive(window->undo_menu, SP_FALSE);
	}
    }
    
    return;
}

void swSetWindowSenseLevel(swWindow window, spBool windows_flag)
{
    spBool process_flag = SP_FALSE;
    
    if (windows_flag == SP_TRUE) {
	swSetWindowsSenseLevel();
	if (swIsClipboardWindow(window) == SP_TRUE) {
	    swWindow first_window, next_window;

	    first_window = swGetNextWindow(window);

	    if (first_window != NULL
		&& swIsClipboardWindow(first_window) == SP_FALSE) {
		next_window = first_window;
		while (next_window != NULL) {
		    swSetWindowSenseLevel(next_window, SP_FALSE);
		
		    next_window = swGetNextWindow(next_window);
		    if (next_window == first_window) {
			break;
		    }
		}
	    }
	}
    }

    if (window != NULL) {
	if (window->wave == NULL) {
	    if (/*window->config->toplevel->num_process > 0 ||*/
		window->num_blocked > 0) {
		spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_PLAY_WAVE);
		spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_PLAY_WAVE);
		spSetWindowSenseLevel(window->window, SW_OPEN_GROUP_ID, SW_STATE_PLAY_WAVE);
		spSetWindowSenseLevel(window->window, SW_SAVE_GROUP_ID, SW_STATE_PLAY_WAVE);
	    } else {
		spSetWindowSenseLevel(window->window, SP_DEFAULT_GROUP_ID, SW_STATE_NO_WAVE);
		spSetWindowSenseLevel(window->window, SW_PAGE_GROUP_ID, SW_STATE_NO_WAVE);
		spSetWindowSenseLevel(window->window, SW_OPEN_GROUP_ID, SW_STATE_NO_WAVE);
		spSetWindowSenseLevel(window->window, SW_SAVE_GROUP_ID, SW_STATE_NO_WAVE);
	    }
	    spSetWindowSenseLevel(window->window, SW_DATA_GROUP_ID, SW_STATE_NO_DATA);
	    spSetWindowSenseLevel(window->window, SW_CLIPBOARD_DATA_GROUP_ID, SW_STATE_NO_DATA);
	    spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_NO_WAVE);
	    spSetWindowSenseLevel(window->window, SW_PLAY_GROUP_ID, SW_STATE_NO_WAVE);

	    spSetWindowSenseLevel(window->window, SW_LABEL_GROUP_ID, SW_STATE_NO_LABEL);
	    spSetWindowSenseLevel(window->window, SW_NORMAL_LABEL_GROUP_ID, SW_STATE_NO_LABEL);
	    spSetWindowSenseLevel(window->window, SW_REGION_LABEL_GROUP_ID, SW_STATE_NO_LABEL);
	    
	    spSetWindowSenseLevel(window->window, SW_CLIPBOARD_GROUP_ID, SW_STATE_NO_CLIPBOARD);
	} else {
	    if (/*window->config->toplevel->num_process > 0 ||*/
		swIsWaveProcessing(window->wave) == SP_TRUE
		|| window->num_blocked > 0) {
		process_flag = SP_TRUE;
		spSetWindowSenseLevel(window->window, SW_OPEN_GROUP_ID, SW_STATE_PLAY_WAVE);
		spSetWindowSenseLevel(window->window, SW_SAVE_GROUP_ID, SW_STATE_PLAY_WAVE);
	    } else {
		spSetWindowSenseLevel(window->window, SW_OPEN_GROUP_ID, SW_STATE_EXIST_WAVE);
		if (swIsWaveEdited(window->wave) == SP_TRUE) {
		    spSetWindowSenseLevel(window->window, SW_SAVE_GROUP_ID, SW_STATE_EDIT_WAVE);
		} else {
		    spSetWindowSenseLevel(window->window, SW_SAVE_GROUP_ID, SW_STATE_EXIST_WAVE);
		}
	    }
	    if (window->data_type == SW_TIME_DATA
		&& window->analysis_flag == SP_FALSE) {
		if (process_flag == SP_TRUE) {
		    spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_PLAY_WAVE);
		} else {
		    if (window->wave->num_channel >= 3) {
			spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_MULTI_CHANNEL_WAVE);
		    } else if (window->wave->num_channel >= 2) {
			spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_STEREO_WAVE);
		    } else {
			spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_MONO_WAVE);
		    }
		}
		spDebug(80, "swSetWindowSenseLevel", "playable = %d, processing = %d\n",
			window->config->toplevel->playable, swIsWaveProcessing(window->wave));
		if (window->config->toplevel->playable == SP_FALSE
		    || swIsWaveProcessing(window->wave) == SP_TRUE
		    || window->num_blocked > 0) {
		    spSetWindowSenseLevel(window->window, SW_PLAY_GROUP_ID, SW_STATE_PLAY_WAVE);
		} else {
		    spSetWindowSenseLevel(window->window, SW_PLAY_GROUP_ID, SW_STATE_NOT_PLAY_WAVE);
		}
	    } else {
		spSetWindowSenseLevel(window->window, SW_CHANNEL_GROUP_ID, SW_STATE_NO_WAVE);
		spSetWindowSenseLevel(window->window, SW_PLAY_GROUP_ID, SW_STATE_NO_WAVE);
	    }
	    
	    /* set sensitive for labels */
	    if (swGetNumLabel(window->wave) <= 0) {
		spSetWindowSenseLevel(window->window, SW_LABEL_GROUP_ID, SW_STATE_NO_LABEL);
	    } else {
		spSetWindowSenseLevel(window->window, SW_LABEL_GROUP_ID, SW_STATE_EXIST_LABEL);
	    }
	    if (swGetNumNormalLabel(window->wave) <= 0) {
		spSetWindowSenseLevel(window->window, SW_NORMAL_LABEL_GROUP_ID, SW_STATE_NO_LABEL);
	    } else {
		spSetWindowSenseLevel(window->window, SW_NORMAL_LABEL_GROUP_ID, SW_STATE_EXIST_LABEL);
	    }
	    if (swGetNumRegionLabel(window->wave) <= 0) {
		spSetWindowSenseLevel(window->window, SW_REGION_LABEL_GROUP_ID, SW_STATE_NO_LABEL);
	    } else {
		spSetWindowSenseLevel(window->window, SW_REGION_LABEL_GROUP_ID, SW_STATE_EXIST_LABEL);
	    }

	    spDebug(80, "swSetWindowSenseLevel", "num_window = %ld, num_blocked = %d\n",
		    window->config->toplevel->num_window, window->num_blocked);
	    
	    swSetSelectSenseLevel(window, ((window->sel_st >= 0 && window->sel_ed >= 0)
					   ? SP_TRUE : SP_FALSE));
	}
	swSetUndoSenseLevel(window);
    }
    
    spDebug(10, "swSetWindowSenseLevel", "done\n");
    
    return;
}

void swSetSenseLevel(swWindow window)
{
    swSetWindowSenseLevel(window, SP_TRUE);
    return;
}

swWindow swCreateWaveWindowAt(swWave wave, swConfig config, swDataType data_type,
			      int x, int y)
{
    swWindow window = NULL;
    
    if (wave != NULL || config->toplevel->num_window == 0) {
	if (wave != NULL && config->toplevel->null_window != NULL) {
	    window = config->toplevel->null_window;
	    window->data_type = data_type;
	    window->wave = wave;
	    swResetWindow(window, SP_FALSE);
	} else {
	    window = swInitWindow(config, data_type);
	    swSetWave(window, wave);
	    swCreateMainWindow(window);
	    swSetWindowTitle(window);
	    
	    if (swIsClipboardWindow(window) == SP_FALSE) {
		swLockMainMutex(window);
		++config->toplevel->num_window;
		swUnlockMainMutex(window);
		
		spDebug(10, "swCreateWaveWindowAt", "num_window = %ld\n",
			config->toplevel->num_window);
	
		window->visible_flag = SP_TRUE;
		swPopupWindow(window, x, y);
	    
		if (wave == NULL) {
		    config->toplevel->null_window = window;
		}
	    }
	}
	
    }
    
    return window;
}

swWindow swCreateWaveWindow(swWave wave, swConfig config, swDataType data_type)
{
    return swCreateWaveWindowAt(wave, config, data_type, -1, -1);
}

void swResetWindow(swWindow window, spBool in_thread)
{
    if (window != NULL) {
	swSetWave(window, window->wave);
	swSetWindowTitle(window);
	swSetSenseLevel(window);
	swReloadWave(window, in_thread);
	swDrawOverview(window);
	temp_output = NULL;
    }
    return;
}

void swRedrawWindow(swWindow window)
{
    if (window != NULL) {
	swSetWave(window, window->wave);
	swSetWindowTitle(window);
	swSetSenseLevel(window);
	swRedrawWave(window);
    }
    return;
}

void swOpenFile(spComponent component, swWindow window, spBool new_flag)
{
    char *filename = NULL;
    swWave wave = NULL;
    swDataType data_type = SW_TIME_DATA;
    spPluginError err;
    static char *file_filters[] =
    {
	"*",
	"*",
#ifdef SW_USE_ANALYSIS
	"*",
#endif
	NULL,
    };
    static char *file_types[] =
    {
	"Auto Detect",
	"Default",
#ifdef SW_USE_ANALYSIS
	"Frequency Data",
#endif
	NULL,
    };
    static int open_file_type_index = -1;

    if (window != NULL) {
	if (window->config->use_def_format == SP_TRUE) {
	    open_file_type_index = 1;
	} else {
	    open_file_type_index = 0;
	}
	
	/* get open file name */
	filename = xspGetOpenFileName(component, SW_OPEN_DIALOG_TITLE,
				      SppFileMustExist, SP_TRUE,
				      SppFileFilters, file_filters,
				      SppFileTypes, file_types,
				      SppFileFilterIndex, &open_file_type_index,
				      NULL);

	if (filename != NULL) {
	    err = SP_PLUGIN_ERROR_SUCCESS;
	    
	    if ((wave = swGetWave(window->config->wave_config, filename, NULL, NULL,
				  window->config->samp_bit,
				  window->config->num_channel,
				  window->config->samp_rate,
				  SP_TRUE, SP_TRUE, &err)) == NULL) {
		spDebug(10, "swOpenFile", "err = %d\n", err);
		
		if (err == SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND
		    || err == SP_PLUGIN_ERROR_FAILURE) {
		    if (open_file_type_index == 0
			|| open_file_type_index == 2) {
			if (swPopupFormatDialog(window->config, filename) == SP_FALSE) {
			    xfree(filename);
			    return;
			}
			data_type = swGetDataType(window->config);
			spDebug(10, "swOpenFile", "data_type = %d, type_index = %d, format_string = %s\n",
				data_type, open_file_type_index, window->config->format_string);
			
#ifdef SW_USE_ANALYSIS
			if (open_file_type_index == 2) {
			    data_type = SW_FREQ_DATA;
			}
#endif
		    } else {
			if (!strnone(window->config->def_format_string)) {
			    spStrCopy(window->config->format_string, SP_MAX_SETUP_VALUE,
				      window->config->def_format_string);
			}
			window->config->samp_bit = window->config->def_samp_bit;
			window->config->num_channel = window->config->def_num_channel;
			window->config->head_size = window->config->def_head_size;
			window->config->samp_rate = window->config->def_samp_rate;
		    }
		}
	    }
	    
	    spDebug(10, "swOpenFile", "filename = %s\n", filename);

	    if (wave == NULL &&
		(err == SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND || err == SP_PLUGIN_ERROR_FAILURE)) {
		/* open wave file */
		wave = swGetWave(window->config->wave_config, filename, NULL, window->config->format_string,
				 window->config->samp_bit, window->config->num_channel,
				 window->config->samp_rate, SP_TRUE, SP_FALSE, &err);
	    }
	    
	    if (wave != NULL) {
		spDebug(10, "swOpenFile", "read wave done: %ld\n", wave->total_length);
		swSetMouseCursor(window, SP_CURSOR_WAIT);

		if (new_flag == SP_FALSE && window->wave != NULL) {
		    swDestroyWave(window->wave);
		    window->wave = NULL;
		    window->offset = 0;
		    window->length = 0;
		    
		    swUpdateLabelList(window->label_list);
		}

		if (window->wave == NULL &&
		    (swIsClipboardWindow(window) == SP_FALSE || new_flag == SP_FALSE)) {
		    /* set wave to current window */
		    window->data_type = data_type;
		    window->analysis_flag = SP_FALSE;
		    window->wave = wave;
		    swResetWindow(window, SP_FALSE);
		    window->config->toplevel->null_window = NULL;
		} else {
		    /* create new window */
		    swCreateWaveWindow(wave, window->config, data_type);
		}
		swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
	    }
	    
	    spDebug(10, "swOpenFile", "filename = %s\n", filename);
	    
	    /* memory free */
	    xfree(filename);
	}
    }
    
    return;
}

void swOpenNewWindow(spComponent component, swWindow window)
{
    swOpenFile(component, window, SP_TRUE);
    return;
}


//
// Start of Record Module
//
/**
 * @author kideki
 * @author Chien-Wen Danny Su
 */
long readData(short *data, long length)
{
    long k;
    long nread;

    if (num_channel > 1) {
		nread = length * num_channel;
    }
	else
	{
		nread = length;
    }
    
	/* read audio data from microphone */
	nread = spReadAudio(source, data, nread);
    
    if (num_channel > 1)
	{
		nread /= num_channel;

		for (k = 0; k < nread; k++) {
			data[k] = data[k * num_channel + default_channel];
		}
    }
    
    return nread;
}


/**
 * @author kideki
 * @author Chien-Wen Danny Su
 */
spThreadReturn inputThread(void *data)
{
    spBool read_flag;
    long nread = 0;
	long nwrite = 0;
	long temp_offset;
    long speech_len = BUFFER_LENGTH;
    short *speech = xalloc(BUFFER_LENGTH * num_channel, short);
	swWindow window = (swWindow)data;
	
	
	offset = 0;

	while (started == SP_TRUE) {
		if (offset > 0) temp_offset = offset;
		else			temp_offset = 0;

		read_flag = SP_FALSE;

		if (temp_offset < speech_len)
		{
			nread = readData(speech + temp_offset, speech_len - temp_offset);
	    
			if (nread < 0)
			{
				spThreadEnter();
				spDisplayError(window->window, NULL, "No data was retrieved from microphone.");
				spSetToggleState(record_button, SP_FALSE);
				spThreadLeave();
				break;
			}

			if (output_plugin != NULL) {
				nwrite = spWritePlugin(output_plugin, speech, offset + nread);
			}
			else
			{
				spThreadEnter();
				spDisplayError(window->window, NULL, "Output plugin is not initialized.");
				spSetToggleState(record_button, SP_FALSE);
				spThreadLeave();
				break;
			}

			read_flag = SP_TRUE;
		}
		else
		{
			spThreadEnter();
			spDisplayError(window->window, NULL, "Nothing was read.");
			spSetToggleState(record_button, SP_FALSE);
			spThreadLeave();
			nread = 0;
		}
	
		spYieldThread();
    }

    xfree(speech);

    if (output_plugin != NULL) {
		spCloseFilePlugin(output_plugin);
		output_plugin = NULL;
	}
    spCloseAudioDevice(source);
    spFreeAudio(source);
    source = NULL;

    started = SP_FALSE;
    
    return SP_THREAD_RETURN_SUCCESS;
}


spThreadReturn removeThread(void *data)
{
	// remove temporary file used for recording if it still exists
	while (spIsExist((char *)data))
	{
		spRemoveFile((char *)data);
		spYieldThread();
	}

	return SP_THREAD_RETURN_SUCCESS;
}

void swRemoveTemporaryFile(char *filename)
{
	spCreateThread(0, SP_THREAD_PRIORITY_NORMAL, removeThread, filename);
}


/**
 * This function...
 * @return true or false depending on whether function ran successfully
 * @author Chien-Wen Danny Su
 */
spBool swStartRecordThread(spComponent component, swWindow window)
{
	spBool flag;
	spWaveInfo wave_info;

	// check whether input device is already open
	if (source != NULL && started == SP_TRUE)
	{
		return SP_TRUE;
	}

	started = SP_FALSE;

	flag = SP_TRUE;

	if (source == NULL)
	{
		if ((source = spInitAudio()) == NULL) {
			flag = SP_FALSE;
		} else {
			spSetAudioSampleRate(source, 16000.0);
			spSetAudioChannel(source, 1);
			spSetAudioSampleBit(source, 16);
			spSetAudioBlockingMode(source, SP_AUDIO_NONBLOCKING);
			spSetAudioBufferSize(source, AUDIO_BUFFER_SIZE);
		}
	}

	if (flag == SP_TRUE)
	{			
		if (spOpenAudioDevice(source, "r") == SP_FALSE)
		{
			spDisplayError(window->window, NULL, "Failed to open audio device.");
			spSetToggleState(record_button, SP_FALSE);
			flag = SP_FALSE;
		}
		else
		{
			spGetAudioChannel(source, &num_channel);
		}
	}

	if (flag == SP_TRUE)
	{
		// NOTE: I need make sure the file doesn't exist already and delete file after user has saved as...
		spInitWaveInfo(&wave_info);
		wave_info.samp_bit = window->config->samp_bit;
		wave_info.num_channel = num_channel;
		wave_info.samp_rate = window->config->samp_rate;

		if ((output_plugin = spOpenFilePluginArgAuto("output_raw", temp_output, "w",
							   SP_PLUGIN_DEVICE_FILE,
							   &wave_info, NULL, 0, NULL, NULL)) == NULL)
		{
			spDisplayError(window->window, NULL, "Error opening output file %s.", temp_output);
			spSetToggleState(record_button, SP_FALSE);
			flag = SP_FALSE;
		}
		else if (spSetPluginFileType(output_plugin, filetype) == SP_FALSE)
		{
			spDisplayWarning(window->window, NULL, "Failed to set recording format.");
		}

		num_channel = MAX(1, num_channel);
		default_channel = MIN(num_channel - 1, default_channel);

		// create thread
		if ((input_thread_handle = spCreateThread(0, SP_THREAD_PRIORITY_NORMAL,
									inputThread, window)) == NULL)
		{
			spDisplayWarning(window->window, NULL, "Unable to create thread... closing audio device.");
			spSetToggleState(record_button, SP_FALSE);

			// unable to create thread
			spCloseAudioDevice(source);
			flag = SP_FALSE;
		}
		else
		{
			started = SP_TRUE;
		}
	}

    if (flag == SP_FALSE)
	{
	    if (output_plugin != NULL) {
			spCloseFilePlugin(output_plugin);
			output_plugin = NULL;
		}

		if (source != NULL) {
			spFreeAudio(source);
			source = NULL;
		}

		started = SP_FALSE;
		spSetToggleState(record_button, SP_FALSE);

		return SP_FALSE;
    }
	else
	{
		return SP_TRUE;
    }
} /* swStartRecordThread */

spBool swStopRecordThread(spComponent component, swWindow window)
{    
    started = SP_FALSE;

    if (input_thread_handle != NULL)
	{
		// wait thread
		spThreadLeave();
		spWaitThread(input_thread_handle);
		spThreadEnter();
	
		spDestroyThread(input_thread_handle);
		input_thread_handle = NULL;

		// Now open up the temporary file so user can do a "save as" or something...
		if (window != NULL)
		{
			err = SP_PLUGIN_ERROR_SUCCESS;

			if ((wave = swGetWave(window->config->wave_config, temp_output, NULL, NULL,
						  window->config->samp_bit,
						  num_channel,
						  window->config->samp_rate,
						SP_TRUE, SP_TRUE, &err)) == NULL)
			{
				if (err == SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND
					|| err == SP_PLUGIN_ERROR_FAILURE)
				{

					if (!strnone(window->config->def_format_string)) {
						spStrCopy(window->config->format_string, SP_MAX_SETUP_VALUE,
							window->config->def_format_string);
					}

					window->config->samp_bit = window->config->def_samp_bit;
					window->config->num_channel = window->config->def_num_channel;
					window->config->head_size = window->config->def_head_size;
					window->config->samp_rate = window->config->def_samp_rate;
				}
			}

			if (wave == NULL &&
				(err == SP_PLUGIN_ERROR_SUITABLE_NOT_FOUND || err == SP_PLUGIN_ERROR_FAILURE))
			{
				// open wave file
				wave = swGetWave(window->config->wave_config, temp_output, NULL, window->config->format_string,
									window->config->samp_bit, num_channel,
									window->config->samp_rate, SP_TRUE, SP_FALSE, &err);
			}
			
			if (wave != NULL)
			{
				swSetMouseCursor(window, SP_CURSOR_WAIT);

				if (window->wave != NULL) {
					swDestroyWave(window->wave);
					window->wave = NULL;
					window->offset = 0;
					window->length = 0;
					
					swUpdateLabelList(window->label_list);
				}

				// set wave to current window
				window->data_type = SW_TIME_DATA;
				window->analysis_flag = SP_FALSE;
				window->wave = wave;
				swResetWindow(window, SP_FALSE);
				window->config->toplevel->null_window = NULL;

				swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
			}
		}
    }
    
	return SP_TRUE;
} /* swStopRecordThread */

void swRecordWindow(spComponent component, swWindow window)
{
    static int num_plugin = 0;
    static char **plugin_names = NULL;
    static char **file_types = NULL;
    static char **file_filters = NULL;
	int i;
	int index;

	if (started == SP_TRUE)
	{
		swStopRecordThread(component, window);
		spSetToggleState(record_button, SP_FALSE);
	}
	else
	{
		if (num_plugin <= 0)
		{
			plugin_names = xalloc(SW_MAX_PLUGIN + 1, char *);
			file_types = xalloc(SW_MAX_PLUGIN + 1, char *);
			file_filters = xalloc(SW_MAX_PLUGIN + 1, char *);
	
			for (i = 0; i < SW_MAX_PLUGIN; i++)
			{
				plugin_names[i] = xalloc(SP_MAX_LINE, char);
				file_types[i] = xalloc(SP_MAX_LINE, char);
				file_filters[i] = xalloc(SP_MAX_LINE, char);
	    
				if (i == 0)
				{
					strcpy(plugin_names[i], "");
					strcpy(file_types[i], "By Extension");
					strcpy(file_filters[i], "*");
				}
				else
				{
					if (spSearchPluginFileType(i - 1, SP_PLUGIN_OUTPUT,
						SP_PLUGIN_DEVICE_FILE, plugin_names[i],
						NULL, file_types[i], file_filters[i]) == SP_FALSE)
					{
						xfree(plugin_names[i]);
						xfree(file_types[i]);
						xfree(file_filters[i]);
						break;
					}
				}
			}

			num_plugin = i;
			plugin_names[i] = NULL;
			file_types[i] = NULL;
			file_filters[i] = NULL;
		}

		temp_output = NULL;
		// get the output file name
		temp_output = xspGetSaveFileName(component, SW_SAVE_DIALOG_TITLE,
						SppInitialFileName, "",
						SppPathMustExist, SP_TRUE,
						SppFileMustExist, SP_FALSE,
						SppOverwritePrompt, SP_TRUE,
						SppFileFilters, file_filters,
						SppFileTypes, file_types,
						SppFileFilterIndex, &index,
						NULL);

		filetype = file_types[index];

		if (temp_output != NULL)
		{
			swStartRecordThread(component, window);
			spSetToggleState(record_button, SP_TRUE);
		}
		else
		{
			spSetToggleState(record_button, SP_FALSE);
		}
	}
	return;
} /* swRecordWindow */
//
// End of Record Module
//

void swOpenWindow(spComponent component, swWindow window)
{
    swOpenFile(component, window, SP_FALSE);
	open = SP_TRUE;
    return;
}

void swSaveAs(spComponent component, swWindow window, int index)
{
    int i;
    char *filename;
    char *plugin_name;
    char *file_type;
    static int num_plugin = 0;
    static char **plugin_names = NULL;
    static char **file_types = NULL;
    static char **file_filters = NULL;

    if (num_plugin <= 0) {
	plugin_names = xalloc(SW_MAX_PLUGIN + 1, char *);
	file_types = xalloc(SW_MAX_PLUGIN + 1, char *);
	file_filters = xalloc(SW_MAX_PLUGIN + 1, char *);
	
	for (i = 0; i < SW_MAX_PLUGIN; i++) {
	    plugin_names[i] = xalloc(SP_MAX_LINE, char);
	    file_types[i] = xalloc(SP_MAX_LINE, char);
	    file_filters[i] = xalloc(SP_MAX_LINE, char);
	    
	    if (i == 0) {
		strcpy(plugin_names[i], "");
		strcpy(file_types[i], "By Extension");
		strcpy(file_filters[i], "*");
	    } else {
		if (spSearchPluginFileType(i - 1, SP_PLUGIN_OUTPUT,
					   SP_PLUGIN_DEVICE_FILE, plugin_names[i],
					   NULL, file_types[i], file_filters[i]) == SP_FALSE) {
		    xfree(plugin_names[i]);
		    xfree(file_types[i]);
		    xfree(file_filters[i]);
		    break;
		}
		spDebug(50, "swSaveAs", "plugin_name = %s, file_type = %s, file_filters = %s\n",
			plugin_names[i], file_types[i], file_filters[i]);
	    }
	}
	num_plugin = i;
	plugin_names[i] = NULL;
	file_types[i] = NULL;
	file_filters[i] = NULL;
    }
    
    if (window != NULL && window->wave != NULL) {
	filename = window->wave->orig_filename;
	
	/* get save file name */
	filename = xspGetSaveFileName(component, SW_SAVE_DIALOG_TITLE,
				      SppInitialFileName, filename,
				      SppPathMustExist, SP_TRUE,
				      SppFileMustExist, SP_FALSE,
				      SppOverwritePrompt, SP_TRUE,
				      SppFileFilters, file_filters,
				      SppFileTypes, file_types,
				      SppFileFilterIndex, &index,
				      NULL);
	spDebug(10, "swSaveAs", "index = %d\n", index);

	if (filename != NULL) {
	    if (index >= 1) {
		plugin_name = plugin_names[index];
		file_type = file_types[index];
		spDebug(10, "swSaveAs", "plugin_name = %s, file_type = %s\n",
			plugin_name, file_type);
	    } else {
		plugin_name = NULL;
		file_type = NULL;
	    }

#ifdef SW_SUPPORT_THREAD_WRITE
	    swWriteWave(filename, plugin_name, file_type, window->wave, SP_TRUE, SP_TRUE);
#else
	    swSetMouseCursor(window, SP_CURSOR_WAIT);
	    if (swWriteWave(filename, plugin_name, file_type, window->wave, SP_TRUE, SP_FALSE) == SP_TRUE) {
		/*window->wave->edit_flag = SP_FALSE;*/
		window->offset = 0;
		window->length = 0;

		swInitWaveRange(window->wave);
		
		/* reset window */
		swResetWindow(window, SP_FALSE);
	    }
	    swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
#endif

	    xfree(filename);
	}
    }

    return;
}

void swSaveAsWindow(spComponent component, swWindow window)
{
    swSaveAs(component, window, 0);
	saveas = SP_TRUE;
    return;
}

void swSaveWindow(spComponent component, swWindow window)
{
    if (window != NULL) {
	if (strnone(window->wave->orig_filename)) {
	    swSaveAs(component, window, 0);
	} else {
#ifdef SW_SUPPORT_THREAD_WRITE
	    swWriteWave(NULL, NULL, NULL, window->wave, SP_TRUE, SP_TRUE);
#else
	    swSetMouseCursor(window, SP_CURSOR_WAIT);
	    if (swWriteWave(NULL, NULL, NULL, window->wave, SP_TRUE, SP_FALSE) == SP_TRUE) {
		swSetWindowTitle(window);
	    }
	    swSetMouseCursor(window, SP_CURSOR_UNKNOWN);
#endif
	}
    }

    return;
}

void swCloseWindowCB(spComponent component, swWindow window)
{
    swCloseWindow(window, SP_TRUE);
    return;
}

void swDestroyWindowCB(spComponent component, swWindow window)
{
    swDestroyWindow(window);
    return;
}

void swQuitCB(spComponent component, swWindow window)
{
#ifdef SW_CLOSE_PROMPT
    swWindow current_window, next_window;

    spDebug(10, "swQuitCB", "in\n");
    
    if (spQuitPrompt(window->window) == SP_FALSE) {
	return;
    }
    
    if ((next_window = swGetNextWindow(window)) == NULL) {
	/*spQuit(0);*/
	swCloseWindow(window, SP_FALSE);
    } else {
	while (next_window != NULL) {
	    current_window = next_window;
	    next_window = swGetNextWindow(next_window);
	    if (swCloseWindow(current_window, SP_FALSE) == SP_FALSE) {
		break;
	    }

	    if (next_window == window) {
		swCloseWindow(window, SP_FALSE);
		break;
	    }
	}
    }
#else
    spQuitCB(component, NULL);
#endif
    
    return;
}

void swDisplayInfoCB(spComponent component, void *data)
{
    spDisplayInformation(NULL, SW_INFO_TITLE, SW_INFO_MESSAGE, SW_VERSION_STRING);
    return;
}

static void createInfoArea(swWindow window)
{
    if (window->display_info_area == SP_TRUE) {
	if (window->info_box == NULL) {
	    /* create container for information display area */
	    window->info_box = spCreateBox(window->window, "infoBox", 0,
					   NULL);

	    window->label_list = swCreateLabelList(window->info_box, window);
	}
    }
    
    return;
}

void swCheckDisplayInfoAreaCB(spComponent component, swWindow window)
{
    spBool set;
    
    if (spGetToggleState(component, &set) == SP_TRUE) {
	if (set != window->display_info_area) {
	    window->display_info_area = set;
	    
	    if (window->display_info_area == SP_TRUE) {
		spSetParams(window->wave_box, SppWidth, -SW_INFO_AREA_WIDTH, NULL);
		if (window->info_box != NULL) {
		    spMapComponent(window->info_box);
		} else {
		    createInfoArea(window);
		}
	    } else {
		spSetParams(window->wave_box, SppWidth, 0, NULL);
		if (window->info_box != NULL) {
		    spUnmapComponent(window->info_box);
		}
	    }
	    spAdjustWindowSize(window->window);
	}
    }
    
    return;
}

void swCheckDrawLabelCB(spComponent component, swWindow window)
{
    spBool set;
    
    if (spGetToggleState(component, &set) == SP_TRUE) {
	if (set != window->draw_label) {
	    window->draw_label = set;
	    swRedrawLabels(window);
	}
    }
    
    return;
}

void swChangeTimeFormatCB(spComponent component, swWindow window)
{
    char *name;
    
    name = spGetName(component);

    spDebug(10, "swChangeTimeFormatCB", "name = %s\n", name);
    
    if (streq(name, "timeFormatSec")) {
	window->config->time_format = SW_TIME_FORMAT_SEC;
    } else if (streq(name, "timeFormatMsec")) {
	window->config->time_format = SW_TIME_FORMAT_MSEC;
    } else if (streq(name, "timeFormatPoint")) {
	window->config->time_format = SW_TIME_FORMAT_POINT;
    } else if (streq(name, "timeFormatSeparatedSec")) {
	window->config->time_format = SW_TIME_FORMAT_SEPARATED_SEC;
    }

    if (window->config->scale_flag == SP_TRUE || window->config->grid_flag == SP_TRUE) {
	/* redraw all window */
	swDrawAllWave(window);
    } else {
	/* redraw all cursor */
	swMoveAllCursor(window);
    }
    
    swUpdateAllLabelList(window);
    
    return;
}

void swCreateToolBar(swWindow window)
{
#ifdef SW_USE_TOOL_BAR
    if (window->tool_bar == NULL) {
#if defined(MACOS) /* MPW doesn't allow the data type. Why?*/
	static char *spwave_xpm[] = {"", ""};
#else
#include "spwave.xpm"
#endif
	
	spComponent tool_item;
	long sense_level;
	
	spDebug(10, "swCreateToolBar", "in\n");

	window->tool_bar = spCreateToolBar(window->window, "spwave", 5,
					   SppBitmapData, spwave_xpm,
					   SppVisible, (window->config->no_tool_bar == SP_FALSE ?
							SP_TRUE : SP_FALSE),
					   NULL);
	if (swIsClipboardWindow(window) == SP_FALSE) {
	    tool_item = spAddToolItem(window->tool_bar, NULL,
				      SppBitmapIndex, 0,
				      SppGroupId, SW_OPEN_GROUP_ID,
				      SppSenseLevel, SW_STATE_NO_WAVE,
				      SppCallbackFunc, swOpenWindow,
				      SppCallbackData, window,
				      SppDescription, SW_MENU_OPEN_DESC,
				      NULL);
	}
	if (swIsClipboardWindow(window) == SP_TRUE) {
	    sense_level = SW_STATE_NO_WAVE;
	} else {
	    sense_level = SW_STATE_EXIST_WAVE;
	}
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 1,
				  SppSenseLevel, sense_level,
				  SppCallbackFunc, swOpenNewWindow,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_OPEN_NEW_DESC,
				  NULL);
#ifdef SW_SUPPORT_SAVE
	if (swIsClipboardWindow(window) == SP_FALSE) {
	    tool_item = spAddToolItem(window->tool_bar, NULL,
				      SppBitmapIndex, 2,
				      SppSenseLevel, SW_STATE_EDIT_WAVE,
				      SppGroupId, SW_SAVE_GROUP_ID,
				      SppCallbackFunc, swSaveWindow,
				      SppCallbackData, window,
				      SppDescription, SW_MENU_SAVE_DESC,
				      NULL);
	    tool_item = spAddToolItem(window->tool_bar, NULL,
				      SppBitmapIndex, 3,
				      SppGroupId, SW_OPEN_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_WAVE,
				      SppCallbackFunc, swSaveAsWindow,
				      SppCallbackData, window,
				      SppDescription, SW_MENU_SAVE_AS_DESC,
				      NULL);
	}
#endif
	spAddToolSeparator(window->tool_bar, NULL, NULL);

	record_button = spAddCheckToolItem(window->tool_bar, NULL,
					SppBitmapIndex, 15,
					SppGroupId, SW_RECORD_GROUP_ID,
					SppSenseLevel, SW_STATE_NO_WAVE,
					SppCallbackFunc, swRecordWindow,
					SppCallbackData, window,
					SppDescription, SW_MENU_RECORD_START,
					NULL);

	spAddToolSeparator(window->tool_bar, NULL, NULL);

	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 7,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_PLAY_TIME_DATA,
				  SppCallbackFunc, swPlayStopCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_PLAY_STOP_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 8,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayRegionCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_PLAY_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 9,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swBackwardCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_BACKWARD_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 10,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swForwardCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_FORWARD_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 11,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swGoHeadCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_GO_HEAD_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 12,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swGoTailCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_GO_TAIL_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 13,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomInCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_ZOOM_IN_DESC,
				  NULL);
	tool_item = spAddToolItem(window->tool_bar, NULL,
				  SppBitmapIndex, 14,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomOutCB,
				  SppCallbackData, window,
				  SppDescription, SW_MENU_ZOOM_OUT_DESC,
				  NULL);
    }
#endif

    return;
}

long swGetScrollCoef(swWave wave)
{
    if (wave->total_length >= 32768) {
	return (long)ceil((double)wave->total_length / 32768.0);
    } else {
	return 1;
    }
}

#ifdef SW_SUPPORT_DRAG_DROP
void swDropCB(spComponent component, swWindow window, int argc, char **argv)
{
    int i;

    if (window == NULL) return;

#if 0
    if (swIsProcessing(window) == SP_TRUE || window->config->toplevel->num_process > 0) {
	spDisplayWarning(window->window, NULL, SW_PROCESSING_WARNING_MESSAGE);
	return;
    }
#endif
    
    for (i = 0; i < argc; i++) {
	swCreateWindow(argv[i], NULL, SP_FALSE, window->config);
    }

    return;
}
#endif

void swCreateMainWindow(swWindow window)
{
    int size;
    int width;
    long sense_level;
    spComponent menu, sub_menu, menu_bar, menu_item;
    spCloseStyle close_style = SW_DEFAULT_CLOSE_STYLE;
    
    if (window == NULL) return;

    window->window = spCreateFrame("mainWindow",
				   SppCallbackFunc, swCloseWindowCB,
				   SppCallbackData, window,
				   SppUserData, window,
				   SppCloseStyle, close_style,
				   SppSimplifiable, SP_TRUE,
				   SppOrientation, SP_HORIZONTAL,
				   NULL);
    spAddCallback(window->window, SP_DESTROY_CALLBACK,
		  (spCallbackFunc)swDestroyWindowCB, (void *)window);
    spAddCallback(window->window, SP_SIMPLIFY_CALLBACK,
		  (spCallbackFunc)swShowAllToolBarCB, (void *)window);
#ifdef SW_SUPPORT_DRAG_DROP
    spAddDropCallback(window->window, (spDropCallbackFunc)swDropCB, (void *)window);
#endif

    menu_bar = spCreateMenuBar(window->window, "menuBar", NULL);
    
    /* `File' menu */
    menu = spCreatePulldownMenu(menu_bar, "fileMenu",
				SppTitle, SW_MENU_FILE_LABEL,
				NULL);

	/*
	menu_item = spAddMenuItem(menu, "record",
				  SppTitle, SW_MENU_RECORD_LABEL,
				  SppGroupId, SW_RECORD_GROUP_ID,
				  SppSenseLevel, SW_STATE_NO_WAVE,
				  SppCallbackFunc, swRecordWindow,
				  SppCallbackData, window,
				  SppShortcut, SW_RECORD_SHORTCUT,
				  NULL);
	*/

    if (swIsClipboardWindow(window) == SP_FALSE) {
	menu_item = spAddMenuItem(menu, "open",
				  SppTitle, SW_MENU_OPEN_LABEL,
				  SppGroupId, SW_OPEN_GROUP_ID,
				  SppSenseLevel, SW_STATE_NO_WAVE,
				  SppCallbackFunc, swOpenWindow,
				  SppCallbackData, window,
				  SppShortcut, SW_OPEN_SHORTCUT,
				  NULL);
    }
    if (swIsClipboardWindow(window) == SP_TRUE) {
	sense_level = SW_STATE_NO_WAVE;
    } else {
	sense_level = SW_STATE_EXIST_WAVE;
    }
    menu_item = spAddMenuItem(menu, "openNew",
			      SppTitle, SW_MENU_OPEN_NEW_LABEL,
			      /*SppGroupId, SW_OPEN_GROUP_ID,*/
			      SppSenseLevel, sense_level,
			      SppCallbackFunc, swOpenNewWindow,
			      SppCallbackData, window,
			      SppShortcut, SW_OPEN_NEW_SHORTCUT,
			      NULL);
#ifdef SW_SUPPORT_SAVE
    if (swIsClipboardWindow(window) == SP_FALSE) {
	menu_item = spAddMenuItem(menu, "save",
				  SppTitle, SW_MENU_SAVE_LABEL,
				  SppSenseLevel, SW_STATE_EDIT_WAVE,
				  SppGroupId, SW_SAVE_GROUP_ID,
				  SppCallbackFunc, swSaveWindow,
				  SppCallbackData, window,
				  SppShortcut, SW_SAVE_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "saveAs",
				  SppTitle, SW_MENU_SAVE_AS_LABEL,
				  SppGroupId, SW_OPEN_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swSaveAsWindow,
				  SppCallbackData, window,
				  SppShortcut, SW_SAVE_AS_SHORTCUT,
				  NULL);
    }
#endif
    {
	spComponent label_menu;
	
	/* create menus for labels */
	spAddMenuSeparator(menu, "labelSeparator", NULL);
	label_menu = spAddSubMenu(menu, "labelMenu",
				  SppTitle, SW_MENU_LABEL_LABEL,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "openLabel",
				  SppTitle, SW_MENU_OPEN_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swOpenLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "saveLabel",
				  SppTitle, SW_MENU_SAVE_LABEL_LABEL,
				  SppGroupId, SW_NORMAL_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swSaveLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "saveAsLabel",
				  SppTitle, SW_MENU_SAVE_AS_LABEL_LABEL,
				  SppGroupId, SW_NORMAL_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swSaveAsLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "clearLabel",
				  SppTitle, SW_MENU_CLEAR_LABEL_LABEL,
				  SppGroupId, SW_NORMAL_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swClearLabelCB,
				  SppCallbackData, window,
				  NULL);
	spAddMenuSeparator(label_menu, "labelSeparator", NULL);
	menu_item = spAddMenuItem(label_menu, "openRegionLabel",
				  SppTitle, SW_MENU_OPEN_REGION_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swOpenRegionLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "saveRegionLabel",
				  SppTitle, SW_MENU_SAVE_REGION_LABEL_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swSaveRegionLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "saveAsRegionLabel",
				  SppTitle, SW_MENU_SAVE_AS_REGION_LABEL_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swSaveAsRegionLabelCB,
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(label_menu, "clearRegionLabel",
				  SppTitle, SW_MENU_CLEAR_REGION_LABEL_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swClearRegionLabelCB,
				  SppCallbackData, window,
				  NULL);
    }
    spAddMenuSeparator(menu, "fileSeparator", NULL);
    menu_item = spAddMenuItem(menu, "close",
			      SppTitle, SW_MENU_CLOSE_LABEL,
			      SppGroupId, SW_WINDOW_GROUP_ID,
			      SppSenseLevel, SW_STATE_STOP_SOME_WINDOWS,
			      SppCallbackFunc, swCloseWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_CLOSE_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "quit",
			      SppTitle, SW_MENU_QUIT_LABEL,
			      SppGroupId, SW_QUIT_GROUP_ID,
			      SppSenseLevel, SW_STATE_ONE_WINDOW,
			      SppCallbackFunc, swQuitCB,
			      SppCallbackData, window,
			      SppShortcut, SW_QUIT_SHORTCUT,
			      NULL);
    
    /* `Edit' menu */
    menu = spCreatePulldownMenu(menu_bar, "editMenu",
				SppTitle, SW_MENU_EDIT_LABEL,
				NULL);
#ifdef SW_SUPPORT_EDIT
    window->undo_menu = spAddMenuItem(menu, "undoWave",
				      SppTitle, SW_MENU_UNDO_LABEL,
				      SppGroupId, SW_UNDO_GROUP_ID,
				      SppCallbackFunc, swUndoWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_UNDO_SHORTCUT,
				      NULL);
    window->redo_menu = spAddMenuItem(menu, "redoWave",
				      SppTitle, SW_MENU_REDO_LABEL,
				      SppGroupId, SW_UNDO_GROUP_ID,
				      SppCallbackFunc, swRedoWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_REDO_SHORTCUT,
				      NULL);
    spAddMenuSeparator(menu, "editSeparator", NULL);
    menu_item = spAddMenuItem(menu, "cropWave",
			      SppTitle, SW_MENU_CROP_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swCropWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_CROP_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "deleteWave",
			      SppTitle, SW_MENU_DELETE_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_CHANNELS,
			      SppCallbackFunc, swDeleteWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_DELETE_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "eraseWave",
			      SppTitle, SW_MENU_ERASE_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swEraseWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_ERASE_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "extractWave",
			      SppTitle, SW_MENU_EXTRACT_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swExtractWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_EXTRACT_SHORTCUT,
			      NULL);
#ifdef SW_SUPPORT_AUTOSAVE
    if (swIsClipboardWindow(window) == SP_FALSE) {
	menu_item = spAddMenuItem(menu, "extractAutosaveWave",
				  SppTitle, SW_MENU_EXTRACT_AUTOSAVE_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
				  SppCallbackFunc, swExtractAutosaveWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_EXTRACT_AUTOSAVE_SHORTCUT,
				  NULL);
    }
#endif
#ifdef SW_SUPPORT_CLIPBOARD
    if (swIsClipboardWindow(window) == SP_FALSE) {
	spAddMenuSeparator(menu, "editClipboardSeparator", NULL);
	menu_item = spAddMenuItem(menu, "cutWave",
				  SppTitle, SW_MENU_CUT_LABEL,
				  SppGroupId, SW_CLIPBOARD_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_TIME_CHANNELS,
				  SppCallbackFunc, swCutWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_CUT_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "copyWave",
				  SppTitle, SW_MENU_COPY_LABEL,
				  SppGroupId, SW_CLIPBOARD_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
				  SppCallbackFunc, swCopyWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_COPY_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "cat",
				  SppTitle, SW_MENU_CAT_LABEL,
				  SppGroupId, SW_CLIPBOARD_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				  SppCallbackFunc, swCatWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_CAT_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "cattop",
				  SppTitle, SW_MENU_CATTOP_LABEL,
				  SppGroupId, SW_CLIPBOARD_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				  SppCallbackFunc, swCatTopWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_CATTOP_SHORTCUT,
				  NULL);
	menu_item = spAddMenuItem(menu, "clipboard",
				  SppTitle, SW_MENU_CLIPBOARD_WINDOW_LABEL,
				  SppCallbackFunc, swPopupClipboardWindowCB,
				  SppCallbackData, window->config,
				  NULL);
    }
#endif
    spAddMenuSeparator(menu, "editGainSeparator", NULL);
    menu_item = spAddMenuItem(menu, "amplifyWave",
			      SppTitle, SW_MENU_AMPLIFY_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swPopupAmplifyDialogCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "invertWave",
			      SppTitle, SW_MENU_INVERT_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swInvertWindowCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "fadeInWave",
			      SppTitle, SW_MENU_FADE_IN_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swFadeInWindowCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "fadeOutWave",
			      SppTitle, SW_MENU_FADE_OUT_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swFadeOutWindowCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "swapWaveChannel",
			      SppTitle, SW_MENU_CHANNEL_SWAP_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_MULTI_CHANNEL,
			      SppCallbackFunc, swSwapWaveChannelCB,
			      SppCallbackData, window,
			      NULL);
    
    spAddMenuSeparator(menu, "editFileSeparator", NULL);
    menu_item = spAddMenuItem(menu, "maximizeWave",
			      SppTitle, SW_MENU_MAXIMIZE_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_TIME_DATA,
			      SppCallbackFunc, swPopupMaximizeDialogCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "bitConv",
			      SppTitle, SW_MENU_BIT_CONV_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_TIME_DATA,
			      SppCallbackFunc, swPopupBitConvDialogCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "sampFreqConv",
			      SppTitle, SW_MENU_SAMP_RATE_CONV_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_TIME_DATA,
			      SppCallbackFunc, swPopupSampFreqConvDialogCB,
			      SppCallbackData, window,
			      NULL);
    menu_item = spAddMenuItem(menu, "monauralize",
			      SppGroupId, SW_CHANNEL_GROUP_ID,
			      SppSenseLevel, SW_STATE_STEREO_WAVE,
			      SppTitle, SW_MENU_MONAURALIZE_LABEL,
			      SppCallbackFunc, swMonauralizeCB,
			      SppCallbackData, window,
			      NULL);
#ifdef SW_SUPPORT_INFO_DIALOG
    menu_item = spAddMenuItem(menu, "infoDialog",
			      SppTitle, SW_MENU_INFO_DIALOG_LABEL,
			      SppCallbackFunc, swPopupInfoDialogCB,
			      SppCallbackData, window,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      NULL);
#endif
    spAddMenuSeparator(menu, "editPrefSeparator", NULL);
#endif
    menu_item = spAddMenuItem(menu, "preferenceDialog",
			      SppTitle, SW_MENU_PREFERENCE_LABEL,
			      SppCallbackFunc, swPopupPreferenceDialogCB,
			      SppCallbackData, window,
			      SppShortcut, SW_PREFERENCE_SHORTCUT,
			      NULL);
#ifdef SW_USE_ANALYSIS
    menu_item = spAddMenuItem(menu, "analysisDialog",
			      SppTitle, SW_MENU_ANALYSIS_PREFERENCE_LABEL,
			      SppCallbackFunc, swPopupAnalysisDialogCB,
			      SppCallbackData, window,
			      NULL);
#endif

    /* `Page' menu */
    menu = spCreatePulldownMenu(menu_bar, "pageMenu",
				SppTitle, SW_MENU_PAGE_LABEL,
				NULL);
    menu_item = spAddMenuItem(menu, "forward",
			      SppTitle, SW_MENU_FORWARD_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swForwardCB,
			      SppCallbackData, window,
			      SppShortcut, SW_FORWARD_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "backward",
			      SppTitle, SW_MENU_BACKWARD_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swBackwardCB,
			      SppCallbackData, window,
			      SppShortcut, SW_BACKWARD_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "goHead",
			      SppTitle, SW_MENU_GO_HEAD_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swGoHeadCB,
			      SppCallbackData, window,
			      SppShortcut, SW_GO_HEAD_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "goTail",
			      SppTitle, SW_MENU_GO_TAIL_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swGoTailCB,
			      SppCallbackData, window,
			      SppShortcut, SW_GO_TAIL_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "nextWindow",
			      SppTitle, SW_MENU_NEXT_WINDOW_LABEL,
			      SppGroupId, SW_WINDOW_GROUP_ID,
			      SppSenseLevel, SW_STATE_SOME_WINDOWS,
			      SppCallbackFunc, swNextWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_NEXT_WINDOW_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "prevWindow",
			      SppTitle, SW_MENU_PREV_WINDOW_LABEL,
			      SppGroupId, SW_WINDOW_GROUP_ID,
			      SppSenseLevel, SW_STATE_SOME_WINDOWS,
			      SppCallbackFunc, swPrevWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_PREV_WINDOW_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "alignWindow",
			      SppTitle, SW_MENU_ALIGN_WINDOW_LABEL,
			      SppGroupId, SW_WINDOW_GROUP_ID,
			      SppSenseLevel, SW_STATE_STOP_SOME_WINDOWS,
			      SppCallbackFunc, swAlignWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_ALIGN_WINDOW_SHORTCUT,
			      NULL);
    spAddMenuSeparator(menu, "pageSeparator", NULL);
    menu_item = spAddMenuItem(menu, "selectRegion",
			      SppTitle, SW_MENU_SELECT_REGION_LABEL,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swSelectRegionCB,
			      SppCallbackData, window,
			      SppShortcut, SW_SELECT_REGION_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "selectAll",
			      SppTitle, SW_MENU_SELECT_ALL_LABEL,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swSelectAllCB,
			      SppCallbackData, window,
			      SppShortcut, SW_SELECT_ALL_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "selectAllRegion",
			      SppTitle, SW_MENU_SELECT_ALL_REGION_LABEL,
			      SppGroupId, SW_WINDOW_GROUP_ID,
			      SppSenseLevel, SW_STATE_SOME_WINDOWS,
			      SppCallbackFunc, swSelectAllRegionCB,
			      SppCallbackData, window,
			      SppShortcut, SW_SELECT_ALL_REGION_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "selectNextChannel",
			      SppTitle, SW_MENU_SELECT_NEXT_CHANNEL_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
			      SppCallbackFunc, swSelectNextChannelCB,
			      SppCallbackData, window,
			      SppShortcut, SW_SELECT_NEXT_CHANNEL_SHORTCUT,
			      NULL);

    /* `View' menu */
    menu = spCreatePulldownMenu(menu_bar, "viewMenu",
				SppTitle, SW_MENU_VIEW_LABEL,
				NULL);
    menu_item = spAddMenuItem(menu, "zoomIn",
			      SppTitle, SW_MENU_ZOOM_IN_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swZoomInCB,
			      SppCallbackData, window,
			      SppShortcut, SW_ZOOM_IN_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "zoomOut",
			      SppTitle, SW_MENU_ZOOM_OUT_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swZoomOutCB,
			      SppCallbackData, window,
			      SppShortcut, SW_ZOOM_OUT_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "zoomFullOut",
			      SppTitle, SW_MENU_ZOOM_FULL_OUT_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_EXIST_WAVE,
			      SppCallbackFunc, swZoomFullOutCB,
			      SppCallbackData, window,
			      SppShortcut, SW_ZOOM_FULL_OUT_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "zoomRegion",
			      SppTitle, SW_MENU_ZOOM_REGION_LABEL,
			      SppGroupId, SW_PAGE_GROUP_ID,
			      SppSenseLevel, SW_STATE_SELECT_WAVE,
			      SppCallbackFunc, swZoomRegionCB,
			      SppCallbackData, window,
			      SppShortcut, SW_ZOOM_REGION_SHORTCUT,
			      NULL);
    
    sub_menu = spAddSubMenu(menu, "timeFormat",
			    SppTitle, SW_MENU_TIME_FORMAT_LABEL,
			    NULL);
    menu_item = spAddRadioButtonMenuItem(sub_menu, "timeFormatSec",
					 SppTitle, SW_MENU_TIME_FORMAT_SEC_LABEL,
					 SppGroupId, SW_DATA_GROUP_ID,
					 SppSenseLevel, SW_STATE_TIME_DATA,
					 SppCallbackFunc, swChangeTimeFormatCB,
					 SppCallbackData, window,
					 SppSet, (window->config->time_format == SW_TIME_FORMAT_SEC ?
						  SP_TRUE : SP_FALSE),
					 NULL);
    menu_item = spAddRadioButtonMenuItem(sub_menu, "timeFormatMsec",
					 SppTitle, SW_MENU_TIME_FORMAT_MSEC_LABEL,
					 SppGroupId, SW_DATA_GROUP_ID,
					 SppSenseLevel, SW_STATE_TIME_DATA,
					 SppCallbackFunc, swChangeTimeFormatCB,
					 SppCallbackData, window,
					 SppSet, (window->config->time_format == SW_TIME_FORMAT_MSEC ?
						  SP_TRUE : SP_FALSE),
					 NULL);
    menu_item = spAddRadioButtonMenuItem(sub_menu, "timeFormatPoint",
					 SppTitle, SW_MENU_TIME_FORMAT_POINT_LABEL,
					 SppGroupId, SW_DATA_GROUP_ID,
					 SppSenseLevel, SW_STATE_TIME_DATA,
					 SppCallbackFunc, swChangeTimeFormatCB,
					 SppCallbackData, window,
					 SppSet, (window->config->time_format == SW_TIME_FORMAT_POINT ?
						  SP_TRUE : SP_FALSE),
					 NULL);
    menu_item = spAddRadioButtonMenuItem(sub_menu, "timeFormatSeparatedSec",
					 SppTitle, SW_MENU_TIME_FORMAT_SEPARATED_SEC_LABEL,
					 SppGroupId, SW_DATA_GROUP_ID,
					 SppSenseLevel, SW_STATE_TIME_DATA,
					 SppCallbackFunc, swChangeTimeFormatCB,
					 SppCallbackData, window,
					 SppSet, (window->config->time_format == SW_TIME_FORMAT_SEPARATED_SEC ?
						  SP_TRUE : SP_FALSE),
					 NULL);
    
    spAddMenuSeparator(menu, "viewSeparator", NULL);
    menu_item = spAddCheckBoxMenuItem(menu, "displayInfoArea",
				      SppTitle, SW_MENU_DISPLAY_INFO_AREA_LABEL,
				      SppSet, window->display_info_area,
				      SppCallbackFunc, swCheckDisplayInfoAreaCB,
				      SppCallbackData, window,
				      NULL);
    menu_item = spAddCheckBoxMenuItem(menu, "drawLabel",
				      SppTitle, SW_MENU_DRAW_LABEL_LABEL,
				      SppSet, window->draw_label,
				      SppCallbackFunc, swCheckDrawLabelCB,
				      SppCallbackData, window,
				      NULL);
    
    /* `Sound' menu */
    menu = spCreatePulldownMenu(menu_bar, "soundMenu",
				SppTitle, SW_MENU_SOUND_LABEL,
				NULL);
    menu_item = spAddMenuItem(menu, "playWave",
			      SppTitle, SW_MENU_PLAY_LABEL,
			      SppGroupId, SW_PLAY_GROUP_ID,
			      SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
			      SppCallbackFunc, swPlayRegionCB,
			      SppCallbackData, window,
			      SppShortcut, SW_PLAY_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "playWindow",
			      SppTitle, SW_MENU_PLAY_WINDOW_LABEL,
			      SppGroupId, SW_PLAY_GROUP_ID,
			      SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
			      SppCallbackFunc, swPlayWindowCB,
			      SppCallbackData, window,
			      SppShortcut, SW_PLAY_WINDOW_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "playFile",
			      SppTitle, SW_MENU_PLAY_FILE_LABEL,
			      SppGroupId, SW_PLAY_GROUP_ID,
			      SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
			      SppCallbackFunc, swPlayFileCB,
			      SppCallbackData, window,
			      SppShortcut, SW_PLAY_FILE_SHORTCUT,
			      NULL);
    menu_item = spAddMenuItem(menu, "playStop",
			      SppTitle, SW_MENU_PLAY_STOP_LABEL,
			      SppGroupId, SW_DATA_GROUP_ID,
			      SppSenseLevel, SW_STATE_PLAY_TIME_DATA,
			      SppCallbackFunc, swPlayStopCB,
			      SppCallbackData, window,
			      SppShortcut, SW_PLAY_STOP_SHORTCUT,
			      NULL);
#ifdef SP_SUPPORT_AUDIO
    spAddMenuSeparator(menu, "soundSeparator", NULL);
    menu_item = spAddCheckBoxMenuItem(menu, "loopPlay",
				      SppTitle, SW_MENU_LOOP_PLAY_LABEL,
				      SppGroupId, SW_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_TIME_DATA,
				      SppSet, window->loop_play,
				      SppCallbackFunc, spGetToggleState,
				      SppCallbackData, &window->loop_play,
				      SppShortcut, SW_LOOP_PLAY_SHORTCUT,
				      NULL);
    menu_item = spAddCheckBoxMenuItem(menu, "syncPlay",
				      SppTitle, SW_MENU_SYNC_PLAY_LABEL,
				      SppGroupId, SW_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_TIME_DATA,
				      SppSet, window->sync_play,
				      SppCallbackFunc, spGetToggleState,
				      SppCallbackData, &window->sync_play,
				      NULL);
#endif

    /* `Help' menu */
    menu = spCreatePulldownMenu(menu_bar, "helpMenu",
				SppTitle, SW_MENU_HELP_LABEL,
				SppMenuHelp, SP_TRUE,
				NULL);
    menu_item = spAddMenuItem(menu, "info",
			      SppTitle, SW_MENU_INFO_LABEL,
			      SppSenseLevel, SW_STATE_NO_WAVE,
			      SppCallbackFunc, swDisplayInfoCB,
			      SppShortcut, SW_INFO_SHORTCUT,
			      NULL);

    swCreateToolBar(window);
    
    /* create container for waveform canvas */
    width = window->width;
    if (window->display_info_area == SP_TRUE) {
	size = -SW_INFO_AREA_WIDTH;
	width = MAX(width - SW_INFO_AREA_WIDTH, SW_MIN_WINDOW_WIDTH);
    } else {
	size = 0;
    }
    window->wave_box = spCreateBox(window->window, "waveBox", size,
				   SppInitialWidth, width,
				   SppInitialHeight, window->height + window->config->overview_height,
				   NULL);
    
    spCreateCanvas(window->wave_box, "overviewCanvas",
		   width, window->config->overview_height,
		   SppBorderOn, SP_TRUE,
		   SppUseArrowKey, SP_TRUE,
		   SppHeight, window->config->overview_height,
		   SppCallbackFunc, swDrawOverviewCB,
		   SppCallbackData, window,
		   NULL);

    spCreateCanvas(window->wave_box, "canvas",
		   width, window->height,
		   SppBorderOn, SP_TRUE,
		   SppUseArrowKey, SP_TRUE,
		   SppHeight, -1 * SW_DEFAULT_SLIDER_WIDTH,
		   SppCallbackFunc, swDrawWaveCB,
		   SppCallbackData, window,
		   NULL);
    
    if (swIsNoWave(window) == SP_FALSE) {
	window->scroll_coef = swGetScrollCoef(window->wave);
	
	window->hscroll = spCreateScrollBar(window->wave_box, "hscroll",
					    SppGroupId, SW_PAGE_GROUP_ID,
					    SppSenseLevel, SW_STATE_EXIST_WAVE,
					    SppHeight, SW_DEFAULT_SLIDER_WIDTH,
					    SppOrientation, SP_HORIZONTAL,
					    SppTrackCallbackOn, SP_TRUE,
					    SppMinimum, 0,
					    SppMaximum, (int)(window->wave->total_length / window->scroll_coef),
					    SppIncrement, (int)MAX(window->length / window->scroll_coef / 4, 1),
					    SppPageIncrement, (int)MAX(window->length / window->scroll_coef, 1),
					    SppSliderSize, (int)MAX(window->length / window->scroll_coef, 1),
					    SppValue, (int)(window->offset / window->scroll_coef),
					    SppCallbackFunc, swScrollCB,
					    SppCallbackData, window,
					    NULL);
    } else {
	window->hscroll = spCreateScrollBar(window->wave_box, "hscroll",
					    SppGroupId, SW_PAGE_GROUP_ID,
					    SppSenseLevel, SW_STATE_EXIST_WAVE,
					    SppHeight, SW_DEFAULT_SLIDER_WIDTH,
					    SppOrientation, SP_HORIZONTAL,
					    SppTrackCallbackOn, SP_TRUE,
					    SppMinimum, 0,
					    SppMaximum, 100,
					    SppIncrement, 10,
					    SppPageIncrement, 100,
					    SppSliderSize, 100,
					    SppValue, 0,
					    SppCallbackFunc, swScrollCB,
					    SppCallbackData, window,
					    NULL);
    }

    createInfoArea(window);

    return;
}

void swSetWindowValue(swWindow window)
{
    spBool track_call_on;
    
    if (window == NULL) return;
    
    if (window->wave != NULL) {
#if 0
	if (window->length >= window->config->wave_config->read_callback_length) {
	    track_call_on = SP_FALSE;
	} else {
	    track_call_on = SP_TRUE;
	}
#else
	track_call_on = SP_TRUE;
#endif

	window->scroll_coef = swGetScrollCoef(window->wave);
	
	spSetParams(window->hscroll,
		    SppMinimum, 0,
		    SppMaximum, (int)(window->wave->total_length / window->scroll_coef),
		    SppIncrement, (int)MAX(window->length / window->scroll_coef / 4, 1),
		    SppPageIncrement, (int)MAX(window->length / window->scroll_coef, 1),
		    SppSliderSize, (int)MAX(window->length / window->scroll_coef, 1),
		    SppValue, (int)(window->offset / window->scroll_coef),
		    SppTrackCallbackOn, track_call_on,
		    NULL);
	
	spDebug(40, NULL, "offset = %ld, length = %ld\n", window->offset, window->length);
	spDebug(10, NULL, "wave length = %ld, scroll_coef = %ld\n",
		window->wave->total_length, window->scroll_coef);
	
	if (window->point >= 0) {
	    window->point_d = swSampToDisp(window, window->point);
	}
	
	if (window->sel_st >= 0 && window->sel_ed >= 0) {
	    window->sel_st_d = swSampToDisp(window, window->sel_st);
	    window->sel_ed_d = swSampToDisp(window, window->sel_ed);

	    if (window->sel_st != window->sel_ed && 
		window->sel_st_d == window->sel_ed_d) {
		window->sel_ed_d += 1;
	    }
	}
    } else {
	spSetParams(window->hscroll,
		    SppMinimum, 0,
		    SppMaximum, 100,
		    SppIncrement, 10,
		    SppPageIncrement, 100,
		    SppSliderSize, 100,
		    SppValue, 0,
		    NULL);
    }
    
    spDebug(10, "swSetWindowValue", "done\n");
    
    return;
}

spBool swIsNoWave(swWindow window)
{
    if (window == NULL) return SP_TRUE;
    
    if (swIsWaveNone(window->wave) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool swIsNoLabel(swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE
	|| swIsWaveLabelNone(window->wave) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool swIsProcessing(swWindow window)
{
    if (window == NULL || window->wave == NULL) return SP_FALSE;

    if (swIsWaveProcessing(window->wave) == SP_TRUE) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool swIsVisible(swWindow window)
{
    if (window == NULL) return SP_FALSE;

    spDebug(10, "swIsVisible", "visible_flag = %d\n", window->visible_flag);
    
    return window->visible_flag;
}

spBool swIsPlayable(swWindow window)
{
    if (window == NULL || window->wave == NULL) return SP_FALSE;

    if (swIsWaveProcessing(window->wave) == SP_TRUE
	|| swIsWavePlayable(window->wave) == SP_FALSE
	|| window->config->toplevel->playable == SP_FALSE) {
	return SP_FALSE;
    }

    return SP_TRUE;
}

spBool swNeedWindowProcessBlock(swWindow src_window, swWindow window)
{
    spBool need_block;
    
    if (src_window == NULL || window == NULL) return SP_FALSE;

    if (window->config->toplevel->using_clipboard == SP_TRUE
	&& swIsClipboardWindow(window) == SP_TRUE) {
	need_block = SP_TRUE;
    } else {
	need_block = swNeedProcessBlock(src_window->wave, window->wave);
    }

    return need_block;
}

void swPopupWindow(swWindow window, int x, int y)
{
    if (window == NULL) return;

    if (x >= 0 && y >= 0) {
	spSetWindowPosition(window->window, x, y);
	spMapWindow(window->window);
    } else {
	spPopupWindow(window->window);
    }
    
    spDebug(10, "swPopupWindow", "done\n");
    
    return;
}

/**/
spBool swIsClipboardWindow(swWindow window)
{
    if (window == NULL) return SP_FALSE;

    if (window->config->toplevel->clipboard_window == window) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool swIsClipboardVisible(swWindow window)
{
    if (window == NULL) return SP_FALSE;

    return spIsVisible(window->config->toplevel->clipboard_window->window);
}

spBool swIsClipboardNone(swWindow window)
{
    if (window == NULL) return SP_FALSE;

    return swIsNoWave(window->config->toplevel->clipboard_window);
}

void swCreateClipboardWindow(swConfig config)
{
    config->toplevel->clipboard_window = swInitWindow(config, SW_TIME_DATA);
    swSetWave(config->toplevel->clipboard_window, NULL);
    swCreateMainWindow(config->toplevel->clipboard_window);
    swSetWindowTitle(config->toplevel->clipboard_window);
    
    return;
}

#ifdef SW_SUPPORT_CLIPBOARD
void swPopupClipboardWindowCB(spComponent component, swConfig config)
{
    if (config->toplevel->clipboard_window != NULL) {
	if (spIsVisible(config->toplevel->clipboard_window->window) == SP_TRUE) {
	    spMapComponent(config->toplevel->clipboard_window->window);
	} else {
	    spDebug(10, "swPopupClipboardWindowCB", "popup clipboard\n");
	    config->toplevel->clipboard_window->visible_flag = SP_TRUE;
	    spPopupWindow(config->toplevel->clipboard_window->window);
	    
	    spLockMutex(config->toplevel->main_mutex);
	    ++config->toplevel->num_window;
	    spUnlockMutex(config->toplevel->main_mutex);
	    
	    swSetSenseLevel(NULL);
	}
    }
    return;
}
#endif

spBool swLockWindowMutex(swWindow window)
{
    if (window == NULL) return SP_FALSE;
    
    return spLockMutex(window->mutex);
}

spBool swUnlockWindowMutex(swWindow window)
{
    if (window == NULL) return SP_FALSE;
    
    return spUnlockMutex(window->mutex);
}

spBool swLockMainMutex(swWindow window)
{
    if (window == NULL) return SP_FALSE;
    
    return spLockMutex(window->config->toplevel->main_mutex);
}

spBool swUnlockMainMutex(swWindow window)
{
    if (window == NULL) return SP_FALSE;
    
    return spUnlockMutex(window->config->toplevel->main_mutex);
}

static swTopLevel sw_toplevel = NULL;

void swInitTopLevel(spTopLevel toplevel, swConfig config)
{
    if (sw_toplevel == NULL) {
	sw_toplevel = xalloc(1, struct _swTopLevel);
	sw_toplevel->toplevel = toplevel;
	sw_toplevel->null_window = NULL;
	sw_toplevel->current_window = NULL;
	sw_toplevel->clipboard_window = NULL;
	
	sw_toplevel->num_window = 0;
	sw_toplevel->num_process = 0;
	sw_toplevel->playable = SP_TRUE;
	sw_toplevel->using_clipboard = SP_FALSE;
	sw_toplevel->main_mutex = spCreateMutex(NULL);
	
	config->toplevel = sw_toplevel;
    }

    return;
}

long swGetNumWindow(void)
{
    if (sw_toplevel == NULL) return 0;
    
    return sw_toplevel->num_window;
}

long swGetNumProcess(void)
{
    if (sw_toplevel == NULL) return 0;
    
    return sw_toplevel->num_process;
}
