/*
 *	swWindow.h
 */

#ifndef __SWWINDOW_H
#define __SWWINDOW_H

#include <sp/spDefs.h>
#include <sp/spComponentLib.h>

#include "swWave.h"
#include "swStringDefs.h"

#ifdef __cplusplus
extern "C" {
#endif

#if defined(MACOS)
#define SW_AUDIO_BUFFER_SIZE 8192
#else
#define SW_AUDIO_BUFFER_SIZE 1024
#endif
#define SW_READ_AFTER

#define SW_SUPPORT_METER
    
#define SW_USE_TOOL_BAR
#define SW_USE_ANALYSIS
#define SW_SUPPORT_EDIT
#define SW_SUPPORT_SAVE
#define SW_SUPPORT_INFO_DIALOG
#define SW_SUPPORT_AUTOSAVE
#define SW_SUPPORT_DRAG_DROP
#define SW_SUPPORT_CLIPBOARD
#define SW_CLOSE_PROMPT
#define SW_SUPPORT_THREAD_WRITE
    
#define SW_VERSION_STRING "0.6.8"
#define SW_HELP_MESSAGE "Speech file editor version %s"
#define SW_SETUP_FILE "~/.spwave"
    
#define SW_INIT_TIME_FORMAT SW_TIME_FORMAT_SEC


/* groups for sense levels */
#define SW_OPEN_GROUP_ID 10
#define SW_SAVE_GROUP_ID 20
#define SW_PLAY_GROUP_ID 30
#define SW_PAGE_GROUP_ID 40
    
#define SW_CHANNEL_GROUP_ID 50
#define SW_WINDOW_GROUP_ID 60
#define SW_QUIT_GROUP_ID 65
#define SW_DATA_GROUP_ID 70
#define SW_CLIPBOARD_DATA_GROUP_ID 75
    
#define SW_LABEL_GROUP_ID 80
#define SW_NORMAL_LABEL_GROUP_ID 90
#define SW_REGION_LABEL_GROUP_ID 100
#define SW_LABEL_LIST_GROUP_ID 105

#define SW_UNDO_GROUP_ID 110
#define SW_SCROLL_GROUP_ID 120
#define SW_CLIPBOARD_GROUP_ID 130

#define SW_RECORD_GROUP_ID 140

/* sense levels for default group */
#define SW_STATE_PLAY_WAVE 0
#define SW_STATE_NO_WAVE 10
#define SW_STATE_EXIST_WAVE 20
#define SW_STATE_NOT_PLAY_WAVE SW_STATE_EXIST_WAVE
#define SW_STATE_SELECT_WAVE 30
#define SW_STATE_EDIT_WAVE 30

/* sense levels for channel group */
#define SW_STATE_MONO_WAVE 20
#define SW_STATE_STEREO_WAVE 30
#define SW_STATE_MULTI_CHANNEL_WAVE 40
    
/* sense levels for window group */
#define SW_STATE_ONE_WINDOW 5
#define SW_STATE_SOME_WINDOWS 10
#define SW_STATE_STOP_SOME_WINDOWS 20

/* sense levels for data group */
#define SW_STATE_NO_DATA 0
#define SW_STATE_FREQ_DATA 10
#define SW_STATE_PLAY_TIME_DATA 15
#define SW_STATE_EDIT_TIME_DATA 18
#define SW_STATE_TIME_DATA 20
#define SW_STATE_SELECT_TIME_DATA 30
#define SW_STATE_SELECT_TIME_CHANNELS 40
#define SW_STATE_SELECT_TIME_MULTI_CHANNEL 50

/* sense levels for label group */
#define SW_STATE_NO_LABEL 0
#define SW_STATE_EXIST_LABEL 10
#define SW_STATE_EXIST_LABEL_HERE 20

/* sense levels for clipboard group */
#define SW_STATE_NO_CLIPBOARD 0
#define SW_STATE_EXIST_CLIPBOARD 30
#define SW_STATE_EXIST_CLIPBOARD_SELECTED 40
    
/* sizes */
#define SW_WINDOW_WIDTH 800
#define SW_WINDOW_HEIGHT 200
#define SW_MIN_WINDOW_WIDTH 50
#define SW_MIN_WINDOW_HEIGHT 20
#define SW_METER_WIDTH 20

    
#define SW_INFO_AREA_WIDTH 340
    
#define SW_FREQ_WINDOW_WIDTH 400
#define SW_FREQ_WINDOW_HEIGHT 300
    
#define SW_DEFAULT_SLIDER_WIDTH 15

#define SW_EDGE_MOTION_RANGE 15
    
#define SW_LABEL_DELETE_RANGE 30

#define SW_MIN_DRAW_WIDTH 40
#define SW_MIN_DRAW_HEIGHT 20

#define SW_TIME_STRING_TOP_OFFSET -15
#define SW_TIME_STRING_LEFT_OFFSET 15

#define SW_AMP_STRING_TOP_OFFSET 25
#define SW_AMP_STRING_LEFT_OFFSET 15

#define SW_WAVE_VSPACE 5
    
    
/* shortcuts */
#define SW_OPEN_SHORTCUT "A-o"
#define SW_RECORD_SHORTCUT "A-r"
#define SW_SAVE_SHORTCUT "A-s"
#define SW_CLOSE_SHORTCUT "A-w"
#define SW_QUIT_SHORTCUT "A-q"
    
#define SW_OPEN_NEW_SHORTCUT "A-n"
#define SW_SAVE_AS_SHORTCUT ""
    
#define SW_FORWARD_SHORTCUT "C-f"
#define SW_BACKWARD_SHORTCUT "C-b"
#define SW_GO_HEAD_SHORTCUT "C-a"
#define SW_GO_TAIL_SHORTCUT "C-e"
#define SW_NEXT_WINDOW_SHORTCUT "C-n"
#define SW_PREV_WINDOW_SHORTCUT "C-p"
#define SW_ALIGN_WINDOW_SHORTCUT "C-l"
#define SW_SELECT_REGION_SHORTCUT "C-Space"
#define SW_SELECT_ALL_REGION_SHORTCUT "S-a"
#define SW_SELECT_ALL_SHORTCUT "A-a"
#define SW_SELECT_NEXT_CHANNEL_SHORTCUT "S-t"
    
#define SW_ZOOM_IN_SHORTCUT "C-i"
#define SW_ZOOM_OUT_SHORTCUT "C-o"
#define SW_ZOOM_FULL_OUT_SHORTCUT "S-f"
#define SW_ZOOM_REGION_SHORTCUT "S-b"
    
#define SW_UNDO_SHORTCUT "A-z"
#define SW_REDO_SHORTCUT "A-y"
#define SW_CROP_SHORTCUT "S-c"
#define SW_DELETE_SHORTCUT "Delete"
#define SW_ERASE_SHORTCUT ""
#define SW_EXTRACT_SHORTCUT "S-x"
#define SW_EXTRACT_AUTOSAVE_SHORTCUT "S-v"
#define SW_PREFERENCE_SHORTCUT ""
    
#define SW_INFO_SHORTCUT ""
    
#define SW_PLAY_SHORTCUT "S-p"
#define SW_LOOP_PLAY_SHORTCUT "S-o"
#define SW_PLAY_WINDOW_SHORTCUT "S-w"
#define SW_PLAY_FILE_SHORTCUT "S-e"
#define SW_PLAY_STOP_SHORTCUT "S-s"

#define SW_ANALYSIS_REGION_SHORTCUT "S-n"
#define SW_INSERT_SIMPLE_LABEL_SHORTCUT "S-l"
#define SW_ERASE_LABEL_SHORTCUT "C-d"
#define SW_SET_REGION_LABEL_SHORTCUT "S-g"
#define SW_CAT_REGION_LABELS_SHORTCUT "A-t"
    
#define SW_CUT_SHORTCUT "A-x"
#define SW_COPY_SHORTCUT "A-c"
#define SW_PASTE_SHORTCUT "A-v"
#define SW_MIX_SHORTCUT "S-m"
#define SW_INSERT_SHORTCUT "S-i"
#define SW_CAT_SHORTCUT "A-d"
#define SW_CATTOP_SHORTCUT "S-d"
#define SW_REPLACE_SHORTCUT "S-r"
    

/* window title */    
#define SW_NO_WAVE_TITLE "No Wave"
#define SW_UNTITLED_TITLE "Untitled"
#define SW_CLIPBOARD_TITLE "Clipboard"

/* parameters for analysis */    
#define SW_ANALYSIS_SPECTRUM_LABEL "Spectrum"
#define SW_ANALYSIS_SMOOTHED_SPECTRUM_LABEL "Smoothed Spectrum"
#define SW_ANALYSIS_PHASE_LABEL "Phase"
#define SW_ANALYSIS_UNWRAPPED_PHASE_LABEL "Unwrapped Phase"
#define SW_ANALYSIS_GROUP_DELAY_LABEL "Group Delay"
#define SW_ANALYSIS_CEPSTRUM_LABEL "Cepstrum"

#define SW_ANALYSIS_SPECTRUM_PARAM_LABEL "spectrum"
#define SW_ANALYSIS_SMOOTHED_SPECTRUM_PARAM_LABEL "smoothed_spectrum"
#define SW_ANALYSIS_PHASE_PARAM_LABEL "phase"
#define SW_ANALYSIS_UNWRAPPED_PHASE_PARAM_LABEL "unwrapped_phase"
#define SW_ANALYSIS_GROUP_DELAY_PARAM_LABEL "group_delay"
#define SW_ANALYSIS_CEPSTRUM_PARAM_LABEL "cepstrum"

#define SW_WINDOW_RECTANGLE_LABEL "Rectangle"
#define SW_WINDOW_HAMMING_LABEL "Hamming"
#define SW_WINDOW_HANNING_LABEL "Hanning"
#define SW_WINDOW_BLACKMAN_LABEL "Blackman"
#define SW_WINDOW_GAUSS_LABEL "Gauss"

#define SW_WINDOW_RECTANGLE_PARAM_LABEL "rectangle"
#define SW_WINDOW_HAMMING_PARAM_LABEL "hamming"
#define SW_WINDOW_HANNING_PARAM_LABEL "hanning"
#define SW_WINDOW_BLACKMAN_PARAM_LABEL "blackman"
#define SW_WINDOW_GAUSS_PARAM_LABEL "gauss"

/* enums */
typedef enum {
    SW_UNKNOWN_DATA = -1,
    SW_TIME_DATA = 0,
    SW_FREQ_DATA = 1,
} swDataType;

/* not used yet */
typedef enum {
    SW_UNKNOWN_PLOT = -1,
    SW_NORMAL_PLOT = 0,
    SW_LOG_PLOT = 1,
    SW_DECIBEL_PLOT = 2,
    SW_DECIBEL10_PLOT = 3,
    SW_ABS_PLOT = 4,
    SW_LOG_ABS_PLOT = 5,
    SW_DECIBEL_ABS_PLOT = 6,
    SW_DECIBEL10_ABS_PLOT = 7,
} swPlotType;

typedef enum {
    SW_FREQ_FORMAT_NONE = -1,
    SW_FREQ_FORMAT_HZ = 0,
    SW_FREQ_FORMAT_KHZ = 1,
    SW_FREQ_FORMAT_POINT = 2,
} swFreqFormat;

typedef enum {
    SW_AUTOSAVE_ID_2DIGIT_SERIAL = 0,
    SW_AUTOSAVE_ID_3DIGIT_SERIAL = 1,
    SW_AUTOSAVE_ID_4DIGIT_SERIAL = 2,
    SW_AUTOSAVE_ID_SERIAL = 3,
    SW_AUTOSAVE_ID_EDGE_POINTS = 4,
} swAutosaveIdType;

typedef enum {
    SW_DRAG_NO_LABEL = 0,
    SW_DRAG_START_LABEL = 1,
    SW_DRAG_END_LABEL = 2,
} swDragLabelType;

typedef enum {
    SW_OVERVIEW_DRAG_NONE = -1,
    SW_OVERVIEW_DRAG_START = 0,
    SW_OVERVIEW_DRAG_END = 1,
    SW_OVERVIEW_DRAG_THUMB = 2,
    SW_OVERVIEW_DRAG_AUDIO = 3,
} swOverviewDragType;

/* structures */
typedef struct _swConfig swConfigRec, *swConfig;
typedef struct _swTopLevel *swTopLevel;
typedef struct _swWindow *swWindow;
typedef struct _swLabelList *swLabelList;

struct _swConfig {
    spBool help_flag;
    spBool prev_flag;
    spBool use_def_format;
    spBool use_lowlevel_thread;
    spBool use_play_command;
    spBool use_loop_play;
    spBool use_sync_play;
    spBool play_draw;
    spBool alt_ctrl_swap;
    spBool no_tool_bar;
    spBool display_overview;
    spBool display_meter;
    spBool draw_selection_times;
    spBool draw_selection_length;
    spBool link_region_label;
    spBool erase_label_prompt;
    spBool display_info_area;
    spBool display_freq_info_area;
    spBool analysis_new_window;
    
    spBool scale_flag;
    spBool grid_flag;
    spBool zero_flag;
    
    swDataType data_type;
    swAnalysisType analysis_type;
    swWindowType window_type;
    int num_channel;
    int def_num_channel;
    int head_size;
    int def_head_size;
    int samp_bit;
    int def_samp_bit;
    int audio_buffer_size;
    
    int width;
    int height;
    int freq_width;
    int freq_height;
    int overview_height;
    int meter_range;
    long fftl;
    long max_fftl;
    double lifterm;
    double shiftm;
    double framem;
    double wide_framem;
    double narrow_framem;
    double samp_rate;
    double def_samp_rate;
    
    char data_type_string[SP_MAX_SETUP_VALUE];
    char analysis_type_string[SP_MAX_SETUP_VALUE];
    char window_type_string[SP_MAX_SETUP_VALUE];
    char format_string[SP_MAX_SETUP_VALUE];
    char def_format_string[SP_MAX_SETUP_VALUE];
    char play_command[SP_MAX_SETUP_VALUE];
    char browser_command[SP_MAX_SETUP_VALUE];
    char doc_dir[SP_MAX_SETUP_VALUE];
    char setup_file[SP_MAX_SETUP_VALUE];
    char temp_dir[SP_MAX_SETUP_VALUE];

    spBool autosave_use_region;
    spBool autosave_label_fullpath;
    spBool autosave_by_drop;
    swAutosaveIdType autosave_id_type;
    char autosave_dir[SP_MAX_SETUP_VALUE];
    char autosave_id_prefix[SP_MAX_SETUP_VALUE];
    char autosave_label_suffix[SP_MAX_SETUP_VALUE];
    
    char wave_fg[SP_MAX_SETUP_VALUE];
    char wave_bg[SP_MAX_SETUP_VALUE];
    char pointer_color[SP_MAX_SETUP_VALUE];
    char string_color[SP_MAX_SETUP_VALUE];
    char label_color[SP_MAX_SETUP_VALUE];
    char scale_color[SP_MAX_SETUP_VALUE];
    
    swTimeFormat time_format;
    swFreqFormat freq_format;

    /* private data */
    spBool format_specified;
    
    swWaveConfig wave_config;
    swTopLevel toplevel;
};

struct _swTopLevel {
    spTopLevel toplevel;
    swWindow null_window;
    swWindow current_window;
    swWindow clipboard_window;
    spBool format_specified;

    long num_window;
    int num_process;
    spBool playable;
    spBool using_clipboard;
    void *main_mutex;
};

struct _swWindow {
    char *title;
    char *name;
    char *data_label;
    char *data_dim;
    swDataType data_type;
    long undo_group_id;
    int index;
    long scroll_coef;
    int width, height;
    int overview_width, overview_height;
    int meter_width;
    double draw_width, draw_height;
    long drawn_pos;

    spBool selecting;
    long sel_st, sel_ed;
    int sel_st_d, sel_ed_d;
    double sel_st_f, sel_ed_f;
    long point;
    int point_d;
    double point_f;
    double prev_point_f;
    int target_channel;
    long active_label_index;
    
    long offset, length;
    swWave wave;

    spBool visible_flag;
    spBool analysis_flag;
    spBool loop_play;
    spBool sync_play;
    spBool drag_region;
    spBool draw_label;
    spBool display_info_area;
    spBool autosave_started;
    spBool process_flag;
    int num_blocked;
    swDragLabelType drag_label_type;
    
    int pointer;
    int direction;
    swConfig config;
    spComponent window;
    spComponent wave_box;
    spComponent overview_canvas;
    spComponent canvas;
    spComponent image;
    spComponent meter_image;
    spComponent meter_bar_image;
    spComponent hscroll;
    spComponent tool_bar;
    spComponent undo_menu;
    spComponent redo_menu;
    spComponent info_box;
    swLabelList label_list;
    struct _swWindow *related_window;
    
    void *mutex;
};

#if defined(MACOS)
#pragma import on
#endif

extern swWindow swGetNextWindow(swWindow window);
extern swWindow swGetPrevWindow(swWindow window);

extern void swInitialize(spTopLevel toplevel, swConfig config);
extern void swCreateWindow(char *filename, char *label_file, spBool is_region_label, swConfig config);
extern void swUpdateMeterWidth(swWindow window);
extern swWindow swInitWindow(swConfig config, swDataType data_type);
extern void swSetWave(swWindow window, swWave wave);
extern void swSetWindowTitle(swWindow window);
extern void swSetSelectSenseLevel(swWindow window, spBool select);
extern void swSetProcessSenseLevel(swWindow window, spBool process);
extern void swSetWindowsSenseLevel(void);
extern void swSetWindowSenseLevel(swWindow window, spBool windows_flag);
extern void swSetSenseLevel(swWindow window);
extern void swResetWindow(swWindow window, spBool in_thread);
extern void swRedrawWindow(swWindow window);




extern void swRemoveTemporaryFile();
extern void swRecordWindow(spComponent component, swWindow window);





extern void swOpenNewWindow(spComponent component, swWindow window);
extern void swOpenWindow(spComponent component, swWindow window);
extern void swSaveAsWindow(spComponent component, swWindow window);
extern void swSaveWindow(spComponent component, swWindow window);

extern swWindow swCreateWaveWindowAt(swWave wave, swConfig config, swDataType data_type,
				     int x, int y);
extern swWindow swCreateWaveWindow(swWave wave, swConfig config, swDataType data_type);
extern void swCreateToolBar(swWindow window);
extern void swCreateMainWindow(swWindow window);

extern void swSetWindowValue(swWindow window);
extern spBool swIsNoWave(swWindow window);
extern spBool swIsNoLabel(swWindow window);
extern spBool swIsProcessing(swWindow window);
extern spBool swIsVisible(swWindow window);
extern spBool swIsPlayable(swWindow window);
extern spBool swNeedWindowProcessBlock(swWindow src_window, swWindow window);
extern void swPopupWindow(swWindow window, int x, int y);

extern spBool swIsClipboardWindow(swWindow window);
extern spBool swIsClipboardVisible(swWindow window);
extern spBool swIsClipboardNone(swWindow window);
extern void swCreateClipboardWindow(swConfig config);
extern void swPopupClipboardWindowCB(spComponent component, swConfig config);

extern spBool swLockWindowMutex(swWindow window);
extern spBool swUnlockWindowMutex(swWindow window);
extern spBool swLockMainMutex(swWindow window);
extern spBool swUnlockMainMutex(swWindow window);
extern void swInitTopLevel(spTopLevel toplevel, swConfig config);
extern long swGetNumWindow(void);
extern long swGetNumProcess(void);

#if defined(MACOS)
#pragma import off
#endif

#ifdef __cplusplus
}  /* Close scope of 'extern "C"' declaration */
#endif

#endif /* __SWWINDOW_H */
